@@ -27,8 +27,19 @@ module CompactedDecl = Context.Compacted.Declaration
 
 let catch_break = ref false
 
+(* tell whether we have a bona fide interrupt *)
+let valid_interrupt () =
+  if Sys.os_type = "Win32" then begin
+    let fname = Shared.get_interrupt_fname (Unix.getpid ()) in
+    let exists = Sys.file_exists fname in
+    if exists then Unix.unlink fname;
+    exists
+  end else
+    true
+
 let init_signal_handler () =
-  let f _ = if !catch_break then raise Sys.Break else Control.interrupt := true in
+  let f _ = if valid_interrupt () then
+              if !catch_break then raise Sys.Break else Control.interrupt := true in
   Sys.set_signal Sys.sigint (Sys.Signal_handle f)
 
 let pr_with_pid s = Printf.eprintf "[pid %d] %s\n%!" (Unix.getpid ()) s
@@ -55,12 +66,13 @@ let coqide_known_option table = List.mem table [
   ["Printing";"Existential";"Instances"];
   ["Printing";"Universes"];
   ["Printing";"Unfocused"];
+  ["Printing";"Goal";"Names"];
   ["Diffs"]]
 
 let is_known_option cmd = match cmd with
-  | VernacSetOption (_, o, OptionSetTrue)
-  | VernacSetOption (_, o, OptionSetString _)
-  | VernacSetOption (_, o, OptionUnset) -> coqide_known_option o
+  VernacSynterp (VernacSetOption (_, o, OptionSetTrue)
+    | VernacSetOption (_, o, OptionSetString _)
+    | VernacSetOption (_, o, OptionUnset)) -> coqide_known_option o
   | _ -> false
 
 let ide_cmd_warns ~id { CAst.loc; v } =
@@ -79,10 +91,16 @@ let set_doc doc = ide_doc := Some doc
 let add ((((s,eid),(sid,verbose)),off),(line_nb,bol_pos)) =
   let doc = get_doc () in
   let open Loc in
-  (* note: this won't yield correct values for bol_pos_last,
-     but the debugger doesn't use that *)
-  let loc = { (initial ToplevelInput) with bp=off; line_nb } in
-  let pa = Pcoq.Parsable.make ~loc (Stream.of_string s) in
+  (* This needs to be akin to what CLexer.after does *)
+  let loc = { (initial ToplevelInput) with
+              bp=off
+            ; line_nb
+            ; bol_pos
+            ; ep=off
+            ; line_nb_last=line_nb
+            ; bol_pos_last=bol_pos } in
+  let r_stream = Gramlib.Stream.of_string ~offset:off s in
+  let pa = Pcoq.Parsable.make ~loc r_stream in
   match Stm.parse_sentence ~doc sid ~entry:Pvernac.main_entry pa with
   | None -> assert false (* s may not be empty *)
   | Some ast ->
@@ -109,13 +127,13 @@ let edit_at id =
  * be removed in the next version of the protocol.
  *)
 let query (route, (s,id)) =
-  let pa = Pcoq.Parsable.make (Stream.of_string s) in
+  let pa = Pcoq.Parsable.make (Gramlib.Stream.of_string s) in
   let doc = get_doc () in
   Stm.query ~at:id ~doc ~route pa
 
 let annotate phrase =
   let doc = get_doc () in
-  let pa = Pcoq.Parsable.make (Stream.of_string phrase) in
+  let pa = Pcoq.Parsable.make (Gramlib.Stream.of_string phrase) in
   match Stm.parse_sentence ~doc (Stm.get_current_state ~doc) ~entry:Pvernac.main_entry pa with
   | None -> Richpp.richpp_of_pp ~width:78 (Pp.mt ())
   | Some ast ->
@@ -173,8 +191,8 @@ let concl_next_tac =
     "right"
   ])
 
-let process_goal sigma g =
-  let evi = Evd.find sigma g in
+let process_goal short sigma g =
+  let evi = Evd.find_undefined sigma g in
   let env = Evd.evar_filtered_env (Global.env ()) evi in
   let min_env = Environ.reset_context env in
   let name = if Printer.print_goal_names () then Some (Names.Id.to_string (Termops.evar_suggested_name env sigma g)) else None in
@@ -183,56 +201,81 @@ let process_goal sigma g =
   in
   let process_hyp d (env,l) =
     let d' = CompactedDecl.to_named_context d in
-      (List.fold_right Environ.push_named d' env,
-       (pr_compacted_decl env sigma d) :: l) in
-  let (_env, hyps) =
-    Context.Compacted.fold process_hyp
-      (Termops.compact_named_context (Environ.named_context env)) ~init:(min_env,[]) in
-  { Interface.goal_hyp = List.rev hyps; Interface.goal_ccl = ccl; Interface.goal_id = Goal.uid g; Interface.goal_name = name }
-
-let process_goal_diffs diff_goal_map oldp nsigma ng =
-  let name = if Printer.print_goal_names () then Some (Names.Id.to_string (Termops.evar_suggested_name (Global.env ()) nsigma ng)) else None in
-  let og_s = match oldp with
-    | Some oldp ->
-      let Proof.{ sigma=osigma } = Proof.data oldp in
-      (try Some (Evar.Map.find ng diff_goal_map, osigma)
-       with Not_found -> None)
-    | None -> None
+      (List.fold_right EConstr.push_named d' env,
+       (pr_ecompacted_decl env sigma d) :: l) in
+  let hyps =
+    if short then [] else
+      let (_env, hyps) =
+        Context.Compacted.fold process_hyp
+        (Termops.compact_named_context sigma (EConstr.named_context env)) ~init:(min_env,[])
+      in
+      hyps
   in
-  let (hyps_pp_list, concl_pp) = Proof_diffs.diff_goal_ide og_s ng nsigma in
+  { Interface.goal_hyp = List.rev hyps; Interface.goal_ccl = ccl; Interface.goal_id = Proof.goal_uid g; Interface.goal_name = name }
+
+let process_goal_diffs ~short diff_goal_map oldp nsigma ng =
+  let env = Global.env () in
+  let name = if Printer.print_goal_names () then Some (Names.Id.to_string (Termops.evar_suggested_name env nsigma ng)) else None in
+  let og_s = match oldp, diff_goal_map with
+  | Some oldp, Some diff_goal_map -> Proof_diffs.map_goal ng diff_goal_map
+  | None, _ | _, None -> None
+  in
+  let (hyps_pp_list, concl_pp) = Proof_diffs.diff_goal ~short ?og_s (Proof_diffs.make_goal env nsigma ng) in
   { Interface.goal_hyp = hyps_pp_list; Interface.goal_ccl = concl_pp;
-    Interface.goal_id = Goal.uid ng; Interface.goal_name = name }
+    Interface.goal_id = Proof.goal_uid ng; Interface.goal_name = name }
 
-let export_pre_goals Proof.{ sigma; goals; stack } process =
-  let process = List.map (process sigma) in
-  { Interface.fg_goals       = process goals
-  ; Interface.bg_goals       = List.(map (fun (lg,rg) -> process lg, process rg)) stack
-  ; Interface.shelved_goals  = process @@ Evd.shelf sigma
-  ; Interface.given_up_goals = process (Evar.Set.elements @@ Evd.given_up sigma)
-  }
+let export_pre_goals flags Proof.{ sigma; goals; stack } process =
+  let open Interface in
+  let process x = List.map (process sigma) x in
+  let fg_goals = if flags.gf_fg then process goals else [] in
+  let bg_goals =
+    if flags.gf_bg then List.(map (fun (lg,rg) -> process lg, process rg)) stack
+    else []
+  in
+  let shelved_goals =
+    if flags.gf_shelved then process @@ Evd.shelf sigma
+    else []
+  in
+  let given_up_goals =
+    if flags.gf_given_up then process (Evar.Set.elements @@ Evd.given_up sigma)
+    else []
+  in
+  { fg_goals; bg_goals; shelved_goals; given_up_goals }
 
-let goals () =
+let subgoals flags =
   let doc = get_doc () in
-  set_doc @@ Stm.finish ~doc;
+  ignore (Stm.finish ~doc : Vernacstate.t);
+  let short = match flags.Interface.gf_mode with
+  | "short" -> true
+  | _ -> false
+  in
   try
     let newp = Vernacstate.Declare.give_me_the_proof () in
     if Proof_diffs.show_diffs () then begin
       let oldp = Stm.get_prev_proof ~doc (Stm.get_current_state ~doc) in
       (try
-        let diff_goal_map = Proof_diffs.make_goal_map oldp newp in
-        Some (export_pre_goals Proof.(data newp) (process_goal_diffs diff_goal_map oldp))
+        let diff_goal_map = match oldp with
+        | None -> None
+        | Some oldp -> Some (Proof_diffs.make_goal_map oldp newp)
+        in
+        Some (export_pre_goals flags Proof.(data newp) (process_goal_diffs ~short diff_goal_map oldp))
        with Pp_diff.Diff_Failure msg ->
          Proof_diffs.notify_proof_diff_failure msg;
-         Some (export_pre_goals Proof.(data newp) process_goal))
+         Some (export_pre_goals flags Proof.(data newp) (process_goal short)))
     end else
-      Some (export_pre_goals Proof.(data newp) process_goal)
+      Some (export_pre_goals flags Proof.(data newp) (process_goal short))
   with Vernacstate.Declare.NoCurrentProof -> None
-  [@@ocaml.warning "-3"];;
+  [@@ocaml.warning "-3"]
+
+let goals () =
+  let open Interface in
+  let all = { gf_mode = "full"; gf_fg = true; gf_bg = true; gf_shelved = true; gf_given_up = true } in
+  subgoals all
 
 let evars () =
   try
     let doc = get_doc () in
-    set_doc @@ Stm.finish ~doc;
+    ignore (Stm.finish ~doc : Vernacstate.t);
     let pfts = Vernacstate.Declare.give_me_the_proof () in
     let Proof.{ sigma } = Proof.data pfts in
     let exl = Evar.Map.bindings (Evd.undefined_map sigma) in
@@ -249,7 +292,7 @@ let hints () =
     match goals with
     | [] -> None
     | g :: _ ->
-      let env = Evd.evar_filtered_env (Global.env ()) (Evd.find sigma g) in
+      let env = Evd.evar_filtered_env (Global.env ()) (Evd.find_undefined sigma g) in
       let get_hint_hyp env d accu = hyp_next_tac sigma env d :: accu in
       let hint_hyps = List.rev (Environ.fold_named_context get_hint_hyp env ~init: []) in
       Some (hint_hyps, concl_next_tac)
@@ -260,15 +303,14 @@ let hints () =
 
 let wait () =
   let doc = get_doc () in
-  set_doc (Stm.wait ~doc)
+  Stm.wait ~doc
 
 let status force =
   (* We remove the initial part of the current [DirPath.t]
      (usually Top in an interactive session, cf "coqtop -top"),
      and display the other parts (opened sections and modules) *)
-  set_doc (Stm.finish ~doc:(get_doc ()));
-  if force then
-    set_doc (Stm.join ~doc:(get_doc ()));
+  ignore (Stm.finish ~doc:(get_doc ()) : Vernacstate.t);
+  if force then Stm.join ~doc:(get_doc ());
   let path =
     let l = Names.DirPath.repr (Lib.cwd ()) in
     List.rev_map Names.Id.to_string l
@@ -350,12 +392,10 @@ let import_option_value = function
 
 let export_option_state s = {
   Interface.opt_sync  = true;
-  Interface.opt_depr  = s.Goptions.opt_depr;
+  Interface.opt_depr  = Option.has_some s.Goptions.opt_depr;
   Interface.opt_value = export_option_value s.Goptions.opt_value;
 }
 
-exception NotSupported of string
-
 let proof_diff (diff_opt, sid) =
   let diff_opt = Proof_diffs.string_to_diffs diff_opt in
   let doc = get_doc () in
@@ -442,7 +482,7 @@ let init =
            get_doc (), init_sid, `NewTip in
          if Filename.check_suffix file ".v" then
            Stm.set_compilation_hints file;
-         set_doc (Stm.finish ~doc);
+         ignore (Stm.finish ~doc : Vernacstate.t);
          initial_id
    end
 
@@ -500,6 +540,7 @@ let eval_call c =
     Interface.stop_worker = Stm.stop_worker;
     Interface.print_ast = print_ast;
     Interface.annotate = interruptible annotate;
+    Interface.subgoals = interruptible subgoals;
   } in
   Xmlprotocol.abstract_eval_call handler c
 
@@ -537,7 +578,7 @@ let msg_format = ref (fun () ->
 let loop ( { Coqtop.run_mode; color_mode },_) ~opts:_ state =
   match run_mode with
   | Coqtop.Batch -> exit 0
-  | Coqtop.(Query PrintTags) -> Coqtop.print_style_tags color_mode; exit 0
+  | Coqtop.(Query PrintTags) -> Colors.print_style_tags color_mode; exit 0
   | Coqtop.(Query _) -> Printf.eprintf "Unknown query"; exit 1
   | Coqtop.Interactive ->
   let open Vernac.State in
@@ -638,7 +679,7 @@ let rec parse = function
        x :: parse rest
   | [] -> []
 
-let coqidetop_specific_usage = Usage.{
+let coqidetop_specific_usage = Boot.Usage.{
   executable_name = "coqidetop";
   extra_args = "";
   extra_options = "\n\
@@ -666,6 +707,7 @@ let islave_default_opts = Coqargs.default
 
 let () =
   let open Coqtop in
+  Shared_os_specific.init ();
   let custom = {
       parse_extra = islave_parse ;
       usage = coqidetop_specific_usage;
