@@ -10,6 +10,7 @@
 
 open Names
 open Notation_term
+open Notationextern
 open Globnames
 
 (** Abbreviations. *)
@@ -17,9 +18,15 @@ open Globnames
 val declare_abbreviation : local:bool -> ?also_in_cases_pattern:bool -> Deprecation.t option -> Id.t ->
   onlyparsing:bool -> interpretation -> unit
 
-val search_abbreviation : ?loc:Loc.t -> abbreviation -> interpretation
+val search_abbreviation : abbreviation -> interpretation
 
-val search_filtered_abbreviation : ?loc:Loc.t ->
+val search_filtered_abbreviation :
   (interpretation -> 'a option) -> abbreviation -> 'a option
 
-val import_abbreviation : int -> Libnames.full_path -> abbreviation -> unit
+val import_abbreviation : int -> Libnames.full_path -> KerName.t -> unit
+
+(** Activate (if on:true) or deactivate (if on:false) an abbreviation assumed to exist *)
+val toggle_abbreviation : on:bool -> use:notation_use -> abbreviation -> unit
+
+(** Activate (if on:true) or deactivate (if on:false) all abbreviations satisfying a criterion *)
+val toggle_abbreviations : on:bool -> use:notation_use -> (Libnames.full_path -> interpretation -> bool) -> unit
