@@ -63,7 +63,7 @@ val compute_internalization_env : env -> evar_map -> ?impls:internalization_env
   internalization_env
 
 val extend_internalization_data :
-  var_internalization_data -> Impargs.implicit_status -> scope_name option -> var_internalization_data
+  var_internalization_data -> Impargs.implicit_status -> scope_name list -> var_internalization_data
 
 type ltac_sign = {
   ltac_vars : Id.Set.t;
@@ -82,9 +82,11 @@ val intern_constr : env -> evar_map -> constr_expr -> glob_constr
 val intern_type : env -> evar_map -> constr_expr -> glob_constr
 
 val intern_gen : typing_constraint -> env -> evar_map ->
-  ?impls:internalization_env -> ?pattern_mode:bool -> ?ltacvars:ltac_sign ->
+  ?impls:internalization_env -> ?strict_check:bool -> ?pattern_mode:bool -> ?ltacvars:ltac_sign ->
   constr_expr -> glob_constr
 
+val intern_unknown_if_term_or_type : env -> evar_map -> constr_expr -> glob_constr
+
 val intern_pattern : env -> cases_pattern_expr ->
   lident list * (Id.t Id.Map.t * cases_pattern) list
 
@@ -139,13 +141,8 @@ val interp_type_evars_impls : ?flags:inference_flags -> env -> evar_map ->
 
 (** Without typing *)
 val intern_constr_pattern :
-  env -> evar_map -> ?as_type:bool -> ?ltacvars:ltac_sign ->
-    constr_pattern_expr -> patvar list * constr_pattern
-
-(** With typing *)
-val interp_constr_pattern :
-  env -> evar_map -> ?expected_type:typing_constraint ->
-    constr_pattern_expr -> constr_pattern
+  env -> evar_map -> ?as_type:bool -> ?strict_check:bool -> ?ltacvars:ltac_sign ->
+    constr_pattern_expr -> Id.Set.t * constr_pattern
 
 (** Returns None if it's an abbreviation not bound to a name, raises an error
     if not existing *)
@@ -194,16 +191,13 @@ val interp_notation_constr : env -> ?impls:internalization_env ->
 (** Idem but to glob_constr (weaker check of binders) *)
 
 val intern_core : typing_constraint ->
-  env -> evar_map -> ?pattern_mode:bool -> ?ltacvars:ltac_sign ->
+  env -> evar_map -> ?strict_check:bool -> ?pattern_mode:bool -> ?ltacvars:ltac_sign ->
   Genintern.intern_variable_status -> constr_expr ->
   glob_constr
 
 (** Globalization options *)
 val parsing_explicit : bool ref
 
-(** Globalization leak for Grammar *)
-val for_grammar : ('a -> 'b) -> 'a -> 'b
-
 (** Placeholder for global option, should be moved to a parameter *)
 val get_asymmetric_patterns : unit -> bool
 
