@@ -65,17 +65,13 @@ let mind_check_names mie =
 (************************************************************************)
 
 type univ_info = { ind_squashed : bool; ind_has_relevant_arg : bool;
-                   ind_min_univ : Sorts.t option; (* Some for template *)
+                   ind_template : bool;
                    ind_univ : Sorts.t;
                    missing : Sorts.t list; (* missing u <= ind_univ constraints *)
                  }
 
-let sup_sort s1 s2 = match s1, s2 with
-| (_, SProp) -> assert false (* template SProp not allowed *)
-| (SProp, s) | (Prop, s) | (s, Prop) -> s
-| (Set, Set) -> Sorts.set
-| (Set, Type u) | (Type u, Set) -> Sorts.sort_of_univ (Universe.sup u Universe.type0)
-| (Type u, Type v) -> Sorts.sort_of_univ (Universe.sup u v)
+let no_sort_variable () =
+  CErrors.anomaly (Pp.str "A sort variable was sent to the kernel")
 
 let check_univ_leq ?(is_real_arg=false) env u info =
   let ind_univ = info.ind_univ in
@@ -84,10 +80,8 @@ let check_univ_leq ?(is_real_arg=false) env u info =
     else info
   in
   (* Inductive types provide explicit lifting from SProp to other universes, so allow SProp <= any. *)
-  if Sorts.is_sprop u || UGraph.check_leq_sort (universes env) u ind_univ
-  then { info with ind_min_univ = Option.map (sup_sort u) info.ind_min_univ }
-  else if is_impredicative_sort env ind_univ
-       && Option.is_empty info.ind_min_univ then { info with ind_squashed = true }
+  if Sorts.is_sprop u || UGraph.check_leq_sort (universes env) u ind_univ then info
+  else if is_impredicative_sort env ind_univ then { info with ind_squashed = true }
   else {info with missing = u :: info.missing}
 
 let check_context_univs ~ctor env info ctx =
@@ -111,11 +105,10 @@ let check_indices_matter env_params info indices =
 let check_arity ~template env_params env_ar ind =
   let {utj_val=arity;utj_type=_} = Typeops.infer_type env_params ind.mind_entry_arity in
   let indices, ind_sort = Reduction.dest_arity env_params arity in
-  let ind_min_univ = if template then Some Sorts.prop else None in
   let univ_info = {
     ind_squashed=false;
     ind_has_relevant_arg=false;
-    ind_min_univ;
+    ind_template = template;
     ind_univ=ind_sort;
     missing=[];
   }
@@ -136,7 +129,7 @@ let check_constructor_univs env_ar_par info (args,_) =
 
 let check_constructors env_ar_par isrecord params lc (arity,indices,univ_info) =
   let lc = Array.map_of_list (fun c -> (Typeops.infer_type env_ar_par c).utj_val) lc in
-  let splayed_lc = Array.map (Reduction.dest_prod_assum env_ar_par) lc in
+  let splayed_lc = Array.map (Reduction.hnf_decompose_prod_decls env_ar_par) lc in
   let univ_info = match Array.length lc with
     (* Empty type: all OK *)
     | 0 -> univ_info
@@ -190,7 +183,7 @@ let check_record data =
 (* - all_sorts in case of small, unitary Prop (not smashed) *)
 (* - logical_sorts in case of large, unitary Prop (smashed) *)
 
-let allowed_sorts {ind_squashed;ind_univ;ind_min_univ=_;ind_has_relevant_arg=_;missing=_} =
+let allowed_sorts {ind_squashed;ind_univ;ind_template=_;ind_has_relevant_arg=_;missing=_} =
   if not ind_squashed then InType
   else Sorts.family ind_univ
 
@@ -206,18 +199,18 @@ let unbounded_from_below u cstrs =
       | Lt | Le -> not (Univ.Level.equal r u))
     cstrs
 
-let get_template univs ~env_params ~env_ar_par ~params data entries =
+let get_template univs ~env_params ~env_ar_par ~params entries =
   match univs with
   | Polymorphic_ind_entry _ | Monomorphic_ind_entry -> None
   | Template_ind_entry ctx ->
-    let is_prop = match data with
-      | [ _, _, info ] -> Sorts.is_prop @@ Option.get info.ind_min_univ
+    let entry = match entries with
+      | [entry] -> entry
       | _ -> CErrors.user_err Pp.(str "Template-polymorphism not allowed with mutual inductives.")
     in
     (* Compute potential template parameters *)
     let map decl = match decl with
     | LocalAssum (_, p) ->
-      let c = Term.strip_prod_assum p in
+      let c = Term.strip_prod_decls p in
       let s = match kind c with
       | Sort (Type u) ->
         begin match Universe.level u with
@@ -230,36 +223,27 @@ let get_template univs ~env_params ~env_ar_par ~params data entries =
     | LocalDef _ -> None
     in
     let params = List.map_filter map params in
-    let params =
-      if is_prop then
-        (* Inductive types in Prop have no template universes, but are still
-           marked as template to please the upper layers. *)
-        List.map (fun _ -> None) params
-      else
-        (* We reuse the same code as the one for variance inference. *)
-        let init_variance = Array.map_of_list (fun l -> l, None) (Level.Set.elements (fst ctx)) in
-        let variance = InferCumulativity.infer_inductive ~env_params ~env_ar_par init_variance
-            ~arities:(List.map (fun e -> e.mind_entry_arity) entries)
-            ~ctors:(List.map (fun e -> e.mind_entry_lc) entries)
-        in
-        let fold accu v (l, _) = match v with
-        | Variance.Irrelevant -> Level.Set.add l accu
-        | Variance.Covariant | Variance.Invariant -> accu
-        in
-        let irrel = Array.fold_left2 fold Level.Set.empty variance init_variance in
-        let () =
-          if Level.Set.is_empty irrel then
-          CErrors.user_err
-            Pp.(strbrk "Ill-formed template inductive declaration: not polymorphic on any universe.")
-        in
-        let map = function
-        | None -> None
-        | Some l ->
-          if Level.Set.mem l irrel && unbounded_from_below l (snd ctx) then Some l
-          else None
-        in
-        List.rev_map map params
+    let fold accu u = match u with None -> accu | Some u -> Level.Set.add u accu in
+    let plevels = List.fold_left fold Level.Set.empty params in
+    let unbound = Level.Set.diff (fst ctx) plevels in
+    let plevels =
+      if not (Level.Set.is_empty unbound) then
+        CErrors.user_err Pp.(strbrk "The following template universes are not \
+          bound by parameters: " ++ pr_sequence Level.raw_pr (Level.Set.elements unbound))
+      else Level.Set.elements plevels
     in
+    let check_bound l =
+      if not (unbounded_from_below l (snd ctx)) then
+        CErrors.user_err Pp.(strbrk "Universe level " ++ Level.raw_pr l ++ strbrk " has a lower bound")
+    in
+    let () = List.iter check_bound plevels in
+    (* We reuse the same code as the one for variance inference. *)
+    let init_variance = Array.map_of_list (fun l -> l, Some Variance.Irrelevant) plevels in
+    let _variance = InferCumulativity.infer_inductive ~env_params ~env_ar_par init_variance
+        ~arities:[entry.mind_entry_arity]
+        ~ctors:[entry.mind_entry_lc]
+    in
+    let params = List.rev params in
     Some { template_param_levels = params; template_context = ctx }
 
 let abstract_packets usubst ((arity,lc),(indices,splayed_lc),univ_info) =
@@ -277,11 +261,14 @@ let abstract_packets usubst ((arity,lc),(indices,splayed_lc),univ_info) =
   let ind_univ = match univ_info.ind_univ with
   | Prop | SProp | Set -> univ_info.ind_univ
   | Type u -> Sorts.sort_of_univ (Univ.subst_univs_level_universe usubst u)
+  | QSort _ -> no_sort_variable ()
   in
 
-  let arity = match univ_info.ind_min_univ with
-    | None -> RegularArity {mind_user_arity = arity; mind_sort = ind_univ}
-    | Some min_univ -> TemplateArity { template_level = min_univ; }
+  let arity =
+    if univ_info.ind_template then
+      TemplateArity { template_level = univ_info.ind_univ; }
+    else
+      RegularArity {mind_user_arity = arity; mind_sort = ind_univ}
   in
 
   let kelim = allowed_sorts univ_info in
@@ -369,7 +356,7 @@ let typecheck_inductive env ~sec_univs (mie:mutual_inductive_entry) =
         Some variances
   in
 
-  let template = get_template mie.mind_entry_universes ~env_params ~env_ar_par ~params data mie.mind_entry_inds in
+  let template = get_template mie.mind_entry_universes ~env_params ~env_ar_par ~params mie.mind_entry_inds in
 
   (* Abstract universes *)
   let usubst, univs = match mie.mind_entry_universes with
