@@ -22,9 +22,6 @@ open Environ
 open Modops
 open Mod_subst
 
-type 'alg translation =
-  module_signature * 'alg * delta_resolver * Univ.Constraints.t
-
 let rec mp_from_mexpr = function
   | MEident mp -> mp
   | MEapply (expr,_) -> mp_from_mexpr expr
@@ -55,12 +52,18 @@ let rec rebuild_mp mp l =
   | i::r -> rebuild_mp (MPdot(mp,Label.of_id i)) r
 
 let infer_gen_conv state env c1 c2 =
-  Reduction.generic_conv Reduction.CONV ~l2r:false (fun _ -> None) TransparentState.full env state c1 c2
+  Conversion.generic_conv Conversion.CONV ~l2r:false Constr.default_evar_handler TransparentState.full env state c1 c2
 
 let infer_gen_conv_leq state env c1 c2 =
-  Reduction.generic_conv Reduction.CUMUL ~l2r:false (fun _ -> None) TransparentState.full env state c1 c2
+  Conversion.generic_conv Conversion.CUMUL ~l2r:false Constr.default_evar_handler TransparentState.full env state c1 c2
 
-let rec check_with_def (cst, ustate) env struc (idl,(c,ctx)) mp reso =
+type with_body = {
+  w_def : Constr.t;
+  w_univs : universes;
+  w_bytecode : Vmemitcodes.body_code option;
+}
+
+let rec check_with_def (cst, ustate) env struc (idl, wth) mp reso =
   let lab,idl = match idl with
     | [] -> assert false
     | id::idl -> Label.of_id id, idl
@@ -78,23 +81,23 @@ let rec check_with_def (cst, ustate) env struc (idl,(c,ctx)) mp reso =
       (* In the spirit of subtyping.check_constant, we accept
          any implementations of parameters and opaque terms,
          as long as they have the right type *)
-      let univs, ctx' =
-        match cb.const_universes, ctx with
-        | Monomorphic, None ->
+      let ctx' =
+        match cb.const_universes, wth.w_univs with
+        | Monomorphic, Monomorphic ->
           let cst = match cb.const_body with
             | Undef _ | OpaqueDef _ ->
-              let j = Typeops.infer env' c in
-              assert (j.uj_val == c); (* relevances should already be correct here *)
+              let j = Typeops.infer env' wth.w_def in
+              assert (j.uj_val == wth.w_def); (* relevances should already be correct here *)
               let typ = cb.const_type in
               let cst = infer_gen_conv_leq (cst, ustate) env' j.uj_type typ in
               cst
             | Def c' ->
-              infer_gen_conv (cst, ustate) env' c c'
+              infer_gen_conv (cst, ustate) env' wth.w_def c'
             | Primitive _ ->
               error_incorrect_with_constraint lab
           in
-          Monomorphic, cst
-        | Polymorphic uctx, Some ctx ->
+          cst
+        | Polymorphic uctx, Polymorphic ctx ->
           let () =
             if not (UGraph.check_subtype (Environ.universes env) uctx ctx) then
               error_incorrect_with_constraint lab
@@ -103,32 +106,29 @@ let rec check_with_def (cst, ustate) env struc (idl,(c,ctx)) mp reso =
           let env' = Environ.push_context ~strict:false (Univ.AbstractContext.repr uctx) env in
           let () = match cb.const_body with
             | Undef _ | OpaqueDef _ ->
-              let j = Typeops.infer env' c in
-              assert (j.uj_val == c); (* relevances should already be correct here *)
+              let j = Typeops.infer env' wth.w_def in
+              assert (j.uj_val == wth.w_def); (* relevances should already be correct here *)
               let typ = cb.const_type in
               begin
-                try Reduction.conv_leq env' j.uj_type typ
-                with Reduction.NotConvertible -> error_incorrect_with_constraint lab
+                try Conversion.conv_leq env' j.uj_type typ
+                with Conversion.NotConvertible -> error_incorrect_with_constraint lab
               end
             | Def c' ->
               begin
-                try Reduction.conv env' c c'
-                with Reduction.NotConvertible -> error_incorrect_with_constraint lab
+                try Conversion.conv env' wth.w_def c'
+                with Conversion.NotConvertible -> error_incorrect_with_constraint lab
               end
             | Primitive _ ->
               error_incorrect_with_constraint lab
           in
-          Polymorphic ctx, cst
+          cst
         | _ -> error_incorrect_with_constraint lab
       in
-      let def = Def c in
-      (*      let ctx' = Univ.UContext.make (newus, cst) in *)
       let cb' =
         { cb with
-          const_body = def;
-          const_universes = univs ;
-          const_body_code =
-              (Vmbytegen.compile_constant_body ~fail_on_error:false env' cb.const_universes def) }
+          const_body = Def wth.w_def;
+          const_universes = wth.w_univs;
+          const_body_code = wth.w_bytecode; }
       in
       before@(lab,SFBconst(cb'))::after, ctx'
     else
@@ -141,7 +141,7 @@ let rec check_with_def (cst, ustate) env struc (idl,(c,ctx)) mp reso =
         | Abstract ->
           let struc = Modops.destr_nofunctor (MPdot (mp,lab)) mb.mod_type in
           let struc', cst =
-            check_with_def (cst, ustate) env' struc (idl,(c,ctx)) (MPdot(mp,lab)) mb.mod_delta
+            check_with_def (cst, ustate) env' struc (idl, wth) (MPdot(mp,lab)) mb.mod_delta
           in
           let mb' = { mb with
                       mod_type = NoFunctor struc';
@@ -152,7 +152,7 @@ let rec check_with_def (cst, ustate) env struc (idl,(c,ctx)) mp reso =
       end
   with
   | Not_found -> error_no_such_label lab mp
-  | Reduction.NotConvertible -> error_incorrect_with_constraint lab
+  | Conversion.NotConvertible -> error_incorrect_with_constraint lab
 
 let rec check_with_mod (cst, ustate) env struc (idl,new_mp) mp reso =
   let lab,idl = match idl with
@@ -175,7 +175,7 @@ let rec check_with_mod (cst, ustate) env struc (idl,new_mp) mp reso =
           let mtb_old = module_type_of_module old in
           let cst = Subtyping.check_subtypes (cst, ustate) env' new_mtb mtb_old in
           cst
-        | Algebraic (NoFunctor (MEident(mp'))) ->
+        | Algebraic (MENoFunctor (MEident(mp'))) ->
           check_modpath_equiv env' new_mp mp';
           cst
         | _ -> error_generative_module_expected lab
@@ -185,7 +185,7 @@ let rec check_with_mod (cst, ustate) env struc (idl,new_mp) mp reso =
       let new_mb' =
         { new_mb with
           mod_mp = mp';
-          mod_expr = Algebraic (NoFunctor (MEident new_mp));
+          mod_expr = Algebraic (MENoFunctor (MEident new_mp));
         }
       in
       let new_reso = add_delta_resolver reso new_mb.mod_delta in
@@ -217,7 +217,7 @@ let rec check_with_mod (cst, ustate) env struc (idl,new_mp) mp reso =
         let id_subst = map_mp mp' mp' reso' in
         let new_after = subst_structure id_subst after in
         before@(lab,SFBmodule new_mb)::new_after, new_reso, cst
-      | Algebraic (NoFunctor (MEident mp0)) ->
+      | Algebraic (MENoFunctor (MEident mp0)) ->
         let mpnew = rebuild_mp mp0 idl in
         check_modpath_equiv env' mpnew mp;
         before@(lab,spec)::after, reso, cst
@@ -225,12 +225,15 @@ let rec check_with_mod (cst, ustate) env struc (idl,new_mp) mp reso =
       end
   with
   | Not_found -> error_no_such_label lab mp
-  | Reduction.NotConvertible -> error_incorrect_with_constraint lab
+  | Conversion.NotConvertible -> error_incorrect_with_constraint lab
 
 let check_with ustate env mp (sign,reso,cst) = function
   | WithDef(idl, (c, ctx)) ->
     let struc = destr_nofunctor mp sign in
-    let struc', cst = check_with_def (cst, ustate) env struc (idl, (c, ctx)) mp reso in
+    let univs = match ctx with None -> Monomorphic | Some uctx -> Polymorphic uctx in
+    let bcode = Vmbytegen.compile_constant_body ~fail_on_error:false env univs (Def c) in
+    let body = { w_def = c; w_univs = univs; w_bytecode = bcode } in
+    let struc', cst = check_with_def (cst, ustate) env struc (idl, body) mp reso in
     NoFunctor struc', reso, cst
   | WithMod(idl,new_mp) ->
     let struc = destr_nofunctor mp sign in
@@ -296,13 +299,13 @@ let rec translate_mse_funct (cst, ustate) env ~is_mod mp inl mse = function
     let sign,reso =
       if is_mod then sign,reso
       else subst_modtype_signature_and_resolver (mp_from_mexpr mse) mp sign reso in
-    sign, NoFunctor alg, reso, cst
+    sign, MENoFunctor alg, reso, cst
   |(mbid, ty, ty_inl) :: params ->
     let mp_id = MPbound mbid in
     let mtb, cst = translate_modtype (cst, ustate) env mp_id ty_inl ([],ty) in
     let env' = add_module_type mp_id mtb env in
     let sign,alg,reso,cst = translate_mse_funct (cst, ustate) env' ~is_mod mp inl mse params in
-    let alg' = MoreFunctor (mbid,mtb,alg) in
+    let alg' = MEMoreFunctor alg in
     MoreFunctor (mbid, mtb, sign), alg',reso, cst
 
 and translate_modtype state env mp inl (params,mte) =
@@ -314,7 +317,7 @@ and translate_modtype state env mp inl (params,mte) =
     from an already-translated (or interactive) implementation and
     an (optional) signature entry, produces a final [module_body] *)
 
-let finalize_module (cst, ustate) env mp (sign,alg,reso) restype = match restype with
+let finalize_module_alg (cst, ustate) env mp (sign,alg,reso) restype = match restype with
   | None ->
     let impl = match alg with Some e -> Algebraic e | None -> FullStruct in
     mk_mod mp impl sign reso, cst
@@ -322,7 +325,15 @@ let finalize_module (cst, ustate) env mp (sign,alg,reso) restype = match restype
     let res_mtb, cst = translate_modtype (cst, ustate) env mp inl params_mte in
     let auto_mtb = mk_modtype mp sign reso in
     let cst = Subtyping.check_subtypes (cst, ustate) env auto_mtb res_mtb in
-    let impl = match alg with Some e -> Algebraic e | None -> Struct sign in
+    let impl = match alg with
+    | Some e -> Algebraic e
+    | None ->
+      let sign = match sign with
+      | NoFunctor s -> s
+      | MoreFunctor _ -> assert false (* All non-algebraic callers enforce this *)
+      in
+      Struct sign
+    in
     { res_mtb with
       mod_mp = mp;
       mod_expr = impl;
@@ -331,6 +342,9 @@ let finalize_module (cst, ustate) env mp (sign,alg,reso) restype = match restype
     (** constraints from module body typing + subtyping + module type. *)
     cst
 
+let finalize_module univs env mp (sign, reso) typ =
+  finalize_module_alg univs env mp (sign, None, reso) typ
+
 let translate_module (cst, ustate) env mp inl = function
   | MType (params,ty) ->
     let mtb, cst = translate_modtype (cst, ustate) env mp inl (params,ty) in
@@ -338,7 +352,7 @@ let translate_module (cst, ustate) env mp inl = function
   |MExpr (params,mse,oty) ->
     let (sg,alg,reso,cst) = translate_mse_funct (cst, ustate) env ~is_mod:true mp inl mse params in
     let restype = Option.map (fun ty -> ((params,ty),inl)) oty in
-    finalize_module (cst, ustate) env mp (sg,Some alg,reso) restype
+    finalize_module_alg (cst, ustate) env mp (sg,Some alg,reso) restype
 
 (** We now forbid any Include of functors with restricted signatures.
     Otherwise, we could end with the creation of undesired axioms
@@ -346,8 +360,8 @@ let translate_module (cst, ustate) env mp inl = function
     thanks to strengthening. *)
 
 let rec unfunct = function
-  | NoFunctor me -> me
-  | MoreFunctor(_,_,me) -> unfunct me
+  | MENoFunctor me -> me
+  | MEMoreFunctor me -> unfunct me
 
 let rec forbid_incl_signed_functor env = function
   | MEapply(fe,_) -> forbid_incl_signed_functor env fe
