@@ -152,8 +152,8 @@ sig
   val is_empty : t -> bool
   (** Test whether a directory path is empty. *)
 
-  val initial : t
-  (** Initial "seed" of the unique identifier generator *)
+  val dummy : t
+  (** Used in [Safe_typing.empty_environment] and similar *)
 
   val hcons : t -> t
   (** Hashconsing of directory paths. *)
@@ -258,8 +258,8 @@ sig
 
   val is_bound : t -> bool
 
-  val initial : t
-  (** Name of the toplevel structure ([= MPfile initial_dir]) *)
+  val dummy : t
+  (** ([= MPfile DirPath.dummy]) *)
 
   val dp : t -> DirPath.t
 
@@ -386,9 +386,6 @@ sig
   val user : t -> KerName.t
   val canonical : t -> KerName.t
 
-  val repr2 : t -> ModPath.t * Label.t
-  (** Shortcut for [KerName.repr (user ...)] *)
-
   val modpath : t -> ModPath.t
   (** Shortcut for [KerName.modpath (user ...)] *)
 
@@ -460,9 +457,6 @@ sig
   val user : t -> KerName.t
   val canonical : t -> KerName.t
 
-  val repr2 : t -> ModPath.t * Label.t
-  (** Shortcut for [KerName.repr (user ...)] *)
-
   val modpath : t -> ModPath.t
   (** Shortcut for [KerName.modpath (user ...)] *)
 
@@ -621,7 +615,7 @@ module Projection : sig
   module Repr : sig
     type t
 
-    val make : inductive -> proj_npars:int -> proj_arg:int -> Label.t -> t
+    val make : inductive -> proj_npars:int -> proj_arg:int -> proj_relevant:bool -> Label.t -> t
 
     include QNameS with type t := t
 
@@ -633,13 +627,14 @@ module Projection : sig
     val npars : t -> int
     val arg : t -> int
     val label : t -> Label.t
+    val relevant : t -> bool
 
     val equal : t -> t -> bool [@@ocaml.deprecated "Use QProjection.equal"]
     val hash : t -> int [@@ocaml.deprecated "Use QProjection.hash"]
     val compare : t -> t -> int [@@ocaml.deprecated "Use QProjection.compare"]
 
     val map : (MutInd.t -> MutInd.t) -> t -> t
-    val map_npars : (MutInd.t -> int -> MutInd.t * int) -> t -> t
+    val map_npars : (int -> int) -> t -> t
 
     val to_string : t -> string
     (** Encode as a string (not to be used for user-facing messages). *)
@@ -679,7 +674,7 @@ module Projection : sig
   [@@ocaml.deprecated "Use QProjection.compare"]
 
   val map : (MutInd.t -> MutInd.t) -> t -> t
-  val map_npars : (MutInd.t -> int -> MutInd.t * int) -> t -> t
+  val map_npars : (int -> int) -> t -> t
 
   val to_string : t -> string
   (** Encode as a string (not to be used for user-facing messages). *)
@@ -691,7 +686,6 @@ end
 
 (** {6 Global reference is a kernel side type for all references together } *)
 
-(* XXX: Should we define GlobRefCan GlobRefUser? *)
 module GlobRef : sig
 
   type t =
@@ -701,6 +695,7 @@ module GlobRef : sig
     | ConstructRef of constructor  (** A reference to a constructor of an inductive type. *)
 
   val equal : t -> t -> bool
+  [@@ocaml.deprecated "Use QGlobRef.equal"]
 
   val is_bound : t -> bool
 
@@ -714,6 +709,9 @@ module GlobRef : sig
   module Map : Map.ExtS
     with type key = t and module Set := Set
 
+  val print : t -> Pp.t
+  (** Print internal representation (not to be used for user-facing messages). *)
+
 end
 
 (** Located identifiers and objects with syntax. *)
