@@ -15,6 +15,7 @@ open Context
 open Declarations
 open Util
 open Nativevalues
+open Genlambda
 open Nativelambda
 open Environ
 
@@ -53,6 +54,21 @@ let fresh_lname n =
   incr lname_ctr;
   { lname = n; luid = !lname_ctr }
 
+type prefix = string
+
+(* Linked code location utilities *)
+let get_mind_prefix env mind =
+   let _,name = lookup_mind_key mind env in
+   match !name with
+   | NotLinked -> ""
+   | Linked s -> s
+
+let get_const_prefix env c =
+   let _,(nameref,_) = lookup_constant_key c env in
+   match !nameref with
+   | NotLinked -> ""
+   | Linked s -> s
+
 (** Global names **)
 type gname =
   | Gind of string * inductive (* prefix, inductive name *)
@@ -156,7 +172,6 @@ let eq_symbol sy1 sy2 =
   | SymbConst kn1, SymbConst kn2 -> Constant.CanOrd.equal kn1 kn2
   | SymbMatch sw1, SymbMatch sw2 -> eq_annot_sw sw1 sw2
   | SymbInd ind1, SymbInd ind2 -> Ind.CanOrd.equal ind1 ind2
-  | SymbMeta m1, SymbMeta m2 -> Int.equal m1 m2
   | SymbEvar evk1, SymbEvar evk2 -> Evar.equal evk1 evk2
   | SymbLevel l1, SymbLevel l2 -> Univ.Level.equal l1 l2
   | SymbProj (i1, k1), SymbProj (i2, k2) -> Ind.CanOrd.equal i1 i2 && Int.equal k1 k2
@@ -170,10 +185,9 @@ let hash_symbol symb =
   | SymbConst c -> combinesmall 4 (Constant.CanOrd.hash c)
   | SymbMatch sw -> combinesmall 5 (hash_annot_sw sw)
   | SymbInd ind -> combinesmall 6 (Ind.CanOrd.hash ind)
-  | SymbMeta m -> combinesmall 7 m
-  | SymbEvar evk -> combinesmall 8 (Evar.hash evk)
-  | SymbLevel l -> combinesmall 9 (Univ.Level.hash l)
-  | SymbProj (i, k) -> combinesmall 10 (combine (Ind.CanOrd.hash i) k)
+  | SymbEvar evk -> combinesmall 7 (Evar.hash evk)
+  | SymbLevel l -> combinesmall 8 (Univ.Level.hash l)
+  | SymbProj (i, k) -> combinesmall 9 (combine (Ind.CanOrd.hash i) k)
 
 module HashedTypeSymbol = struct
   type t = symbol
@@ -217,11 +231,6 @@ let get_ind tbl i =
     | SymbInd ind -> ind
     | _ -> anomaly (Pp.str "get_ind failed.")
 
-let get_meta tbl i =
-  match tbl.(i) with
-    | SymbMeta m -> m
-    | _ -> anomaly (Pp.str "get_meta failed.")
-
 let get_evar tbl i =
   match tbl.(i) with
     | SymbEvar ev -> ev
@@ -273,7 +282,6 @@ type primitive =
   | Mk_int
   | Mk_bool
   | Val_to_int
-  | Mk_meta
   | Mk_evar
   | MLand
   | MLnot
@@ -292,6 +300,17 @@ type primitive =
   | MLarrayget
   | Mk_empty_instance
   | MLparray_of_array
+  | Get_value
+  | Get_sort
+  | Get_name
+  | Get_const
+  | Get_match
+  | Get_ind
+  | Get_evar
+  | Get_level
+  | Get_proj
+  | Get_symbols
+  | Lazy
   | Coq_primitive of CPrimitives.t * bool (* check for accu *)
 
 let eq_primitive p1 p2 =
@@ -308,7 +327,6 @@ let eq_primitive p1 p2 =
   | Cast_accu, Cast_accu -> true
   | Upd_cofix, Upd_cofix -> true
   | Force_cofix, Force_cofix -> true
-  | Mk_meta, Mk_meta -> true
   | Mk_evar, Mk_evar -> true
   | Mk_proj, Mk_proj -> true
   | MLarrayget, MLarrayget -> true
@@ -338,35 +356,45 @@ let primitive_hash = function
   | Mk_int -> 16
   | Mk_bool -> 17
   | Val_to_int -> 18
-  | Mk_meta -> 19
-  | Mk_evar -> 20
-  | MLand -> 21
-  | MLle -> 22
-  | MLlt -> 23
-  | MLinteq -> 24
-  | MLlsl -> 25
-  | MLlsr -> 26
-  | MLland -> 27
-  | MLlor -> 28
-  | MLlxor -> 29
-  | MLadd -> 30
-  | MLsub -> 31
-  | MLmul -> 32
-  | MLmagic -> 33
-  | Coq_primitive (prim, b) -> combinesmall 34 (combine (CPrimitives.hash prim) (Hashtbl.hash b))
-  | Mk_proj -> 35
-  | MLarrayget -> 36
-  | Mk_empty_instance -> 37
-  | Mk_float -> 38
-  | Is_float -> 39
-  | Is_parray -> 41
-  | MLnot -> 41
-  | MLparray_of_array -> 42
+  | Mk_evar -> 19
+  | MLand -> 20
+  | MLle -> 21
+  | MLlt -> 22
+  | MLinteq -> 23
+  | MLlsl -> 24
+  | MLlsr -> 25
+  | MLland -> 26
+  | MLlor -> 27
+  | MLlxor -> 28
+  | MLadd -> 29
+  | MLsub -> 30
+  | MLmul -> 31
+  | MLmagic -> 32
+  | Coq_primitive (prim, b) -> combinesmall 33 (combine (CPrimitives.hash prim) (Hashtbl.hash b))
+  | Mk_proj -> 34
+  | MLarrayget -> 35
+  | Mk_empty_instance -> 36
+  | Mk_float -> 37
+  | Is_float -> 38
+  | Is_parray -> 39
+  | MLnot -> 40
+  | MLparray_of_array -> 41
+  | Get_value -> 42
+  | Get_sort -> 43
+  | Get_name -> 44
+  | Get_const -> 45
+  | Get_match -> 46
+  | Get_ind -> 47
+  | Get_evar -> 48
+  | Get_level -> 49
+  | Get_proj -> 50
+  | Get_symbols -> 51
+  | Lazy -> 52
 
 type mllambda =
   | MLlocal        of lname
   | MLglobal       of gname
-  | MLprimitive    of primitive
+  | MLprimitive    of primitive * mllambda array
   | MLlam          of lname array * mllambda
   | MLletrec       of (lname * lname array * mllambda) array * mllambda
   | MLlet          of lname * mllambda * mllambda
@@ -409,7 +437,9 @@ let rec eq_mllambda gn1 gn2 n env1 env2 t1 t2 =
   | MLglobal gn1', MLglobal gn2' ->
       eq_gname gn1' gn2' || (eq_gname gn1 gn1' && eq_gname gn2 gn2')
       || (eq_gname gn1 gn2' && eq_gname gn2 gn1')
-  | MLprimitive prim1, MLprimitive prim2 -> eq_primitive prim1 prim2
+  | MLprimitive (prim1, args1), MLprimitive (prim2, args2) ->
+      eq_primitive prim1 prim2 &&
+      Array.equal (eq_mllambda gn1 gn2 n env1 env2) args1 args2
   | MLlam (lns1, ml1), MLlam (lns2, ml2) ->
       Int.equal (Array.length lns1) (Array.length lns2) &&
       let env1 = push_lnames n env1 lns1 in
@@ -503,7 +533,9 @@ let rec hash_mllambda gn n env t =
   match t with
   | MLlocal ln -> combinesmall 1 (LNmap.find ln env)
   | MLglobal gn' -> combinesmall 2 (if eq_gname gn gn' then 0 else gname_hash gn')
-  | MLprimitive prim -> combinesmall 3 (primitive_hash prim)
+  | MLprimitive (prim, args) ->
+    let h = primitive_hash prim in
+    combinesmall 3 (hash_mllambda_array gn n env h args)
   | MLlam (lns, ml) ->
       let env = push_lnames n env lns in
       combinesmall 4 (combine (Array.length lns) (hash_mllambda gn (n+1) env ml))
@@ -586,7 +618,10 @@ let fv_lam l =
     match l with
     | MLlocal l ->
         if LNset.mem l bind then fv else LNset.add l fv
-    | MLglobal _ | MLprimitive _  | MLint _ | MLuint _ | MLfloat _ -> fv
+    | MLglobal _ | MLint _ | MLuint _ | MLfloat _ -> fv
+    | MLprimitive (_, args) ->
+        let fv_arg arg fv = aux arg bind fv in
+        Array.fold_right fv_arg args fv
     | MLlam (ln,body) ->
         let bind = Array.fold_right LNset.add ln bind in
         aux body bind fv
@@ -649,7 +684,7 @@ let mkForceCofix prefix ind arg =
   MLlet (name, arg,
     MLif (
       MLisaccu (prefix, ind, MLlocal name),
-      MLapp (MLprimitive Force_cofix, [|MLlocal name|]),
+      MLprimitive (Force_cofix, [|MLlocal name|]),
       MLlocal name))
 
 let empty_params = [||]
@@ -773,14 +808,19 @@ type env =
       (* free variables *)
       env_urel : (int * mllambda) list ref; (* list of unbound rel *)
       env_named : (Id.t * mllambda) list ref;
-      env_univ : lname option}
+      env_univ : lname option;
+      env_const_prefix : Constant.t -> prefix;
+      env_mind_prefix : MutInd.t -> prefix;
+    }
 
-let empty_env univ =
+let empty_env univ get_const get_mind =
   { env_rel = [];
     env_bound = 0;
     env_urel = ref [];
     env_named = ref [];
-    env_univ = univ
+    env_univ = univ;
+    env_const_prefix = get_const;
+    env_mind_prefix = get_mind;
   }
 
 let push_rel env id =
@@ -892,43 +932,39 @@ let fv_args env fvn fvr =
     end
 
 let get_value_code i =
-  MLapp (MLglobal (Ginternal "get_value"),
+  MLprimitive (Get_value,
     [|MLglobal symbols_tbl_name; MLint i|])
 
 let get_sort_code i =
-  MLapp (MLglobal (Ginternal "get_sort"),
+  MLprimitive (Get_sort,
     [|MLglobal symbols_tbl_name; MLint i|])
 
 let get_name_code i =
-  MLapp (MLglobal (Ginternal "get_name"),
+  MLprimitive (Get_name,
     [|MLglobal symbols_tbl_name; MLint i|])
 
 let get_const_code i =
-  MLapp (MLglobal (Ginternal "get_const"),
+  MLprimitive (Get_const,
     [|MLglobal symbols_tbl_name; MLint i|])
 
 let get_match_code i =
-  MLapp (MLglobal (Ginternal "get_match"),
+  MLprimitive (Get_match,
     [|MLglobal symbols_tbl_name; MLint i|])
 
 let get_ind_code i =
-  MLapp (MLglobal (Ginternal "get_ind"),
-    [|MLglobal symbols_tbl_name; MLint i|])
-
-let get_meta_code i =
-  MLapp (MLglobal (Ginternal "get_meta"),
+  MLprimitive (Get_ind,
     [|MLglobal symbols_tbl_name; MLint i|])
 
 let get_evar_code i =
-  MLapp (MLglobal (Ginternal "get_evar"),
+  MLprimitive (Get_evar,
     [|MLglobal symbols_tbl_name; MLint i|])
 
 let get_level_code i =
-  MLapp (MLglobal (Ginternal "get_level"),
+  MLprimitive (Get_level,
     [|MLglobal symbols_tbl_name; MLint i|])
 
 let get_proj_code i =
-  MLapp (MLglobal (Ginternal "get_proj"),
+  MLprimitive (Get_proj,
     [|MLglobal symbols_tbl_name; MLint i|])
 
 type rlist =
@@ -970,7 +1006,7 @@ let merge_branches t =
   Array.iter (fun (pat,body) -> insert pat body newt) t;
   Array.of_list (to_list newt)
 
-let app_prim p args = MLapp(MLprimitive p, args)
+let app_prim p args = MLprimitive (p, args)
 
 type prim_aux =
   | PAprim of string * pconstant * CPrimitives.t * prim_aux array
@@ -979,9 +1015,9 @@ type prim_aux =
 let add_check cond targs args =
   let aux cond t a =
     match t, a with
-    | CPrimitives.(PITT_type (PT_int63, _)), PAml(MLapp(MLprimitive Mk_uint, _)) -> cond
-    | CPrimitives.(PITT_type (PT_array, _)), PAml(MLapp(MLprimitive MLparray_of_array, _)) -> cond
-    | CPrimitives.(PITT_type (PT_array, _)), PAml(MLapp (MLglobal (Ginternal "get_value"),_)) -> cond
+    | CPrimitives.(PITT_type (PT_int63, _)), PAml(MLprimitive (Mk_uint, _)) -> cond
+    | CPrimitives.(PITT_type (PT_array, _)), PAml(MLprimitive (MLparray_of_array, _)) -> cond
+    | CPrimitives.(PITT_type (PT_array, _)), PAml(MLprimitive (Get_value ,_)) -> cond
     | CPrimitives.(PITT_type (prim_ty, _)), PAml ml ->
        (* FIXME: use explicit equality function *)
        let c = (CPrimitives.PTE prim_ty, ml) in
@@ -990,7 +1026,7 @@ let add_check cond targs args =
   in
   Array.fold_left2 aux cond targs args
 
-let extract_prim ml_of l =
+let extract_prim env ml_of l =
   let decl = ref [] in
   let cond = ref [] in
   let type_args p =
@@ -998,12 +1034,13 @@ let extract_prim ml_of l =
     List.length params, Array.of_list args_ty in
   let rec aux l =
     match l with
-    | Lprim(prefix,kn,p,args) ->
+    | Lprim (kn, p, args) ->
+      let prefix = env.env_const_prefix (fst kn) in
       let nparams, targs = type_args p in
       let args = Array.map aux args in
       let checked_args = Array.init (Array.length args - nparams) (fun i -> args.(i+nparams)) in
       cond := add_check !cond targs checked_args;
-      PAprim(prefix,kn,p,args)
+      PAprim (prefix, kn, p, args)
     | Lrel _ | Lvar _ | Luint _ | Lval _ | Lconst _ -> PAml (ml_of l)
     | _ ->
       let x = fresh_lname Anonymous in
@@ -1015,20 +1052,22 @@ let extract_prim ml_of l =
 let cast_to_int v =
   match v with
   | MLint _ -> v
-  | _ -> MLapp(MLprimitive Val_to_int, [|v|])
+  | _ -> MLprimitive (Val_to_int, [|v|])
 
 let ml_of_instance instance u =
   let ml_of_level l =
     match Univ.Level.var_index l with
     | Some i ->
-       let univ = MLapp(MLprimitive MLmagic, [|MLlocal (Option.get instance)|]) in
-       mkMLapp (MLprimitive MLarrayget) [|univ; MLint i|]
+      (* FIXME: use a proper cast function *)
+       let univ = MLprimitive (MLmagic, [|MLlocal (Option.get instance)|]) in
+       MLprimitive (MLarrayget, [|univ; MLint i|])
     | None -> let i = push_symbol (SymbLevel l) in get_level_code i
   in
   let u = Univ.Instance.to_array u in
   if Array.is_empty u then [||]
   else let u = Array.map ml_of_level u in
-       [|MLapp (MLprimitive MLmagic, [|MLarray u|])|]
+      (* FIXME: use a proper cast function *)
+       [|MLprimitive (MLmagic, [|MLarray u|])|]
 
 let compile_prim env decl cond paux =
 
@@ -1045,7 +1084,7 @@ let compile_prim env decl cond paux =
     | PAprim(prefix, (kn,u), op, args) ->
       let uarg = ml_of_instance env.env_univ u in
       let prim_const = mkMLapp (MLglobal (Gconstant(prefix,kn))) uarg in
-      let prim = mkMLapp (MLprimitive(Coq_primitive(op, true))) [|prim_const|] in
+      let prim = MLprimitive ((Coq_primitive(op, true)), [|prim_const|]) in
       mkMLapp prim (Array.map naive_prim_aux args)
     | PAml ml -> ml
   in
@@ -1101,29 +1140,20 @@ let compile_prim env decl cond paux =
   match t with
   | Lrel(id ,i) -> get_rel env id i
   | Lvar id -> get_var env id
-  | Lmeta(mv,_ty) ->
-     let tyn = fresh_lname Anonymous in
-     let i = push_symbol (SymbMeta mv) in
-     MLapp(MLprimitive Mk_meta, [|get_meta_code i; MLlocal tyn|])
   | Levar(evk, args) ->
      let i = push_symbol (SymbEvar evk) in
      (** Arguments are *not* reversed in evar instances in native compilation *)
      let args = MLarray(Array.map (ml_of_lam env l) args) in
-     MLapp(MLprimitive Mk_evar, [|get_evar_code i; args|])
+     MLprimitive (Mk_evar, [|get_evar_code i; args|])
   | Lprod(dom,codom) ->
       let dom = ml_of_lam env l dom in
       let codom = ml_of_lam env l codom in
       let n = get_prod_name codom in
       let i = push_symbol (SymbName n) in
-      MLapp(MLprimitive Mk_prod, [|get_name_code i;dom;codom|])
+      MLprimitive (Mk_prod, [|get_name_code i;dom;codom|])
   | Llam(ids,body) ->
     let lnames,env = push_rels env ids in
     MLlam(lnames, ml_of_lam env l body)
-  | Lrec(id,body) ->
-      let ids,body = decompose_Llam body in
-      let lname, env = push_rel env id in
-      let lnames, env = push_rels env ids in
-      MLletrec([|lname, lnames, ml_of_lam env l body|], MLlocal lname)
   | Llet(id,def,body) ->
       let def = ml_of_lam env l def in
       let lname, env = push_rel env id in
@@ -1131,12 +1161,17 @@ let compile_prim env decl cond paux =
       MLlet(lname,def,body)
   | Lapp(f,args) ->
       MLapp(ml_of_lam env l f, Array.map (ml_of_lam env l) args)
-  | Lconst (prefix, (c, u)) ->
+  | Lconst (c, u) ->
+     let prefix = env.env_const_prefix c in
      let args = ml_of_instance env.env_univ u in
      mkMLapp (MLglobal(Gconstant (prefix, c))) args
-  | Lproj (prefix, ind, i) -> MLglobal(Gproj (prefix, ind, i))
+  | Lproj (p, c) ->
+    let ind = Projection.Repr.inductive p in
+    let i = Projection.Repr.arg p in
+    let prefix = env.env_mind_prefix (fst ind) in
+    MLapp (MLglobal(Gproj (prefix, ind, i)), [| ml_of_lam env l c |])
   | Lprim _ ->
-    let decl,cond,paux = extract_prim (ml_of_lam env l) t in
+    let decl,cond,paux = extract_prim env (ml_of_lam env l) t in
     compile_prim env decl cond paux
   | Lcase (annot,p,a,bs) ->
       (* let predicate_uid fv_pred = compilation of p
@@ -1149,14 +1184,22 @@ let compile_prim env decl cond paux =
          (* Remark: if we do not want to compile the predicate we
             should a least compute the fv, then store the lambda representation
             of the predicate (not the mllambda) *)
-      let env_p = empty_env env.env_univ in
+      let annot =
+        let (ci, tbl, knd) = annot in {
+          asw_ind = ci.ci_ind;
+          asw_ci = ci;
+          asw_reloc = tbl;
+          asw_finite = knd <> CoFinite;
+          asw_prefix = env.env_mind_prefix (fst ci.ci_ind);
+      } in
+      let env_p = empty_env env.env_univ env.env_const_prefix env.env_mind_prefix in
       let pn = fresh_gpred l in
       let mlp = ml_of_lam env_p l p in
       let mlp = generalize_fv env_p mlp in
       let (pfvn,pfvr) = !(env_p.env_named), !(env_p.env_urel) in
       let pn = push_global_let pn mlp in
       (* Compilation of the case *)
-      let env_c = empty_env env.env_univ in
+      let env_c = empty_env env.env_univ env.env_const_prefix env.env_mind_prefix in
       let a_uid = fresh_lname Anonymous in
       let la_uid = MLlocal a_uid in
       (* compilation of branches *)
@@ -1178,8 +1221,8 @@ let compile_prim env decl cond paux =
          (* remark : the call to fv_args does not add free variables in env_c *)
       let i = push_symbol (SymbMatch annot) in
       let accu =
-        MLapp(MLprimitive Mk_sw,
-              [| get_match_code i; MLapp (MLprimitive Cast_accu, [|la_uid|]);
+        MLprimitive (Mk_sw,
+              [| get_match_code i; MLprimitive (Cast_accu, [|la_uid|]);
                  pred;
                  cn_fv |]) in
 (*      let body = MLlam([|a_uid|], MLmatch(annot, la_uid, accu, bs)) in
@@ -1193,8 +1236,6 @@ let compile_prim env decl cond paux =
         if annot.asw_finite then arg
         else mkForceCofix annot.asw_prefix annot.asw_ind arg in
       mkMLapp (MLapp (MLglobal cn, fv_args env fvn fvr)) [|force|]
-  | Lif(t,bt,bf) ->
-      MLif(ml_of_lam env l t, ml_of_lam env l bt, ml_of_lam env l bf)
   | Lfix ((rec_pos, inds, start), (ids, tt, tb)) ->
       (* let type_f fvt = [| type fix |]
          let norm_f1 fv f1 .. fn params1 = body1
@@ -1212,7 +1253,7 @@ let compile_prim env decl cond paux =
            start
       *)
       (* Compilation of type *)
-      let env_t = empty_env env.env_univ in
+      let env_t = empty_env env.env_univ env.env_const_prefix env.env_mind_prefix in
       let ml_t = Array.map (ml_of_lam env_t l) tt in
       let params_t = fv_params env_t in
       let args_t = fv_args env !(env_t.env_named) !(env_t.env_urel) in
@@ -1221,7 +1262,7 @@ let compile_prim env decl cond paux =
       let mk_type = MLapp(MLglobal gft, args_t) in
       (* Compilation of norm_i *)
       let ndef = Array.length ids in
-      let lf,env_n = push_rels (empty_env env.env_univ) ids in
+      let lf,env_n = push_rels (empty_env env.env_univ env.env_const_prefix env.env_mind_prefix) ids in
       let t_params = Array.make ndef [||] in
       let t_norm_f = Array.make ndef (Gnorm (l,-1)) in
       let mk_let _envi (id,def) t = MLlet (id,def,t) in
@@ -1264,11 +1305,12 @@ let compile_prim env decl cond paux =
         let paramsi = t_params.(i) in
         let reci = MLlocal (paramsi.(rec_pos.(i))) in
         let pargsi = Array.map (fun id -> MLlocal id) paramsi in
-        let (prefix, ind) = inds.(i) in
+        let ind = inds.(i) in
+        let prefix = env.env_mind_prefix (fst ind) in
         let body =
           MLif(MLisaccu (prefix, ind, reci),
                mkMLapp
-                 (MLapp(MLprimitive (Mk_fix(rec_pos,i)),
+                 (MLprimitive ((Mk_fix(rec_pos,i)),
                         [|mk_type; mk_norm|]))
                  pargsi,
                MLapp(MLglobal t_norm_f.(i),
@@ -1278,7 +1320,7 @@ let compile_prim env decl cond paux =
       MLletrec(Array.mapi mkrec lf, lf_args.(start))
   | Lcofix (start, (ids, tt, tb)) ->
       (* Compilation of type *)
-      let env_t = empty_env env.env_univ in
+      let env_t = empty_env env.env_univ env.env_const_prefix env.env_mind_prefix in
       let ml_t = Array.map (ml_of_lam env_t l) tt in
       let params_t = fv_params env_t in
       let args_t = fv_args env !(env_t.env_named) !(env_t.env_urel) in
@@ -1287,7 +1329,7 @@ let compile_prim env decl cond paux =
       let mk_type = MLapp(MLglobal gft, args_t) in
       (* Compilation of norm_i *)
       let ndef = Array.length ids in
-      let lf,env_n = push_rels (empty_env env.env_univ) ids in
+      let lf,env_n = push_rels (empty_env env.env_univ env.env_const_prefix env.env_mind_prefix) ids in
       let t_params = Array.make ndef [||] in
       let t_norm_f = Array.make ndef (Gnorm (l,-1)) in
       let ml_of_fix i body =
@@ -1322,12 +1364,12 @@ let compile_prim env decl cond paux =
           MLlam(Array.append paramsi [|uniti|],
                 MLapp(MLglobal t_norm_f.(i),
                       Array.concat [fv_args;lf_args;pargsi])) in
-        MLsequence(MLapp(MLprimitive Upd_cofix, [|lf_args.(i);body|]),
+        MLsequence(MLprimitive (Upd_cofix, [|lf_args.(i);body|]),
                    cont) in
       let upd = Array.fold_right_i upd lf lf_args.(start) in
       let mk_let i lname cont =
         MLlet(lname,
-              MLapp(MLprimitive(Mk_cofix i),[| MLlocal ltype; MLlocal lnorm|]),
+              MLprimitive ((Mk_cofix i),[| MLlocal ltype; MLlocal lnorm|]),
               cont) in
       let init = Array.fold_right_i mk_let lf upd in
       MLlet(lnorm, mk_norm, MLlet(ltype, mk_type, init))
@@ -1345,16 +1387,17 @@ let compile_prim env decl cond paux =
         (lname, paramsi, body) in
       MLletrec(Array.mapi mkrec lf, lf_args.(start)) *)
 
-  | Lint tag -> MLapp(MLprimitive Mk_int, [|MLint tag|])
+  | Lint tag -> MLprimitive (Mk_int, [|MLint tag|])
 
-  | Lmakeblock (prefix,cn,tag,args) ->
+  | Lmakeblock (cn,tag,args) ->
+     let prefix = env.env_mind_prefix (fst cn) in
      let args = Array.map (ml_of_lam env l) args in
      MLconstruct(prefix,cn,tag,args)
-  | Luint i -> MLapp(MLprimitive Mk_uint, [|MLuint i|])
-  | Lfloat f -> MLapp(MLprimitive Mk_float, [|MLfloat f|])
+  | Luint i -> MLprimitive (Mk_uint, [|MLuint i|])
+  | Lfloat f -> MLprimitive (Mk_float, [|MLfloat f|])
   | Lparray (t,def) ->
     let def = ml_of_lam env l def in
-    MLapp(MLprimitive MLparray_of_array, [| MLarray (Array.map (ml_of_lam env l) t); def |])
+    MLprimitive (MLparray_of_array, [| MLarray (Array.map (ml_of_lam env l) t); def |])
   | Lval v ->
       let i = push_symbol (SymbValue v) in get_value_code i
   | Lsort s ->
@@ -1363,16 +1406,17 @@ let compile_prim env decl cond paux =
       | None -> MLarray [||]
       | Some u -> MLlocal u
     in
-    let uarg = MLapp(MLprimitive MLmagic, [|uarg|]) in
-    MLapp(MLprimitive Mk_sort, [|get_sort_code i; uarg|])
-  | Lind (prefix, (ind, u)) ->
+    (* FIXME: use a dedicated cast function *)
+    let uarg = MLprimitive (MLmagic, [|uarg|]) in
+    MLprimitive (Mk_sort, [|get_sort_code i; uarg|])
+  | Lind (ind, u) ->
+     let prefix = env.env_mind_prefix (fst ind) in
      let uargs = ml_of_instance env.env_univ u in
      mkMLapp (MLglobal (Gind (prefix, ind))) uargs
-  | Llazy -> MLglobal (Ginternal "lazy")
   | Lforce -> MLglobal (Ginternal "Lazy.force")
 
-let mllambda_of_lambda univ auxdefs l t =
-  let env = empty_env univ in
+let mllambda_of_lambda univ constpref mindpref auxdefs l t =
+  let env = empty_env univ constpref mindpref in
   global_stack := auxdefs;
   let ml = ml_of_lam env l t in
   let fv_rel = !(env.env_urel) in
@@ -1405,7 +1449,8 @@ let subst s l =
     let rec aux l =
       match l with
       | MLlocal id -> (try LNmap.find id s with Not_found -> l)
-      | MLglobal _ | MLprimitive _ | MLint _ | MLuint _ | MLfloat _ -> l
+      | MLglobal _ | MLint _ | MLuint _ | MLfloat _ -> l
+      | MLprimitive (p, args) -> MLprimitive (p, Array.map aux args)
       | MLlam(params,body) -> MLlam(params, aux body)
       | MLletrec(defs,body) ->
         let arec (f,params,body) = (f,params,aux body) in
@@ -1475,7 +1520,9 @@ let optimize gdef l =
   let rec optimize s l =
     match l with
     | MLlocal id -> (try LNmap.find id s with Not_found -> l)
-    | MLglobal _ | MLprimitive _ | MLint _ | MLuint _ | MLfloat _ -> l
+    | MLglobal _ | MLint _ | MLuint _ | MLfloat _ -> l
+    | MLprimitive (p, args) ->
+        MLprimitive (p, Array.map (optimize s) args)
     | MLlam(params,body) ->
         MLlam(params, optimize s body)
     | MLletrec(decls,body) ->
@@ -1653,12 +1700,13 @@ let pp_mllam fmt l =
     match l with
     | MLlocal ln -> Format.fprintf fmt "@[%a@]" pp_lname ln
     | MLglobal g -> Format.fprintf fmt "@[%a@]" pp_gname g
-    | MLprimitive p -> Format.fprintf fmt "@[%a@]" pp_primitive p
+    | MLprimitive (p, args) ->
+      Format.fprintf fmt "@[%a@ %a@]" pp_primitive p (pp_args true) args
     | MLlam(ids,body) ->
         Format.fprintf fmt "@[(fun%a@ ->@\n %a)@]"
           pp_ldecls ids pp_mllam body
     | MLletrec(defs, body) ->
-        Format.fprintf fmt "@[%a@ in@\n%a@]" pp_letrec defs
+        Format.fprintf fmt "@[(%a@ in@\n%a)@]" pp_letrec defs
           pp_mllam body
     | MLlet(id,def,body) ->
         Format.fprintf fmt "@[(let@ %a@ =@\n %a@ in@\n%a)@]"
@@ -1724,8 +1772,7 @@ let pp_mllam fmt l =
 
   and pp_blam fmt l =
     match l with
-    | MLprimitive (Mk_prod | Mk_sort) (* FIXME: why this special case? *)
-    | MLlam _ | MLletrec _ | MLlet _ | MLapp _ | MLif _ ->
+    | MLprimitive (_, _) | MLlam _ | MLletrec _ | MLlet _ | MLapp _ | MLif _ ->
         Format.fprintf fmt "(%a)" pp_mllam l
     | MLconstruct(_,_,_,args) when Array.length args > 0 ->
         Format.fprintf fmt "(%a)" pp_mllam l
@@ -1786,7 +1833,7 @@ let pp_mllam fmt l =
     Array.iter pp_branch bs
 
   and pp_primitive fmt = function
-    | Mk_prod -> Format.fprintf fmt "mk_prod_accu"
+    | Mk_prod -> Format.fprintf fmt "mk_prod"
     | Mk_sort -> Format.fprintf fmt "mk_sort_accu"
     | Mk_ind -> Format.fprintf fmt "mk_ind_accu"
     | Mk_const -> Format.fprintf fmt "mk_constant_accu"
@@ -1815,7 +1862,6 @@ let pp_mllam fmt l =
     | Mk_int -> Format.fprintf fmt "mk_int"
     | Mk_bool -> Format.fprintf fmt "mk_bool"
     | Val_to_int -> Format.fprintf fmt "val_to_int"
-    | Mk_meta -> Format.fprintf fmt "mk_meta_accu"
     | Mk_evar -> Format.fprintf fmt "mk_evar_accu"
     | MLand -> Format.fprintf fmt "(&&)"
     | MLnot -> Format.fprintf fmt "not"
@@ -1837,6 +1883,17 @@ let pp_mllam fmt l =
     | Coq_primitive (op, false) ->
        Format.fprintf fmt "no_check_%s" (CPrimitives.to_string op)
     | Coq_primitive (op, true) -> Format.fprintf fmt "%s" (CPrimitives.to_string op)
+    | Get_value -> Format.fprintf fmt "get_value"
+    | Get_sort -> Format.fprintf fmt "get_sort"
+    | Get_name -> Format.fprintf fmt "get_name"
+    | Get_const -> Format.fprintf fmt "get_const"
+    | Get_match -> Format.fprintf fmt "get_match"
+    | Get_ind -> Format.fprintf fmt "get_ind"
+    | Get_evar -> Format.fprintf fmt "get_evar"
+    | Get_level -> Format.fprintf fmt "get_level"
+    | Get_proj -> Format.fprintf fmt "get_proj"
+    | Get_symbols -> Format.fprintf fmt "get_symbols"
+    | Lazy -> Format.fprintf fmt "lazy"
   in
   Format.fprintf fmt "@[%a@]" pp_mllam l
 
@@ -1850,12 +1907,20 @@ let pp_array fmt t =
     Format.fprintf fmt "%a" pp_mllam t.(len - 1);
   Format.fprintf fmt "|]@]"
 
+let type_of_global gn c = match gn with
+  | Ginternal "symbols_tbl" -> ""
+  | _ -> match c with
+    | MLprimitive (Lazy, _) -> " : Nativevalues.t Lazy.t"
+    | MLlam ([|_|], MLprimitive (Lazy, _)) -> " : Nativevalues.t -> Nativevalues.t Lazy.t"
+    | MLprimitive ((Mk_ind | Mk_const), [|_|]) -> " : Univ.Level.t array -> Nativevalues.t"
+    | MLsetref (_,_) -> " : unit"
+    | _ -> " : Nativevalues.t"
+
 let pp_global fmt g =
   match g with
   | Glet (gn, c) ->
-      let ids, c = decompose_MLlam c in
-      Format.fprintf fmt "@[let %a%a =@\n  %a@]@\n@." pp_gname gn
-        pp_ldecls ids
+      Format.fprintf fmt "@[let %a%s = let Refl = Nativevalues.t_eq in@\n  %a@]@\n@." pp_gname gn
+        (type_of_global gn c)
         pp_mllam c
   | Gopen s ->
       Format.fprintf fmt "@[open %s@]@." s
@@ -1878,13 +1943,13 @@ let pp_global fmt g =
     in
     Format.fprintf fmt "@[type ind_%s =@\n%a@]@\n@." (string_of_ind ind) pp_const_sigs lar
   | Gtblfixtype (g, params, t) ->
-      Format.fprintf fmt "@[let %a %a =@\n  %a@]@\n@." pp_gname g
+      Format.fprintf fmt "@[let %a %a : Nativevalues.t array = let Refl = Nativevalues.t_eq in@\n  %a@]@\n@." pp_gname g
         pp_ldecls params pp_array t
   | Gtblnorm (g, params, t) ->
-      Format.fprintf fmt "@[let %a %a =@\n  %a@]@\n@." pp_gname g
+      Format.fprintf fmt "@[let %a %a : Nativevalues.t array = let Refl = Nativevalues.t_eq in@\n  %a@]@\n@." pp_gname g
         pp_ldecls params pp_array t
   | Gletcase(gn,params,annot,a,accu,bs) ->
-      Format.fprintf fmt "@[(* Hash = %i *)@\nlet rec %a %a =@\n  %a@]@\n@."
+      Format.fprintf fmt "@[(* Hash = %i *)@\nlet rec %a %a : Nativevalues.t = let Refl = Nativevalues.t_eq in@\n  %a@]@\n@."
       (hash_global g)
         pp_gname gn pp_ldecls params
         pp_mllam (MLmatch(annot,a,accu,bs))
@@ -1892,8 +1957,11 @@ let pp_global fmt g =
       Format.fprintf fmt "@[(* %s *)@]@." s
 
 (** Compilation of elements in environment **)
-let rec compile_with_fv env sigma univ auxdefs l t =
-  let (auxdefs,(fv_named,fv_rel),ml) = mllambda_of_lambda univ auxdefs l t in
+let rec compile_with_fv ?(wrap = fun t -> t) env sigma univ auxdefs l t =
+  let const_prefix c = get_const_prefix env c in
+  let mind_prefix c = get_mind_prefix env c in
+  let (auxdefs,(fv_named,fv_rel),ml) = mllambda_of_lambda univ const_prefix mind_prefix auxdefs l t in
+  let ml = wrap ml in
   if List.is_empty fv_named && List.is_empty fv_rel then (auxdefs,ml)
   else apply_fv env sigma univ (fv_named,fv_rel) auxdefs ml
 
@@ -1932,7 +2000,7 @@ and compile_rel env sigma univ auxdefs n =
       let auxdefs,code = compile_with_fv env sigma univ auxdefs None code in
       Glet(Grel n, code)::auxdefs
   | LocalAssum _ ->
-      Glet(Grel n, MLprimitive (Mk_rel n))::auxdefs
+      Glet(Grel n, MLprimitive (Mk_rel n, [||]))::auxdefs
 
 and compile_named env sigma univ auxdefs id =
   let open Context.Named.Declaration in
@@ -1942,7 +2010,7 @@ and compile_named env sigma univ auxdefs id =
       let auxdefs,code = compile_with_fv env sigma univ auxdefs None code in
       Glet(Gnamed id, code)::auxdefs
   | LocalAssum _ ->
-      Glet(Gnamed id, MLprimitive (Mk_var id))::auxdefs
+      Glet(Gnamed id, MLprimitive (Mk_var id, [||]))::auxdefs
 
 let compile_constant env sigma con cb =
     let no_univs = 0 = Univ.AbstractContext.size (Declareops.constant_polymorphic_context cb) in
@@ -1951,13 +2019,14 @@ let compile_constant env sigma con cb =
       let code = lambda_of_constr env sigma t in
       debug_native_compiler (fun () -> Pp.str "Generated lambda code");
       let is_lazy = is_lazy t in
-      let code = if is_lazy then mk_lazy code else code in
+      let wrap t = if is_lazy then MLprimitive (Lazy, [|t|]) else t in
       let l = Constant.label con in
       let auxdefs,code =
-        if no_univs then compile_with_fv env sigma None [] (Some l) code
+        if no_univs then
+          compile_with_fv ~wrap env sigma None [] (Some l) code
         else
           let univ = fresh_univ () in
-          let (auxdefs,code) = compile_with_fv env sigma (Some univ) [] (Some l) code in
+          let (auxdefs,code) = compile_with_fv ~wrap env sigma (Some univ) [] (Some l) code in
           (auxdefs,mkMLlam [|univ|] code)
       in
       debug_native_compiler (fun () -> Pp.str "Generated mllambda code");
@@ -1975,7 +2044,7 @@ let compile_constant env sigma con cb =
         (*
         let t = mkMLlam [|univ|] (mkMLapp (MLprimitive Mk_const)
          *)
-        [Glet(Gconstant ("", con), mkMLapp (MLprimitive Mk_const) args)]
+        [Glet(Gconstant ("", con), MLprimitive (Mk_const, args))]
     end
 
 module StringOrd = struct type t = string let compare = String.compare end
@@ -2009,7 +2078,8 @@ let compile_mind mb mind stack =
           [|get_ind_code j; MLarray [||]|]
         else [|get_ind_code j|]
       in
-      Glet(name, MLapp (MLprimitive Mk_ind, args))
+      (* FIXME: pass universes here *)
+      Glet(name, MLprimitive (Mk_ind, args))
     in
     let nparams = mb.mind_nparams in
     let add_proj proj_arg acc _pb =
@@ -2030,8 +2100,8 @@ let compile_mind mb mind stack =
         (fun i -> if Int.equal i proj_arg then Some ci_uid else None)
       in
       let i = push_symbol (SymbProj (ind, proj_arg)) in
-      let accu = MLapp (MLprimitive Cast_accu, [|MLlocal cf_uid|]) in
-      let accu_br = MLapp (MLprimitive Mk_proj, [|get_proj_code i;accu|]) in
+      let accu = MLprimitive (Cast_accu, [|MLlocal cf_uid|]) in
+      let accu_br = MLprimitive (Mk_proj, [|get_proj_code i;accu|]) in
       let code = MLmatch(asw,MLlocal cf_uid,accu_br,[|[NonConstPattern (tag,cargs)],MLlocal ci_uid|]) in
       let code = MLlet(cf_uid, mkForceCofix "" ind (MLlocal c_uid), code) in
       let gn = Gproj ("", ind, proj_arg) in
@@ -2082,8 +2152,8 @@ let compile_deps env sigma prefix init t =
   let rec aux env lvl init t =
   match kind t with
   | Ind ((mind,_),_u) -> compile_mind_deps env prefix init mind
-  | Const c ->
-    let c,_u = get_alias env c in
+  | Const (c, _u) ->
+    let c = get_alias env c in
     let cb,(nameref,_) = lookup_constant_key c env in
     let (_, (_, const_updates)) = init in
     if is_code_loaded nameref
@@ -2169,7 +2239,7 @@ let mk_conv_code env sigma prefix t1 t2 =
   let setref2 = Glet(Ginternal "_", MLsetref("rt2",g2)) in
   let gl = List.rev (setref2 :: setref1 :: t2 :: t1 :: gl) in
   let header = Glet(Ginternal "symbols_tbl",
-    MLapp (MLglobal (Ginternal "get_symbols"),
+    MLprimitive (Get_symbols,
       [|MLglobal (Ginternal "()")|])) in
   header::gl, (mind_updates, const_updates)
 
@@ -2187,7 +2257,7 @@ let mk_norm_code env sigma prefix t =
   let setref = Glet(Ginternal "_", MLsetref("rt1",g1)) in
   let gl = List.rev (setref :: t1 :: gl) in
   let header = Glet(Ginternal "symbols_tbl",
-    MLapp (MLglobal (Ginternal "get_symbols"),
+    MLprimitive (Get_symbols,
       [|MLglobal (Ginternal "()")|])) in
   header::gl, (mind_updates, const_updates)
 
