@@ -14,7 +14,17 @@ open Names
 the native compiler. Be careful when removing apparently dead code from this
 interface, as it may be used by programs generated at runtime. *)
 
-type t = t -> t
+type t
+
+val apply : t -> t -> t
+val of_fun : (t -> t) -> t
+
+val eta_expand : t -> t
+
+type ('a,'b) eq = ('a,'b) Util.eq = Refl : ('a,'a) eq
+val t_eq : (t, t -> t) eq
+(** When -rectypes, matching on this makes [t = ('a -> 'a) as 'a].
+    When not -rectypes, it does nothing AFAICT so you have to generalize your problem to use this. *)
 
 type accumulator
 
@@ -41,18 +51,17 @@ type rec_pos = int array
 
 val eq_rec_pos : rec_pos -> rec_pos -> bool
 
+type vcofix
+
 type atom =
   | Arel of int
   | Aconstant of pconstant
   | Aind of pinductive
   | Asort of Sorts.t
   | Avar of Id.t
-  | Acase of annot_sw * accumulator * t * (t -> t)
+  | Acase of annot_sw * accumulator * t * t
   | Afix of t array * t array * rec_pos * int
-  | Acofix of t array * t array * int * t
-  | Acofixe of t array * t array * int * t
-  | Aprod of Name.t * t * (t -> t)
-  | Ameta of metavariable * t
+  | Acofix of t array * t array * int * vcofix
   | Aevar of Evar.t * t array (* arguments *)
   | Aproj of (inductive * int) * accumulator
 
@@ -63,7 +72,6 @@ type symbol =
   | SymbConst of Constant.t
   | SymbMatch of annot_sw
   | SymbInd of inductive
-  | SymbMeta of metavariable
   | SymbEvar of Evar.t
   | SymbLevel of Univ.Level.t
   | SymbProj of (inductive * int)
@@ -82,16 +90,15 @@ val mk_ind_accu : inductive -> Univ.Level.t array -> t
 val mk_sort_accu : Sorts.t -> Univ.Level.t array -> t
 val mk_var_accu : Id.t -> t
 val mk_sw_accu : annot_sw -> accumulator -> t -> (t -> t)
-val mk_prod_accu : Name.t -> t -> t -> t
 val mk_fix_accu : rec_pos  -> int -> t array -> t array -> t
 val mk_cofix_accu : int -> t array -> t array -> t
-val mk_meta_accu : metavariable -> t
 val mk_evar_accu : Evar.t -> t array -> t
 val mk_proj_accu : (inductive * int) -> accumulator -> t
 val upd_cofix : t -> t -> unit
 val force_cofix : t -> t
 val mk_const : tag -> t
 val mk_block : tag -> t array -> t
+val mk_prod : Name.t -> t -> t -> t
 
 val mk_bool : bool -> t
 [@@ocaml.inline always]
@@ -110,7 +117,7 @@ val napply : t -> t array -> t
 
 val dummy_value : unit -> t
 val atom_of_accu : accumulator -> atom
-val args_of_accu : accumulator -> t array
+val args_of_accu : accumulator -> t list
 val accu_nargs : accumulator -> int
 
 val cast_accu : t -> accumulator
@@ -124,20 +131,11 @@ val block_size : block -> int
 val block_field : block -> int -> t
 val block_tag : block -> int
 
-
-
 (* kind_of_value *)
 
-type kind_of_value =
-  | Vaccu of accumulator
-  | Vfun of (t -> t)
-  | Vconst of int
-  | Vint64 of int64
-  | Vfloat64 of float
-  | Varray of t Parray.t
-  | Vblock of block
+type kind = (t, accumulator, t -> t, Name.t * t * t, Util.Empty.t, Util.Empty.t, block) Values.kind
 
-val kind_of_value : t -> kind_of_value
+val kind_of_value : t -> kind
 
 val str_encode : 'a -> string
 val str_decode : string -> 'a
@@ -368,17 +366,15 @@ val is_parray : t -> bool
 
 val arraymake : t -> t -> t -> t -> t (* accu A n def *)
 val arrayget : t -> t -> t -> t -> t (* accu A t n *)
-val arraydefault : t -> t -> t (* accu A t *)
+val arraydefault : t -> t -> t -> t (* accu A t *)
 val arrayset : t -> t -> t -> t -> t -> t (* accu A t n v *)
 val arraycopy : t -> t -> t -> t (* accu A t *)
 val arraylength : t -> t -> t -> t (* accu A t *)
-val arrayinit : t -> t -> t -> t (* accu A n f def *)
-val arraymap : t -> t -> t (* accu A B f t *)
 
 val no_check_arraymake : t -> t -> t
 [@@ocaml.inline always]
 
-val no_check_arrayget : t -> t -> t -> t
+val no_check_arrayget : t -> t -> t
 [@@ocaml.inline always]
 
 val no_check_arraydefault : t -> t
