@@ -28,8 +28,7 @@ type reloc_info =
   | Reloc_annot of annot_switch
   | Reloc_const of structured_constant
   | Reloc_getglobal of Names.Constant.t
-  | Reloc_proj_name of Projection.Repr.t
-  | Reloc_caml_prim of CPrimitives.t
+  | Reloc_caml_prim of caml_prim
 
 let eq_reloc_info r1 r2 = match r1, r2 with
 | Reloc_annot sw1, Reloc_annot sw2 -> eq_annot_switch sw1 sw2
@@ -38,9 +37,7 @@ let eq_reloc_info r1 r2 = match r1, r2 with
 | Reloc_const _, _ -> false
 | Reloc_getglobal c1, Reloc_getglobal c2 -> Constant.CanOrd.equal c1 c2
 | Reloc_getglobal _, _ -> false
-| Reloc_proj_name p1, Reloc_proj_name p2 -> Projection.Repr.CanOrd.equal p1 p2
-| Reloc_proj_name _, _ -> false
-| Reloc_caml_prim p1, Reloc_caml_prim p2 -> CPrimitives.equal p1 p2
+| Reloc_caml_prim p1, Reloc_caml_prim p2 -> CPrimitives.equal (caml_prim_to_prim p1) (caml_prim_to_prim p2)
 | Reloc_caml_prim _, _ -> false
 
 let hash_reloc_info r =
@@ -49,8 +46,7 @@ let hash_reloc_info r =
   | Reloc_annot sw -> combinesmall 1 (hash_annot_switch sw)
   | Reloc_const c -> combinesmall 2 (hash_structured_constant c)
   | Reloc_getglobal c -> combinesmall 3 (Constant.CanOrd.hash c)
-  | Reloc_proj_name p -> combinesmall 4 (Projection.Repr.CanOrd.hash p)
-  | Reloc_caml_prim p -> combinesmall 5 (CPrimitives.hash p)
+  | Reloc_caml_prim p -> combinesmall 4 (CPrimitives.hash (caml_prim_to_prim p))
 
 module RelocTable = Hashtbl.Make(struct
   type t = reloc_info
@@ -58,39 +54,88 @@ module RelocTable = Hashtbl.Make(struct
   let hash = hash_reloc_info
 end)
 
-(** We use arrays for on-disk representation. On 32-bit machines, this means we
-    can only have a maximum amount of about 4.10^6 relocations, which seems
-    quite a lot, but potentially reachable if e.g. compiling big proofs. This
-    would prevent VM computing with these terms on 32-bit architectures. Maybe
-    we should use a more robust data structure? *)
+module Positions :
+sig
+  type t
+  val of_list : int list -> t
+  val iter : (int -> unit) -> t -> unit
+end =
+struct
+
+type t = string
+(* Represent an ordered set of 32-bit integers as an array of successive diffs.
+   We use furthermore an approach where smaller integers use less bytes. Numbers
+   smaller than 255 are stored into one byte. Otherwise we use the byte 0x00 to
+   signal that we store the integer in the next 4 bytes. This is a cheap way to
+   compact this data. *)
+
+let output buf n =
+  if n <= 0xFF then Buffer.add_uint8 buf n
+  else
+    let () = Buffer.add_uint8 buf 0x00 in
+    Buffer.add_int32_be buf (Int32.of_int n)
+
+let input s pos =
+  let c = Char.code s.[!pos] in
+  if Int.equal c 0x00 then
+    (* TODO: use String.get_int32_be when available in OCaml 4.13.0 *)
+    let c1 = Char.code s.[!pos + 1] in
+    let c2 = Char.code s.[!pos + 2] in
+    let c3 = Char.code s.[!pos + 3] in
+    let c4 = Char.code s.[!pos + 4] in
+    let () = pos := !pos + 5 in
+    (c1 lsl 24) lor (c2 lsl 16) lor (c3 lsl 8) lor c4
+  else
+    let () = pos := !pos + 1 in
+    c
+
+let of_list l = match l with
+| [] -> ""
+| n :: l ->
+  let buf = Buffer.create 16 in
+  let () = assert (0 < n) in
+  let () = output buf n in
+  let rec aux cur l = match l with
+  | [] -> ()
+  | n :: l ->
+    let () = assert (cur < n) in
+    let () = output buf (n - cur) in
+    aux n l
+  in
+  let () = aux n l in
+  Buffer.contents buf
+
+let iter f s =
+  let pos = ref 0 in
+  let len = String.length s in
+  let cur = ref 0 in
+  while !pos < len do
+    let n = input s pos in
+    let () = cur := n + !cur in
+    f !cur
+  done
+
+end
+
+(** This data type is stored in vo files. *)
 type patches = {
-  reloc_infos : (reloc_info * int array) array;
+  reloc_infos : reloc_info array;
+  reloc_positions : Positions.t;
 }
 
-let patch_char4 buff pos c1 c2 c3 c4 =
-  Bytes.unsafe_set buff pos       c1;
-  Bytes.unsafe_set buff (pos + 1) c2;
-  Bytes.unsafe_set buff (pos + 2) c3;
-  Bytes.unsafe_set buff (pos + 3) c4
-
-let patch1 buff pos n =
-  patch_char4 buff pos
-    (Char.unsafe_chr n) (Char.unsafe_chr (n asr 8))  (Char.unsafe_chr (n asr 16))
-    (Char.unsafe_chr (n asr 24))
-
-let patch_int buff reloc =
-  (* copy code *before* patching because of nested evaluations:
-     the code we are patching might be called (and thus "concurrently" patched)
-     and results in wrong results. Side-effects... *)
+let patch_int buff reloc positions =
   let buff = Bytes.of_string buff in
-  let iter (reloc, npos) = Array.iter (fun pos -> patch1 buff pos reloc) npos in
-  let () = CArray.iter iter reloc in
+  let iter pos =
+    let id = Bytes.get_int32_le buff pos in
+    let reloc = reloc.(Int32.to_int id) in
+    Bytes.set_int32_le buff pos (Int32.of_int reloc)
+  in
+  let () = Positions.iter iter positions in
   buff
 
-let patch buff pl f =
-  (** Order seems important here? *)
-  let reloc = CArray.map (fun (r, pos) -> (f r, pos)) pl.reloc_infos in
-  let buff = patch_int buff reloc in
+let patch (buff, pl) f =
+  let reloc = CArray.map_left f pl.reloc_infos in
+  let buff = patch_int buff reloc pl.reloc_positions in
   tcode_of_code buff
 
 (* Buffering of bytecode *)
@@ -102,32 +147,35 @@ type label_definition =
 type env = {
   mutable out_buffer : Bytes.t;
   mutable out_position : int;
+  mutable reloc_pos : int list;
+  mutable reloc_id : int;
   mutable label_table : label_definition array;
-  (* le ieme element de la table = Label_defined n signifie que l'on a
-    deja rencontrer le label i et qu'il est a l'offset n.
-                                = Label_undefined l signifie que l'on a
-    pas encore rencontrer ce label, le premier entier indique ou est l'entier
-    a patcher dans la string, le deuxieme son origine  *)
-  reloc_info : int list RelocTable.t;
+  (* i-th table element = Label_defined n means that label i was already
+     encountered and lives at offset n
+     i-th table element = Label_undefined l means that the label was not
+     encountered yet, first integer is the location of the value to be patched
+     in the string, seconed one is its origin *)
+  reloc_info : int RelocTable.t;
 }
 
 let out_word env b1 b2 b3 b4 =
   let p = env.out_position in
-  if p >= Bytes.length env.out_buffer then begin
+  let buf =
     let len = Bytes.length env.out_buffer in
-    let new_len =
-      if len <= Sys.max_string_length / 2
-      then 2 * len
-      else
-        if len = Sys.max_string_length
-        then invalid_arg "String.create"  (* Pas la bonne exception .... *)
-        else Sys.max_string_length in
-    let new_buffer = Bytes.create new_len in
-    Bytes.blit env.out_buffer 0 new_buffer 0 len;
-    env.out_buffer <- new_buffer
-  end;
-  patch_char4 env.out_buffer p (Char.unsafe_chr b1)
-   (Char.unsafe_chr b2) (Char.unsafe_chr b3) (Char.unsafe_chr b4);
+    if p + 3 < len then env.out_buffer
+    else
+      let new_len = min (Sys.max_string_length) (2 * len) in
+      (* Not the right exception... *)
+      let () = if not (p + 3 < new_len) then invalid_arg "String.create" in
+      let new_buffer = Bytes.create new_len in
+      let () = Bytes.blit env.out_buffer 0 new_buffer 0 len in
+      let () = env.out_buffer <- new_buffer in
+      new_buffer
+  in
+  let () = Bytes.set_uint8 buf p b1 in
+  let () = Bytes.set_uint8 buf (p + 1) b2 in
+  let () = Bytes.set_uint8 buf (p + 2) b3 in
+  let () = Bytes.set_uint8 buf (p + 3) b4 in
   env.out_position <- p + 4
 
 let out env opcode =
@@ -159,10 +207,7 @@ let extend_label_table env needed =
 
 let backpatch env (pos, orig) =
   let displ = (env.out_position - orig) asr 2 in
-  Bytes.set env.out_buffer  pos    @@ Char.unsafe_chr displ;
-  Bytes.set env.out_buffer (pos+1) @@ Char.unsafe_chr (displ asr 8);
-  Bytes.set env.out_buffer (pos+2) @@ Char.unsafe_chr (displ asr 16);
-  Bytes.set env.out_buffer (pos+3) @@ Char.unsafe_chr (displ asr 24)
+  Bytes.set_int32_le env.out_buffer pos (Int32.of_int displ)
 
 let define_label env lbl =
   if lbl >= Array.length env.label_table then extend_label_table env lbl;
@@ -179,10 +224,6 @@ let out_label_with_orig env orig lbl =
     Label_defined def ->
       out_int env ((def - orig) asr 2)
   | Label_undefined patchlist ->
-      (* spiwack: patchlist is supposed to be non-empty all the time
-         thus I commented that out. If there is no problem I suggest
-         removing it for next release (cur: 8.1) *)
-      (*if patchlist = [] then *)
         (env.label_table).(lbl) <-
           Label_undefined((env.out_position, orig) :: patchlist);
       out_int env 0
@@ -193,28 +234,20 @@ let out_label env l = out_label_with_orig env env.out_position l
 
 let enter env info =
   let pos = env.out_position in
-  let old = try RelocTable.find env.reloc_info info with Not_found -> [] in
-  RelocTable.replace env.reloc_info info (pos :: old)
+  let () = env.reloc_pos <- pos :: env.reloc_pos in
+  try RelocTable.find env.reloc_info info
+  with Not_found ->
+    let id = env.reloc_id in
+    let () = env.reloc_id <- id + 1 in
+    let () = RelocTable.add env.reloc_info info id in
+    id
 
-let slot_for_const env c =
-  enter env (Reloc_const c);
-  out_int env 0
+let slot_for env r = out_int env (enter env r)
 
-let slot_for_annot env a =
-  enter env (Reloc_annot a);
-  out_int env 0
-
-let slot_for_getglobal env p =
-  enter env (Reloc_getglobal p);
-  out_int env 0
-
-let slot_for_proj_name env p =
-  enter env (Reloc_proj_name p);
-  out_int env 0
-
-let slot_for_caml_prim env op =
-  enter env (Reloc_caml_prim op);
-  out_int env 0
+let slot_for_const env c = slot_for env (Reloc_const c)
+let slot_for_annot env a = slot_for env (Reloc_annot a)
+let slot_for_getglobal env p = slot_for env (Reloc_getglobal p)
+let slot_for_caml_prim env op = slot_for env (Reloc_caml_prim op)
 
 (* Emission of one instruction *)
 
@@ -255,6 +288,7 @@ let check_prim_op = function
   | Float64lt      -> opCHECKLTFLOAT
   | Float64le      -> opCHECKLEFLOAT
   | Float64compare -> opCHECKCOMPAREFLOAT
+  | Float64equal   -> opCHECKEQUALFLOAT
   | Float64classify -> opCHECKCLASSIFYFLOAT
   | Float64add     -> opCHECKADDFLOAT
   | Float64sub     -> opCHECKSUBFLOAT
@@ -267,11 +301,14 @@ let check_prim_op = function
   | Float64ldshiftexp -> opCHECKLDSHIFTEXP
   | Float64next_up    -> opCHECKNEXTUPFLOAT
   | Float64next_down  -> opCHECKNEXTDOWNFLOAT
-  | Arraymake -> opCHECKCAMLCALL2_1
-  | Arrayget -> opCHECKCAMLCALL2
-  | Arrayset -> opCHECKCAMLCALL3_1
-  | Arraydefault | Arraycopy | Arraylength ->
-      opCHECKCAMLCALL1
+  | Arraymake | Arrayget | Arrayset | Arraydefault | Arraycopy | Arraylength ->
+    assert false
+
+let check_caml_prim_op = function
+| CAML_Arraymake -> opCHECKCAMLCALL2_1
+| CAML_Arrayget -> opCHECKCAMLCALL2
+| CAML_Arrayset -> opCHECKCAMLCALL3_1
+| CAML_Arraydefault | CAML_Arraycopy | CAML_Arraylength -> opCHECKCAMLCALL1
 
 let inplace_prim_op = function
   | Float64next_up | Float64next_down -> true
@@ -343,8 +380,7 @@ let emit_instr env = function
   | Kconst c ->
       out env opGETGLOBAL; slot_for_const env c
   | Kmakeblock(n, t) ->
-      if Int.equal n 0 then invalid_arg "emit_instr : block size = 0"
-      else if n < 4 then (out env(opMAKEBLOCK1 + n - 1); out_int env t)
+      if 0 < n && n < 4 then (out env(opMAKEBLOCK1 + n - 1); out_int env t)
       else (out env opMAKEBLOCK; out_int env n; out_int env t)
   | Kmakeswitchblock(typlbl,swlbl,annot,sz) ->
       out env opMAKESWITCHBLOCK;
@@ -368,7 +404,7 @@ let emit_instr env = function
   | Ksetfield n ->
       out env opSETFIELD; out_int env n
   | Ksequence _ -> invalid_arg "Vmemitcodes.emit_instr"
-  | Kproj p -> out env opPROJ; out_int env (Projection.Repr.arg p); slot_for_proj_name env p
+  | Kproj p -> out env opPROJ; out_int env p
   | Kensurestackcapacity size -> out env opENSURESTACKCAPACITY; out_int env size
   | Kbranch lbl -> out env opBRANCH; out_label env lbl
   | Kprim (op, (q,_u)) ->
@@ -376,7 +412,7 @@ let emit_instr env = function
       slot_for_getglobal env q
 
   | Kcamlprim (op,lbl) ->
-    out env (check_prim_op op);
+    out env (check_caml_prim_op op);
     out_label env lbl;
     slot_for_caml_prim env op
 
@@ -434,7 +470,7 @@ let rec emit env insns remaining = match insns with
 
 (* Initialization *)
 
-type to_patch = emitcodes * patches * fv
+type to_patch = emitcodes * patches
 
 (* Substitution *)
 let subst_strcst s sc =
@@ -450,44 +486,47 @@ let subst_reloc s ri =
   | Reloc_annot a -> Reloc_annot (subst_annot s a)
   | Reloc_const sc -> Reloc_const (subst_strcst s sc)
   | Reloc_getglobal kn -> Reloc_getglobal (subst_constant s kn)
-  | Reloc_proj_name p -> Reloc_proj_name (subst_proj_repr s p)
   | Reloc_caml_prim _ -> ri
 
 let subst_patches subst p =
-  let infos = CArray.map (fun (r, pos) -> (subst_reloc subst r, pos)) p.reloc_infos in
-  { reloc_infos = infos; }
+  let infos = CArray.Smart.map (fun r -> subst_reloc subst r) p.reloc_infos in
+  { reloc_infos = infos; reloc_positions = p.reloc_positions }
 
-let subst_to_patch s (code,pl,fv) =
-  code, subst_patches s pl, fv
+let subst_to_patch s (code, pl) =
+  (code, subst_patches s pl)
 
 type body_code =
-  | BCdefined of to_patch
+  | BCdefined of to_patch * fv
   | BCalias of Names.Constant.t
   | BCconstant
 
 let subst_body_code s = function
-| BCdefined tp -> BCdefined (subst_to_patch s tp)
+| BCdefined (tp, fv) -> BCdefined (subst_to_patch s tp, fv)
 | BCalias cu -> BCalias (subst_constant s cu)
 | BCconstant -> BCconstant
 
-let to_memory (init_code, fun_code, fv) =
+let to_memory code =
   let env = {
     out_buffer = Bytes.create 1024;
     out_position = 0;
+    reloc_id = 0;
+    reloc_pos = [];
     label_table = Array.make 16 (Label_undefined []);
     reloc_info = RelocTable.create 91;
   } in
-  emit env init_code [];
-  emit env fun_code [];
+  emit env code [];
   (** Later uses of this string are all purely functional *)
   let code = Bytes.sub_string env.out_buffer 0 env.out_position in
   let code = CString.hcons code in
-  let fold reloc npos accu = (reloc, Array.of_list npos) :: accu in
+  let fold reloc id accu = (id, reloc) :: accu in
   let reloc = RelocTable.fold fold env.reloc_info [] in
-  let reloc = { reloc_infos = CArray.of_list reloc } in
+  let reloc = List.sort (fun (id1, _) (id2, _) -> Int.compare id1 id2) reloc in
+  let reloc_infos = CArray.map_of_list snd reloc in
+  let reloc_positions = Positions.of_list (List.rev env.reloc_pos) in
+  let reloc = { reloc_infos; reloc_positions } in
   Array.iter (fun lbl ->
     (match lbl with
       Label_defined _ -> assert true
     | Label_undefined patchlist ->
         assert (patchlist = []))) env.label_table;
-  (code, reloc, fv)
+  (code, reloc)
