@@ -15,23 +15,18 @@ type reloc_info =
   | Reloc_annot of annot_switch
   | Reloc_const of structured_constant
   | Reloc_getglobal of Constant.t
-  | Reloc_proj_name of Projection.Repr.t
-  | Reloc_caml_prim of CPrimitives.t
+  | Reloc_caml_prim of caml_prim
 
-type patches
-type emitcodes
+type to_patch
 
-val patch : emitcodes -> patches -> (reloc_info -> int) -> Vmvalues.tcode
-
-type to_patch = emitcodes * patches * fv
+val patch : to_patch -> (reloc_info -> int) -> Vmvalues.tcode
 
 type body_code =
-  | BCdefined of to_patch
+  | BCdefined of to_patch * fv
   | BCalias of Constant.t
   | BCconstant
 
 
 val subst_body_code : Mod_subst.substitution -> body_code -> body_code
 
-val to_memory : bytecodes * bytecodes * fv -> to_patch
-               (** init code, fun code, fv *)
+val to_memory : bytecodes -> to_patch
