@@ -77,11 +77,12 @@ let accept (sr,sw) =
 
 let spawn_sock env prog args =
   let main_sock, main_sock_name = mk_socket_channel () in
-  let extra = [| prog; "-main-channel"; main_sock_name |] in
-  let args = Array.append extra args in
+  let extra = [| "-main-channel"; main_sock_name |] in
+  let args = Array.append [| prog |] (Array.append args extra) in
   prerr_endline ("EXEC: " ^ String.concat " " (Array.to_list args));
   let pid =
     Unix.create_process_env prog args env Unix.stdin Unix.stdout Unix.stderr in
+  prerr_endline ("PID " ^ string_of_int pid);
   if pid = 0 then begin
     Unix.sleep 1; (* to avoid respawning like crazy *)
     raise (Failure "create_process failed")
@@ -92,8 +93,8 @@ let spawn_sock env prog args =
 let spawn_pipe env prog args =
   let master2worker_r,master2worker_w = Unix.pipe () in
   let worker2master_r,worker2master_w = Unix.pipe () in
-  let extra = [| prog; "-main-channel"; "stdfds" |] in
-  let args = Array.append extra args in
+  let extra = [| "-main-channel"; "stdfds" |] in
+  let args = Array.append [|prog|] (Array.append args extra) in
   Unix.set_close_on_exec master2worker_w;
   Unix.set_close_on_exec worker2master_r;
   prerr_endline ("EXEC: " ^ String.concat " " (Array.to_list args));
@@ -129,7 +130,7 @@ let spawn_with_control prefer_sock env prog args =
     if not_Unix then
       let control_sock, control_sock_name = mk_socket_channel () in
       let extra = [| "-control-channel"; control_sock_name |] in
-      Array.append extra args, Some control_sock
+      Array.append args extra, Some control_sock
     else
       args, None in
   let (pid, cin, cout, s), is_sock =
@@ -180,7 +181,7 @@ let kill ({ pid = unixpid; oob_resp; oob_req; cin; cout; alive; watch } as p) =
     close_out_noerr cout;
     Option.iter close_in_noerr oob_resp;
     Option.iter close_out_noerr oob_req;
-    if Sys.os_type = "Unix" then Unix.kill unixpid 9;
+    if Sys.os_type = "Unix" then Unix.kill unixpid Sys.sigkill;
     p.watch <- None
   with e -> prerr_endline ("kill: "^Printexc.to_string e) end
 
@@ -250,7 +251,7 @@ let kill ({ pid = unixpid; oob_req; oob_resp; cin; cout; alive } as p) =
     close_out_noerr cout;
     Option.iter close_in_noerr oob_resp;
     Option.iter close_out_noerr oob_req;
-    if Sys.os_type = "Unix" then Unix.kill unixpid 9;
+    if Sys.os_type = "Unix" then Unix.kill unixpid Sys.sigkill;
   with e -> prerr_endline ("kill: "^Printexc.to_string e) end
 
 let rec wait p =
