@@ -9,15 +9,6 @@
 (************************************************************************)
 
 open Names
-open Constr
-open Mod_subst
-
-type global_reference = GlobRef.t =
-  | VarRef of variable     [@ocaml.deprecated "Use Names.GlobRef.VarRef"]
-  | ConstRef of Constant.t [@ocaml.deprecated "Use Names.GlobRef.ConstRef"]
-  | IndRef of inductive    [@ocaml.deprecated "Use Names.GlobRef.IndRef"]
-  | ConstructRef of constructor [@ocaml.deprecated "Use Names.GlobRef.ConstructRef"]
-[@@ocaml.deprecated "Use Names.GlobRef.t"]
 
 open GlobRef
 
@@ -35,25 +26,25 @@ let destConstructRef = function ConstructRef ind -> ind | _ -> failwith "destCon
 let subst_global_reference subst ref = match ref with
   | VarRef var -> ref
   | ConstRef kn ->
-    let kn' = subst_constant subst kn in
+    let kn' = Mod_subst.subst_constant subst kn in
       if kn==kn' then ref else ConstRef kn'
   | IndRef ind ->
-    let ind' = subst_ind subst ind in
+    let ind' = Mod_subst.subst_ind subst ind in
       if ind==ind' then ref else IndRef ind'
   | ConstructRef ((kn,i),j as c) ->
-    let c' = subst_constructor subst c in
+    let c' = Mod_subst.subst_constructor subst c in
     if c'==c then ref else ConstructRef c'
 
 let subst_global subst ref = match ref with
   | VarRef var -> ref, None
   | ConstRef kn ->
-     let kn',t = subst_con subst kn in
+     let kn',t = Mod_subst.subst_con subst kn in
       if kn==kn' then ref, None else ConstRef kn', t
   | IndRef ind ->
-      let ind' = subst_ind subst ind in
+      let ind' = Mod_subst.subst_ind subst ind in
       if ind==ind' then ref, None else IndRef ind', None
   | ConstructRef ((kn,i),j as c) ->
-      let c' = subst_constructor subst c in
+      let c' = Mod_subst.subst_constructor subst c in
       if c'==c then ref,None else ConstructRef c', None
 
 let canonical_gr = function
@@ -62,28 +53,15 @@ let canonical_gr = function
   | ConstructRef ((kn,i),j )-> ConstructRef((MutInd.make1(MutInd.canonical kn),i),j)
   | VarRef id -> VarRef id
 
-let global_of_constr c = match kind c with
-  | Const (sp,u) -> ConstRef sp
-  | Ind (ind_sp,u) -> IndRef ind_sp
-  | Construct (cstr_cp,u) -> ConstructRef cstr_cp
-  | Var id -> VarRef id
-  |  _ -> raise Not_found
-
-let is_global = Constr.isRefX
-
-let printable_constr_of_global = function
-  | VarRef id -> mkVar id
-  | ConstRef sp -> mkConst sp
-  | ConstructRef sp -> mkConstruct sp
-  | IndRef sp -> mkInd sp
-
 (* Extended global references *)
 
-type syndef_name = KerName.t
+type abbreviation = KerName.t
 
 type extended_global_reference =
   | TrueGlobal of GlobRef.t
-  | SynDef of syndef_name
+  | Abbrev of abbreviation
+
+let abbreviation_eq = KerName.equal
 
 (* We order [extended_global_reference] via their user part
    (cf. pretty printer) *)
@@ -95,22 +73,22 @@ module ExtRefOrdered = struct
     x == y ||
     match x, y with
     | TrueGlobal rx, TrueGlobal ry -> GlobRef.UserOrd.equal rx ry
-    | SynDef knx, SynDef kny -> KerName.equal knx kny
-    | (TrueGlobal _ | SynDef _), _ -> false
+    | Abbrev knx, Abbrev kny -> KerName.equal knx kny
+    | (TrueGlobal _ | Abbrev _), _ -> false
 
   let compare x y =
     if x == y then 0
     else match x, y with
       | TrueGlobal rx, TrueGlobal ry -> GlobRef.UserOrd.compare rx ry
-      | SynDef knx, SynDef kny -> KerName.compare knx kny
-      | TrueGlobal _, SynDef _ -> -1
-      | SynDef _, TrueGlobal _ -> 1
+      | Abbrev knx, Abbrev kny -> KerName.compare knx kny
+      | TrueGlobal _, Abbrev _ -> -1
+      | Abbrev _, TrueGlobal _ -> 1
 
   open Hashset.Combine
 
   let hash = function
   | TrueGlobal gr -> combinesmall 1 (GlobRef.UserOrd.hash gr)
-  | SynDef kn -> combinesmall 2 (KerName.hash kn)
+  | Abbrev kn -> combinesmall 2 (KerName.hash kn)
 
 end
 
@@ -118,5 +96,5 @@ module ExtRefMap = HMap.Make(ExtRefOrdered)
 module ExtRefSet = ExtRefMap.Set
 
 let subst_extended_reference sub = function
-  | SynDef kn -> SynDef (subst_kn sub kn)
+  | Abbrev kn -> Abbrev (Mod_subst.subst_kn sub kn)
   | TrueGlobal gr -> TrueGlobal (subst_global_reference sub gr)
