@@ -10,12 +10,8 @@
 
 (* This module manages customization parameters at the vernacular level     *)
 
-open Pp
-open CErrors
 open Util
-open Libobject
-open Libnames
-open Mod_subst
+open Summary.Stage
 
 type option_name = string list
 type option_value =
@@ -26,11 +22,11 @@ type option_value =
 
 type table_value =
   | StringRefValue of string
-  | QualidRefValue of qualid
+  | QualidRefValue of Libnames.qualid
 
 (** Summary of an option status *)
 type option_state = {
-  opt_depr  : bool;
+  opt_depr  : Deprecation.t option;
   opt_value : option_value;
 }
 
@@ -40,12 +36,14 @@ type option_state = {
 let nickname table = String.concat " " table
 
 let error_no_table_of_this_type ~kind key =
-  user_err
-    (str ("There is no " ^ kind ^ "-valued table with this name: \"" ^ nickname key ^ "\"."))
+  CErrors.user_err
+    Pp.(str ("There is no " ^ kind ^ "-valued table with this name: \""
+      ^ nickname key ^ "\"."))
 
 let error_undeclared_key key =
-  user_err
-    (str ("There is no flag, option or table with this name: \"" ^ nickname key ^ "\"."))
+  CErrors.user_err
+    Pp.(str ("There is no flag, option or table with this name: \""
+      ^ nickname key ^ "\"."))
 
 (****************************************************************************)
 (* 1- Tables                                                                *)
@@ -65,7 +63,7 @@ module MakeTable =
           module Set : CSig.SetS with type elt = t
           val table : (string * key table_of_A) list ref
           val encode : Environ.env -> key -> t
-          val subst : substitution -> t -> t
+          val subst : Mod_subst.substitution -> t -> t
           val printer : t -> Pp.t
           val key : option_name
           val title : string
@@ -78,13 +76,15 @@ module MakeTable =
 
     let nick = nickname A.key
 
-    let _ =
+    let () =
       if String.List.mem_assoc nick !A.table then
-        user_err Pp.(str "Sorry, this table name (" ++ str nick ++ str ") is already used.")
+        CErrors.user_err
+          Pp.(strbrk "Sorry, this table name (" ++ str nick
+            ++ strbrk ") is already used.")
 
     module MySet = A.Set
 
-    let t = Summary.ref MySet.empty ~name:nick
+    let t = Summary.ref ~stage:Summary.Stage.Synterp MySet.empty ~name:nick
 
     let (add_option,remove_option) =
         let cache_options (f,p) = match f with
@@ -96,10 +96,11 @@ module MakeTable =
             if p' == p then obj else
               (f,p')
         in
-        let inGo : option_mark * A.t -> obj =
+        let inGo : option_mark * A.t -> Libobject.obj =
           Libobject.declare_object {(Libobject.default_object nick) with
+                Libobject.object_stage = Summary.Stage.Synterp;
                 Libobject.load_function = load_options;
-                Libobject.open_function = simple_open load_options;
+                Libobject.open_function = Libobject.simple_open load_options;
                 Libobject.cache_function = cache_options;
                 Libobject.subst_function = subst_options;
                 Libobject.classify_function = (fun x -> Substitute)}
@@ -108,13 +109,14 @@ module MakeTable =
          (fun c -> Lib.add_leaf (inGo (GOrmv, c))))
 
     let print_table table_name printer table =
-      Feedback.msg_notice
-        (str table_name ++
-           (hov 0
-              (if MySet.is_empty table then str " None" ++ fnl ()
-               else MySet.fold
-                 (fun a b -> spc () ++ printer a ++ b)
-                 table (mt ()) ++ str "." ++ fnl ())))
+      let open Pp in
+      let pp = if MySet.is_empty table then str table_name ++ str " is empty."
+        else
+          str table_name ++ str ":" ++ spc() ++
+          (hov 0 (prlist_with_sep spc printer (MySet.elements table))) ++
+          str "."
+      in
+      Feedback.msg_notice pp
 
     let table_of_A = {
        add = (fun env x -> add_option (A.encode env x));
@@ -152,7 +154,7 @@ struct
   let table = string_table
   let encode _env x = x
   let subst _ x = x
-  let printer = str
+  let printer = Pp.str
   let key = A.key
   let title = A.title
   let member_message = A.member_message
@@ -169,8 +171,8 @@ module type RefConvertArg =
 sig
   type t
   module Set : CSig.SetS with type elt = t
-  val encode : Environ.env -> qualid -> t
-  val subst : substitution -> t -> t
+  val encode : Environ.env -> Libnames.qualid -> t
+  val subst : Mod_subst.substitution -> t -> t
   val printer : t -> Pp.t
   val key : option_name
   val title : string
@@ -180,7 +182,7 @@ end
 module RefConvert = functor (A : RefConvertArg) ->
 struct
   type t = A.t
-  type key = qualid
+  type key = Libnames.qualid
   module Set = A.Set
   let table = ref_table
   let encode = A.encode
@@ -196,16 +198,16 @@ module MakeRefTable =
 
 type iter_table_aux = { aux : 'a. 'a table_of_A -> Environ.env -> 'a -> unit }
 
-let iter_table f key lv =
+let iter_table env f key lv =
   let aux = function
     | StringRefValue s ->
        begin
-         try f.aux (get_string_table key) (Global.env()) s
+         try f.aux (get_string_table key) env s
          with Not_found -> error_no_table_of_this_type ~kind:"string" key
        end
     | QualidRefValue locqid ->
        begin
-         try f.aux (get_ref_table key) (Global.env()) locqid
+         try f.aux (get_ref_table key) env locqid
          with Not_found -> error_no_table_of_this_type ~kind:"qualid" key
        end
   in
@@ -215,7 +217,8 @@ let iter_table f key lv =
 (* 2- Flags.                                                              *)
 
 type 'a option_sig = {
-  optdepr  : bool;
+  optstage : Summary.Stage.t;
+  optdepr  : Deprecation.t option;
   optkey   : option_name;
   optread  : unit -> 'a;
   optwrite : 'a -> unit }
@@ -240,26 +243,28 @@ let get_option key = OptionMap.find key !value_tab
 
 let check_key key = try
   let _ = get_option key in
-  user_err Pp.(str "Sorry, this option name ("++ str (nickname key) ++ str ") is already used.")
+  CErrors.user_err Pp.(str "Sorry, this option name ("
+    ++ str (nickname key) ++ str ") is already used.")
 with Not_found ->
   if String.List.mem_assoc (nickname key) !string_table
     || String.List.mem_assoc (nickname key) !ref_table
-  then user_err Pp.(str "Sorry, this option name (" ++ str (nickname key) ++ str ") is already used.")
+  then CErrors.user_err Pp.(str "Sorry, this option name ("
+    ++ str (nickname key) ++ str ") is already used.")
 
 open Libobject
 
 let warn_deprecated_option =
-  CWarnings.create ~name:"deprecated-option" ~category:"deprecated"
-         (fun key -> str "Option" ++ spc () ++ str (nickname key) ++
-                       strbrk " is deprecated")
+  Deprecation.create_warning ~object_name:"Option" ~warning_name_if_no_since:"deprecated-option"
+    (fun key -> Pp.str (nickname key))
 
 let declare_option cast uncast append ?(preprocess = fun x -> x)
-  { optdepr=depr; optkey=key; optread=read; optwrite=write } =
+  { optstage=stage; optdepr=depr; optkey=key; optread=read; optwrite=write } =
   check_key key;
   let default = read() in
   let change =
-      let _ = Summary.declare_summary (nickname key)
-        { Summary.freeze_function = (fun ~marshallable -> read ());
+      let () = Summary.declare_summary (nickname key)
+        { stage;
+          Summary.freeze_function = read;
           Summary.unfreeze_function = write;
           Summary.init_function = (fun () -> write default) } in
       let cache_options (l,m,v) =
@@ -288,6 +293,7 @@ let declare_option cast uncast append ?(preprocess = fun x -> x)
       let options : option_locality * option_mod * _ -> obj =
         declare_object
           { (default_object (nickname key)) with
+            object_stage = stage;
             load_function = load_options;
             open_function = simple_open open_options;
             cache_function = cache_options;
@@ -296,59 +302,64 @@ let declare_option cast uncast append ?(preprocess = fun x -> x)
             classify_function = classify_options } in
       (fun l m v -> let v = preprocess v in Lib.add_leaf (options (l, m, v)))
   in
-  let warn () = if depr then warn_deprecated_option key in
+  let warn () = depr |> Option.iter (fun depr -> warn_deprecated_option (key,depr)) in
   let cread () = cast (read ()) in
   let cwrite l v = warn (); change l OptSet (uncast v) in
   let cappend l v = warn (); change l OptAppend (uncast v) in
-  value_tab := OptionMap.add key (depr, (cread,cwrite,cappend)) !value_tab
+  value_tab := OptionMap.add key (depr, stage, (cread,cwrite,cappend)) !value_tab
 
 let declare_int_option =
   declare_option
     (fun v -> IntValue v)
-    (function IntValue v -> v | _ -> anomaly (Pp.str "async_option."))
-    (fun _ _ -> anomaly (Pp.str "async_option."))
+    (function IntValue v -> v | _ -> CErrors.anomaly (Pp.str "async_option."))
+    (fun _ _ -> CErrors.anomaly (Pp.str "async_option."))
 let declare_bool_option =
   declare_option
     (fun v -> BoolValue v)
-    (function BoolValue v -> v | _ -> anomaly (Pp.str "async_option."))
-    (fun _ _ -> anomaly (Pp.str "async_option."))
+    (function BoolValue v -> v | _ -> CErrors.anomaly (Pp.str "async_option."))
+    (fun _ _ -> CErrors.anomaly (Pp.str "async_option."))
 let declare_string_option =
   declare_option
     (fun v -> StringValue v)
-    (function StringValue v -> v | _ -> anomaly (Pp.str "async_option."))
+    (function StringValue v -> v | _ -> CErrors.anomaly (Pp.str "async_option."))
     (fun x y -> x^","^y)
+
 let declare_stringopt_option =
   declare_option
     (fun v -> StringOptValue v)
-    (function StringOptValue v -> v | _ -> anomaly (Pp.str "async_option."))
-    (fun _ _ -> anomaly (Pp.str "async_option."))
+    (function StringOptValue v -> v | _ -> CErrors.anomaly (Pp.str "async_option."))
+    (fun _ _ -> CErrors.anomaly (Pp.str "async_option."))
 
+type 'a getter = { get : unit -> 'a }
 
-type 'a opt_decl = depr:bool -> key:option_name -> 'a
+type 'a opt_decl = ?stage:Summary.Stage.t -> ?depr:Deprecation.t -> key:option_name -> value:'a -> unit -> 'a getter
 
-let declare_int_option_and_ref ~depr ~key ~(value:int) =
+let declare_int_option_and_ref ?(stage=Interp) ?depr ~key ~(value:int) () =
   let r_opt = ref value in
   let optwrite v = r_opt := Option.default value v in
   let optread () = Some !r_opt in
-  let _ = declare_int_option {
+  let () = declare_int_option {
+      optstage = stage;
       optdepr = depr;
       optkey = key;
-      optread; optwrite
+      optread;
+      optwrite;
     } in
-  fun () -> !r_opt
+  { get = fun () -> !r_opt }
 
-let declare_intopt_option_and_ref ~depr ~key =
-  let r_opt = ref None in
+let declare_intopt_option_and_ref ?(stage=Interp) ?depr ~key ~value () =
+  let r_opt = ref value in
   let optwrite v = r_opt := v in
   let optread () = !r_opt in
-  let _ = declare_int_option {
+  let () = declare_int_option {
+      optstage = stage;
       optdepr = depr;
       optkey = key;
       optread; optwrite
     } in
-  optread
+  { get = optread }
 
-let declare_nat_option_and_ref ~depr ~key ~(value:int) =
+let declare_nat_option_and_ref ?(stage=Interp) ?depr ~key ~(value:int) () =
   assert (value >= 0);
   let r_opt = ref value in
   let optwrite v =
@@ -358,56 +369,61 @@ let declare_nat_option_and_ref ~depr ~key ~(value:int) =
     r_opt := v
   in
   let optread () = Some !r_opt in
-  let _ = declare_int_option {
+  let () = declare_int_option {
+      optstage = stage;
       optdepr = depr;
       optkey = key;
       optread; optwrite
     } in
-  fun () -> !r_opt
+  { get = fun () -> !r_opt }
 
-let declare_bool_option_and_ref ~depr ~key ~(value:bool) =
+let declare_bool_option_and_ref ?(stage=Interp) ?depr ~key ~(value:bool) () =
   let r_opt = ref value in
   let optwrite v = r_opt := v in
   let optread () = !r_opt in
-  let _ = declare_bool_option {
+  let () = declare_bool_option {
+      optstage = stage;
       optdepr = depr;
       optkey = key;
       optread; optwrite
     } in
-  optread
+  { get = optread }
 
-let declare_string_option_and_ref ~depr ~key ~(value:string) =
+let declare_string_option_and_ref ?(stage=Interp) ?depr ~key ~(value:string) () =
   let r_opt = ref value in
   let optwrite v = r_opt := Option.default value v in
   let optread () = Some !r_opt in
-  let _ = declare_stringopt_option {
+  let () = declare_stringopt_option {
+      optstage = stage;
       optdepr = depr;
       optkey = key;
       optread; optwrite
     } in
-  fun () -> !r_opt
+  { get = fun () -> !r_opt }
 
-let declare_stringopt_option_and_ref ~depr ~key =
-  let r_opt = ref None in
+let declare_stringopt_option_and_ref ?(stage=Interp) ?depr ~key ~value () =
+  let r_opt = ref value in
   let optwrite v = r_opt := v in
   let optread () = !r_opt in
-  let _ = declare_stringopt_option {
+  let () = declare_stringopt_option {
+      optstage = stage;
       optdepr = depr;
       optkey = key;
       optread; optwrite
     } in
-  optread
+  { get = optread }
 
-let declare_interpreted_string_option_and_ref ~depr ~key ~(value:'a) from_string to_string =
+let declare_interpreted_string_option_and_ref from_string to_string ?(stage=Interp) ?depr ~key ~(value:'a) () =
   let r_opt = ref value in
   let optwrite v = r_opt := Option.default value @@ Option.map from_string v in
   let optread () = Some (to_string !r_opt) in
-  let _ = declare_stringopt_option {
+  let () = declare_stringopt_option {
+      optstage = stage;
       optdepr = depr;
       optkey = key;
       optread; optwrite
     } in
-  fun () -> !r_opt
+  { get = fun () -> !r_opt }
 
 (* 3- User accessible commands *)
 
@@ -415,24 +431,33 @@ let declare_interpreted_string_option_and_ref ~depr ~key ~(value:'a) from_string
 
 let warn_unknown_option =
   CWarnings.create
-    ~name:"unknown-option" ~category:"option"
-    (fun key -> strbrk "There is no flag or option with this name: \"" ++
+    ~name:"unknown-option"
+    Pp.(fun key -> strbrk "There is no flag or option with this name: \"" ++
                   str (nickname key) ++ str "\".")
 
-let set_option_value ?(locality = OptDefault) check_and_cast key v =
+let get_option_value key =
+  try
+    let (_,_,(read,write,append)) = get_option key in
+    Some read
+  with Not_found -> None
+
+(** Sets the option only if [stage] matches the option declaration or if [stage]
+  is omitted. If the option is not found, a warning is emitted only if the stage
+  is [Interp] or omitted. *)
+let set_option_value ?(locality = OptDefault) ?stage check_and_cast key v =
   let opt = try Some (get_option key) with Not_found -> None in
   match opt with
-  | None -> warn_unknown_option key
-  | Some (depr, (read,write,append)) ->
-    write locality (check_and_cast v (read ()))
+  | None -> begin match stage with None | Some Summary.Stage.Interp -> warn_unknown_option key | _ -> () end
+  | Some (depr, option_stage, (read,write,append)) ->
+    if Option.cata (fun s -> s = option_stage) true stage then
+      write locality (check_and_cast v (read ()))
 
 let bad_type_error ~expected ~got =
-  user_err Pp.(str "Bad type of value for this option:" ++ spc() ++
-               str "expected " ++ str expected ++
-               str ", got " ++ str got ++ str ".")
+  CErrors.user_err Pp.(strbrk "Bad type of value for this option:" ++ spc()
+    ++ str "expected " ++ str expected ++ str ", got " ++ str got ++ str ".")
 
 let error_flag () =
-  user_err Pp.(str "This is a flag. It does not take a value.")
+  CErrors.user_err Pp.(str "This is a flag. It does not take a value.")
 
 let check_int_value v = function
   | BoolValue _ -> error_flag ()
@@ -442,7 +467,9 @@ let check_int_value v = function
 
 let check_bool_value v = function
   | BoolValue _ -> BoolValue v
-  | _ -> user_err Pp.(str "This is an option. A value must be provided.")
+  | _ ->
+      CErrors.user_err
+        Pp.(str "This is an option. A value must be provided.")
 
 let check_string_value v = function
   | BoolValue _ -> error_flag ()
@@ -454,57 +481,60 @@ let check_unset_value v = function
   | BoolValue _ -> BoolValue false
   | IntValue _ -> IntValue None
   | StringOptValue _ -> StringOptValue None
-  | StringValue _ -> user_err Pp.(str "This option does not support the \"Unset\" command.")
+  | StringValue _ ->
+      CErrors.user_err
+        Pp.(str "This option does not support the \"Unset\" command.")
 
 (* Nota: For compatibility reasons, some errors are treated as
    warnings. This allows a script to refer to an option that doesn't
    exist anymore *)
 
-let set_int_option_value_gen ?locality =
-  set_option_value ?locality check_int_value
-let set_bool_option_value_gen ?locality key v =
-  set_option_value ?locality check_bool_value key v
-let set_string_option_value_gen ?locality =
-  set_option_value ?locality check_string_value
-let unset_option_value_gen ?locality key =
-  set_option_value ?locality check_unset_value key ()
+let set_int_option_value_gen ?locality ?stage =
+  set_option_value ?locality ?stage check_int_value
+let set_bool_option_value_gen ?locality ?stage key v =
+  set_option_value ?locality ?stage check_bool_value key v
+let set_string_option_value_gen ?locality ?stage =
+  set_option_value ?locality ?stage check_string_value
+let unset_option_value_gen ?locality ?stage key =
+  set_option_value ?locality ?stage check_unset_value key ()
 
-let set_string_option_append_value_gen ?(locality = OptDefault) key v =
+let set_string_option_append_value_gen ?(locality = OptDefault) ?stage key v =
   let opt = try Some (get_option key) with Not_found -> None in
   match opt with
   | None -> warn_unknown_option key
-  | Some (depr, (read,write,append)) ->
-    append locality (check_string_value v (read ()))
+  | Some (depr, option_stage, (read,write,append)) ->
+    if Option.cata (fun s -> s = option_stage) true stage then
+      append locality (check_string_value v (read ()))
 
-let set_int_option_value opt v = set_int_option_value_gen opt v
-let set_bool_option_value opt v = set_bool_option_value_gen opt v
-let set_string_option_value opt v = set_string_option_value_gen opt v
+let set_int_option_value ?stage opt v = set_int_option_value_gen ?stage opt v
+let set_bool_option_value ?stage opt v = set_bool_option_value_gen ?stage opt v
+let set_string_option_value ?stage opt v = set_string_option_value_gen ?stage opt v
 
 (* Printing options/tables *)
 
-let msg_option_value v =
-  match v with
-    | BoolValue true  -> str "on"
-    | BoolValue false -> str "off"
-    | IntValue (Some n) -> int n
-    | IntValue None   -> str "undefined"
-    | StringValue s   -> quote (str s)
-    | StringOptValue None   -> str"undefined"
-    | StringOptValue (Some s)   -> quote (str s)
-(*     | IdentValue r    -> pr_global_env Id.Set.empty r *)
+let msg_option_value = Pp.(function
+  | BoolValue true -> str "on"
+  | BoolValue false -> str "off"
+  | IntValue (Some n) -> int n
+  | IntValue None -> str "undefined"
+  | StringValue s -> quote (str s)
+  | StringOptValue None -> str "undefined"
+  | StringOptValue (Some s) -> quote (str s))
 
 let print_option_value key =
-  let (depr, (read,_,_)) = get_option key in
+  let (depr, _stage, (read,_,_)) = get_option key in
   let s = read () in
   match s with
     | BoolValue b ->
-        Feedback.msg_notice (prlist_with_sep spc str key ++ str " is " ++ str (if b then "on" else "off"))
+        Feedback.msg_notice Pp.(prlist_with_sep spc str key ++ str " is "
+          ++ str (if b then "on" else "off"))
     | _ ->
-        Feedback.msg_notice (str "Current value of " ++ prlist_with_sep spc str key ++ str " is " ++ msg_option_value s)
+        Feedback.msg_notice Pp.(str "Current value of "
+          ++ prlist_with_sep spc str key ++ str " is " ++ msg_option_value s)
 
 let get_tables () =
   let tables = !value_tab in
-  let fold key (depr, (read,_,_)) accu =
+  let fold key (depr, _stage, (read,_,_)) accu =
     let state = {
       opt_depr = depr;
       opt_value = read ();
@@ -514,14 +544,22 @@ let get_tables () =
   OptionMap.fold fold tables OptionMap.empty
 
 let print_tables () =
+  let open Pp in
   let print_option key value depr =
     let msg = str "  " ++ str (nickname key) ++ str ": " ++ msg_option_value value in
-    if depr then msg ++ str " [DEPRECATED]" ++ fnl ()
-    else msg ++ fnl ()
+    let depr = pr_opt (fun depr ->
+        hov 2
+          (str "[DEPRECATED" ++
+           pr_opt (fun since -> str "since " ++ str since) depr.Deprecation.since ++
+           pr_opt str depr.Deprecation.note ++
+           str "]"))
+        depr
+    in
+    msg ++ depr ++ fnl()
   in
   str "Options:" ++ fnl () ++
     OptionMap.fold
-      (fun key (depr, (read,_,_)) p ->
+      (fun key (depr, _stage, (read,_,_)) p ->
         p ++ print_option key (read ()) depr)
       !value_tab (mt ()) ++
   str "Tables:" ++ fnl () ++
