@@ -24,29 +24,17 @@ val make_oname : Nametab.object_prefix -> Names.Id.t -> Libobject.object_name
 val make_foname : Names.Id.t -> Libobject.object_name
 val oname_prefix : Libobject.object_name -> Nametab.object_prefix
 
-type node =
+type 'summary node =
   | CompilingLibrary of Nametab.object_prefix
-  | OpenedModule of is_type * export * Nametab.object_prefix * Summary.frozen
-  | OpenedSection of Nametab.object_prefix * Summary.frozen
+  | OpenedModule of is_type * export * Nametab.object_prefix * 'summary
+  | OpenedSection of Nametab.object_prefix * 'summary
 
 (** Extract the [object_prefix] component. Note that it is the prefix
    of the objects *inside* this node, eg in [Module M.] we have
    [OpenedModule] with prefix containing [M]. *)
-val node_prefix : node -> Nametab.object_prefix
+val node_prefix : 'summary node -> Nametab.object_prefix
 
-type library_segment = (node * Libobject.t list) list
-
-type classified_objects = {
-  substobjs : Libobject.t list;
-  keepobjs : Libobject.t list;
-  anticipateobjs : Libobject.t list;
-}
-
-(** {6 ... } *)
-(** Low-level adding operations (does not cache) *)
-
-val add_entry : node -> unit
-val add_leaf_entry : Libobject.t -> unit
+type 'summary library_segment = ('summary node * Libobject.t list) list
 
 (** {6 ... } *)
 (** Adding operations (which call the [cache] method, and getting the
@@ -58,11 +46,26 @@ val add_leaf : Libobject.obj -> unit
 
 (** The function [contents] gives access to the current entire segment *)
 
-val contents : unit -> library_segment
+val contents : unit -> Summary.Interp.frozen library_segment
 
 (** {6 Functions relative to current path } *)
 
-(** User-side names *)
+(** User-side names
+
+    [cwd()] is [(prefix()).obj_dir]
+    [current_mp()] is [(prefix()).obj_mp]
+
+    Inside a library A.B module M section S, we have
+    - library_dp = A.B
+    - cwd = A.B.M.S
+    - cwd_except_section = A.B.M
+    - current_dirpath true = M.S
+    - current_dirpath false = S
+    - current_mp = MPdot(MPfile A.B, M)
+
+    make_path (resp make_path_except_section) uses cwd (resp cwd_except_section)
+    make_kn uses current_mp
+*)
 val prefix : unit -> Nametab.object_prefix
 val cwd : unit -> DirPath.t
 val cwd_except_section : unit -> DirPath.t
@@ -76,7 +79,6 @@ val make_kn : Id.t -> KerName.t
 
 (** Are we inside an opened section *)
 val sections_are_opened : unit -> bool
-[@@ocaml.deprecated "Use Global.sections_are_opened"]
 val sections_depth : unit -> int
 
 (** Are we inside an opened module type *)
@@ -87,58 +89,80 @@ val is_modtype : unit -> bool
 val is_modtype_strict : unit -> bool
 val is_module : unit -> bool
 
-(** Returns the opening node of a given name *)
-val find_opening_node : Id.t -> node
+(** The [StagedLibS] abstraction describes operations and traversal for Lib at a
+    given stage. *)
+module type StagedLibS = sig
 
-(** {6 Modules and module types } *)
+  type summary
 
-val start_module :
-  export -> module_ident -> ModPath.t ->
-  Summary.frozen -> Nametab.object_prefix
+  type classified_objects = {
+    substobjs : Libobject.t list;
+    keepobjs : Libobject.t list;
+    anticipateobjs : Libobject.t list;
+  }
+  val classify_segment : Libobject.t list -> classified_objects
 
-val start_modtype :
-  module_ident -> ModPath.t ->
-  Summary.frozen -> Nametab.object_prefix
+  (** Returns the opening node of a given name *)
+  val find_opening_node : Id.t -> summary node
 
-val end_module :
-  unit ->
-  Nametab.object_prefix * Summary.frozen * classified_objects
+  val add_entry : summary node -> unit
+  val add_leaf_entry : Libobject.t -> unit
 
-val end_modtype :
-  unit ->
-  Nametab.object_prefix * Summary.frozen * classified_objects
+  (** {6 Sections } *)
+  val open_section : Id.t -> unit
+  val close_section : unit -> unit
+
+  (** {6 Modules and module types } *)
+
+  val start_module :
+    export -> module_ident -> ModPath.t ->
+    summary -> Nametab.object_prefix
+
+  val start_modtype :
+    module_ident -> ModPath.t ->
+    summary -> Nametab.object_prefix
+
+  val end_module :
+    unit ->
+    Nametab.object_prefix * summary * classified_objects
+
+  val end_modtype :
+    unit ->
+    Nametab.object_prefix * summary * classified_objects
+
+  type frozen
+
+  val freeze : unit -> frozen
+  val unfreeze : frozen -> unit
+  val init : unit -> unit
+
+  (** Keep only the libobject structure, not the objects themselves *)
+  val drop_objects : frozen -> frozen
+
+end
+
+(** We provide two instances of [StagedLibS], corresponding to the Synterp and
+    Interp stages. *)
+
+module Synterp : StagedLibS with type summary = Summary.Synterp.frozen
+module Interp : StagedLibS with type summary = Summary.Interp.frozen
 
 (** {6 Compilation units } *)
 
 val start_compilation : DirPath.t -> ModPath.t -> unit
-val end_compilation : DirPath.t -> Nametab.object_prefix * classified_objects
+
+(** Finalize the compilation of a library and return respectively the library
+    prefix, the regular objects, and the syntax-related objects. *)
+val end_compilation : DirPath.t -> Nametab.object_prefix * Interp.classified_objects * Synterp.classified_objects
 
 (** The function [library_dp] returns the [DirPath.t] of the current
    compiling library (or [default_library]) *)
 val library_dp : unit -> DirPath.t
 
 (** Extract the library part of a name even if in a section *)
-val dp_of_mp : ModPath.t -> DirPath.t
 val split_modpath : ModPath.t -> DirPath.t * Id.t list
 val library_part :  GlobRef.t -> DirPath.t
 
-(** {6 Sections } *)
-
-val open_section : Id.t -> unit
-val close_section : unit -> unit
-
-(** {6 We can get and set the state of the operations (used in [States]). } *)
-
-type frozen
-
-val freeze : unit -> frozen
-val unfreeze : frozen -> unit
-
-(** Keep only the libobject structure, not the objects themselves *)
-val drop_objects : frozen -> frozen
-
-val init : unit -> unit
-
 (** {6 Section management for discharge } *)
 val section_segment_of_constant : Constant.t -> Cooking.cooking_info
 val section_segment_of_inductive: MutInd.t -> Cooking.cooking_info
@@ -150,3 +174,10 @@ val is_in_section : GlobRef.t -> bool
 (** {6 Discharge: decrease the section level if in the current section } *)
 
 val discharge_proj_repr : Projection.Repr.t -> Projection.Repr.t
+
+(** Compatibility layer *)
+
+val init : unit -> unit
+
+val open_section : Id.t -> unit
+val close_section : unit -> unit
