@@ -10,10 +10,6 @@
 
 open EConstr
 
-val proof_tac: Ccproof.proof -> unit Proofview.tactic
-
-val cc_tactic : int -> constr list -> bool -> unit Proofview.tactic
-
 val congruence_tac : int -> constr list -> unit Proofview.tactic
 
 val simple_congruence_tac : int -> constr list -> unit Proofview.tactic
