@@ -61,14 +61,15 @@ let raw_string_of_modfile = function
   | MPfile f -> String.capitalize_ascii (Id.to_string (List.hd (DirPath.repr f)))
   | _ -> assert false
 
-let is_toplevel mp =
-  ModPath.equal mp ModPath.initial || ModPath.equal mp (Lib.current_mp ())
+let extraction_current_mp () = fst (Safe_typing.flatten_env (Global.safe_env ()))
+
+let is_toplevel mp = ModPath.equal mp (extraction_current_mp ())
 
 let at_toplevel mp =
   is_modfile mp || is_toplevel mp
 
 let mp_length mp =
-  let mp0 = Lib.current_mp () in
+  let mp0 = extraction_current_mp () in
   let rec len = function
     | mp when ModPath.equal mp mp0 -> 1
     | MPdot (mp,_) -> 1 + len mp
@@ -90,15 +91,13 @@ let common_prefix_from_list mp0 mpl =
     | mp :: l -> if MPset.mem mp prefixes then Some mp else f l
   in f mpl
 
-let rec parse_labels2 ll mp1 = function
-  | mp when ModPath.equal mp1 mp -> mp,ll
-  | MPdot (mp,l) -> parse_labels2 (l::ll) mp1 mp
+let rec parse_labels2 ll = function
+  | MPdot (mp,l) -> parse_labels2 (l::ll) mp
   | mp -> mp,ll
 
 let labels_of_ref r =
-  let mp_top = Lib.current_mp () in
   let mp,l = KerName.repr (repr_of_r r) in
-  parse_labels2 [l] mp_top mp
+  parse_labels2 [l] mp
 
 
 (*S The main tables: constants, inductives, records, ... *)
@@ -305,7 +304,7 @@ let pr_long_global ref = pr_path (Nametab.path_of_global ref)
 let err ?loc s = user_err ?loc s
 
 let warn_extraction_axiom_to_realize =
-  CWarnings.create ~name:"extraction-axiom-to-realize" ~category:"extraction"
+  CWarnings.create ~name:"extraction-axiom-to-realize" ~category:CWarnings.CoreCategories.extraction
          (fun axioms ->
           let s = if Int.equal (List.length axioms) 1 then "axiom" else "axioms" in
           strbrk ("The following "^s^" must be realized in the extracted code:")
@@ -313,7 +312,7 @@ let warn_extraction_axiom_to_realize =
                    ++ str "." ++ fnl ())
 
 let warn_extraction_logical_axiom =
-  CWarnings.create ~name:"extraction-logical-axiom" ~category:"extraction"
+  CWarnings.create ~name:"extraction-logical-axiom" ~category:CWarnings.CoreCategories.extraction
          (fun axioms ->
           let s =
             if Int.equal (List.length axioms) 1 then "axiom was" else "axioms were"
@@ -333,13 +332,13 @@ let warning_axioms () =
     warn_extraction_logical_axiom log_axioms
 
 let warn_extraction_opaque_accessed =
-  CWarnings.create ~name:"extraction-opaque-accessed" ~category:"extraction"
+  CWarnings.create ~name:"extraction-opaque-accessed" ~category:CWarnings.CoreCategories.extraction
     (fun lst -> strbrk "The extraction is currently set to bypass opacity, " ++
                   strbrk "the following opaque constant bodies have been accessed :" ++
                   lst ++ str "." ++ fnl ())
 
 let warn_extraction_opaque_as_axiom =
-  CWarnings.create ~name:"extraction-opaque-as-axiom" ~category:"extraction"
+  CWarnings.create ~name:"extraction-opaque-as-axiom" ~category:CWarnings.CoreCategories.extraction
     (fun lst -> strbrk "The extraction now honors the opacity constraints by default, " ++
          strbrk "the following opaque constants have been extracted as axioms :" ++
          lst ++ str "." ++ fnl () ++
@@ -354,7 +353,7 @@ let warning_opaques accessed =
     else warn_extraction_opaque_as_axiom lst
 
 let warning_ambiguous_name =
-  CWarnings.create ~name:"extraction-ambiguous-name" ~category:"extraction"
+  CWarnings.create ~name:"extraction-ambiguous-name" ~category:CWarnings.CoreCategories.extraction
     (fun (q,mp,r) -> strbrk "The name " ++ pr_qualid q ++ strbrk " is ambiguous, " ++
                        strbrk "do you mean module " ++
                        pr_long_mp mp ++
@@ -368,26 +367,13 @@ let error_axiom_scheme ?loc r i =
        safe_pr_global r ++ spc () ++ str "needs " ++ int i ++
        str " type variable(s).")
 
-let warn_extraction_inside_module =
-  CWarnings.create ~name:"extraction-inside-module" ~category:"extraction"
-      (fun () -> strbrk "Extraction inside an opened module is experimental." ++ spc () ++
-       strbrk "In case of problem, close it first.")
-
-
-let check_inside_module () =
-  if Lib.is_modtype () then
-    err (str "You can't do that within a Module Type." ++ fnl () ++
-         str "Close it and try again.")
-  else if Lib.is_module () then
-    warn_extraction_inside_module ()
-
 let check_inside_section () =
-  if Global.sections_are_opened () then
+  if Lib.sections_are_opened () then
     err (str "You can't do that within a section." ++ fnl () ++
          str "Close it and try again.")
 
 let warn_extraction_reserved_identifier =
-  CWarnings.create ~name:"extraction-reserved-identifier" ~category:"extraction"
+  CWarnings.create ~name:"extraction-reserved-identifier" ~category:CWarnings.CoreCategories.extraction
     (fun s -> strbrk ("The identifier "^s^
                 " contains __ which is reserved for the extraction"))
 
@@ -471,7 +457,7 @@ let error_remaining_implicit k =
        fnl() ++ str "the extraction of unsafe code and review it manually.")
 
 let warn_extraction_remaining_implicit =
-  CWarnings.create ~name:"extraction-remaining-implicit" ~category:"extraction"
+  CWarnings.create ~name:"extraction-remaining-implicit" ~category:CWarnings.CoreCategories.extraction
     (fun s -> strbrk ("At least an implicit occurs after extraction : "^s^".") ++ fnl () ++
      strbrk "Extraction SafeImplicits is unset, extracting nonetheless,"
      ++ strbrk "but this code is potentially unsafe, please review it manually.")
@@ -483,7 +469,7 @@ let warning_remaining_implicit k =
 let check_loaded_modfile mp = match base_mp mp with
   | MPfile dp ->
       if not (Library.library_is_loaded dp) then begin
-        match base_mp (Lib.current_mp ()) with
+        match base_mp (extraction_current_mp ()) with
           | MPfile dp' when not (DirPath.equal dp dp') ->
             err (str "Please load library " ++ DirPath.print dp ++ str " first.")
           | _ -> ()
@@ -501,7 +487,13 @@ let info_file f =
    so we register them to coq save/undo mechanism. *)
 
 let my_bool_option name value =
-  declare_bool_option_and_ref ~depr:false ~key:["Extraction"; name] ~value
+  let { Goptions.get } =
+    declare_bool_option_and_ref
+    ~key:["Extraction"; name]
+    ~value
+    ()
+  in
+  get
 
 (*s Extraction AccessOpaque *)
 
@@ -567,13 +559,15 @@ let chg_flag n = int_flag_ref := n; opt_flag_ref := flag_of_int n
 let optims () = !opt_flag_ref
 
 let () = declare_bool_option
-          {optdepr = false;
+          {optstage = Summary.Stage.Interp;
+           optdepr = None;
            optkey = ["Extraction"; "Optimize"];
            optread = (fun () -> not (Int.equal !int_flag_ref 0));
            optwrite = (fun b -> chg_flag (if b then int_flag_init else 0))}
 
 let () = declare_int_option
-          { optdepr = false;
+          { optstage = Summary.Stage.Interp;
+            optdepr = None;
             optkey = ["Extraction";"Flag"];
             optread = (fun _ -> Some !int_flag_ref);
             optwrite = (function
@@ -582,18 +576,18 @@ let () = declare_int_option
 
 (* This option controls whether "dummy lambda" are removed when a
    toplevel constant is defined. *)
-let conservative_types =
+let { Goptions.get = conservative_types } =
   declare_bool_option_and_ref
-    ~depr:false
     ~key:["Extraction"; "Conservative"; "Types"]
     ~value:false
+    ()
 
 (* Allows to print a comment at the beginning of the output files *)
-let file_comment =
+let { Goptions.get = file_comment } =
   declare_string_option_and_ref
-    ~depr:false
     ~key:["Extraction"; "File"; "Comment"]
     ~value:""
+    ()
 
 (*s Extraction Lang *)
 
