@@ -8,16 +8,8 @@
 (*         *     (see LICENSE file for the text of the license)         *)
 (************************************************************************)
 
-open Names
 open Rules
 
-val collect_quantified : Evd.evar_map -> Sequent.t -> Formula.t list * Sequent.t
-
-val give_instances : Environ.env -> Evd.evar_map -> Formula.t list -> Sequent.t ->
-  (Unify.instance * GlobRef.t) list
-
-val quantified_tac : Formula.t list -> seqtac with_backtracking
-
-
-
+val collect_quantified : Evd.evar_map -> Sequent.t -> Formula.any_formula list * Sequent.t
 
+val quantified_tac : flags:Formula.flags -> Formula.any_formula list -> seqtac with_backtracking
