@@ -13,18 +13,21 @@ open CErrors
 open Util
 open Constr
 open Context
+open Termops
 open EConstr
 open Vars
 open Namegen
 open Names
 open Pp
-open Termops
 open Tactics
 open Indfun_common
 open Libnames
 open Context.Rel.Declaration
 module RelDecl = Context.Rel.Declaration
 
+(* funind does not support univ poly *)
+open UnsafeMonomorphic
+
 let list_chop ?(msg = "") n l =
   try List.chop n l with Failure msg' -> failwith (msg ^ msg')
 
@@ -45,10 +48,9 @@ type 'a dynamic_info =
 type body_info = constr dynamic_info
 
 let observe_tac s =
-  New.observe_tac ~header:(str "observation") (fun _ _ -> Pp.str s)
+  observe_tac ~header:(str "observation") (fun _ _ -> Pp.str s)
 
 let finish_proof dynamic_infos = observe_tac "finish" assumption
-let refine c = Logic.refiner ~check:true EConstr.Unsafe.(to_constr c)
 let thin = clear
 let eq_constr sigma u v = EConstr.eq_constr_nounivs sigma u v
 
@@ -67,7 +69,7 @@ let is_trivial_eq sigma t =
   res
 
 let rec incompatible_constructor_terms sigma t1 t2 =
-  let c1, arg1 = decompose_app sigma t1 and c2, arg2 = decompose_app sigma t2 in
+  let c1, arg1 = decompose_app_list sigma t1 and c2, arg2 = decompose_app_list sigma t2 in
   (not (eq_constr sigma t1 t2))
   && isConstruct sigma c1 && isConstruct sigma c2
   && ( (not (eq_constr sigma c1 c2))
@@ -122,10 +124,10 @@ let prove_trivial_eq h_id context (constructor, type_of_term, term) =
             :: List.map mkVar context_hyps
           in
           let to_refine = applist (mkVar h_id, List.rev context_hyps') in
-          refine to_refine) ]
+          Tactics.exact_check to_refine) ]
 
 let find_rectype env sigma c =
-  let t, l = decompose_app sigma (Reductionops.whd_betaiotazeta env sigma c) in
+  let t, l = decompose_app_list sigma (Reductionops.whd_betaiotazeta env sigma c) in
   match EConstr.kind sigma t with
   | Ind ind -> (t, l)
   | Construct _ -> (t, l)
@@ -250,7 +252,7 @@ let change_eq env sigma hyp_id (context : rel_context) x t end_of_type =
   in
   let new_type_of_hyp = Reductionops.nf_betaiota env sigma new_type_of_hyp in
   let new_ctxt, new_end_of_type =
-    decompose_prod_n_assum sigma ctxt_size new_type_of_hyp
+    decompose_prod_n_decls sigma ctxt_size new_type_of_hyp
   in
   let prove_new_hyp =
     let open Tacticals in
@@ -264,7 +266,7 @@ let change_eq env sigma hyp_id (context : rel_context) x t end_of_type =
              Typing.type_of (Proofview.Goal.env g) (Proofview.Goal.sigma g)
                to_refine
            in
-           tclTHEN (Proofview.Unsafe.tclEVARS evm) (refine to_refine)))
+           tclTHEN (Proofview.Unsafe.tclEVARS evm) (Tactics.exact_check to_refine)))
   in
   let simpl_eq_tac =
     change_hyp_with_using "prove_pattern_simplification" hyp_id new_type_of_hyp
@@ -353,7 +355,7 @@ let clean_hyp_with_heq ptes_infos eq_hyps hyp_id env sigma =
       in
       (* length of context didn't change ? *)
       let new_context, new_typ_of_hyp =
-        decompose_prod_n_assum sigma (List.length context) reduced_type_of_hyp
+        decompose_prod_n_decls sigma (List.length context) reduced_type_of_hyp
       in
       tclTHENLIST
         [ h_reduce_with_zeta (Locusops.onHyp hyp_id)
@@ -391,7 +393,7 @@ let clean_hyp_with_heq ptes_infos eq_hyps hyp_id env sigma =
                     [ (* observe_tac "prove rec hyp" *)
                       prove_rec_hyp eq_hyps
                     ; (*                      observe_tac "prove rec hyp" *)
-                      refine to_refine ]) ]
+                      Tactics.exact_check to_refine ]) ]
         in
         tclTHENLIST
           [ (*              observe_tac "hyp rec"  *)
@@ -429,7 +431,7 @@ let clean_hyp_with_heq ptes_infos eq_hyps hyp_id env sigma =
                       ( mkVar hyp_id
                       , List.rev (coq_I :: List.map mkVar context_hyps) )
                   in
-                  refine to_refine) ]
+                  Tactics.exact_check to_refine) ]
         in
         tclTHENLIST
           [ change_hyp_with_using "prove_trivial" hyp_id real_type_of_hyp
@@ -663,7 +665,7 @@ let build_proof (interactive_proof : bool) (fnames : Constant.t list) ptes_infos
          |Int _ | Float _ ->
           do_finalize dyn_infos
         | App (_, _) -> (
-          let f, args = decompose_app sigma dyn_infos.info in
+          let f, args = decompose_app_list sigma dyn_infos.info in
           match EConstr.kind sigma f with
           | Int _ -> user_err Pp.(str "integer cannot be applied")
           | Float _ -> user_err Pp.(str "float cannot be applied")
@@ -675,7 +677,7 @@ let build_proof (interactive_proof : bool) (fnames : Constant.t list) ptes_infos
            |Prod _ ->
             let new_infos = {dyn_infos with info = (f, args)} in
             build_proof_args env sigma do_finalize new_infos
-          | Const (c, _) when not (List.mem_f Constant.CanOrd.equal c fnames) ->
+          | Const (c, _) when not (List.exists (fun kn -> Environ.QConstant.equal env c kn) fnames) ->
             let new_infos = {dyn_infos with info = (f, args)} in
             (*                    Pp.msgnl (str "proving in " ++ pr_lconstr_env (pf_env g) dyn_infos.info); *)
             build_proof_args env sigma do_finalize new_infos
@@ -720,7 +722,7 @@ let build_proof (interactive_proof : bool) (fnames : Constant.t list) ptes_infos
         | Array _ -> CErrors.user_err Pp.(str "Arrays not handled yet"))
   and build_proof do_finalize dyn_infos =
     (*     observe (str "proving with "++Printer.pr_lconstr dyn_infos.info++ str " on goal " ++ pr_gls g); *)
-    Indfun_common.New.observe_tac ~header:(str "observation")
+    Indfun_common.observe_tac ~header:(str "observation")
       (fun env sigma ->
         str "build_proof with " ++ pr_leconstr_env env sigma dyn_infos.info)
       (build_proof_aux do_finalize dyn_infos)
@@ -774,7 +776,7 @@ let prove_rec_hyp_for_struct fix_info eq_hyps =
          let rec_hyp_proof =
            mkApp (mkVar fix_info.name, array_get_start pte_args)
          in
-         refine rec_hyp_proof))
+         Tactics.exact_check rec_hyp_proof))
 
 let prove_rec_hyp fix_info =
   {proving_tac = prove_rec_hyp_for_struct fix_info; is_valid = (fun _ -> true)}
@@ -831,7 +833,7 @@ let generate_equation_lemma env evd fnames f fun_num nb_params nb_args rec_args_
     Option.get (Global.body_of_constant_body Library.indirect_accessor f_def)
   in
   let f_body = EConstr.of_constr f_body in
-  let params, f_body_with_params = decompose_lam_n evd nb_params f_body in
+  let params, f_body_with_params = decompose_lambda_n evd nb_params f_body in
   let (_, num), (_, _, bodies) = destFix evd f_body_with_params in
   let fnames_with_params =
     let params = Array.init nb_params (fun i -> mkRel (nb_params - i)) in
@@ -856,7 +858,7 @@ let generate_equation_lemma env evd fnames f fun_num nb_params nb_args rec_args_
   (*   observe (str "eq_rhs " ++  pr_lconstr eq_rhs); *)
   let (type_ctxt, type_of_f), evd =
     let evd, t = Typing.type_of ~refresh:true env evd f in
-    (decompose_prod_n_assum evd (nb_params + nb_args) t, evd)
+    (decompose_prod_n_decls evd (nb_params + nb_args) t, evd)
   in
   let eqn = mkApp (Lazy.force eq, [|type_of_f; eq_lhs; eq_rhs|]) in
   let lemma_type = it_mkProd_or_LetIn eqn type_ctxt in
@@ -999,7 +1001,7 @@ let prove_princ_for_struct (evd : Evd.evar_map ref) interactive_proof fun_num
         | None -> user_err Pp.(str "Cannot define a principle over an axiom ")
       in
       let fbody = get_body fnames.(fun_num) in
-      let f_ctxt, f_body = decompose_lam sigma fbody in
+      let f_ctxt, f_body = decompose_lambda sigma fbody in
       let f_ctxt_length = List.length f_ctxt in
       let diff_params = princ_info.nparams - f_ctxt_length in
       let full_params, princ_params, fbody_with_full_params =
@@ -1070,7 +1072,7 @@ let prove_princ_for_struct (evd : Evd.evar_map ref) interactive_proof fun_num
                 ; types
                 ; offset = fix_offset
                 ; nb_realargs =
-                    List.length (fst (decompose_lam sigma bodies.(i)))
+                    List.length (fst (decompose_lambda sigma bodies.(i)))
                     - fix_offset
                 ; body_with_param = bodies_with_all_params.(i)
                 ; num_in_block = i })
@@ -1144,7 +1146,7 @@ let prove_princ_for_struct (evd : Evd.evar_map ref) interactive_proof fun_num
             if this_fix_info.idx + 1 = 0 then Proofview.tclUNIT ()
               (* Someone  tries to defined a principle on a fully parametric definition declared as a fixpoint (strange but ....) *)
             else
-              Indfun_common.New.observe_tac ~header:(str "observation")
+              Indfun_common.observe_tac ~header:(str "observation")
                 (fun _ _ -> str "h_fix " ++ int (this_fix_info.idx + 1))
                 (fix this_fix_info.name (this_fix_info.idx + 1))
           else
@@ -1169,8 +1171,8 @@ let prove_princ_for_struct (evd : Evd.evar_map ref) interactive_proof fun_num
         Proofview.Goal.enter (fun gl ->
             let sigma = Proofview.Goal.sigma gl in
             let ccl = Proofview.Goal.concl gl in
-            let ctxt, pte_app = decompose_prod_assum sigma ccl in
-            let pte, pte_args = decompose_app sigma pte_app in
+            let ctxt, pte_app = decompose_prod_decls sigma ccl in
+            let pte, pte_args = decompose_app_list sigma pte_app in
             try
               let pte =
                 try destVar sigma pte
@@ -1257,7 +1259,7 @@ let prove_princ_for_struct (evd : Evd.evar_map ref) interactive_proof fun_num
                       let fname =
                         destConst sigma
                           (fst
-                             (decompose_app sigma (List.hd (List.rev pte_args))))
+                             (decompose_app_list sigma (List.hd (List.rev pte_args))))
                       in
                       tclTHENLIST
                         [ unfold_in_concl
