@@ -14,7 +14,7 @@ open Indfun_common
 module RelDecl = Context.Rel.Declaration
 
 let observe_tac s =
-  New.observe_tac ~header:(Pp.str "observation") (fun _ _ -> Pp.str s)
+  observe_tac ~header:(Pp.str "observation") (fun _ _ -> Pp.str s)
 
 (*
    Construct a fixpoint as a Glob_term
@@ -83,7 +83,7 @@ let is_rec names =
   let rec lookup names gt =
     match DAst.get gt with
     | GVar id -> check_id id names
-    | GRef _ | GEvar _ | GPatVar _ | GSort _ | GHole _ | GInt _ | GFloat _ ->
+    | GRef _ | GEvar _ | GPatVar _ | GSort _ | GHole _ | GGenarg _ | GInt _ | GFloat _ ->
       false
     | GCast (b, _, _) -> lookup names b
     | GRec _ -> CErrors.user_err (Pp.str "GRec not handled")
@@ -204,7 +204,7 @@ let build_functional_principle env (sigma : Evd.evar_map) old_princ_type sorts f
   let new_principle_type =
     Functional_principles_types.compute_new_princ_type_from_rel (Global.env ())
       (Array.map Constr.mkConstU funs)
-      sorts old_princ_type
+      (Array.map (fun s -> EConstr.ESorts.kind sigma s) sorts) old_princ_type
   in
   let sigma, _ =
     Typing.type_of ~refresh:true env sigma
@@ -223,6 +223,7 @@ let build_functional_principle env (sigma : Evd.evar_map) old_princ_type sorts f
 
 let change_property_sort evd toSort princ princName =
   let open Context.Rel.Declaration in
+  let toSort = EConstr.ESorts.kind evd toSort in
   let princ = EConstr.of_constr princ in
   let princ_info = Tactics.compute_elim_sig evd princ in
   let change_sort_in_predicate decl =
@@ -273,7 +274,7 @@ let generate_functional_principle (evd : Evd.evar_map ref) old_princ_type sorts
       | Some id -> (id, id)
       | None ->
         let id_of_f = Label.to_id (Constant.label (fst f)) in
-        (id_of_f, Indrec.make_elimination_ident id_of_f (Sorts.family type_sort))
+        (id_of_f, Indrec.make_elimination_ident id_of_f (EConstr.ESorts.family !evd type_sort))
     in
     let names = ref [new_princ_name] in
     let hook new_principle_type _ =
@@ -464,7 +465,7 @@ let generate_type evd g_to_f f graph =
   evd := evd';
   let sigma, graph_arity = Typing.type_of (Global.env ()) !evd graph in
   evd := sigma;
-  let ctxt, _ = decompose_prod_assum !evd graph_arity in
+  let ctxt, _ = decompose_prod_decls !evd graph_arity in
   let fun_ctxt, res_type =
     match ctxt with
     | [] | [_] -> CErrors.anomaly (Pp.str "Not a valid context.")
@@ -632,7 +633,7 @@ let prove_fun_correct evd graphs_constr schemes lemmas_types_infos i :
                 CAst.make @@ Tactypes.IntroNaming (Namegen.IntroIdentifier id))
               (generate_fresh_id (Id.of_string "y") ids
                  (List.length
-                    (fst (decompose_prod_assum evd (RelDecl.get_type decl))))))
+                    (fst (decompose_prod_decls evd (RelDecl.get_type decl))))))
           branches
       in
       (* before building the full intro pattern for the principle *)
@@ -1027,19 +1028,14 @@ let rec reflexivity_with_destruct_cases () =
                   with
                   | App (eq, [|_; t1; t2|])
                     when EConstr.eq_constr (Proofview.Goal.sigma g) eq eq_ind ->
-                    if
-                      Equality.discriminable (Proofview.Goal.env g)
-                        (Proofview.Goal.sigma g) t1 t2
-                    then Equality.discrHyp id
-                    else if
-                      Equality.injectable (Proofview.Goal.env g)
-                        (Proofview.Goal.sigma g) ~keep_proofs:None t1 t2
-                    then
+                    tclFIRST [
+                      Equality.discrHyp id;
                       tclTHENLIST
                         [ Equality.injHyp my_inj_flags ~injection_in_context:false None id
                         ; thin [id]
-                        ; intros_with_rewrite () ]
-                    else Proofview.tclUNIT ()
+                        ; intros_with_rewrite () ];
+                      Proofview.tclUNIT ()
+                    ]
                   | _ -> Proofview.tclUNIT ()))
       in
       tclFIRST
@@ -1137,7 +1133,7 @@ let prove_fun_complete funcs graphs schemes lemmas_types_infos i :
               in
               tclTHENLIST
                 [ tclMAP Simple.intro ids
-                ; Equality.rewriteLR (mkConst eq_lemma)
+                ; Equality.rewriteLR (UnsafeMonomorphic.mkConst eq_lemma)
                 ; (* Don't forget to $\zeta$ normlize the term since the principles
                      have been $\zeta$-normalized *)
                   reduce
@@ -1246,7 +1242,7 @@ let get_funs_constant mp =
       List.map find_constant_body (Array.to_list (Array.map fst l_const))
     in
     let l_params, _l_fixes =
-      List.split (List.map Term.decompose_lam l_bodies)
+      List.split (List.map Term.decompose_lambda l_bodies)
     in
     (* all the parameters must be equal*)
     let _check_params =
@@ -1373,12 +1369,13 @@ let make_scheme evd (fas : (Constr.pconstant * Sorts.family) list) : _ list =
     let other_fun_princ_types =
       let funs = Array.map Constr.mkConstU this_block_funs in
       let sorts = Array.of_list sorts in
+      let sorts = Array.map (fun s -> EConstr.ESorts.kind sigma s) sorts in
       List.map
         (Functional_principles_types.compute_new_princ_type_from_rel (Global.env ()) funs sorts)
         other_princ_types
     in
     let first_princ_body = body in
-    let ctxt, fix = Term.decompose_lam_assum first_princ_body in
+    let ctxt, fix = Term.decompose_lambda_decls first_princ_body in
     (* the principle has for forall ...., fix .*)
     let (idxs, _), ((_, ta, _) as decl) = Constr.destFix fix in
     let other_result =
@@ -1386,18 +1383,18 @@ let make_scheme evd (fas : (Constr.pconstant * Sorts.family) list) : _ list =
         (fun scheme_type ->
           incr i;
           observe (Printer.pr_lconstr_env env sigma scheme_type);
-          let type_concl = Term.strip_prod_assum scheme_type in
+          let type_concl = Term.strip_prod_decls scheme_type in
           let applied_f =
-            List.hd (List.rev (snd (Constr.decompose_app type_concl)))
+            List.hd (List.rev (snd (Constr.decompose_app_list type_concl)))
           in
           let f = fst (Constr.decompose_app applied_f) in
           try
             (* we search the number of the function in the fix block (name of the function) *)
             Array.iteri
               (fun j t ->
-                let t = Term.strip_prod_assum t in
+                let t = Term.strip_prod_decls t in
                 let applied_g =
-                  List.hd (List.rev (snd (Constr.decompose_app t)))
+                  List.hd (List.rev (snd (Constr.decompose_app_list t)))
                 in
                 let g = fst (Constr.decompose_app applied_g) in
                 if Constr.equal f g then raise (Found_type j);
@@ -1423,7 +1420,7 @@ let make_scheme evd (fas : (Constr.pconstant * Sorts.family) list) : _ list =
             (body, typ, univs, opaque)
           with Found_type i ->
             let princ_body =
-              Termops.it_mkLambda_or_LetIn (Constr.mkFix ((idxs, i), decl)) ctxt
+              Term.it_mkLambda_or_LetIn (Constr.mkFix ((idxs, i), decl)) ctxt
             in
             (princ_body, Some scheme_type, univs, opaque))
         other_fun_princ_types
@@ -1447,7 +1444,7 @@ let derive_correctness (funs : Constr.pconstant list) (graphs : inductive list)
     (fun () ->
       let env = Global.env () in
       let evd = ref (Evd.from_env env) in
-      let graphs_constr = Array.map mkInd graphs in
+      let graphs_constr = Array.map UnsafeMonomorphic.mkInd graphs in
       let lemmas_types_infos =
         Util.Array.map2_i
           (fun i f_constr graph ->
@@ -1591,7 +1588,7 @@ let derive_correctness (funs : Constr.pconstant list) (graphs : inductive list)
     ()
 
 let warn_funind_cannot_build_inversion =
-  CWarnings.create ~name:"funind-cannot-build-inversion" ~category:"funind"
+  CWarnings.create ~name:"funind-cannot-build-inversion" ~category:CWarnings.CoreCategories.funind
     Pp.(
       fun e' ->
         strbrk "Cannot build inversion information"
@@ -1849,12 +1846,12 @@ let do_generate_principle_aux pconstants on_error register_built
   lemma
 
 let warn_cannot_define_graph =
-  CWarnings.create ~name:"funind-cannot-define-graph" ~category:"funind"
+  CWarnings.create ~name:"funind-cannot-define-graph" ~category:CWarnings.CoreCategories.funind
     (fun (names, error) ->
       Pp.(strbrk "Cannot define graph(s) for " ++ hv 1 names ++ error))
 
 let warn_cannot_define_principle =
-  CWarnings.create ~name:"funind-cannot-define-principle" ~category:"funind"
+  CWarnings.create ~name:"funind-cannot-define-principle" ~category:CWarnings.CoreCategories.funind
     (fun (names, error) ->
       Pp.(
         strbrk "Cannot define induction principle(s) for " ++ hv 1 names ++ error))
@@ -2017,7 +2014,7 @@ let rec add_args id new_args =
         , (na, Option.map (add_args id new_args) b_option)
         , add_args id new_args b2
         , add_args id new_args b3 )
-    | (CHole _ | CPatVar _ | CEvar _ | CPrim _ | CSort _) as b -> b
+    | (CHole _ | CGenarg _ | CGenargGlob _ | CPatVar _ | CEvar _ | CPrim _ | CSort _) as b -> b
     | CCast (b1, k, b2) ->
       CCast (add_args id new_args b1, k, add_args id new_args b2)
     | CRecord pars ->
@@ -2057,7 +2054,7 @@ let make_graph (f_ref : GlobRef.t) =
         CErrors.user_err
           Pp.(
             str "Cannot find "
-            ++ Printer.pr_leconstr_env env sigma (EConstr.mkConst c))
+            ++ Termops.pr_global_env env (ConstRef c))
     | _ -> CErrors.user_err Pp.(str "Not a function reference")
   in
   match Global.body_of_constant_body Library.indirect_accessor c_body with
@@ -2228,10 +2225,11 @@ let build_case_scheme fa =
     let ind = (first_fun_kn, funs_indexes) in
     ((ind, Univ.Instance.empty) (*FIXME*), prop_sort)
   in
-  let sigma, scheme, scheme_type =
+  let sigma, scheme =
     Indrec.build_case_analysis_scheme_default env sigma ind sf
   in
-  let sorts = (fun (_, _, x) -> fst @@ UnivGen.fresh_sort_in_family x) fa in
+  let scheme, scheme_type = Indrec.eval_case_analysis scheme in
+  let sorts = (fun (_, _, x) -> EConstr.ESorts.make @@ fst @@ UnivGen.fresh_sort_in_family x) fa in
   let princ_name = (fun (x, _, _) -> x) fa in
   let (_ : unit) =
     (* Pp.msgnl (str "Generating " ++ Ppconstr.pr_id princ_name ++str " with " ++
@@ -2240,7 +2238,7 @@ let build_case_scheme fa =
     *)
     generate_functional_principle
       (ref (Evd.from_env (Global.env ())))
-      scheme_type
+      (EConstr.Unsafe.to_constr scheme_type)
       (Some [|sorts|])
       (Some princ_name) this_block_funs 0
       (Functional_principles_proofs.prove_princ_for_struct
