@@ -24,36 +24,36 @@ open Tac2intern
 
 module Pltac =
 struct
-let ltac2_expr = Pcoq.Entry.create "ltac2_expr"
-let tac2expr_in_env = Pcoq.Entry.create "tac2expr_in_env"
+let ltac2_expr = Pcoq.Entry.make "ltac2_expr"
+let tac2expr_in_env = Pcoq.Entry.make "tac2expr_in_env"
 
-let q_ident = Pcoq.Entry.create "q_ident"
-let q_bindings = Pcoq.Entry.create "q_bindings"
-let q_with_bindings = Pcoq.Entry.create "q_with_bindings"
-let q_intropattern = Pcoq.Entry.create "q_intropattern"
-let q_intropatterns = Pcoq.Entry.create "q_intropatterns"
-let q_destruction_arg = Pcoq.Entry.create "q_destruction_arg"
-let q_induction_clause = Pcoq.Entry.create "q_induction_clause"
-let q_conversion = Pcoq.Entry.create "q_conversion"
-let q_rewriting = Pcoq.Entry.create "q_rewriting"
-let q_clause = Pcoq.Entry.create "q_clause"
-let q_dispatch = Pcoq.Entry.create "q_dispatch"
-let q_occurrences = Pcoq.Entry.create "q_occurrences"
-let q_reference = Pcoq.Entry.create "q_reference"
-let q_strategy_flag = Pcoq.Entry.create "q_strategy_flag"
-let q_constr_matching = Pcoq.Entry.create "q_constr_matching"
-let q_goal_matching = Pcoq.Entry.create "q_goal_matching"
-let q_hintdb = Pcoq.Entry.create "q_hintdb"
-let q_move_location = Pcoq.Entry.create "q_move_location"
-let q_pose = Pcoq.Entry.create "q_pose"
-let q_assert = Pcoq.Entry.create "q_assert"
+let q_ident = Pcoq.Entry.make "q_ident"
+let q_bindings = Pcoq.Entry.make "q_bindings"
+let q_with_bindings = Pcoq.Entry.make "q_with_bindings"
+let q_intropattern = Pcoq.Entry.make "q_intropattern"
+let q_intropatterns = Pcoq.Entry.make "q_intropatterns"
+let q_destruction_arg = Pcoq.Entry.make "q_destruction_arg"
+let q_induction_clause = Pcoq.Entry.make "q_induction_clause"
+let q_conversion = Pcoq.Entry.make "q_conversion"
+let q_rewriting = Pcoq.Entry.make "q_rewriting"
+let q_clause = Pcoq.Entry.make "q_clause"
+let q_dispatch = Pcoq.Entry.make "q_dispatch"
+let q_occurrences = Pcoq.Entry.make "q_occurrences"
+let q_reference = Pcoq.Entry.make "q_reference"
+let q_strategy_flag = Pcoq.Entry.make "q_strategy_flag"
+let q_constr_matching = Pcoq.Entry.make "q_constr_matching"
+let q_goal_matching = Pcoq.Entry.make "q_goal_matching"
+let q_hintdb = Pcoq.Entry.make "q_hintdb"
+let q_move_location = Pcoq.Entry.make "q_move_location"
+let q_pose = Pcoq.Entry.make "q_pose"
+let q_assert = Pcoq.Entry.make "q_assert"
 end
 
 let () =
   let entries = [
-    Pcoq.AnyEntry Pltac.ltac2_expr;
+    Pcoq.Entry.Any Pltac.ltac2_expr;
   ] in
-  Pcoq.register_grammars_by_name "coq-core.plugins.ltac2" entries
+  Pcoq.register_grammars_by_name "ltac2" entries
 
 (** Tactic definition *)
 
@@ -292,7 +292,7 @@ let fresh_var avoid x =
 
 let extract_pattern_type ({loc;v=p} as pat) = match p with
 | CPatCnv (pat, ty) -> pat, Some ty
-| CPatVar _ | CPatRef _ -> pat, None
+| CPatAtm _ | CPatVar _ | CPatRef _ | CPatOr _ | CPatAs _ | CPatRecord _ -> pat, None
 
 (** Mangle recursive tactics *)
 let inline_rec_tactic tactics =
@@ -330,6 +330,12 @@ let check_lowercase {loc;v=id} =
   if Tac2env.is_constructor (Libnames.qualid_of_ident id) then
     user_err ?loc (str "The identifier " ++ Id.print id ++ str " must be lowercase")
 
+let check_value ?loc e =
+  if not (is_value e) then
+    user_err ?loc
+      (str "Tactic definition must be a syntactical value." ++ spc() ++
+       str "Consider using a thunk.")
+
 let register_ltac ?deprecation ?(local = false) ?(mut = false) isrec tactics =
   let map ({loc;v=na}, e) =
     let id = match na with
@@ -344,12 +350,9 @@ let register_ltac ?deprecation ?(local = false) ?(mut = false) isrec tactics =
   let tactics =
     if isrec then inline_rec_tactic tactics else tactics
   in
-  let map ({loc;v=id}, e) =
+  let map ({loc;v=id}, ({loc=eloc} as e)) =
     let (e, t) = intern ~strict:true [] e in
-    let () =
-      if not (is_value e) then
-        user_err ?loc (str "Tactic definition must be a syntactical value")
-    in
+    let () = check_value ?loc:eloc e in
     let kn = Lib.make_kn id in
     let exists =
       try let _ = Tac2env.interp_global kn in true with Not_found -> false
@@ -465,23 +468,12 @@ let register_typedef ?(local = false) isrec types =
 
 let register_primitive ?deprecation ?(local = false) {loc;v=id} t ml =
   let t = intern_open_type t in
-  let rec count_arrow = function
-  | GTypArrow (_, t) -> 1 + count_arrow t
-  | _ -> 0
-  in
-  let arrows = count_arrow (snd t) in
-  let () = if Int.equal arrows 0 then
-    user_err ?loc (str "External tactic must have at least one argument") in
   let () =
     try let _ = Tac2env.interp_primitive ml in () with Not_found ->
       user_err ?loc (str "Unregistered primitive " ++
         quote (str ml.mltac_plugin) ++ spc () ++ quote (str ml.mltac_tactic))
   in
-  let init i = Id.of_string (Printf.sprintf "x%i" i) in
-  let names = List.init arrows init in
-  let bnd = List.map (fun id -> Name id) names in
-  let arg = List.map (fun id -> GTacVar id) names in
-  let e = GTacFun (bnd, GTacPrm (ml, arg)) in
+  let e = GTacPrm ml in
   let def = {
     tacdef_local = local;
     tacdef_mutable = false;
@@ -603,8 +595,13 @@ let parse_scope = function
 
 let parse_token = function
 | SexprStr {v=s} -> TacTerm s
-| SexprRec (_, {v=na}, [tok]) ->
-  let na = match na with None -> Anonymous | Some id -> Name id in
+| SexprRec (_, na, [tok]) ->
+  let na = match na.CAst.v with
+  | None -> Anonymous
+  | Some id ->
+    let () = check_lowercase (CAst.make ?loc:na.CAst.loc id) in
+    Name id
+  in
   let scope = parse_scope tok in
   TacNonTerm (na, scope)
 | tok ->
@@ -628,8 +625,8 @@ end
 let parse_scope = ParseToken.parse_scope
 
 type synext = {
+  synext_kn : KerName.t;
   synext_tok : sexpr list;
-  synext_exp : raw_tacexpr;
   synext_lev : int;
   synext_loc : bool;
   synext_depr : Deprecation.t option;
@@ -649,14 +646,14 @@ let rec get_rule (tok : scope_rule token list) : krule = match tok with
   KRule (rule, act)
 | TacTerm t :: tok ->
   let KRule (rule, act) = get_rule tok in
-  let rule = Pcoq.(Rule.next rule (Symbol.token (CLexer.terminal t))) in
+  let rule = Pcoq.(Rule.next rule (Symbol.token (Pcoq.terminal t))) in
   let act k _ = act k in
   KRule (rule, act)
 
 let deprecated_ltac2_notation =
   Deprecation.create_warning
     ~object_name:"Ltac2 notation"
-    ~warning_name:"deprecated-ltac2-notation"
+    ~warning_name_if_no_since:"deprecated-ltac2-notation"
     (fun (toks : sexpr list) -> pr_sequence ParseToken.print_token toks)
 
 (* This is a hack to preserve the level 4 entry which is initially empty. The
@@ -679,7 +676,7 @@ let perform_notation syn st =
       ((CAst.make ?loc:e.loc @@ CPatVar na), e)
     in
     let bnd = List.map map args in
-    CAst.make ~loc @@ CTacLet (false, bnd, syn.synext_exp)
+    CAst.make ~loc @@ CTacSyn (bnd, syn.synext_kn)
   in
   let rule = Pcoq.Production.make rule (act mk) in
   let pos = Some (string_of_int syn.synext_lev) in
@@ -691,7 +688,7 @@ let perform_notation syn st =
     ([Pcoq.ExtendRuleReinit (Pltac.ltac2_expr, reinit, rule)], st)
 
 let ltac2_notation =
-  Pcoq.create_grammar_command "ltac2-notation" perform_notation
+  Pcoq.create_grammar_command "ltac2-notation" { gext_fun = perform_notation; gext_eq = (==) (* FIXME *) }
 
 let cache_synext syn =
   Pcoq.extend_grammar_command ltac2_notation syn
@@ -700,8 +697,9 @@ let open_synext i syn =
   if Int.equal i 1 then Pcoq.extend_grammar_command ltac2_notation syn
 
 let subst_synext (subst, syn) =
-  let e = Tac2intern.subst_rawexpr subst syn.synext_exp in
-  if e == syn.synext_exp then syn else { syn with synext_exp = e }
+  let kn = Mod_subst.subst_kn subst syn.synext_kn in
+  if kn == syn.synext_kn then syn
+  else { syn with synext_kn = kn }
 
 let classify_synext o =
   if o.synext_loc then Dispose else Substitute
@@ -710,11 +708,34 @@ let ltac2_notation_cat = Libobject.create_category "ltac2.notations"
 
 let inTac2Notation : synext -> obj =
   declare_object {(default_object "TAC2-NOTATION") with
+     object_stage = Summary.Stage.Synterp;
      cache_function  = cache_synext;
      open_function   = simple_open ~cat:ltac2_notation_cat open_synext;
      subst_function = subst_synext;
      classify_function = classify_synext}
 
+let cache_synext_interp (local,kn,tac) =
+  Tac2env.define_notation kn tac
+
+let open_synext_interp i o =
+  if Int.equal i 1 then cache_synext_interp o
+
+let subst_synext_interp (subst, (local,kn,tac as o)) =
+  let tac' = Tac2intern.subst_rawexpr subst tac in
+  let kn' = Mod_subst.subst_kn subst kn in
+  if kn' == kn && tac' == tac then o else
+  (local, kn', tac')
+
+let classify_synext_interp (local,_,_) =
+  if local then Dispose else Substitute
+
+let inTac2NotationInterp : (bool*KerName.t*raw_tacexpr) -> obj =
+  declare_object {(default_object "TAC2-NOTATION-INTERP") with
+     cache_function  = cache_synext_interp;
+     open_function   = simple_open ~cat:ltac2_notation_cat open_synext_interp;
+     subst_function = subst_synext_interp;
+     classify_function = classify_synext_interp}
+
 type abbreviation = {
   abbr_body : raw_tacexpr;
   abbr_depr : Deprecation.t option;
@@ -746,41 +767,77 @@ let inTac2Abbreviation : Id.t -> abbreviation -> obj =
      subst_function = subst_abbreviation;
      classify_function = classify_abbreviation}
 
-let register_notation ?deprecation ?(local = false) tkn lev body = match tkn, lev with
-| [SexprRec (_, {loc;v=Some id}, [])], None ->
-  (* Tactic abbreviation *)
-  let () = check_lowercase CAst.(make ?loc id) in
-  let body = Tac2intern.globalize Id.Set.empty body in
-  let abbr = { abbr_body = body; abbr_depr = deprecation } in
-  Lib.add_leaf (inTac2Abbreviation id abbr)
-| _ ->
-  (* Check that the tokens make sense *)
-  let entries = List.map ParseToken.parse_token tkn in
-  let fold accu tok = match tok with
-  | TacTerm _ -> accu
-  | TacNonTerm (Name id, _) -> Id.Set.add id accu
-  | TacNonTerm (Anonymous, _) -> accu
-  in
-  let ids = List.fold_left fold Id.Set.empty entries in
-  (* Globalize so that names are absolute *)
-  let body = Tac2intern.globalize ids body in
-  let lev = match lev with
-  | Some n ->
-    let () =
-      if n < 0 || n > 6 then
-        user_err (str "Notation levels must range between 0 and 6")
+let rec string_of_scope = function
+| SexprStr s -> Printf.sprintf "str(%s)" s.CAst.v
+| SexprInt i -> Printf.sprintf "int(%i)" i.CAst.v
+| SexprRec (_, {v=na}, []) -> Option.cata Id.to_string "_" na
+| SexprRec (_, {v=na}, e) ->
+  Printf.sprintf "%s(%s)" (Option.cata Id.to_string "_" na) (String.concat " " (List.map string_of_scope e))
+
+let string_of_token = function
+| SexprStr {v=s} -> Printf.sprintf "str(%s)" s
+| SexprRec (_, {v=na}, [tok]) -> string_of_scope tok
+| _ -> assert false
+
+let make_fresh_key tokens =
+  let prods = String.concat "_" (List.map string_of_token tokens) in
+  (* We embed the hash of the kernel name in the label so that the identifier
+      should be mostly unique. This ensures that including two modules
+      together won't confuse the corresponding labels. *)
+  let hash = (ModPath.hash (Lib.current_mp ())) land 0x7FFFFFFF in
+  let lbl = Id.of_string_soft (Printf.sprintf "%s_%08X" prods hash) in
+  Lib.make_kn lbl
+
+type notation_interpretation_data =
+| Abbreviation of Id.t * Deprecation.t option * raw_tacexpr
+| Synext of bool * KerName.t * Id.Set.t * raw_tacexpr
+
+let register_notation atts tkn lev body =
+  let deprecation, local = Attributes.(parse Notations.(deprecation ++ locality)) atts in
+  let local = Option.default false local in
+  match tkn, lev with
+  | [SexprRec (_, {loc;v=Some id}, [])], None ->
+    (* Tactic abbreviation *)
+    let () = check_lowercase CAst.(make ?loc id) in
+    Abbreviation(id, deprecation, body)
+  | _ ->
+    (* Check that the tokens make sense *)
+    let entries = List.map ParseToken.parse_token tkn in
+    let fold accu tok = match tok with
+    | TacTerm _ -> accu
+    | TacNonTerm (Name id, _) -> Id.Set.add id accu
+    | TacNonTerm (Anonymous, _) -> accu
     in
-    n
-  | None -> 5
-  in
-  let ext = {
-    synext_tok = tkn;
-    synext_exp = body;
-    synext_lev = lev;
-    synext_loc = local;
-    synext_depr = deprecation;
-  } in
-  Lib.add_leaf (inTac2Notation ext)
+    let ids = List.fold_left fold Id.Set.empty entries in
+    (* Globalize so that names are absolute *)
+    let lev = match lev with
+    | Some n ->
+      let () =
+        if n < 0 || n > 6 then
+          user_err (str "Notation levels must range between 0 and 6")
+      in
+      n
+    | None -> 5
+    in
+    let key = make_fresh_key tkn in
+    let ext = {
+      synext_kn = key;
+      synext_tok = tkn;
+      synext_lev = lev;
+      synext_loc = local;
+      synext_depr = deprecation;
+    } in
+    Lib.add_leaf (inTac2Notation ext);
+    Synext (local,key,ids,body)
+
+let register_notation_interpretation = function
+  | Abbreviation (id, deprecation, body) ->
+    let body = Tac2intern.globalize Id.Set.empty body in
+    let abbr = { abbr_body = body; abbr_depr = deprecation } in
+    Lib.add_leaf (inTac2Abbreviation id abbr)
+  | Synext (local,kn,ids,body) ->
+    let body = Tac2intern.globalize ids body in
+    Lib.add_leaf (inTac2NotationInterp (local,kn,body))
 
 type redefinition = {
   redef_kn : ltac_constant;
@@ -817,7 +874,7 @@ let inTac2Redefinition : redefinition -> obj =
      classify_function = classify_redefinition;
     }
 
-let register_redefinition qid old e =
+let register_redefinition qid old ({loc=eloc} as e) =
   let kn =
     try Tac2env.locate_ltac qid
     with Not_found -> user_err ?loc:qid.CAst.loc (str "Unknown tactic " ++ pr_qualid qid)
@@ -837,10 +894,7 @@ let register_redefinition qid old e =
   | Some { CAst.v = id } -> [id, data.Tac2env.gdata_type]
   in
   let (e, t) = intern ~strict:true ctx e in
-  let () =
-    if not (is_value e) then
-      user_err ?loc:qid.CAst.loc (str "Tactic definition must be a syntactical value")
-  in
+  let () = check_value ?loc:eloc e in
   let () =
     if not (Tac2intern.check_subtype t data.Tac2env.gdata_type) then
       let name = int_name () in
@@ -883,38 +937,45 @@ let perform_eval ~pstate e =
 
 (** Toplevel entries *)
 
+let warn_modtype = CWarnings.create ~name:"ltac2-in-modtype" ~category:CWarnings.CoreCategories.ltac2 ~default:AsError
+    Pp.(fun what -> strbrk "Ltac2 " ++ str what ++ strbrk " should not be defined inside module types: functor application to arguments of this module type will be unchecked")
+
+
+let check_modtype what =
+  if Lib.is_modtype ()
+  then warn_modtype what
+
 let register_struct atts str = match str with
 | StrVal (mut, isrec, e) ->
+  check_modtype "definitions";
   let deprecation, local = Attributes.(parse Notations.(deprecation ++ locality)) atts in
   register_ltac ?deprecation ?local ~mut isrec e
 | StrTyp (isrec, t) ->
+  check_modtype "types";
   let local = Attributes.(parse locality) atts in
   register_type ?local isrec t
 | StrPrm (id, t, ml) ->
+  check_modtype "externals";
   let deprecation, local = Attributes.(parse Notations.(deprecation ++ locality)) atts in
   register_primitive ?deprecation ?local id t ml
-| StrSyn (tok, lev, e) ->
-  let deprecation, local = Attributes.(parse Notations.(deprecation ++ locality)) atts in
-  register_notation ?deprecation ?local tok lev e
 | StrMut (qid, old, e) ->
   let () = Attributes.unsupported_attributes atts in
   register_redefinition qid old e
 
 (** Toplevel exception *)
 
-let _ = Goptions.declare_bool_option {
-  Goptions.optdepr = false;
+let () = Goptions.declare_bool_option {
+  Goptions.optstage = Summary.Stage.Interp;
+  Goptions.optdepr = None;
   Goptions.optkey = ["Ltac2"; "Backtrace"];
-  Goptions.optread = (fun () -> !Tac2interp.print_ltac2_backtrace);
-  Goptions.optwrite = (fun b -> Tac2interp.print_ltac2_backtrace := b);
+  Goptions.optread = (fun () -> !Tac2bt.print_ltac2_backtrace);
+  Goptions.optwrite = (fun b -> Tac2bt.print_ltac2_backtrace := b);
 }
 
-let backtrace : backtrace Exninfo.t = Exninfo.make ()
-
 let pr_frame = function
 | FrAnon e -> str "Call {" ++ pr_glbexpr e ++ str "}"
 | FrLtac kn ->
-  str "Call " ++ Libnames.pr_qualid (Tac2env.shortest_qualid_of_ltac (TacConstant kn))
+  str "Call " ++ pr_tacref kn
 | FrPrim ml ->
   str "Prim <" ++ str ml.mltac_plugin ++ str ":" ++ str ml.mltac_tactic ++ str ">"
 | FrExtn (tag, arg) ->
@@ -935,22 +996,97 @@ let () = register_handler begin function
 end
 
 let () = CErrors.register_additional_error_info begin fun info ->
-  if !Tac2interp.print_ltac2_backtrace then
-    let bt = Exninfo.get info backtrace in
+  if !Tac2bt.print_ltac2_backtrace then
+    let bt = Exninfo.get info Tac2bt.backtrace in
     let bt = match bt with
-    | Some bt -> bt
+    | Some bt -> List.rev bt
     | None -> []
     in
     let bt =
       str "Backtrace:" ++ fnl () ++ prlist_with_sep fnl pr_frame bt ++ fnl ()
     in
-    Some (Loc.tag bt)
+    Some bt
   else None
 end
 
 (** Printing *)
 
-let print_ltac qid =
+let print_constant ~print_def qid ?info data =
+  let e = data.Tac2env.gdata_expr in
+  let (_, t) = data.Tac2env.gdata_type in
+  let name = int_name () in
+  let def = if print_def then fnl () ++ hov 2 (pr_qualid qid ++ spc () ++ str ":=" ++ spc () ++ pr_glbexpr e) else mt() in
+  let info = match info with
+    | None -> mt()
+    | Some info -> fnl() ++ fnl() ++ hov 2 (str "Compiled as" ++ spc() ++ str info.Tac2env.source)
+  in
+  hov 0 (
+    hov 2 (pr_qualid qid ++ spc () ++ str ":" ++ spc () ++ pr_glbtype name t) ++ def ++ info
+  )
+
+let print_tacref ~print_def qid = function
+  | TacConstant kn ->
+    let data = Tac2env.interp_global kn in
+    let info = Option.map fst (Tac2env.get_compiled_global kn) in
+    print_constant ~print_def qid data ?info
+  | TacAlias kn -> str "Alias to ..."
+
+let locatable_ltac2 = "Ltac2"
+
+type ltac2_object =
+  | Constructor of ltac_constructor
+  | TacRef of tacref
+
+let locate_object qid =
+  try Constructor (Tac2env.locate_constructor qid)
+  with Not_found ->
+    TacRef (Tac2env.locate_ltac qid)
+
+let locate_all_object qid =
+  let open Tac2env in
+  (List.map (fun x -> Constructor x) (locate_extended_all_constructor qid))
+   @ (List.map (fun x -> TacRef x) (locate_extended_all_ltac qid))
+
+let shortest_qualid_of_object = function
+  | Constructor kn -> Tac2env.shortest_qualid_of_constructor kn
+  | TacRef kn -> Tac2env.shortest_qualid_of_ltac kn
+
+let path_of_object = function
+  | Constructor kn -> Tac2env.path_of_constructor kn
+  | TacRef kn -> Tac2env.path_of_ltac kn
+
+let print_object ~print_def qid = function
+  | Constructor _ -> str "Ltac2 Constructor" ++ spc() ++ pr_qualid qid
+  | TacRef kn -> str "Ltac2 " ++ print_tacref ~print_def qid kn
+
+let () =
+  let open Prettyp in
+  let locate qid = try Some (qid, locate_object qid) with Not_found -> None in
+  let locate_all qid = List.map (fun x -> qid, x) (locate_all_object qid) in
+  let shortest_qualid (_,kn) = shortest_qualid_of_object kn in
+  let name (_,kn) =
+    let hdr = match kn with
+      | TacRef (TacConstant _) -> str "Ltac2"
+      | TacRef (TacAlias _) -> str "Ltac2 Notation"
+      | Constructor _ -> str "Ltac2 Constructor"
+    in
+    hdr ++ spc () ++ pr_path (path_of_object kn)
+  in
+  let print (qid,kn) = print_object ~print_def:true qid kn in
+  let about (qid,kn) = print_object ~print_def:false qid kn in
+  register_locatable locatable_ltac2 {
+    locate;
+    locate_all;
+    shortest_qualid;
+    name;
+    print;
+    about;
+  }
+
+let print_located_tactic qid =
+  Feedback.msg_notice (Prettyp.print_located_other locatable_ltac2 qid)
+
+let print_ltac2 qid =
   if Tac2env.is_constructor qid then
     let kn =
       try Tac2env.locate_constructor qid
@@ -963,20 +1099,26 @@ let print_ltac qid =
       try Tac2env.locate_ltac qid
       with Not_found -> user_err ?loc:qid.CAst.loc (str "Unknown tactic " ++ pr_qualid qid)
     in
-    match kn with
-    | TacConstant kn ->
-      let data = Tac2env.interp_global kn in
-      let e = data.Tac2env.gdata_expr in
-      let (_, t) = data.Tac2env.gdata_type in
-      let name = int_name () in
-      Feedback.msg_notice (
-        hov 0 (
-          hov 2 (pr_qualid qid ++ spc () ++ str ":" ++ spc () ++ pr_glbtype name t) ++ fnl () ++
-          hov 2 (pr_qualid qid ++ spc () ++ str ":=" ++ spc () ++ pr_glbexpr e)
-        )
-      )
-    | TacAlias kn ->
-      Feedback.msg_notice (str "Alias to ...")
+    Feedback.msg_notice (print_tacref ~print_def:true qid kn)
+
+let print_signatures () =
+  let entries = KNmap.bindings (Tac2env.globals ()) in
+  let sort (kn1, _) (kn2, _) = KerName.compare kn1 kn2 in
+  let entries = List.sort sort entries in
+  let map (kn, entry) =
+    let qid =
+      try Some (Tac2env.shortest_qualid_of_ltac (TacConstant kn))
+      with Not_found -> None
+    in
+    match qid with
+    | None -> None
+    | Some qid -> Some (qid, entry)
+  in
+  let entries = List.map_filter map entries in
+  let pr_entry (qid, data) =
+    hov 2 (print_constant ~print_def:false qid data)
+  in
+  Feedback.msg_notice (prlist_with_sep fnl pr_entry entries)
 
 (** Calling tactics *)
 
@@ -1024,21 +1166,6 @@ let def_unit = {
 
 let t_list = coq_def "list"
 
-let (f_register_constr_quotations, register_constr_quotations) = Hook.make ()
-
-let cache_ltac2_init () =
-  Hook.get f_register_constr_quotations ()
-
-let load_ltac2_init _ () =
-  Hook.get f_register_constr_quotations ()
-
-(** Dummy object that register global rules when Require is called *)
-let inTac2Init : unit -> obj =
-  declare_object {(default_object "TAC2-INIT") with
-    cache_function = cache_ltac2_init;
-    load_function = load_ltac2_init;
-  }
-
 let () = Mltop.declare_cache_obj begin fun () ->
   let unit = Id.of_string "unit" in
   Lib.add_leaf (inTypDef unit def_unit);
@@ -1046,5 +1173,4 @@ let () = Mltop.declare_cache_obj begin fun () ->
     ("[]", []);
     ("::", [GTypVar 0; GTypRef (Other t_list, [GTypVar 0])]);
   ];
-  Lib.add_leaf (inTac2Init ());
 end "coq-core.plugins.ltac2"
