@@ -213,33 +213,33 @@ let generalize_arg = make_to_repr to_generalize_arg
 
 (** Standard tactics sharing their implementation with Ltac1 *)
 
-let pname s = { mltac_plugin = "ltac2"; mltac_tactic = s }
+let pname s = { mltac_plugin = "coq-core.plugins.ltac2"; mltac_tactic = s }
 
 let lift tac = tac <*> return v_unit
 
 let define_prim0 name tac =
   let tac _ = lift tac in
-  Tac2env.define_primitive (pname name) (mk_closure arity_one tac)
+  Tac2env.define_primitive (pname name) (mk_closure_val arity_one tac)
 
 let define_prim1 name r0 f =
   let tac x = lift (f (Value.repr_to r0 x)) in
-  Tac2env.define_primitive (pname name) (mk_closure arity_one tac)
+  Tac2env.define_primitive (pname name) (mk_closure_val arity_one tac)
 
 let define_prim2 name r0 r1 f =
   let tac x y = lift (f (Value.repr_to r0 x) (Value.repr_to r1 y)) in
-  Tac2env.define_primitive (pname name) (mk_closure (arity_suc arity_one) tac)
+  Tac2env.define_primitive (pname name) (mk_closure_val (arity_suc arity_one) tac)
 
 let define_prim3 name r0 r1 r2 f =
   let tac x y z = lift (f (Value.repr_to r0 x) (Value.repr_to r1 y) (Value.repr_to r2 z)) in
-  Tac2env.define_primitive (pname name) (mk_closure (arity_suc (arity_suc arity_one)) tac)
+  Tac2env.define_primitive (pname name) (mk_closure_val (arity_suc (arity_suc arity_one)) tac)
 
 let define_prim4 name r0 r1 r2 r3 f =
   let tac x y z u = lift (f (Value.repr_to r0 x) (Value.repr_to r1 y) (Value.repr_to r2 z) (Value.repr_to r3 u)) in
-  Tac2env.define_primitive (pname name) (mk_closure (arity_suc (arity_suc (arity_suc arity_one))) tac)
+  Tac2env.define_primitive (pname name) (mk_closure_val (arity_suc (arity_suc (arity_suc arity_one))) tac)
 
 let define_prim5 name r0 r1 r2 r3 r4 f =
   let tac x y z u v = lift (f (Value.repr_to r0 x) (Value.repr_to r1 y) (Value.repr_to r2 z) (Value.repr_to r3 u) (Value.repr_to r4 v)) in
-  Tac2env.define_primitive (pname name) (mk_closure (arity_suc (arity_suc (arity_suc (arity_suc arity_one)))) tac)
+  Tac2env.define_primitive (pname name) (mk_closure_val (arity_suc (arity_suc (arity_suc (arity_suc arity_one)))) tac)
 
 (** Tactics from Tacexpr *)
 
@@ -279,7 +279,7 @@ end
 let () = define_prim3 "tac_set" bool (thunk (pair name constr)) clause begin fun ev p cl ->
   Proofview.tclEVARMAP >>= fun sigma ->
   thaw (pair name constr) p >>= fun (na, c) ->
-  Tac2tactics.letin_pat_tac ev None na (sigma, c) cl
+  Tac2tactics.letin_pat_tac ev None na (Some sigma, c) cl
 end
 
 let () = define_prim5 "tac_remember" bool name (thunk constr) (option intro_pattern) clause begin fun ev na c eqpat cl ->
@@ -288,7 +288,7 @@ let () = define_prim5 "tac_remember" bool name (thunk constr) (option intro_patt
   | IntroNaming eqpat ->
     Proofview.tclEVARMAP >>= fun sigma ->
     thaw constr c >>= fun c ->
-    Tac2tactics.letin_pat_tac ev (Some (true, eqpat)) na (sigma, c) cl
+    Tac2tactics.letin_pat_tac ev (Some (true, eqpat)) na (Some sigma, c) cl
   | _ ->
     Tacticals.tclZEROMSG (Pp.str "Invalid pattern for remember")
 end
@@ -351,15 +351,15 @@ let lift tac = tac >>= fun c -> Proofview.tclUNIT (Value.of_constr c)
 
 let define_red1 name r0 f =
   let tac x = lift (f (Value.repr_to r0 x)) in
-  Tac2env.define_primitive (pname name) (mk_closure arity_one tac)
+  Tac2env.define_primitive (pname name) (mk_closure_val arity_one tac)
 
 let define_red2 name r0 r1 f =
   let tac x y = lift (f (Value.repr_to r0 x) (Value.repr_to r1 y)) in
-  Tac2env.define_primitive (pname name) (mk_closure (arity_suc arity_one) tac)
+  Tac2env.define_primitive (pname name) (mk_closure_val (arity_suc arity_one) tac)
 
 let define_red3 name r0 r1 r2 f =
   let tac x y z = lift (f (Value.repr_to r0 x) (Value.repr_to r1 y) (Value.repr_to r2 z)) in
-  Tac2env.define_primitive (pname name) (mk_closure (arity_suc (arity_suc arity_one)) tac)
+  Tac2env.define_primitive (pname name) (mk_closure_val (arity_suc (arity_suc arity_one)) tac)
 
 let () = define_red1 "eval_red" constr begin fun c ->
   Tac2tactics.eval_red c
@@ -557,8 +557,8 @@ let () = define_prim3 "tac_trivial" debug (list (thunk constr)) (option (list id
   Tac2tactics.trivial dbg lems dbs
 end
 
-let () = define_prim5 "tac_eauto" debug (option int) (option int) (list (thunk constr)) (option (list ident)) begin fun dbg n p lems dbs ->
-  Tac2tactics.eauto dbg n p lems dbs
+let () = define_prim4 "tac_eauto" debug (option int) (list (thunk constr)) (option (list ident)) begin fun dbg n lems dbs ->
+  Tac2tactics.eauto dbg n lems dbs
 end
 
 let () = define_prim4 "tac_auto" debug (option int) (list (thunk constr)) (option (list ident)) begin fun dbg n lems dbs ->
