@@ -44,26 +44,28 @@ let global_head_of_constr sigma c =
     try fst (EConstr.destRef sigma f)
     with DestKO -> CErrors.anomaly (str "global_head_of_constr.")
 
-let global_of_constr_nofail c =
-  try fst @@ Constr.destRef c
+let global_of_constr_nofail sigma c =
+  try fst @@ EConstr.destRef sigma c
   with DestKO -> GlobRef.VarRef (Id.of_string "dummy")
 
-let rec mk_clos_but f_map n t =
-  let (f, args) = Constr.decompose_appvect t in
-  match f_map (global_of_constr_nofail f) with
+let mk_atom c = CClosure.mk_atom (EConstr.Unsafe.to_constr c)
+
+let rec mk_clos_but sigma f_map n t =
+  let (f, args) = EConstr.decompose_app sigma t in
+  match f_map (global_of_constr_nofail sigma f) with
   | Some tag ->
-      let map i t = tag_arg f_map n (tag i) t in
+      let map i t = tag_arg sigma f_map n (tag i) t in
       if Array.is_empty args then map (-1) f
       else mk_red (FApp (map (-1) f, Array.mapi map args))
   | None -> mk_atom t
 
-and tag_arg f_map n tag c = match tag with
-| Eval -> mk_clos (Esubst.subs_id n) c
+and tag_arg sigma f_map n tag c = match tag with
+| Eval -> mk_clos (Esubst.subs_id n, Univ.Instance.empty) (EConstr.Unsafe.to_constr c)
 | Prot -> mk_atom c
-| Rec -> mk_clos_but f_map n c
+| Rec -> mk_clos_but sigma f_map n c
 
 let interp_map l t =
-  try Some(List.assoc_f GlobRef.equal t l) with Not_found -> None
+  try Some(List.assoc_f GlobRef.CanOrd.equal t l) with Not_found -> None
 
 let protect_maps : protection String.Map.t ref = ref String.Map.empty
 let add_map s m = protect_maps := String.Map.add s m !protect_maps
@@ -72,16 +74,15 @@ let lookup_map map =
   with Not_found ->
     CErrors.user_err (str "Map " ++ qs map ++ str "not found.")
 
-let protect_red map env sigma c0 =
-  let c = EConstr.Unsafe.to_constr c0 in
+let protect_red map env sigma c =
   let tab = create_tab () in
   let infos = Evarutil.create_clos_infos env sigma all in
-  let map = lookup_map map sigma c0 in
-  let rec eval n c = match Constr.kind c with
-  | Prod (na, t, u) -> Constr.mkProd (na, eval n t, eval (n + 1) u)
-  | _ -> kl infos tab (mk_clos_but map n c)
+  let map = lookup_map map sigma c in
+  let rec eval n c = match EConstr.kind sigma c with
+  | Prod (na, t, u) -> EConstr.mkProd (na, eval n t, eval (n + 1) u)
+  | _ -> EConstr.of_constr @@ kl infos tab (mk_clos_but sigma map n c)
   in
-  EConstr.of_constr (eval 0 c)
+  eval 0 c
 
 let protect_tac map =
   Tactics.reduct_option ~check:false (protect_red map,DEFAULTcast) None
@@ -148,9 +149,11 @@ let decl_constant name univs c =
   let () = DeclareUctx.declare_universe_context ~poly:false univs in
   let types = (Typeops.infer (Global.env ()) c).uj_type in
   let univs = UState.Monomorphic_entry Univ.ContextSet.empty, UnivNames.empty_binders in
-  mkConst(declare_constant ~name
-            ~kind:Decls.(IsProof Lemma)
-            (DefinitionEntry (definition_entry ~opaque:true ~types ~univs c)))
+  (* UnsafeMonomorphic: we always do poly:false *)
+  UnsafeMonomorphic.mkConst
+    (declare_constant ~name
+       ~kind:Decls.(IsProof Lemma)
+       (DefinitionEntry (definition_entry ~opaque:true ~types ~univs c)))
 
 let decl_constant na suff univs c =
   let na = Namegen.next_global_ident_away (Nameops.add_suffix na suff) Id.Set.empty in
@@ -344,7 +347,7 @@ let find_ring_structure env sigma l =
               (str"Arguments of ring_simplify do not have all the same type.")
         in
         List.iter check cl';
-        (try ring_for_carrier (EConstr.to_constr sigma ty)
+        (try ring_for_carrier (EConstr.to_constr ~abort_on_undefined_evars:false sigma ty)
         with Not_found ->
           CErrors.user_err
             (str"Cannot find a declared ring structure over"++
@@ -514,8 +517,7 @@ let make_hyp_list env sigma lH =
         (plapp sigma coq_nil [|carrier|])
   in
   let sigma, l' = Typing.solve_evars env sigma l in
-  let l' = EConstr.Unsafe.to_constr l' in
-  sigma, Evarutil.nf_evars_universes sigma l'
+  sigma, l'
 
 let interp_power env sigma pow =
   let sigma, carrier = Evd.fresh_global env sigma (Lazy.force coq_hypo) in
@@ -683,7 +685,7 @@ let ring_lookup (f : Value.t) lH rl t =
     let sigma, l = make_term_list env sigma (EConstr.of_constr e.ring_carrier) rl in
     let rl = Value.of_constr l in
     let sigma, l = make_hyp_list env sigma lH in
-    let lH = carg l in
+    let lH = Value.of_constr l in
     let ring = ltac_ring_structure e in
     Proofview.tclTHEN (Proofview.Unsafe.tclEVARS sigma) (Value.apply f (ring@[lH;rl]))
   end
@@ -747,18 +749,18 @@ let dest_field env sigma th_spec =
   let th_typ = Retyping.get_type_of env sigma th_spec in
   match EConstr.kind sigma th_typ with
     | App(f,[|r;zero;one;add;mul;sub;opp;div;inv;req|])
-        when isRefX sigma (Lazy.force afield_theory) f ->
+        when isRefX env sigma (Lazy.force afield_theory) f ->
         let sigma, rth = plapp sigma af_ar
           [|r;zero;one;add;mul;sub;opp;div;inv;req;th_spec|] in
         (None,r,zero,one,add,mul,Some sub,Some opp,div,inv,req,rth)
     | App(f,[|r;zero;one;add;mul;sub;opp;div;inv;req|])
-        when isRefX sigma (Lazy.force field_theory) f ->
+        when isRefX env sigma (Lazy.force field_theory) f ->
         let sigma, rth =
           plapp sigma f_r
             [|r;zero;one;add;mul;sub;opp;div;inv;req;th_spec|] in
         (Some false,r,zero,one,add,mul,Some sub,Some opp,div,inv,req,rth)
     | App(f,[|r;zero;one;add;mul;div;inv;req|])
-        when isRefX sigma (Lazy.force sfield_theory) f ->
+        when isRefX env sigma (Lazy.force sfield_theory) f ->
         let sigma, rth = plapp sigma sf_sr
           [|r;zero;one;add;mul;div;inv;req;th_spec|] in
         (Some true,r,zero,one,add,mul,None,None,div,inv,req,rth)
@@ -975,7 +977,7 @@ let field_lookup (f : Value.t) lH rl t =
     let sigma, c = make_term_list env sigma (EConstr.of_constr e.field_carrier) rl in
     let rl = Value.of_constr c in
     let sigma, l = make_hyp_list env sigma lH in
-    let lH = carg l in
+    let lH = Value.of_constr l in
     let field = ltac_field_structure e in
     Proofview.tclTHEN (Proofview.Unsafe.tclEVARS sigma) (Value.apply f (field@[lH;rl]))
   end
