@@ -131,7 +131,7 @@ let rec make_form env sigma atom_env term =
   | Cast(a,_,_) ->
     make_form env sigma atom_env a
   | Ind (ind, _) ->
-    if Names.Ind.CanOrd.equal ind (fst (Lazy.force li_False)) then
+    if Environ.QInd.equal env ind (fst (Lazy.force li_False)) then
       Bot
     else
       make_atom atom_env (normalize term)
@@ -139,11 +139,11 @@ let rec make_form env sigma atom_env term =
     begin
       try
         let ind, _ = destInd sigma hd in
-        if Names.Ind.CanOrd.equal ind (fst (Lazy.force li_and)) then
+        if Environ.QInd.equal env ind (fst (Lazy.force li_and)) then
           let fa = make_form env sigma atom_env argv.(0) in
           let fb = make_form env sigma atom_env argv.(1) in
           Conjunct (fa,fb)
-        else if Names.Ind.CanOrd.equal ind (fst (Lazy.force li_or)) then
+        else if Environ.QInd.equal env ind (fst (Lazy.force li_or)) then
           let fa = make_form env sigma atom_env argv.(0) in
           let fb = make_form env sigma atom_env argv.(1) in
           Disjunct (fa,fb)
@@ -159,7 +159,7 @@ let rec make_hyps env sigma atom_env lenv = function
   | LocalAssum (id,typ)::rest ->
     let hrec=
       make_hyps env sigma atom_env (typ::lenv) rest in
-    if List.exists (fun c -> Termops.local_occur_var Evd.empty (* FIXME *) id.binder_name c) lenv ||
+    if List.exists (fun c -> Termops.local_occur_var sigma id.binder_name c) lenv ||
        (Retyping.get_sort_family_of env sigma typ != InProp)
     then
       hrec
@@ -252,17 +252,17 @@ let build_env gamma=
                      mkApp(force node_count l_push,[|mkProp;p;e|]))
     gamma.env (mkApp (force node_count l_empty,[|mkProp|]))
 
-let verbose =
+let { Goptions.get = verbose } =
   Goptions.declare_bool_option_and_ref
-    ~depr:false
     ~key:["Rtauto";"Verbose"]
     ~value:false
+    ()
 
-let check =
+let { Goptions.get = check } =
   Goptions.declare_bool_option_and_ref
-    ~depr:false
     ~key:["Rtauto";"Check"]
     ~value:false
+    ()
 
 open Pp
 
