@@ -56,6 +56,7 @@ let analyze_eliminator elimty env sigma =
        let count = ref 0 in
        let rec occur_rec n c = match EConstr.kind sigma c with
          | Rel m -> if m = n then incr count
+         | Evar (_, args) -> SList.Skip.iter (fun c -> occur_rec n c) args
          | _ -> EConstr.iter_with_binders sigma succ occur_rec n c
        in
        occur_rec n term; !count in
@@ -144,24 +145,24 @@ let fire_subst sigma t = Reductionops.nf_evar sigma t
 let mkTpat env sigma0 (sigma, t) = (* takes a term, refreshes it and makes a T pattern *)
   let t, evs, ucst = abs_evars env sigma0 (sigma, fire_subst sigma t) in
   let t, _, _, sigma = saturate ~beta:true env sigma t (List.length evs) in
-  Evd.merge_universe_context sigma ucst, T t
+  { pat_sigma = Evd.merge_universe_context sigma ucst; pat_pat = T t }
 
 let redex_of_pattern env p = match redex_of_pattern p with
 | None -> CErrors.anomaly (Pp.str "pattern without redex.")
 | Some (sigma, e) -> Evarutil.nf_evar sigma e, Evd.evar_universe_context sigma
 
-let unif_redex env sigma0 nsigma (sigma, r as p) t = (* t is a hint for the redex of p *)
+let unif_redex env sigma0 nsigma p t = (* t is a hint for the redex of p *)
   let t, evs, ucst = abs_evars env sigma0 (nsigma, fire_subst nsigma t) in
-  let t, _, _, sigma = saturate ~beta:true env sigma t (List.length evs) in
+  let t, _, _, sigma = saturate ~beta:true env p.pat_sigma t (List.length evs) in
   let sigma = Evd.merge_universe_context sigma ucst in
-  match r with
-  | X_In_T (e, p) -> sigma, E_As_X_In_T (t, e, p)
+  match p.pat_pat with
+  | X_In_T (e, p) -> { pat_sigma = sigma; pat_pat = E_As_X_In_T (t, e, p) }
   | _ ->
-      try unify_HO env sigma t (fst (redex_of_pattern env p)), r
-      with e when CErrors.noncritical e -> p
+    try { p with pat_sigma = unify_HO env sigma t (fst (redex_of_pattern env p)) }
+    with e when CErrors.noncritical e -> p
 
 let get_nth_arg n args =
-  List.find_map (fun (i, x, _) -> if Int.equal i n then Some x else None) args
+  List.find_map_exn (fun (i, x, _) -> if Int.equal i n then Some x else None) args
 
 let find_eliminator env sigma ~concl ~is_case ?elim oc c_gen =
   match elim with
@@ -208,8 +209,9 @@ let find_eliminator env sigma ~concl ~is_case ?elim oc c_gen =
         sigma, elim, elimty
       else
         let indu = (fst indu, EConstr.EInstance.kind sigma (snd indu)) in
-        let (sigma, ind, indty) = Indrec.build_case_analysis_scheme env sigma indu true sort in
-        (sigma, EConstr.of_constr ind, EConstr.of_constr indty)
+        let sigma, case = Indrec.build_case_analysis_scheme env sigma indu true sort in
+        let (ind, indty) = Indrec.eval_case_analysis case in
+        (sigma, ind, indty)
     in
     let pred_id,n_elim_args,is_rec,elim_is_dep,n_pred_args,ctx_concl =
       analyze_eliminator elimty env sigma in
@@ -228,13 +230,13 @@ let find_eliminator env sigma ~concl ~is_case ?elim oc c_gen =
           tys
         in
         let drop_params x =
-          snd @@ EConstr.decompose_prod_n_assum sigma
+          snd @@ EConstr.decompose_prod_n_decls sigma
             mind.Declarations.mind_nparams (EConstr.of_constr x) in
         Array.map drop_params renamed_tys
       else
         subgoals_tys sigma ctx_concl
     in
-    let rctx = fst (EConstr.decompose_prod_assum sigma unfolded_c_ty) in
+    let rctx = fst (EConstr.decompose_prod_decls sigma unfolded_c_ty) in
     let n_c_args = Context.Rel.length rctx in
     let c, c_ty, t_args, sigma = saturate env sigma c ~ty:c_ty n_c_args in
     let elim, elimty, elim_args, sigma =
@@ -306,7 +308,7 @@ let check_pattern_instantiated env sigma patterns =
   let ev = List.fold_left Evar.Set.union Evar.Set.empty patterns_ev in
   let ty_ev = Evar.Set.fold (fun i e ->
         let ex = i in
-        let i_ty = Evd.evar_concl (Evd.find sigma ex) in
+        let i_ty = Evd.evar_concl (Evd.find_undefined sigma ex) in
         Evar.Set.union e (evars_of_term i_ty))
     ev Evar.Set.empty in
   let inter = Evar.Set.inter ev ty_ev in
@@ -318,8 +320,8 @@ let check_pattern_instantiated env sigma patterns =
       str"occurs in the type of another non-instantiated pattern variable");
   end
 
-let is_undef_pat = function
-| sigma, T t -> EConstr.isEvar sigma t
+let is_undef_pat { pat_sigma = sigma; pat_pat = pat } = match pat with
+| T t -> EConstr.isEvar sigma t
 | _ -> false
 
 let generate_pred env sigma0 ~concl patterns predty eqid is_rec deps elim_args n_elim_args c_is_head_p clr sigma =
@@ -342,7 +344,7 @@ let generate_pred env sigma0 ~concl patterns predty eqid is_rec deps elim_args n
     | NoMatch | NoProgress ->
         let e, ucst = redex_of_pattern env p in
         let sigma = Evd.merge_universe_context sigma ucst in
-        let e, evs, _ucst = abs_evars env sigma (fst p, e) in
+        let e, evs, _ucst = abs_evars env sigma (p.pat_sigma, e) in
         let e, _, _, sigma = saturate ~beta:true env sigma e (List.length evs) in
         let sigma = try unify_HO env sigma inf_t e
                   with exn when CErrors.noncritical exn -> error sigma e inf_t in
@@ -357,7 +359,7 @@ let generate_pred env sigma0 ~concl patterns predty eqid is_rec deps elim_args n
         str"the defined ones matched")
     else match_all concl sigma postponed in
   let concl, sigma = match_all concl sigma patterns in
-  let pred_rctx, _ = EConstr.decompose_prod_assum sigma (fire_subst sigma predty) in
+  let pred_rctx, _ = EConstr.decompose_prod_decls sigma (fire_subst sigma predty) in
   let sigma, concl, gen_eq_tac, clr = match eqid with
   | Some (IPatId _) when not is_rec ->
       let k = List.length deps in
@@ -373,12 +375,20 @@ let generate_pred env sigma0 ~concl patterns predty eqid is_rec deps elim_args n
         let erefl_ty = Retyping.get_type_of env sigma erefl in
         let eq_ty = Retyping.get_type_of env sigma erefl_ty in
         let ucst = Evd.evar_universe_context sigma in
+        let evds = Evar.Map.bind (Evd.find_undefined sigma) @@
+          Evd.evars_of_term sigma new_concl in
         let gen_eq_tac =
           let open Proofview.Notations in
           Proofview.Goal.enter begin fun s ->
           let sigma = Proofview.Goal.sigma s in
           let sigma = Evd.merge_universe_context sigma ucst in
+          let sigma, shelve = Evar.Map.fold (fun e info (sigma, shelve) ->
+              if not @@ Evd.mem sigma e then
+                Evd.add sigma e info, e::shelve else
+                (sigma, shelve))
+              evds (sigma, []) in
           Proofview.Unsafe.tclEVARS sigma <*>
+          Proofview.shelve_goals @@ shelve <*>
           Tactics.apply_type ~typecheck:true new_concl [erefl]
           end
         in
@@ -490,7 +500,7 @@ let ssrelim ?(is_case=false) deps what ?elim eqid elim_intro_tac =
   let elim = sigma, elim in
   let seed =
     Array.map (fun ty ->
-    let ctx,_ = EConstr.decompose_prod_assum sigma ty in
+    let ctx,_ = EConstr.decompose_prod_decls sigma ty in
     CList.rev_map Context.Rel.Declaration.get_name ctx) seed in
 
   let elim_tac =
@@ -517,7 +527,7 @@ let revtoptac n0 =
   let concl = Proofview.Goal.concl gl in
   let env = Proofview.Goal.env gl in
   let n = nb_prod sigma concl - n0 in
-  let dc, cl = EConstr.decompose_prod_n_assum sigma n concl in
+  let dc, cl = EConstr.decompose_prod_n_decls sigma n concl in
   let ty = EConstr.it_mkProd_or_LetIn cl (List.rev dc) in
   let dc' = dc @ [Context.Rel.Declaration.LocalAssum(make_annot (Name rev_id) Sorts.Relevant, ty)] in
   Refine.refine ~typecheck:true begin fun sigma ->
@@ -528,7 +538,7 @@ let revtoptac n0 =
   end
 
 let nothing_to_inject =
-   CWarnings.create ~name:"spurious-ssr-injection" ~category:"ssr"
+   CWarnings.create ~name:"spurious-ssr-injection" ~category:CWarnings.CoreCategories.ssr
      (fun (sigma, env, ty) ->
          Pp.(str "SSReflect: cannot obtain new equations out of" ++ fnl() ++
              str"  " ++ Printer.pr_econstr_env env sigma ty ++ fnl() ++
@@ -561,7 +571,7 @@ let injectl2rtac sigma c = match EConstr.kind sigma c with
 
 let is_injection_case env sigma c =
   let sigma, cty = Typing.type_of env sigma c in
-  let (mind,_), _ = Tacred.reduce_to_quantified_ind env sigma cty in
+  let (mind,_) = Tacred.eval_to_quantified_ind env sigma cty in
   Coqlib.check_ind_ref "core.eq.type" mind
 
 let perform_injection c =
@@ -583,7 +593,7 @@ let perform_injection c =
   let id_with_ebind = (EConstr.mkVar id, NoBindings) in
   let injtac = Tacticals.tclTHEN (introid id) (injectidl2rtac id id_with_ebind) in
   Proofview.Unsafe.tclEVARS sigma <*>
-  Tacticals.tclTHENLAST (Tactics.apply (EConstr.compose_lam dc cl1)) injtac
+  Tacticals.tclTHENLAST (Tactics.apply (EConstr.it_mkLambda cl1 dc)) injtac
   end
 
 let ssrscase_or_inj_tac c =
