@@ -16,7 +16,6 @@ open Term
 open Constr
 open Vars
 open Context
-open Termops
 open Declarations
 open Declareops
 open Environ
@@ -29,14 +28,29 @@ open Context.Rel.Declaration
 let type_of_inductive env (ind,u) =
  let (mib,_ as specif) = Inductive.lookup_mind_specif env ind in
  Typeops.check_hyps_inclusion env (GlobRef.IndRef ind) mib.mind_hyps;
- Inductive.type_of_inductive (specif,u)
+ let t = Inductive.type_of_inductive (specif,u) in
+ Arguments_renaming.rename_type t (IndRef ind)
+
+let e_type_of_inductive env sigma (ind,u) =
+ let (mib,_ as specif) = Inductive.lookup_mind_specif env ind in
+ Reductionops.check_hyps_inclusion env sigma (GlobRef.IndRef ind) mib.mind_hyps;
+ let t = Inductive.type_of_inductive (specif, EConstr.Unsafe.to_instance u) in
+ EConstr.of_constr (Arguments_renaming.rename_type t (IndRef ind))
 
 (* Return type as quoted by the user *)
 let type_of_constructor env (cstr,u) =
  let (mib,_ as specif) =
    Inductive.lookup_mind_specif env (inductive_of_constructor cstr) in
  Typeops.check_hyps_inclusion env (GlobRef.ConstructRef cstr) mib.mind_hyps;
- Inductive.type_of_constructor (cstr,u) specif
+ let t = Inductive.type_of_constructor (cstr,u) specif in
+ Arguments_renaming.rename_type t (ConstructRef cstr)
+
+let e_type_of_constructor env sigma (cstr,u) =
+ let (mib,_ as specif) =
+   Inductive.lookup_mind_specif env (inductive_of_constructor cstr) in
+ Reductionops.check_hyps_inclusion env sigma (GlobRef.ConstructRef cstr) mib.mind_hyps;
+ let t = Inductive.type_of_constructor (cstr,EConstr.Unsafe.to_instance u) specif in
+ EConstr.of_constr (Arguments_renaming.rename_type t (ConstructRef cstr))
 
 (* Return constructor types in user form *)
 let type_of_constructors env (ind,u as indu) =
@@ -103,7 +117,7 @@ let mis_is_recursive (ind,mib,mip) =
   mis_is_recursive_subset (List.interval 0 (mib.mind_ntypes - 1))
     mip.mind_recargs
 
-let mis_nf_constructor_type ((ind,u),mib,mip) j =
+let mis_nf_constructor_type ((_,j),u) (mib,mip) =
   let nconstr = Array.length mip.mind_consnames in
   if j > nconstr then user_err Pp.(str "Not enough constructors in the type.");
   let (ctx, cty) = mip.mind_nf_lc.(j - 1) in
@@ -269,9 +283,9 @@ let top_allowed_sort env (kn,i as ind) =
 let sorts_below top =
   List.filter (fun s -> Sorts.family_leq s top) Sorts.[InSProp;InProp;InSet;InType]
 
-let has_dependent_elim mib =
+let has_dependent_elim (mib,mip) =
   match mib.mind_record with
-  | PrimRecord _ -> mib.mind_finite == BiFinite
+  | PrimRecord _ -> mib.mind_finite == BiFinite || mip.mind_relevance == Irrelevant
   | NotRecord | FakeRecord -> true
 
 (* Annotation for cases *)
@@ -313,18 +327,21 @@ let lift_constructor n cs = {
 let instantiate_params t params sign =
   let nnonrecpar = Context.Rel.nhyps sign - List.length params in
   (* Adjust the signature if recursively non-uniform parameters are not here *)
-  let _,sign = context_chop nnonrecpar sign in
-  let _,t = decompose_prod_n_assum (Context.Rel.length sign) t in
+  let _,sign = Context.Rel.chop_nhyps nnonrecpar sign in
+  let _,t = decompose_prod_n_decls (Context.Rel.length sign) t in
   let subst = subst_of_rel_context_instance_list sign params in
   substl subst t
 
-let get_constructor ((ind,u as indu),mib,mip,params) j =
-  assert (j <= Array.length mip.mind_consnames);
-  let typi = mis_nf_constructor_type (indu,mib,mip) j in
+let instantiate_constructor_params (_,u as cstru) (mib,_ as mind_specif) params =
+  let typi = mis_nf_constructor_type cstru mind_specif in
   let ctx = Vars.subst_instance_context u mib.mind_params_ctxt in
-  let typi = instantiate_params typi params ctx in
-  let (args,ccl) = decompose_prod_assum typi in
-  let (_,allargs) = decompose_app ccl in
+  instantiate_params typi params ctx
+
+let get_constructor ((ind,u),mib,mip,params) j =
+  assert (j <= Array.length mip.mind_consnames);
+  let typi = instantiate_constructor_params ((ind,j),u) (mib,mip) params in
+  let (args,ccl) = decompose_prod_decls typi in
+  let (_,allargs) = decompose_app_list ccl in
   let vargs = List.skipn (List.length params) allargs in
   { cs_cstr = (ith_constructor_of_inductive ind j,u);
     cs_params = params;
@@ -348,24 +365,44 @@ let make_project env sigma ind pred c branches ps =
   let open EConstr in
   assert(Array.length branches == 1);
   let na, ty, t = destLambda sigma pred in
+  let mib, mip as specif = Inductive.lookup_mind_specif env ind in
   let () =
-    let mib, _ = Inductive.lookup_mind_specif env ind in
     if (* dependent *) not (Vars.noccurn sigma 1 t) &&
-         not (has_dependent_elim mib) then
+         not (has_dependent_elim specif) then
       user_err
         Pp.(str"Dependent case analysis not allowed" ++
-              str" on inductive type " ++ Termops.Internal.print_constr_env env sigma (mkInd ind) ++ str ".")
+              str" on inductive type " ++ Termops.pr_global_env env (IndRef ind) ++ str ".")
   in
   let branch = branches.(0) in
-  let ctx, br = decompose_lam_n_assum sigma (Array.length ps) branch in
+  let ctx, br = decompose_lambda_n_decls sigma mip.mind_consnrealdecls.(0) branch in
+  let proj = match EConstr.destRel sigma br with
+    | exception DestKO -> None
+    | i ->
+      begin match List.skipn (i-1) ctx with
+      | exception Failure _ -> None
+      | ctx -> match ctx with
+        | [] -> None
+        | LocalDef _ :: _ ->
+          (* XXX Maybe we should produce the applied constant for this letin pseudoprojection?
+             We would have to get the params etc*)
+          None
+        | LocalAssum _ :: ctx ->
+          (* This match is just a projection *)
+          let p = ps.(Context.Rel.nhyps ctx) in
+          Some (mkProj (Projection.make p true, c))
+      end
+  in
+  match proj with
+  | Some proj -> proj
+  | None ->
   let n, len, ctx =
     List.fold_right
       (fun decl (i, j, ctx) ->
-        match decl with
-        | LocalAssum (na, ty) ->
+         match decl with
+         | LocalAssum (na, ty) ->
            let t = mkProj (Projection.make ps.(i) true, mkRel j) in
            (i + 1, j + 1, LocalDef (na, t, Vars.liftn 1 j ty) :: ctx)
-        | LocalDef (na, b, ty) ->
+         | LocalDef (na, b, ty) ->
            (i, j + 1, LocalDef (na, Vars.liftn 1 j b, Vars.liftn 1 j ty) :: ctx))
       ctx (0, 1, [])
   in
@@ -407,18 +444,16 @@ let get_arity env ((ind,u),params) =
        parameters *)
     let nparams = List.length params in
     if Int.equal nparams mib.mind_nparams then
-      mib.mind_params_ctxt
+      Inductive.inductive_paramdecls (mib,u)
     else begin
       assert (Int.equal nparams mib.mind_nparams_rec);
-      let nnonrecparamdecls = mib.mind_nparams - mib.mind_nparams_rec in
-      snd (Termops.context_chop nnonrecparamdecls mib.mind_params_ctxt)
+      snd (Inductive.inductive_nonrec_rec_paramdecls (mib,u))
     end in
-  let parsign = Vars.subst_instance_context u parsign in
   let arproperlength = List.length mip.mind_arity_ctxt - List.length parsign in
   let arsign,_ = List.chop arproperlength mip.mind_arity_ctxt in
   let subst = subst_of_rel_context_instance_list parsign params in
   let arsign = Vars.subst_instance_context u arsign in
-  (substl_rel_context subst arsign, Inductive.inductive_sort_family mip)
+  substl_rel_context subst arsign
 
 (* Functions to build standard types related to inductive *)
 let build_dependent_constructor cs =
@@ -428,7 +463,7 @@ let build_dependent_constructor cs =
       @(Context.Rel.instance_list mkRel 0 cs.cs_args))
 
 let build_dependent_inductive env ((ind, params) as indf) =
-  let arsign,_ = get_arity env indf in
+  let arsign = get_arity env indf in
   let nrealargs = List.length arsign in
   applist
     (mkIndU ind,
@@ -436,9 +471,9 @@ let build_dependent_inductive env ((ind, params) as indf) =
 
 (* builds the arity of an elimination predicate in sort [s] *)
 
-let make_arity_signature env sigma dep indf =
-  let (arsign,s) = get_arity env indf in
-  let r = Sorts.relevance_of_sort_family s in
+let make_arity_signature env sigma dep ((ind,_), _ as indf) =
+  let arsign = get_arity env indf in
+  let r = Inductive.relevance_of_inductive env ind in
   let anon = make_annot Anonymous r in
   let arsign = List.map (fun d -> Termops.map_rel_decl EConstr.of_constr d) arsign in
   if dep then
@@ -454,16 +489,6 @@ let make_arity env sigma dep indf s =
   let open EConstr in
   it_mkProd_or_LetIn (mkSort s) (make_arity_signature env sigma dep indf)
 
-(* [p] is the predicate and [cs] a constructor summary *)
-let build_branch_type env sigma dep p cs =
-  let base = appvect (lift cs.cs_nargs p, cs.cs_concl_realargs) in
-  if dep then
-    EConstr.Unsafe.to_constr (Namegen.it_mkProd_or_LetIn_name env sigma
-      (EConstr.of_constr (applist (base,[build_dependent_constructor cs])))
-      (List.map (fun d -> Termops.map_rel_decl EConstr.of_constr d) cs.cs_args))
-  else
-    Term.it_mkProd_or_LetIn base cs.cs_args
-
 (**************************************************)
 
 (** From a rel context describing the constructor arguments,
@@ -514,7 +539,12 @@ let compute_projections env (kn, i as ind) =
       match na.binder_name with
       | Name id ->
         let lab = Label.of_id id in
-        let kn = Projection.Repr.make ind ~proj_npars:mib.mind_nparams ~proj_arg lab in
+        let proj_relevant = match na.binder_relevance with
+        | Sorts.Irrelevant -> false
+        | Sorts.Relevant -> true
+        | Sorts.RelevanceVar _ -> assert false
+        in
+        let kn = Projection.Repr.make ind ~proj_relevant ~proj_npars:mib.mind_nparams ~proj_arg lab in
         (* from [params, field1,..,fieldj |- t(params,field1,..,fieldj)]
            to [params, x:I, field1,..,fieldj |- t(params,field1,..,fieldj] *)
         let t = liftn 1 j t in
@@ -541,13 +571,13 @@ let compute_projections env (kn, i as ind) =
 
 let extract_mrectype sigma t =
   let open EConstr in
-  let (t, l) = decompose_app sigma t in
+  let (t, l) = decompose_app_list sigma t in
   match EConstr.kind sigma t with
     | Ind ind -> (ind, l)
     | _ -> raise Not_found
 
 let find_mrectype_vect env sigma c =
-  let (t, l) = Termops.decompose_app_vect sigma (whd_all env sigma c) in
+  let (t, l) = EConstr.decompose_app sigma (whd_all env sigma c) in
   match EConstr.kind sigma t with
     | Ind ind -> (ind, l)
     | _ -> raise Not_found
@@ -557,7 +587,7 @@ let find_mrectype env sigma c =
 
 let find_rectype env sigma c =
   let open EConstr in
-  let (t, l) = decompose_app sigma (whd_all env sigma c) in
+  let (t, l) = decompose_app_list sigma (whd_all env sigma c) in
   match EConstr.kind sigma t with
     | Ind (ind,u) ->
         let (mib,mip) = Inductive.lookup_mind_specif env ind in
@@ -570,7 +600,7 @@ let find_rectype env sigma c =
 
 let find_inductive env sigma c =
   let open EConstr in
-  let (t, l) = decompose_app sigma (whd_all env sigma c) in
+  let (t, l) = decompose_app_list sigma (whd_all env sigma c) in
   match EConstr.kind sigma t with
     | Ind ind
         when (fst (Inductive.lookup_mind_specif env (fst ind))).mind_finite <> CoFinite ->
@@ -580,7 +610,7 @@ let find_inductive env sigma c =
 
 let find_coinductive env sigma c =
   let open EConstr in
-  let (t, l) = decompose_app sigma (whd_all env sigma c) in
+  let (t, l) = decompose_app_list sigma (whd_all env sigma c) in
   match EConstr.kind sigma t with
     | Ind ind
         when (fst (Inductive.lookup_mind_specif env (fst ind))).mind_finite == CoFinite ->
@@ -589,83 +619,10 @@ let find_coinductive env sigma c =
     | _ -> raise Not_found
 
 
-(***********************************************)
-(* find appropriate names for pattern variables. Useful in the Case
-   and Inversion (case_then_using et case_nodep_then_using) tactics. *)
-
-let is_predicate_explicitly_dep env sigma pred arsign =
-  let rec srec env pval arsign =
-    let pv' = whd_all env sigma pval in
-    match EConstr.kind sigma pv', arsign with
-      | Lambda (na,t,b), (LocalAssum _)::arsign ->
-          srec (push_rel_assum (na, t) env) b arsign
-      | Lambda (na,_,t), _ ->
-
-       (* The following code has an impact on the introduction names
-          given by the tactics "case" and "inversion": when the
-          elimination is not dependent, "case" uses Anonymous for
-          inductive types in Prop and names created by mkProd_name for
-          inductive types in Set/Type while "inversion" uses anonymous
-          for inductive types both in Prop and Set/Type !!
-
-          Previously, whether names were created or not relied on
-          whether the predicate created in Indrec.make_case_com had a
-          dependent arity or not. To avoid different predicates
-          printed the same in v8, all predicates built in indrec.ml
-          got a dependent arity (Aug 2004). The new way to decide
-          whether names have to be created or not is to use an
-          Anonymous or Named variable to enforce the expected
-          dependency status (of course, Anonymous implies non
-          dependent, but not conversely).
-
-          From Coq > 8.2, using or not the effective dependency of
-          the predicate is parametrable! *)
-
-          begin match na.binder_name with
-          | Anonymous -> false
-          | Name _ -> true
-          end
-
-      | _ -> anomaly (Pp.str "Non eta-expanded dep-expanded \"match\" predicate.")
-  in
-  srec env (EConstr.of_constr pred) arsign
-
-let is_elim_predicate_explicitly_dependent env sigma pred indf =
-  let arsign,_ = get_arity env indf in
-  is_predicate_explicitly_dep env sigma pred arsign
-
-let set_names env sigma n brty =
-  let open EConstr in
-  let (ctxt,cl) = decompose_prod_n_assum sigma n brty in
-  Namegen.it_mkProd_or_LetIn_name env sigma cl ctxt
-
-let set_pattern_names env sigma ind brv =
-  let (mib,mip) = Inductive.lookup_mind_specif env ind in
-  let arities =
-    Array.map
-      (fun (d, _) -> List.length d - mib.mind_nparams)
-      mip.mind_nf_lc in
-  Array.map2 (set_names env sigma) arities brv
-
-let type_case_branches_with_names env sigma indspec p c =
-  let (ind,args) = indspec in
-  let args = List.map EConstr.Unsafe.to_constr args in
-  let (mib,mip as specif) = Inductive.lookup_mind_specif env (fst ind) in
-  let nparams = mib.mind_nparams in
-  let (params,realargs) = List.chop nparams args in
-  let lbrty = Inductive.build_branches_type ind specif params p in
-  let lbrty = Array.map EConstr.of_constr lbrty in
-  (* Build case type *)
-  let conclty = lambda_appvect_assum (mip.mind_nrealdecls+1) p (Array.of_list (realargs@[c])) in
-  (* Adjust names *)
-  if is_elim_predicate_explicitly_dependent env sigma p (ind,params) then
-    (set_pattern_names env sigma (fst ind) lbrty, conclty)
-  else (lbrty, conclty)
-
 (* Type of Case predicates *)
 let arity_of_case_predicate env (ind,params) dep k =
-  let arsign,s = get_arity env (ind,params) in
-  let r = Sorts.relevance_of_sort_family s in
+  let arsign = get_arity env (ind,params) in
+  let r = Inductive.relevance_of_inductive env (fst ind) in
   let mind = build_dependent_inductive env (ind,params) in
   let concl = if dep then mkArrow mind r (mkSort k) else mkSort k in
   Term.it_mkProd_or_LetIn concl arsign
@@ -674,6 +631,10 @@ let arity_of_case_predicate env (ind,params) dep k =
 (* Inferring the sort of parameters of a polymorphic inductive type
    knowing the sort of the conclusion *)
 
+let univ_level_mem l s = match s with
+| Prop | Set | SProp -> false
+| Type u -> univ_level_mem l u
+| QSort (_, u) -> assert false (* template cannot contain sort variables *)
 
 (* Compute the inductive argument types: replace the sorts
    that appear in the type of the inductive by the sort of the
@@ -694,9 +655,10 @@ let rec instantiate_universes env evdref scl is = function
         else
           (* unconstrained sort: replace by fresh universe *)
           let evm, s = Evd.new_sort_variable Evd.univ_flexible !evdref in
-          let evm = Evd.set_leq_sort env evm s (Sorts.sort_of_univ u) in
+          let evm = Evd.set_leq_sort env evm s (EConstr.ESorts.make (Sorts.sort_of_univ u)) in
             evdref := evm; s
       in
+      let s = EConstr.ESorts.kind !evdref s in
       (LocalAssum (na,mkArity(ctx,s))) :: instantiate_universes env evdref scl is (sign, exp)
   | sign, [] -> sign (* Uniform parameters are exhausted *)
   | [], _ -> assert false
@@ -710,12 +672,12 @@ let type_of_inductive_knowing_conclusion env sigma ((mib,mip),u) conclty =
     | Some t -> t
     in
     let _,scl = splay_arity env sigma conclty in
-    let scl = EConstr.ESorts.kind sigma scl in
     let ctx = List.rev mip.mind_arity_ctxt in
     let evdref = ref sigma in
     let ctx =
       instantiate_universes
         env evdref scl ar.template_level (ctx,templ.template_param_levels) in
+    let scl = EConstr.ESorts.kind !evdref scl in
       !evdref, EConstr.of_constr (mkArity (List.rev ctx,scl))
 
 let type_of_projection_constant env (p,u) =
