@@ -12,7 +12,6 @@ open CErrors
 open Util
 open Names
 open Constr
-open Context
 open Termops
 open Univ
 open Evd
@@ -79,6 +78,21 @@ module ReductionBehaviour = struct
   type t = NeverUnfold | UnfoldWhen of when_flags | UnfoldWhenNoMatch of when_flags
   and when_flags = { recargs : int list ; nargs : int option }
 
+  let equal t1 t2 =
+    match t1, t2 with
+      NeverUnfold, NeverUnfold -> true
+    | (UnfoldWhen {recargs = r1; nargs = n1}),
+      (UnfoldWhen {recargs = r2; nargs = n2}) ->
+       List.length r1 = List.length r2 &&
+         List.for_all2 Int.equal r1 r2 &&
+           Option.equal Int.equal n1 n2
+    | UnfoldWhenNoMatch {recargs = r1; nargs = n1},
+      UnfoldWhenNoMatch {recargs = r2; nargs = n2} ->
+       List.length r1 = List.length r2 &&
+         List.for_all2 Int.equal r1 r2 &&
+           Option.equal Int.equal n1 n2
+    | _, _ -> false
+
   let more_args_when k { recargs; nargs } =
     { nargs = Option.map ((+) k) nargs;
       recargs = List.map ((+) k) recargs;
@@ -89,11 +103,17 @@ module ReductionBehaviour = struct
     | UnfoldWhen x -> UnfoldWhen (more_args_when k x)
     | UnfoldWhenNoMatch x -> UnfoldWhenNoMatch (more_args_when k x)
 
+(* We need to have a fast way to know the set of all constants that
+  have the NeverUnfold flag.  Therefore, the table has a distinct subpart
+  that is this set. *)
   let table =
-    Summary.ref (GlobRef.Map.empty : t GlobRef.Map.t) ~name:"reductionbehaviour"
+    Summary.ref ((GlobRef.Set.empty, GlobRef.Map.empty) :
+                   GlobRef.Set.t * t GlobRef.Map.t) ~name:"reductionbehaviour"
 
   let load _ (_,(r, b)) =
-    table := GlobRef.Map.add r b !table
+    table := (match b with
+                | NeverUnfold -> GlobRef.Set.add r (fst !table), GlobRef.Map.remove r (snd !table)
+                | _ -> GlobRef.Set.remove r (fst !table), GlobRef.Map.add r b (snd !table))
 
   let cache o = load 1 o
 
@@ -132,7 +152,22 @@ module ReductionBehaviour = struct
   let set ~local r b =
     Lib.add_leaf (inRedBehaviour (local, (r, b)))
 
-  let get r = GlobRef.Map.find_opt r !table
+  let get r =
+    if GlobRef.Set.mem r (fst !table) then
+      Some NeverUnfold
+    else
+      GlobRef.Map.find_opt r (snd !table)
+
+  let all_tagged t =
+    match t with
+      NeverUnfold -> fst !table
+    | _ ->
+       GlobRef.Map.fold
+         (fun a u s ->
+           if equal t u then
+             GlobRef.Set.add a s
+           else
+             s) (snd !table) GlobRef.Set.empty
 
   let print ref =
     let open Pp in
@@ -513,6 +548,8 @@ open Primred
 module CNativeEntries =
 struct
 
+  open UnsafeMonomorphic
+
   type elem = EConstr.t
   type args = EConstr.t array
   type evd = evar_map
@@ -874,7 +911,7 @@ let raw_whd_state_gen flags env =
   f
 
 let stack_red_of_state_red f =
-  let f env sigma x = EConstr.decompose_app sigma (Stack.zip sigma (f env sigma (x, Stack.empty))) in
+  let f env sigma x = EConstr.decompose_app_list sigma (Stack.zip sigma (f env sigma (x, Stack.empty))) in
   f
 
 let red_of_state_red f env sigma x =
@@ -927,6 +964,8 @@ let whd_allnolet_state = raw_whd_state_gen CClosure.allnolet
 let whd_allnolet_stack = stack_red_of_state_red whd_allnolet_state
 let whd_allnolet = red_of_state_red whd_allnolet_state
 
+let whd_stack_gen reds = stack_red_of_state_red (whd_state_gen reds)
+
 let is_head_evar env sigma c =
   let head, _ = whd_all_state env sigma (c,Stack.empty) in
   EConstr.isEvar sigma head
@@ -937,7 +976,7 @@ let shrink_eta sigma c =
   let rec whrec x = match EConstr.kind sigma x with
     | Cast (c, _, _) -> whrec c
     | Lambda (_, _, c) ->
-      let (f, cl) = decompose_app_vect sigma (whrec c) in
+      let (f, cl) = decompose_app sigma (whrec c) in
       let napp = Array.length cl in
       if napp > 0 then
         let x' = whrec (Array.last cl) in
@@ -1011,92 +1050,90 @@ let is_transparent e k =
 
 type conversion_test = Constraints.t -> Constraints.t
 
-let pb_is_equal pb = pb == Reduction.CONV
+module CheckUnivs =
+struct
 
-let pb_equal = function
-  | Reduction.CUMUL -> Reduction.CONV
-  | Reduction.CONV -> Reduction.CONV
+open Conversion
 
-(* NOTE: We absorb anomalies happening in the conversion tactic, which
-   is a bit ugly. This is mostly due to efficiency both in tactics and
-   in the conversion machinery itself. It is not uncommon for a tactic
-   to send some ill-typed term to the engine.
+let check_eq univs u u' =
+  if not (Evd.check_eq univs u u') then raise NotConvertible
 
-   We would usually say that a tactic that converts ill-typed terms is
-   buggy, but fixing the tactic could have a very large runtime cost
-   *)
-exception AnomalyInConversion of exn
+let check_leq univs u u' =
+  if not (Evd.check_leq univs u u') then raise NotConvertible
 
-let _ = CErrors.register_handler (function
-    | AnomalyInConversion e ->
-      Some Pp.(str "Conversion test raised an anomaly:" ++
-               spc () ++ CErrors.print e)
-    | _ -> None)
+let check_sort_cmp_universes pb s0 s1 univs =
+  let s0 = ESorts.make s0 in
+  let s1 = ESorts.make s1 in
+  match pb with
+  | CUMUL -> check_leq univs s0 s1
+  | CONV -> check_eq univs s0 s1
 
-let report_anomaly (e, info) =
-  let e =
-    if is_anomaly e then AnomalyInConversion e
-    else e
+let checked_sort_cmp_universes _env pb s0 s1 univs =
+  check_sort_cmp_universes pb s0 s1 univs; univs
+
+let check_convert_instances ~flex:_ u u' univs =
+  let u = Instance.to_array u in
+  let u' = Instance.to_array u' in
+  let fold accu l1 l2 = Constraints.add (l1, Eq, l2) accu in
+  let cst = Array.fold_left2 fold Constraints.empty u u' in
+  if Evd.check_constraints univs cst then univs else raise NotConvertible
+
+(* general conversion and inference functions *)
+let check_inductive_instances cv_pb variance u1 u2 univs =
+  let csts = get_cumulativity_constraints cv_pb variance u1 u2 in
+  if (Evd.check_constraints univs csts) then univs
+  else raise NotConvertible
+
+let checked_universes =
+  { compare_sorts = checked_sort_cmp_universes;
+    compare_instances = check_convert_instances;
+    compare_cumul_instances = check_inductive_instances; }
+
+end
+
+let is_fconv ?(reds=TransparentState.full) pb env sigma t1 t2 =
+  let univs = Evd.universes sigma in
+  let t1 = EConstr.Unsafe.to_constr t1 in
+  let t2 = EConstr.Unsafe.to_constr t2 in
+  let b = match pb with
+  | Conversion.CUMUL -> leq_constr_univs univs t1 t2
+  | Conversion.CONV -> eq_constr_univs univs t1 t2
   in
-  Exninfo.iraise (e, info)
-
-let f_conv ?l2r ?reds env ?evars x y =
-  let inj = EConstr.Unsafe.to_constr in
-  Reduction.conv ?l2r ?reds env ?evars (inj x) (inj y)
-
-let f_conv_leq ?l2r ?reds env ?evars x y =
-  let inj = EConstr.Unsafe.to_constr in
-  Reduction.conv_leq ?l2r ?reds env ?evars (inj x) (inj y)
-
-let test_trans_conversion (f: constr Reduction.extended_conversion_function) reds env sigma x y =
-  try
+  if b then true
+  else
     let evars = Evd.evar_handler sigma in
-    let env = Environ.set_universes (Evd.universes sigma) env in
-    let _ = f ~reds env ~evars x y in
-    true
-  with Reduction.NotConvertible -> false
-    | e ->
-      let e = Exninfo.capture e in
-      report_anomaly e
+    try
+      let env = Environ.set_universes (Evd.universes sigma) env in
+      let _ = Conversion.generic_conv ~l2r:false pb evars reds env (sigma, CheckUnivs.checked_universes) t1 t2 in
+      true
+    with Conversion.NotConvertible -> false
 
-let is_conv ?(reds=TransparentState.full) env sigma = test_trans_conversion f_conv reds env sigma
-let is_conv_leq ?(reds=TransparentState.full) env sigma = test_trans_conversion f_conv_leq reds env sigma
-let is_fconv ?(reds=TransparentState.full) = function
-  | Reduction.CONV -> is_conv ~reds
-  | Reduction.CUMUL -> is_conv_leq ~reds
-
-let check_conv ?(pb=Reduction.CUMUL) ?(ts=TransparentState.full) env sigma x y =
-  let f = match pb with
-    | Reduction.CONV -> f_conv
-    | Reduction.CUMUL -> f_conv_leq
-  in
-    let env = Environ.set_universes (Evd.universes sigma) env in
-    try f ~reds:ts env ~evars:(Evd.evar_handler sigma) x y; true
-    with Reduction.NotConvertible -> false
-    | UGraph.UniverseInconsistency _ -> false
-    | e ->
-      let e = Exninfo.capture e in
-      report_anomaly e
+let is_conv ?(reds=TransparentState.full) env sigma x y =
+  is_fconv ~reds Conversion.CONV env sigma x y
+let is_conv_leq ?(reds=TransparentState.full) env sigma x y =
+  is_fconv ~reds Conversion.CUMUL env sigma x y
+let check_conv ?(pb=Conversion.CUMUL) ?(ts=TransparentState.full) env sigma x y =
+  is_fconv ~reds:ts pb env sigma x y
 
 let sigma_compare_sorts env pb s0 s1 sigma =
   match pb with
-  | Reduction.CONV -> Evd.set_eq_sort env sigma s0 s1
-  | Reduction.CUMUL -> Evd.set_leq_sort env sigma s0 s1
+  | Conversion.CONV -> Evd.set_eq_sort env sigma (ESorts.make s0) (ESorts.make s1)
+  | Conversion.CUMUL -> Evd.set_leq_sort env sigma (ESorts.make s0) (ESorts.make s1)
 
 let sigma_compare_instances ~flex i0 i1 sigma =
   try Evd.set_eq_instances ~flex sigma i0 i1
   with Evd.UniversesDiffer
      | UGraph.UniverseInconsistency _ ->
-        raise Reduction.NotConvertible
+        raise Conversion.NotConvertible
 
 let sigma_check_inductive_instances cv_pb variance u1 u2 sigma =
   match Evarutil.compare_cumulative_instances cv_pb variance u1 u2 sigma with
   | Inl sigma -> sigma
   | Inr _ ->
-    raise Reduction.NotConvertible
+    raise Conversion.NotConvertible
 
 let sigma_univ_state =
-  let open Reduction in
+  let open Conversion in
   { compare_sorts = sigma_compare_sorts;
     compare_instances = sigma_compare_instances;
     compare_cumul_instances = sigma_check_inductive_instances; }
@@ -1104,8 +1141,8 @@ let sigma_univ_state =
 let univproblem_compare_sorts env pb s0 s1 uset =
   let open UnivProblem in
   match pb with
-  | Reduction.CONV -> UnivProblem.Set.add (UEq (s0, s1)) uset
-  | Reduction.CUMUL -> UnivProblem.Set.add (ULe (s0, s1)) uset
+  | Conversion.CONV -> UnivProblem.Set.add (UEq (s0, s1)) uset
+  | Conversion.CUMUL -> UnivProblem.Set.add (ULe (s0, s1)) uset
 
 let univproblem_compare_instances ~flex i0 i1 uset =
   UnivProblem.enforce_eq_instances_univs flex i0 i1 uset
@@ -1116,25 +1153,25 @@ let univproblem_check_inductive_instances cv_pb variances u u' uset =
   let mk u = Sorts.sort_of_univ @@ Univ.Universe.make u in
   let fold cstr v u u' = match v with
   | Irrelevant -> Set.add (UWeak (u,u')) cstr
-  | Covariant when cv_pb == Reduction.CUMUL -> Set.add (ULe (mk u, mk u')) cstr
+  | Covariant when cv_pb == Conversion.CUMUL -> Set.add (ULe (mk u, mk u')) cstr
   | Covariant | Invariant -> Set.add (UEq (mk u, mk u')) cstr
   in
   Array.fold_left3 fold uset
     variances (Univ.Instance.to_array u) (Univ.Instance.to_array u')
 
 let univproblem_univ_state =
-  let open Reduction in
+  let open Conversion in
   { compare_sorts = univproblem_compare_sorts;
     compare_instances = univproblem_compare_instances;
     compare_cumul_instances = univproblem_check_inductive_instances; }
 
-let infer_conv_gen conv_fun ?(catch_incon=true) ?(pb=Reduction.CUMUL)
+let infer_conv_gen conv_fun ?(catch_incon=true) ?(pb=Conversion.CUMUL)
     ?(ts=TransparentState.full) env sigma x y =
   try
       let ans = match pb with
-      | Reduction.CUMUL ->
+      | Conversion.CUMUL ->
           EConstr.leq_constr_universes env sigma x y
-      | Reduction.CONV ->
+      | Conversion.CONV ->
           EConstr.eq_constr_universes env sigma x y
       in
       let ans = match ans with
@@ -1154,22 +1191,19 @@ let infer_conv_gen conv_fun ?(catch_incon=true) ?(pb=Reduction.CUMUL)
             env (sigma, sigma_univ_state) x y in
         Some sigma'
   with
-  | Reduction.NotConvertible -> None
+  | Conversion.NotConvertible -> None
   | UGraph.UniverseInconsistency _ when catch_incon -> None
-  | e ->
-    let e = Exninfo.capture e in
-    report_anomaly e
 
 let infer_conv = infer_conv_gen (fun pb ~l2r sigma ->
-      Reduction.generic_conv pb ~l2r (Evd.evar_handler sigma))
+      Conversion.generic_conv pb ~l2r (Evd.evar_handler sigma))
 
-let infer_conv_ustate ?(catch_incon=true) ?(pb=Reduction.CUMUL)
+let infer_conv_ustate ?(catch_incon=true) ?(pb=Conversion.CUMUL)
     ?(ts=TransparentState.full) env sigma x y =
   try
       let ans = match pb with
-      | Reduction.CUMUL ->
+      | Conversion.CUMUL ->
           EConstr.leq_constr_universes env sigma x y
-      | Reduction.CONV ->
+      | Conversion.CONV ->
           EConstr.eq_constr_universes env sigma x y
       in
       match ans with
@@ -1179,20 +1213,17 @@ let infer_conv_ustate ?(catch_incon=true) ?(pb=Reduction.CUMUL)
         let y = EConstr.Unsafe.to_constr y in
         let env = Environ.set_universes (Evd.universes sigma) env in
         let cstr =
-          Reduction.generic_conv pb ~l2r:false (Evd.evar_handler sigma) ts
+          Conversion.generic_conv pb ~l2r:false (Evd.evar_handler sigma) ts
             env (UnivProblem.Set.empty, univproblem_univ_state) x y in
         Some cstr
   with
-  | Reduction.NotConvertible -> None
+  | Conversion.NotConvertible -> None
   | UGraph.UniverseInconsistency _ when catch_incon -> None
-  | e ->
-    let e = Exninfo.capture e in
-    report_anomaly e
 
 let evars_of_evar_map sigma =
   { Genlambda.evars_val = Evd.evar_handler sigma }
 
-let vm_infer_conv ?(pb=Reduction.CUMUL) env sigma t1 t2 =
+let vm_infer_conv ?(pb=Conversion.CUMUL) env sigma t1 t2 =
   infer_conv_gen (fun pb ~l2r sigma ts ->
       Vconv.vm_conv_gen pb (evars_of_evar_map sigma))
     ~catch_incon:true ~pb env sigma t1 t2
@@ -1200,50 +1231,19 @@ let vm_infer_conv ?(pb=Reduction.CUMUL) env sigma t1 t2 =
 let native_conv_generic pb sigma t =
   Nativeconv.native_conv_gen pb (evars_of_evar_map sigma) t
 
-let native_infer_conv ?(pb=Reduction.CUMUL) env sigma t1 t2 =
+let native_infer_conv ?(pb=Conversion.CUMUL) env sigma t1 t2 =
   infer_conv_gen (fun pb ~l2r sigma ts -> native_conv_generic pb sigma)
     ~catch_incon:true ~pb env sigma t1 t2
 
+let check_hyps_inclusion env sigma x hyps =
+  let env = Environ.set_universes (Evd.universes sigma) env in
+  let evars = Evd.evar_handler sigma in
+  Typeops.check_hyps_inclusion env ~evars x hyps
+
 (********************************************************************)
 (*             Special-Purpose Reduction                            *)
 (********************************************************************)
 
-let default_plain_instance_ident = Id.of_string "H"
-
-type subst_fun = { sfun : metavariable -> EConstr.t }
-
-(* Try to replace all metas. Does not replace metas in the metas' values
- * Differs from (strong whd_meta). *)
-let plain_instance sigma s c = match s with
-| None -> c
-| Some s ->
-  let rec irec n u = match EConstr.kind sigma u with
-    | Meta p -> (try lift n (s.sfun p) with Not_found -> u)
-    | App (f,l) when isCast sigma f ->
-        let (f,_,t) = destCast sigma f in
-        let l' = Array.Fun1.Smart.map irec n l in
-        (match EConstr.kind sigma f with
-        | Meta p ->
-            (* Don't flatten application nodes: this is used to extract a
-               proof-term from a proof-tree and we want to keep the structure
-               of the proof-tree *)
-            (try let g = s.sfun p in
-            match EConstr.kind sigma g with
-            | App _ ->
-                let l' = Array.Fun1.Smart.map lift 1 l' in
-                let r = Sorts.Relevant in (* TODO fix relevance *)
-                let na = make_annot (Name default_plain_instance_ident) r in
-                mkLetIn (na,g,t,mkApp(mkRel 1, l'))
-            | _ -> mkApp (g,l')
-            with Not_found -> mkApp (f,l'))
-        | _ -> mkApp (irec n f,l'))
-    | Cast (m,_,_) when isMeta sigma m ->
-        (try lift n (s.sfun (destMeta sigma m)) with Not_found -> u)
-    | _ ->
-        map_with_binders sigma succ irec n u
-  in
-  irec 0 c
-
 (* [instance] is used for [res_pf]; the call to [local_strong whd_betaiota]
    has (unfortunately) different subtle side effects:
 
@@ -1278,11 +1278,11 @@ let plain_instance sigma s c = match s with
      empty map).
  *)
 
-let instance env sigma s c =
+let instance env sigma c =
   (* if s = [] then c else *)
   (* No need to compute contexts under binders as whd_betaiota is local *)
   let rec strongrec t = EConstr.map sigma strongrec (whd_betaiota env sigma t) in
-  strongrec (plain_instance sigma s c)
+  strongrec c
 
 (* pseudo-reduction rule:
  * [hnf_prod_app env s (Prod(_,B)) N --> B[N]
@@ -1312,7 +1312,7 @@ let hnf_lam_appvect env sigma t nl =
 let hnf_lam_applist env sigma t nl =
   List.fold_left (fun acc t -> hnf_lam_app env sigma acc t) t nl
 
-let splay_prod env sigma =
+let hnf_decompose_prod env sigma =
   let rec decrec env m c =
     let t = whd_all env sigma c in
     match EConstr.kind sigma t with
@@ -1322,7 +1322,7 @@ let splay_prod env sigma =
   in
   decrec env []
 
-let splay_lam env sigma =
+let hnf_decompose_lambda env sigma =
   let rec decrec env m c =
     let t = whd_all env sigma c in
     match EConstr.kind sigma t with
@@ -1332,7 +1332,7 @@ let splay_lam env sigma =
   in
   decrec env []
 
-let splay_prod_assum env sigma =
+let hnf_decompose_prod_decls env sigma =
   let rec prodec_rec env l c =
     let t = whd_allnolet env sigma c in
     match EConstr.kind sigma t with
@@ -1351,30 +1351,30 @@ let splay_prod_assum env sigma =
   prodec_rec env Context.Rel.empty
 
 let splay_arity env sigma c =
-  let l, c = splay_prod env sigma c in
+  let l, c = hnf_decompose_prod env sigma c in
   match EConstr.kind sigma c with
     | Sort s -> l,s
     | _ -> raise Reduction.NotArity
 
 let sort_of_arity env sigma c = snd (splay_arity env sigma c)
 
-let splay_prod_n env sigma n =
+let hnf_decompose_prod_n_decls env sigma n =
   let rec decrec env m ln c = if Int.equal m 0 then (ln,c) else
     match EConstr.kind sigma (whd_all env sigma c) with
       | Prod (n,a,c0) ->
           decrec (push_rel (LocalAssum (n,a)) env)
             (m-1) (Context.Rel.add (LocalAssum (n,a)) ln) c0
-      | _                      -> invalid_arg "splay_prod_n"
+      | _                      -> invalid_arg "hnf_decompose_prod_n_decls"
   in
   decrec env n Context.Rel.empty
 
-let splay_lam_n env sigma n =
+let hnf_decompose_lambda_n_assum env sigma n =
   let rec decrec env m ln c = if Int.equal m 0 then (ln,c) else
     match EConstr.kind sigma (whd_all env sigma c) with
       | Lambda (n,a,c0) ->
           decrec (push_rel (LocalAssum (n,a)) env)
             (m-1) (Context.Rel.add (LocalAssum (n,a)) ln) c0
-      | _                      -> invalid_arg "splay_lam_n"
+      | _                      -> invalid_arg "hnf_decompose_lambda_n_assum"
   in
   decrec env n Context.Rel.empty
 
@@ -1424,7 +1424,7 @@ let whd_betaiota_deltazeta_for_iota_state ts env sigma s =
     | (FConstruct _ | FCoFix _) ->
       (* Non-neutral normal, can trigger reduction below *)
       let c = EConstr.of_constr (term_of_process c stk) in
-      Some (decompose_app_vect sigma c)
+      Some (decompose_app sigma c)
     | _ -> None
   in
   let rec whrec s =
@@ -1468,76 +1468,44 @@ let is_arity env sigma c =
 (*************************************)
 (* Metas *)
 
-type meta_instance_subst = {
-  sigma : Evd.evar_map;
-  mutable cache : EConstr.t Metamap.t;
-}
-
-let create_meta_instance_subst sigma = {
-  sigma;
-  cache = Metamap.empty;
-}
-
-let eval_subst env subst =
-  let rec ans mv =
-    try Metamap.find mv subst.cache
-    with Not_found ->
-      match meta_opt_fvalue subst.sigma mv with
-      | None -> mkMeta mv
-      | Some (b, _) ->
-        let metas =
-          if Metaset.is_empty b.freemetas then None
-          else Some { sfun = ans }
-        in
-        let res = instance env subst.sigma metas b.rebus in
-        let () = subst.cache <- Metamap.add mv res subst.cache in
-        res
-  in
-  { sfun = ans }
-
-let meta_instance env subst b =
+let meta_instance env sigma b =
   let fm = b.freemetas in
   if Metaset.is_empty fm then b.rebus
   else
-    let sfun = eval_subst env subst in
-    instance env subst.sigma (Some sfun) b.rebus
-
-let nf_meta env sigma c =
-  let sigma = create_meta_instance_subst sigma in
-  let cl = mk_freelisted c in
-  meta_instance env sigma { cl with rebus = cl.rebus }
+    instance env sigma b.rebus
 
 module Infer = struct
 
-open Reduction
+open Conversion
 
 let infer_eq (univs, cstrs as cuniv) u u' =
   if UGraph.check_eq_sort univs u u' then cuniv
   else
-    univs, (UnivSubst.enforce_eq_sort u u' cstrs)
+    let cstrs' = UnivSubst.enforce_eq_sort u u' Constraints.empty in
+    UGraph.merge_constraints cstrs' univs, Constraints.union cstrs cstrs'
 
 let infer_leq (univs, cstrs as cuniv) u u' =
   if UGraph.check_leq_sort univs u u' then cuniv
   else
-    let cstrs', _ = UnivSubst.enforce_leq_alg_sort u u' univs in
-      univs, Univ.Constraints.union cstrs cstrs'
+    let cstrs', univs = UnivSubst.enforce_leq_alg_sort u u' univs in
+    univs, Univ.Constraints.union cstrs cstrs'
 
 let infer_cmp_universes _env pb s0 s1 univs =
   match pb with
   | CUMUL -> infer_leq univs s0 s1
   | CONV -> infer_eq univs s0 s1
 
-let infer_convert_instances ~flex u u' (univs,cstrs) =
-  let cstrs' =
-    if flex then
-      if UGraph.check_eq_instances univs u u' then cstrs
-      else raise NotConvertible
-    else Univ.enforce_eq_instances u u' cstrs
-  in (univs, cstrs')
+let infer_convert_instances ~flex u u' (univs,cstrs as cuniv) =
+  if flex then
+    if UGraph.check_eq_instances univs u u' then cuniv
+    else raise NotConvertible
+  else
+    let cstrs' = Univ.enforce_eq_instances u u' Constraints.empty in
+    (univs, Constraints.union cstrs cstrs')
 
-let infer_inductive_instances cv_pb variance u1 u2 (univs,csts') =
-  let csts = get_cumulativity_constraints cv_pb variance u1 u2 in
-  (univs, Univ.Constraints.union csts csts')
+let infer_inductive_instances cv_pb variance u1 u2 (univs,csts) =
+  let csts' = get_cumulativity_constraints cv_pb variance u1 u2 in
+  (UGraph.merge_constraints csts' univs, Univ.Constraints.union csts csts')
 
 let inferred_universes : (UGraph.t * Univ.Constraints.t) universe_compare =
   { compare_sorts = infer_cmp_universes;
@@ -1547,3 +1515,11 @@ let inferred_universes : (UGraph.t * Univ.Constraints.t) universe_compare =
 end
 
 let inferred_universes = Infer.inferred_universes
+
+(* Deprecated *)
+
+let splay_prod = hnf_decompose_prod
+let splay_lam = hnf_decompose_lambda
+let splay_prod_assum = hnf_decompose_prod_decls
+let splay_prod_n = hnf_decompose_prod_n_decls
+let splay_lam_n = hnf_decompose_lambda_n_assum
