@@ -27,23 +27,23 @@ module CompactedDecl = Context.Compacted.Declaration
 (* This is set on by proofgeneral proof-tree mode. But may be used for
    other purposes *)
 let print_goal_tag_opt_name = ["Printing";"Goal";"Tags"]
-let should_tag =
+let { Goptions.get = should_tag } =
   Goptions.declare_bool_option_and_ref
-    ~depr:false
     ~key:print_goal_tag_opt_name
     ~value:false
+    ()
 
-let should_unfoc =
+let { Goptions.get = should_unfoc } =
   Goptions.declare_bool_option_and_ref
-    ~depr:false
     ~key:["Printing";"Unfocused"]
     ~value:false
+    ()
 
-let should_gname =
+let { Goptions.get = should_gname } =
   Goptions.declare_bool_option_and_ref
-    ~depr:false
     ~key:["Printing";"Goal";"Names"]
     ~value:false
+    ()
 
 let print_goal_names = should_gname (* for export *)
 
@@ -58,24 +58,24 @@ let print_goal_names = should_gname (* for export *)
    and only names of goal/section variables and rel names that do
    _not_ occur in the scope of the binder to be printed are avoided. *)
 
-let pr_econstr_n_env ?lax ?inctx ?scope env sigma n t =
-  pr_constr_expr_n env sigma n (extern_constr ?lax ?inctx ?scope env sigma t)
-let pr_econstr_env ?lax ?inctx ?scope env sigma t =
-  pr_constr_expr env sigma (extern_constr ?lax ?inctx ?scope env sigma t)
-let pr_leconstr_env ?lax ?inctx ?scope env sigma t =
-  Ppconstr.pr_lconstr_expr env sigma (extern_constr ?lax ?inctx ?scope env sigma t)
+let pr_econstr_n_env ?inctx ?scope env sigma n t =
+  pr_constr_expr_n env sigma n (extern_constr ?inctx ?scope env sigma t)
+let pr_econstr_env ?inctx ?scope env sigma t =
+  pr_constr_expr env sigma (extern_constr ?inctx ?scope env sigma t)
+let pr_leconstr_env ?inctx ?scope env sigma t =
+  Ppconstr.pr_lconstr_expr env sigma (extern_constr ?inctx ?scope env sigma t)
 
-let pr_constr_n_env ?lax ?inctx ?scope env sigma n c =
-  pr_econstr_n_env ?lax ?inctx ?scope env sigma n (EConstr.of_constr c)
-let pr_constr_env ?lax ?inctx ?scope env sigma c =
-  pr_econstr_env ?lax ?inctx ?scope env sigma (EConstr.of_constr c)
-let pr_lconstr_env ?lax ?inctx ?scope env sigma c =
-  pr_leconstr_env ?lax ?inctx ?scope env sigma (EConstr.of_constr c)
+let pr_constr_n_env ?inctx ?scope env sigma n c =
+  pr_econstr_n_env ?inctx ?scope env sigma n (EConstr.of_constr c)
+let pr_constr_env ?inctx ?scope env sigma c =
+  pr_econstr_env ?inctx ?scope env sigma (EConstr.of_constr c)
+let pr_lconstr_env ?inctx ?scope env sigma c =
+  pr_leconstr_env ?inctx ?scope env sigma (EConstr.of_constr c)
 
-let pr_open_lconstr_env ?lax ?inctx ?scope env sigma (_,c) =
-  pr_leconstr_env ?lax ?inctx ?scope env sigma c
-let pr_open_constr_env ?lax ?inctx ?scope env sigma (_,c) =
-  pr_econstr_env ?lax ?inctx ?scope env sigma c
+let pr_open_lconstr_env ?inctx ?scope env sigma (_,c) =
+  pr_leconstr_env ?inctx ?scope env sigma c
+let pr_open_constr_env ?inctx ?scope env sigma (_,c) =
+  pr_econstr_env ?inctx ?scope env sigma c
 
 let pr_constr_under_binders_env_gen pr env sigma (ids,c) =
   (* Warning: clashes can occur with variables of same name in env but *)
@@ -87,15 +87,15 @@ let pr_constr_under_binders_env_gen pr env sigma (ids,c) =
 let pr_constr_under_binders_env = pr_constr_under_binders_env_gen pr_econstr_env
 let pr_lconstr_under_binders_env = pr_constr_under_binders_env_gen pr_leconstr_env
 
-let pr_etype_env ?lax ?goal_concl_style env sigma t =
-  pr_constr_expr env sigma (extern_type ?lax ?goal_concl_style env sigma t)
-let pr_letype_env ?lax ?goal_concl_style env sigma ?impargs t =
-  pr_lconstr_expr env sigma (extern_type ?lax ?goal_concl_style env sigma ?impargs t)
+let pr_etype_env ?goal_concl_style env sigma t =
+  pr_constr_expr env sigma (extern_type ?goal_concl_style env sigma t)
+let pr_letype_env ?goal_concl_style env sigma ?impargs t =
+  pr_lconstr_expr env sigma (extern_type ?goal_concl_style env sigma ?impargs t)
 
-let pr_type_env ?lax ?goal_concl_style env sigma c =
-  pr_etype_env ?lax ?goal_concl_style env sigma (EConstr.of_constr c)
-let pr_ltype_env ?lax ?goal_concl_style env sigma ?impargs c =
-  pr_letype_env ?lax ?goal_concl_style env sigma ?impargs (EConstr.of_constr c)
+let pr_type_env ?goal_concl_style env sigma c =
+  pr_etype_env ?goal_concl_style env sigma (EConstr.of_constr c)
+let pr_ltype_env ?goal_concl_style env sigma ?impargs c =
+  pr_letype_env ?goal_concl_style env sigma ?impargs (EConstr.of_constr c)
 
 let pr_ljudge_env env sigma j =
   (pr_leconstr_env env sigma j.uj_val, pr_leconstr_env env sigma j.uj_type)
@@ -105,10 +105,10 @@ let pr_lglob_constr_env env sigma c =
 let pr_glob_constr_env env sigma c =
   pr_constr_expr env sigma (extern_glob_constr (extern_env env sigma) c)
 
-let pr_closed_glob_n_env ?lax ?goal_concl_style ?inctx ?scope env sigma n c =
-  pr_constr_expr_n env sigma n (extern_closed_glob ?lax ?goal_concl_style ?inctx ?scope env sigma c)
-let pr_closed_glob_env ?lax ?goal_concl_style ?inctx ?scope env sigma c =
-  pr_constr_expr env sigma (extern_closed_glob ?lax ?goal_concl_style ?inctx ?scope env sigma c)
+let pr_closed_glob_n_env ?goal_concl_style ?inctx ?scope env sigma n c =
+  pr_constr_expr_n env sigma n (extern_closed_glob ?goal_concl_style ?inctx ?scope env sigma c)
+let pr_closed_glob_env ?goal_concl_style ?inctx ?scope env sigma c =
+  pr_constr_expr env sigma (extern_closed_glob ?goal_concl_style ?inctx ?scope env sigma c)
 
 let pr_lconstr_pattern_env env sigma c =
   pr_lconstr_pattern_expr env sigma (extern_constr_pattern (Termops.names_of_rel_context env) sigma c)
@@ -121,7 +121,7 @@ let pr_cases_pattern t =
 let pr_sort sigma s = pr_sort_expr (extern_sort sigma s)
 
 let () = Termops.Internal.set_print_constr
-  (fun env sigma t -> pr_lconstr_expr env sigma (extern_constr ~lax:true env sigma t))
+  (fun env sigma t -> pr_lconstr_expr env sigma (extern_constr env sigma t))
 
 let pr_in_comment x = str "(* " ++ x ++ str " *)"
 
@@ -180,6 +180,7 @@ let safe_pr_constr_env = safe_gen pr_constr_env
 let u_ident = Id.of_string "u"
 
 let universe_binders_with_opt_names orig names =
+  let open Univ in
   let orig = Univ.AbstractContext.names orig in
   let orig = Array.to_list orig in
   let udecl = match names with
@@ -197,19 +198,21 @@ let universe_binders_with_opt_names orig names =
           expect = List.length udecl;
         })
   in
-  let fold_named i ubind = function
-    | Name id -> Id.Map.add id (Univ.Level.var i) ubind
-    | Anonymous -> ubind
+  let fold_named i (ubind,revubind as o) = function
+    | Name id -> let ui = Level.var i in
+      Id.Map.add id ui ubind, Level.Map.add ui id revubind
+    | Anonymous -> o
   in
-  let ubind = List.fold_left_i fold_named 0 UnivNames.empty_binders udecl in
-  let fold_anons i (u_ident, ubind) = function
-    | Name _ -> u_ident, ubind
+  let names = List.fold_left_i fold_named 0 (UnivNames.empty_binders,Level.Map.empty) udecl in
+  let fold_anons i (u_ident, (ubind, revubind) as o) = function
+    | Name _ -> o
     | Anonymous ->
+      let ui = Level.var i in
       let id = Namegen.next_ident_away_from u_ident (fun id -> Id.Map.mem id ubind) in
-      (id, Id.Map.add id (Univ.Level.var i) ubind)
+      (id, (Id.Map.add id ui ubind, Level.Map.add ui id revubind))
   in
-  let (_, ubind) = List.fold_left_i fold_anons 0 (u_ident, ubind) udecl in
-  ubind
+  let (_, names) = List.fold_left_i fold_anons 0 (u_ident, names) udecl in
+  names
 
 let pr_universe_ctx_set sigma c =
   if !Detyping.print_universes && not (Univ.ContextSet.is_empty c) then
@@ -310,9 +313,17 @@ let pr_compacted_decl env sigma decl =
   let ptyp = (str" : " ++ pt) in
   hov 0 (pids ++ pbody ++ ptyp)
 
+let pr_ecompacted_decl env sigma (decl:EConstr.compacted_declaration) =
+  let Refl = EConstr.Unsafe.eq in
+  pr_compacted_decl env sigma decl
+
 let pr_named_decl env sigma decl =
   decl |> CompactedDecl.of_named_decl |> pr_compacted_decl env sigma
 
+let pr_enamed_decl env sigma (decl:EConstr.named_declaration) =
+  let Refl = EConstr.Unsafe.eq in
+  pr_named_decl env sigma decl
+
 let pr_rel_decl env sigma decl =
   let na = RelDecl.get_name decl in
   let typ = RelDecl.get_type decl in
@@ -328,6 +339,11 @@ let pr_rel_decl env sigma decl =
   | Anonymous -> hov 0 (str"<>" ++ spc () ++ pbody ++ str":" ++ spc () ++ ptyp)
   | Name id -> hov 0 (pr_id id ++ spc () ++ pbody ++ str":" ++ spc () ++ ptyp)
 
+let pr_erel_decl env sigma (decl:EConstr.rel_declaration) =
+  let Refl = EConstr.Unsafe.eq in
+  pr_rel_decl env sigma decl
+
+
 
 (* Prints out an "env" in a nice format.  We print out the
  * signature,then a horizontal bar, then the debruijn environment.
@@ -340,7 +356,7 @@ let pr_named_context_of env sigma =
   hv 0 (prlist_with_sep (fun _ -> ws 2) (fun x -> x) psl)
 
 let pr_var_list_decl env sigma decl =
-  hov 0 (pr_compacted_decl env sigma decl)
+  hov 0 (pr_ecompacted_decl env sigma decl)
 
 let pr_named_context env sigma ne_context =
   hv 0 (Context.Named.fold_outside
@@ -359,9 +375,9 @@ let pr_context_unlimited env sigma =
   let sign_env =
     Context.Compacted.fold
       (fun d pps ->
-         let pidt =  pr_compacted_decl env sigma d in
+         let pidt =  pr_ecompacted_decl env sigma d in
          (pps ++ fnl () ++ pidt))
-      (Termops.compact_named_context (named_context env)) ~init:(mt ())
+      (Termops.compact_named_context sigma (EConstr.named_context env)) ~init:(mt ())
   in
   let db_env =
     fold_rel_context
@@ -380,8 +396,8 @@ let pr_ne_context_of header env sigma =
    considers as "variables": An hypothesis H:T where T:S and S<>Prop. *)
 let should_compact env sigma typ =
   get_compact_context() &&
-    let type_of_typ = Retyping.get_type_of env sigma (EConstr.of_constr typ) in
-    not (is_Prop (EConstr.to_constr sigma type_of_typ))
+    let type_of_typ = Retyping.get_type_of env sigma typ in
+    not (Termops.is_Prop sigma type_of_typ)
 
 
 (* If option Compact Contexts is set, we pack "simple" hypothesis in a
@@ -411,7 +427,8 @@ and bld_sign_env_id env sigma ctxt pps is_start =
 (* compact printing an env (variables and de Bruijn). Separator: three
    spaces between simple hyps, and newline otherwise *)
 let pr_context_limit_compact ?n env sigma =
-  let ctxt = Termops.compact_named_context (named_context env) in
+  let ctxt = EConstr.named_context env in
+  let ctxt = Termops.compact_named_context sigma ctxt in
   let lgth = List.length ctxt in
   let n_capped =
     match n with
@@ -430,8 +447,11 @@ let pr_context_limit_compact ?n env sigma =
 
 (* The number of printed hypothesis in a goal *)
 (* If [None], no limit *)
-let print_hyps_limit =
-  Goptions.declare_intopt_option_and_ref ~depr:false ~key:["Hyps";"Limit"]
+let { Goptions.get = print_hyps_limit } =
+  Goptions.declare_intopt_option_and_ref
+    ~key:["Hyps";"Limit"]
+    ~value:None
+    ()
 
 let pr_context_of env sigma = match print_hyps_limit () with
   | None -> hv 0 (pr_context_limit_compact env sigma)
@@ -461,8 +481,13 @@ let pr_transparent_state ts =
         str"CONSTANTS: " ++ pr_cpred ts.TransparentState.tr_cst ++ fnl ())
 
 let goal_repr sigma g =
-  let evi = Evd.find sigma g in
-  Evd.evar_filtered_env (Global.env ()) evi, Evd.evar_concl evi
+  let EvarInfo evi = Evd.find sigma g in
+  let env = Evd.evar_filtered_env (Global.env ()) evi in
+  let concl = match Evd.evar_body evi with
+  | Evd.Evar_empty -> Evd.evar_concl evi
+  | Evd.Evar_defined b -> Retyping.get_type_of env sigma b
+  in
+  env, concl
 
 (* display complete goal
  og_s has goal+sigma on the previous proof step for diffs
@@ -517,7 +542,7 @@ let pr_concl n ?diffs sigma g =
   header ++ str " is:" ++ cut () ++ str" "  ++ pc
 
 (* display evar type: a context and a type *)
-let pr_evgl_sign env sigma evi =
+let pr_evgl_sign env sigma (evi : undefined evar_info) =
   let env = evar_env env evi in
   let ps = pr_named_context_of env sigma in
   let _, l = match Filter.repr (evar_filter evi) with
@@ -529,14 +554,15 @@ let pr_evgl_sign env sigma evi =
     if List.is_empty ids then mt () else
       (str " (" ++ prlist_with_sep pr_comma pr_id ids ++ str " cannot be used)")
   in
-  let pc = pr_leconstr_env env sigma evi.evar_concl in
+  let concl = Evd.evar_concl evi in
+  let pc = pr_leconstr_env env sigma concl in
   let candidates =
-    match evi.evar_body, evi.evar_candidates with
-    | Evar_empty, Some l ->
-       spc () ++ str "= {" ++
-         prlist_with_sep (fun () -> str "|") (pr_leconstr_env env sigma) l ++ str "}"
-    | _ ->
-       mt ()
+    begin match Evd.evar_candidates evi with
+    | None -> mt ()
+    | Some l ->
+      spc () ++ str "= {" ++
+      prlist_with_sep (fun () -> str "|") (pr_leconstr_env env sigma) l ++ str "}"
+    end
   in
   hov 0 (str"[" ++ ps ++ spc () ++ str"|- "  ++ pc ++ str"]" ++
            candidates ++ warn)
@@ -572,9 +598,11 @@ let pr_evars sigma evs =
 (* Display a list of evars given by their name, with a prefix *)
 let pr_ne_evar_set hd tl sigma l =
   if l != Evar.Set.empty then
-    let l = Evar.Set.fold (fun ev ->
-      Evar.Map.add ev (Evarutil.nf_evar_info sigma (Evd.find sigma ev)))
-      l Evar.Map.empty in
+    let l = Evar.Map.bind (fun ev ->
+        let evi = Evd.find_undefined sigma ev in
+        Evarutil.nf_evar_info sigma evi)
+        l
+    in
     hd ++ pr_evars sigma l ++ tl
   else
     mt ()
@@ -624,12 +652,12 @@ let print_evar_constraints gl sigma =
     str" " ++
       hov 2 (pr_env env ++ pr_leconstr_env env sigma t1 ++ spc () ++
              str (match pbty with
-                  | Reduction.CONV -> "=="
-                  | Reduction.CUMUL -> "<=") ++
+                  | Conversion.CONV -> "=="
+                  | Conversion.CUMUL -> "<=") ++
              spc () ++ pr_leconstr_env env sigma t2)
   in
   let pr_candidate ev evi (candidates,acc) =
-    if Option.has_some evi.evar_candidates then
+    if Option.has_some (Evd.evar_candidates evi) then
       (succ candidates, acc ++ pr_evar sigma (ev,evi) ++ fnl ())
     else (candidates, acc)
   in
@@ -646,18 +674,84 @@ let print_evar_constraints gl sigma =
         str" with candidates:" ++ fnl () ++ hov 0 ppcandidates
     else mt ()
 
-let should_print_dependent_evars =
+let { Goptions.get = should_print_dependent_evars } =
   Goptions.declare_bool_option_and_ref
-    ~depr:false
     ~key:["Printing";"Dependent";"Evars";"Line"]
     ~value:false
+    ()
+
+let evar_nodes_of_term c =
+  let rec evrec acc c =
+    match kind c with
+    | Evar (n, l) -> Evar.Set.add n (SList.Skip.fold evrec acc l)
+    | _ -> Constr.fold evrec acc c
+  in
+  evrec Evar.Set.empty (EConstr.Unsafe.to_constr c)
+
+(* spiwack: a few functions to gather evars on which goals depend. *)
+let queue_set q is_dependent set =
+  Evar.Set.iter (fun a -> Queue.push (is_dependent,a) q) set
+let queue_term q is_dependent c =
+  queue_set q is_dependent (evar_nodes_of_term c)
+
+let process_dependent_evar q acc evm is_dependent e =
+  let EvarInfo evi = Evd.find evm e in
+  (* Queues evars appearing in the types of the goal (conclusion, then
+     hypotheses), they are all dependent. *)
+  let () = match Evd.evar_body evi with
+  | Evar_empty ->
+    queue_term q true (Evd.evar_concl evi)
+  | Evar_defined b -> ()
+  in
+  List.iter begin fun decl ->
+    let open NamedDecl in
+    queue_term q true (NamedDecl.get_type decl);
+    match decl with
+    | LocalAssum _ -> ()
+    | LocalDef (_,b,_) -> queue_term q true b
+  end (EConstr.named_context_of_val (Evd.evar_hyps evi));
+  match Evd.evar_body evi with
+  | Evar_empty ->
+      if is_dependent then Evar.Map.add e None acc else acc
+  | Evar_defined b ->
+      let subevars = evar_nodes_of_term b in
+      (* evars appearing in the definition of an evar [e] are marked
+         as dependent when [e] is dependent itself: if [e] is a
+         non-dependent goal, then, unless they are reach from another
+         path, these evars are just other non-dependent goals. *)
+      queue_set q is_dependent subevars;
+      if is_dependent then Evar.Map.add e (Some subevars) acc else acc
+
+(** [gather_dependent_evars evm seeds] classifies the evars in [evm]
+    as dependent_evars and goals (these may overlap). A goal is an evar
+    appearing in the (partial) definition [seeds] (including defined evars). A
+    dependent evar is an evar appearing in the type
+    (hypotheses and conclusion) of a goal, or in the type or (partial)
+    definition of a dependent evar.  The value return is a map
+    associating to each dependent evar [None] if it has no (partial)
+    definition or [Some s] if [s] is the list of evars appearing in
+    its (partial) definition. This completely breaks the EConstr abstraction. *)
+let gather_dependent_evars evm l =
+  let q = Queue.create () in
+  List.iter (queue_term q false) l;
+  let acc = ref Evar.Map.empty in
+  while not (Queue.is_empty q) do
+    let (is_dependent,e) = Queue.pop q in
+    (* checks if [e] has already been added to [!acc] *)
+    begin if not (Evar.Map.mem e !acc) then
+        acc := process_dependent_evar q !acc evm is_dependent e
+    end
+  done;
+  !acc
+
+(* /spiwack *)
 
 let gather_dependent_evars_goal sigma goals =
   let map evk =
-    let evi = Evd.find sigma evk in
+    let EvarInfo evi = Evd.find sigma evk in
     EConstr.mkEvar (evk, Evd.evar_identity_subst evi)
   in
-  Evarutil.gather_dependent_evars sigma (List.map map goals)
+  gather_dependent_evars sigma (List.map map goals)
 
 let print_dependent_evars_core gl sigma evars =
   let mt_pp = mt () in
@@ -698,7 +792,7 @@ let print_dependent_evars_entry gl sigma = function
   | Some entry ->
     if should_print_dependent_evars () then
       let terms = List.map pi2 (Proofview.initial_goals entry) in
-      let evars = Evarutil.gather_dependent_evars sigma terms in
+      let evars = gather_dependent_evars sigma terms in
       print_dependent_evars_core gl sigma evars
     else mt ()
 
@@ -874,7 +968,8 @@ let pr_goal_emacs ~proof gid sid =
     in
     try
       let { Proof.sigma } = Proof.data proof in
-      pr sigma (Evar.unsafe_of_int gid)
+      let gl = Evar.unsafe_of_int gid in
+      v 0 (pr sigma gl ++ print_dependent_evars (Some gl) sigma [ gl ])
     with Not_found -> user_err Pp.(str "No such goal.")
 
 (* Printer function for sets of Assumptions.assumptions.
@@ -1038,7 +1133,6 @@ let pr_typing_flags flags =
   str "check_guarded: " ++ bool flags.check_guarded ++ fnl ()
   ++ str "check_positive: " ++ bool flags.check_positive ++ fnl ()
   ++ str "check_universes: " ++ bool flags.check_universes ++ fnl ()
-  ++ str "cumulative sprop: " ++ bool flags.cumulative_sprop ++ fnl ()
   ++ str "definitional uip: " ++ bool flags.allow_uip
 
 module Debug =
