@@ -14,7 +14,6 @@ open Constr
 open Context
 open Termops
 open Univ
-open Evd
 open Environ
 open EConstr
 open Vars
@@ -29,6 +28,8 @@ open Context.Rel.Declaration
 (** Machinery to custom the behavior of the reduction *)
 module ReductionBehaviour = Reductionops.ReductionBehaviour
 
+type volatile = { volatile : bool } [@@unboxed]
+
 (** Machinery about stack of unfolded constants *)
 module Cst_stack = struct
   open EConstr
@@ -38,34 +39,34 @@ module Cst_stack = struct
 - constant applied to params = term in head applied to args
 - there is at most one arguments with an empty list of args, it must be the first.
 - in args, the int represents the indice of the first arg to consider *)
-  type t = (constr * constr list * (int * constr array) list)  list
+  type t = (constr * constr list * (int * constr array) list * volatile)  list
 
   let empty = []
-  let is_empty = CList.is_empty
+  let all_volatile = CList.for_all (fun (_,_,_,{volatile}) -> volatile)
 
   let drop_useless = function
-    | _ :: ((_,_,[])::_ as q) -> q
+    | _ :: ((_,_,[],_)::_ as q) -> q
     | l -> l
 
   let add_param h cst_l =
     let append2cst = function
-      | (c,params,[]) -> (c, h::params, [])
-      | (c,params,((i,t)::q)) when i = pred (Array.length t) ->
-        (c, params, q)
-      | (c,params,(i,t)::q) ->
-        (c, params, (succ i,t)::q)
+      | (c,params,[],vol) -> (c, h::params, [], vol)
+      | (c,params,((i,t)::q),vol) when i = pred (Array.length t) ->
+        (c, params, q, vol)
+      | (c,params,(i,t)::q, vol) ->
+        (c, params, (succ i,t)::q, vol)
     in
       drop_useless (List.map append2cst cst_l)
 
   let add_args cl =
-    List.map (fun (a,b,args) -> (a,b,(0,cl)::args))
+    List.map (fun (a,b,args,vol) -> (a,b,(0,cl)::args,vol))
 
-  let add_cst cst = function
-    | (_,_,[]) :: q as l -> l
-    | l -> (cst,[],[])::l
+  let add_cst ?(volatile=false) cst = function
+    | (_,_,[],_) :: q as l -> l
+    | l -> (cst,[],[],{volatile})::l
 
   let best_cst = function
-    | (cst,params,[])::_ -> Some(cst,params)
+    | (cst,params,[],_)::_ -> Some(cst,params)
     | _ -> None
 
   let reference sigma t = match best_cst t with
@@ -78,7 +79,7 @@ module Cst_stack = struct
     let reconstruct_head = List.fold_left
       (fun t (i,args) -> mkApp (t,Array.sub args i (Array.length args - i))) in
     List.fold_right
-      (fun (cst,params,args) t -> Termops.replace_term sigma
+      (fun (cst,params,args,_) t -> Termops.replace_term sigma
         (reconstruct_head d args)
         (applist (cst, List.rev params))
         t) cst_l c
@@ -87,10 +88,11 @@ module Cst_stack = struct
     let open Pp in
     let p_c c = Termops.Internal.print_constr_env env sigma c in
     prlist_with_sep pr_semicolon
-      (fun (c,params,args) ->
+      (fun (c,params,args,{volatile}) ->
         hov 1 (str"(" ++ p_c c ++ str ")" ++ spc () ++ pr_sequence p_c params ++ spc () ++ str "(args:" ++
                  pr_sequence (fun (i,el) -> prvect_with_sep spc p_c (Array.sub el i (Array.length el - i))) args ++
-                 str ")")) l
+               str ")" ++
+              (if volatile then str " (volatile)" else mt()))) l
 end
 
 
@@ -122,8 +124,6 @@ sig
 
   and 'a t = 'a member list
 
-  exception IncompatibleFold2
-
   val pr : ('a -> Pp.t) -> 'a t -> Pp.t
   val empty : 'a t
   val append_app : 'a array -> 'a t -> 'a t
@@ -137,8 +137,8 @@ sig
   val args_size : 'a t -> int
   val tail : int -> 'a t -> 'a t
   val nth : 'a t -> int -> 'a
-  val best_state : evar_map -> constr * constr t -> Cst_stack.t -> constr * constr t
-  val zip : ?refold:bool -> evar_map -> constr * constr t -> constr
+  val best_state : Evd.evar_map -> constr * constr t -> Cst_stack.t -> constr * constr t
+  val zip : ?refold:bool -> Evd.evar_map -> constr * constr t -> constr
   val check_native_args : CPrimitives.t -> 'a t -> bool
   val get_next_primitive_args : CPrimitives.args_red -> 'a t -> CPrimitives.args_red * ('a t * 'a * 'a t) option
 end =
@@ -215,7 +215,7 @@ struct
       | Cst_const (c, u) ->
         if Univ.Instance.is_empty u then Constant.debug_print c
         else str"(" ++ Constant.debug_print c ++ str ", " ++
-          Univ.Instance.pr Univ.Level.pr u ++ str")"
+          Univ.Instance.pr Univ.Level.raw_pr u ++ str")"
       | Cst_proj p ->
         str".(" ++ Constant.debug_print (Projection.constant p) ++ str")"
 
@@ -268,8 +268,6 @@ struct
       | ((App _|Case _|Proj _|Fix _|Cst _|Primitive _)::_|[]), _ -> false
     in equal_rec (List.rev sk1) (List.rev sk2)
 
-  exception IncompatibleFold2
-
   let append_app_list l s =
     let a = Array.of_list l in
     append_app a s
@@ -301,7 +299,7 @@ struct
   let will_expose_iota args =
     List.exists
       (function (Fix (_,_,l) | Case (_,l) |
-                 Proj (_,l) | Cst {cst_l=l}) when Cst_stack.is_empty l -> true | _ -> false)
+                 Proj (_,l) | Cst {cst_l=l}) when Cst_stack.all_volatile l -> true | _ -> false)
       args
 
   let list_of_app_stack s =
@@ -337,12 +335,13 @@ struct
   (** This function breaks the abstraction of Cst_stack ! *)
   let best_state sigma (_,sk as s) l =
     let rec aux sk def = function
-      |(cst, params, []) -> (cst, append_app_list (List.rev params) sk)
-      |(cst, params, (i,t)::q) -> match decomp sk with
+      |(_,_,_,{volatile=true}) -> def
+      |(cst, params, [], _) -> (cst, append_app_list (List.rev params) sk)
+      |(cst, params, (i,t)::q, vol) -> match decomp sk with
         | Some (el,sk') when EConstr.eq_constr sigma el t.(i) ->
           if i = pred (Array.length t)
-          then aux sk' def (cst, params, q)
-          else aux sk' def (cst, params, (succ i,t)::q)
+          then aux sk' def (cst, params, q, vol)
+          else aux sk' def (cst, params, (succ i,t)::q, vol)
         | _ -> def
     in List.fold_left (aux sk) s l
 
@@ -440,7 +439,7 @@ let magically_constant_of_fixbody env sigma reference bd = function
   | Name.Anonymous -> bd
   | Name.Name id ->
     let open UnivProblem in
-    let (cst_mod,_) = Constant.repr2 reference in
+    let cst_mod = KerName.modpath (Constant.user reference) in
     let cst = Constant.make2 cst_mod (Label.of_id id) in
     if not (Environ.mem_constant cst env) then bd
     else
@@ -455,13 +454,17 @@ let magically_constant_of_fixbody env sigma reference bd = function
                 let l, r = match cst with
                   | ULub (u, v) | UWeak (u, v) -> u, v
                   | UEq (u, v) | ULe (u, v) ->
-                    let get u = Option.get (Universe.level u) in
+                    let get u = match u with
+                    | Sorts.SProp | Sorts.Prop -> assert false
+                    | Sorts.Set -> Level.set
+                    | Sorts.Type u | Sorts.QSort (_, u) -> Option.get (Universe.level u)
+                    in
                     get u, get v
                 in
                 Univ.Level.Map.add l r acc)
                 csts Univ.Level.Map.empty
             in
-            let inst = Instance.subst_fn (fun u -> Univ.Level.Map.find u subst) u in
+            let inst = subst_univs_level_instance subst u in
             mkConstU (cst, EInstance.make inst)
           | None -> bd
         end
@@ -607,10 +610,23 @@ let whd_state_gen ?csts flags env sigma =
                  UnfoldWhenNoMatch { nargs = Some n } )
                 when Stack.args_size stack < n ->
                 fold ()
-              | UnfoldWhenNoMatch { recargs } -> (* maybe unfolds *)
+              | UnfoldWhenNoMatch { recargs; nargs } -> (* maybe unfolds *)
                   let app_sk,sk = Stack.strip_app stack in
+                  let volatile = Option.has_some nargs in
                   let (tm',sk'),cst_l' =
-                    whrec (Cst_stack.add_cst (mkConstU const) cst_l) (body, app_sk)
+                    match recargs with
+                    | [] ->
+                      whrec (Cst_stack.add_cst ~volatile (mkConstU const) cst_l) (body, app_sk)
+                    | curr :: remains -> match Stack.strip_n_app curr app_sk with
+                      | None -> (x,app_sk), cst_l
+                      | Some (bef,arg,app_sk') ->
+                        let cst_l = Stack.Cst
+                            { const = Stack.Cst_const (fst const, u');
+                              volatile;
+                              curr; remains; params=bef; cst_l;
+                            }
+                        in
+                        whrec Cst_stack.empty (arg,cst_l :: app_sk')
                   in
                   let rec is_case x = match EConstr.kind sigma x with
                     | Lambda (_,_, x) | LetIn (_,_,_, x) | Cast (x, _,_) -> is_case x
@@ -753,7 +769,7 @@ let whd_state_gen ?csts flags env sigma =
               | Some body ->
                 let const = (fst const, EInstance.make (snd const)) in
                 let body = EConstr.of_constr body in
-                let cst_l = if volatile then cst_l else Cst_stack.add_cst (mkConstU const) cst_l in
+                let cst_l = Cst_stack.add_cst ~volatile (mkConstU const) cst_l in
                 whrec cst_l (body, s' @ (Stack.append_app [|x'|] s'')))
             | Stack.Cst_proj p ->
               let stack = s' @ (Stack.append_app [|x'|] s'') in
