@@ -113,19 +113,20 @@ let nondep_elim_scheme from_kind to_kind =
   | InProp, InProp  -> ind_scheme_kind_from_prop
   | InProp, InSProp -> sind_scheme_kind_from_prop
   | InProp, InSet   -> rec_scheme_kind_from_prop
-  | InProp, InType  -> rect_scheme_kind_from_prop
+  | InProp, (InType | InQSort) -> rect_scheme_kind_from_prop
   | _     , InProp  -> ind_scheme_kind_from_type
   | _     , InSProp -> sind_scheme_kind_from_type
   | _     , InSet   -> rec_scheme_kind_from_type
-  | _     , InType  -> rect_scheme_kind_from_type
+  | _     , (InType | InQSort) -> rect_scheme_kind_from_type
 
 (* Case analysis *)
 
 let build_case_analysis_scheme_in_type env dep sort ind =
   let sigma = Evd.from_env env in
   let (sigma, indu) = Evd.fresh_inductive_instance env sigma ind in
-  let (sigma, c, _) = build_case_analysis_scheme env sigma indu dep sort in
-    c, Evd.evar_universe_context sigma
+  let (sigma, c) = build_case_analysis_scheme env sigma indu dep sort in
+  let (c, _) = Indrec.eval_case_analysis c in
+  EConstr.Unsafe.to_constr c, Evd.evar_universe_context sigma
 
 let case_scheme_kind_from_type =
   declare_individual_scheme_object "_case_nodep"
