@@ -113,7 +113,7 @@ val force_destruction_arg : evars_flag -> env -> evar_map ->
     evar_map * constr with_bindings destruction_arg
 
 val finish_evar_resolution : ?flags:Pretyping.inference_flags ->
-  env -> evar_map -> (evar_map * constr) -> evar_map * constr
+  env -> evar_map -> (evar_map option * constr) -> evar_map * constr
 
 (** Tell if a used hypothesis should be cleared by default or not *)
 
@@ -189,7 +189,7 @@ val clear         : Id.t list -> unit Proofview.tactic
 val clear_body    : Id.t list -> unit Proofview.tactic
 val unfold_body   : Id.t -> unit Proofview.tactic
 val keep          : Id.t list -> unit Proofview.tactic
-val apply_clear_request : clear_flag -> bool -> constr -> unit Proofview.tactic
+val apply_clear_request : clear_flag -> bool -> Id.t option -> unit Proofview.tactic
 
 val specialize    : constr with_bindings -> intro_pattern option -> unit Proofview.tactic
 
@@ -204,16 +204,16 @@ val apply_type : typecheck:bool -> constr -> constr list -> unit Proofview.tacti
 val bring_hyps : named_context -> unit Proofview.tactic
 
 val apply                 : constr -> unit Proofview.tactic
-val eapply                : constr -> unit Proofview.tactic
+val eapply : ?with_classes:bool -> constr -> unit Proofview.tactic
 
 val apply_with_bindings_gen :
-  advanced_flag -> evars_flag -> (clear_flag * constr with_bindings CAst.t) list -> unit Proofview.tactic
+  ?with_classes:bool -> advanced_flag -> evars_flag -> (clear_flag * constr with_bindings CAst.t) list -> unit Proofview.tactic
 
 val apply_with_delayed_bindings_gen :
   advanced_flag -> evars_flag -> (clear_flag * delayed_open_constr_with_bindings CAst.t) list -> unit Proofview.tactic
 
 val apply_with_bindings   : constr with_bindings -> unit Proofview.tactic
-val eapply_with_bindings  : constr with_bindings -> unit Proofview.tactic
+val eapply_with_bindings : ?with_classes:bool -> constr with_bindings -> unit Proofview.tactic
 
 val cut_and_apply         : constr -> unit Proofview.tactic
 
@@ -275,15 +275,8 @@ type elim_scheme = {
 
 val compute_elim_sig : evar_map -> types -> elim_scheme
 
-type eliminator =
-| ElimTerm of EConstr.constr
-| ElimClause of EConstr.constr with_bindings
-
-val general_elim  : evars_flag -> clear_flag ->
-  constr with_bindings -> eliminator -> unit Proofview.tactic
-
 val general_elim_clause : evars_flag -> unify_flags -> Id.t option ->
-  (EConstr.t * EConstr.t) -> eliminator -> unit Proofview.tactic
+  ((metavariable * Evd.clbinding) list * EConstr.t * EConstr.types) -> Constant.t -> unit Proofview.tactic
 
 val default_elim  : evars_flag -> clear_flag -> constr with_bindings ->
   unit Proofview.tactic
@@ -316,6 +309,7 @@ val induction_destruct : rec_flag -> evars_flag ->
 
 val case_type         : types -> unit Proofview.tactic
 val elim_type         : types -> unit Proofview.tactic
+val exfalso : unit Proofview.tactic
 
 (** {6 Constructor tactics. } *)
 
@@ -363,10 +357,6 @@ val assert_after_replacing : Id.t -> types -> unit Proofview.tactic
 val assert_before : Name.t -> types -> unit Proofview.tactic
 val assert_after  : Name.t -> types -> unit Proofview.tactic
 
-val assert_as     : (* true = before *) bool ->
-  (* optionally tell if a specialization of some hyp: *) Id.t option ->
-  intro_pattern option -> constr -> unit Proofview.tactic
-
 (** Implements the tactics assert, enough and pose proof; note that "by"
     applies on the first goal for both assert and enough *)
 
@@ -396,7 +386,7 @@ val letin_tac : (bool * intro_pattern_naming) option ->
 (** Common entry point for user-level "set", "pose" and "remember" *)
 
 val letin_pat_tac : evars_flag -> (bool * intro_pattern_naming) option ->
-  Name.t -> (evar_map * constr) -> clause -> unit Proofview.tactic
+  Name.t -> (evar_map option * constr) -> clause -> unit Proofview.tactic
 
 (** {6 Generalize tactics. } *)
 
@@ -452,13 +442,9 @@ end
 
 (** {6 Tacticals defined directly in term of Proofview} *)
 
-module New : sig
+val refine : typecheck:bool -> (evar_map -> evar_map * constr) -> unit Proofview.tactic
+(** [refine ~typecheck c] is [Refine.refine ~typecheck c]
+    followed by beta-iota-reduction of the conclusion. *)
 
-  val refine : typecheck:bool -> (evar_map -> evar_map * constr) -> unit Proofview.tactic
-  (** [refine ~typecheck c] is [Refine.refine ~typecheck c]
-      followed by beta-iota-reduction of the conclusion. *)
-
-  val reduce_after_refine : unit Proofview.tactic
-  (** The reducing tactic called after {!refine}. *)
-
-end
+val reduce_after_refine : unit Proofview.tactic
+(** The reducing tactic called after {!refine}. *)
