@@ -29,7 +29,7 @@ module RelDecl = Context.Rel.Declaration
 (* 3b| Mutual inductive definitions *)
 
 let warn_auto_template =
-  CWarnings.create ~name:"auto-template" ~category:"vernacular" ~default:CWarnings.Disabled
+  CWarnings.create ~name:"auto-template" ~default:CWarnings.Disabled
     (fun id ->
        Pp.(strbrk "Automatically declaring " ++ Id.print id ++
            strbrk " as template polymorphic. Use attributes or " ++
@@ -39,7 +39,8 @@ let should_auto_template =
   let open Goptions in
   let auto = ref true in
   let () = declare_bool_option
-      { optdepr  = false;
+      { optstage = Summary.Stage.Interp;
+        optdepr  = None;
         optkey   = ["Auto";"Template";"Polymorphism"];
         optread  = (fun () -> !auto);
         optwrite = (fun b -> auto := b); }
@@ -128,7 +129,7 @@ let pretype_ind_arity env sigma (loc, c, impls, pseudo_poly) =
       | Some b -> Some (b, s)
       | None -> None
     in
-    sigma, (t, Retyping.relevance_of_sort s, concl, impls)
+    sigma, (t, Retyping.relevance_of_sort sigma s, concl, impls)
 
 (* ind_rel is the Rel for this inductive in the context without params.
    n is how many arguments there are in the constructor. *)
@@ -155,14 +156,14 @@ let interp_cstrs env (sigma, ind_rel) impls params ind arity =
                   solve_unification_constraints = false }
     in
     let sigma, (ctyp, cimpl) = interp_type_evars_impls ~flags env sigma ~impls ctyp in
-    let ctx, concl = Reductionops.splay_prod_assum env sigma ctyp in
+    let ctx, concl = Reductionops.hnf_decompose_prod_decls env sigma ctyp in
     let concl_env = EConstr.push_rel_context ctx env in
     let sigma_with_model_evars, model =
       model_conclusion concl_env sigma ind_rel params (Context.Rel.length ctx) arity_indices
     in
     (* unify the expected with the provided conclusion *)
     let sigma =
-      try Evarconv.unify concl_env sigma_with_model_evars Reduction.CONV concl model
+      try Evarconv.unify concl_env sigma_with_model_evars Conversion.CONV concl model
       with Evarconv.UnableToUnify (sigma,e) ->
         user_err (Himsg.explain_pretype_error concl_env sigma
                     (Pretype_errors.CannotUnify (concl, model, (Some e))))
@@ -182,6 +183,7 @@ let max_sort s1 s2 = match s1, s2 with
 | (Prop, (Set | Type _ as s)) | ((Set | Type _) as s, Prop) -> s
 | (Set, Type u) | (Type u, Set) -> Sorts.sort_of_univ (Univ.Universe.sup Univ.Universe.type0 u)
 | (Type u, Type v) -> Sorts.sort_of_univ (Univ.Universe.sup u v)
+| (QSort _, _) | (_, QSort _) -> assert false
 
 let compute_constructor_level env evd sign =
   fst (List.fold_right
@@ -190,6 +192,7 @@ let compute_constructor_level env evd sign =
       | LocalDef _ -> lev, EConstr.push_rel d env
       | LocalAssum _ ->
         let s = Retyping.get_sort_of env evd (RelDecl.get_type d) in
+        let s = EConstr.ESorts.kind evd s in
           (max_sort s lev, EConstr.push_rel d env))
     sign (Sorts.sprop,env))
 
@@ -200,11 +203,12 @@ let sup_list min = List.fold_left max_sort min
 
 let extract_level env evd min tys =
   let sorts = List.map (fun ty ->
-    let ctx, concl = Reduction.dest_prod_assum env ty in
+    let ctx, concl = Reduction.hnf_decompose_prod_decls env ty in
       sign_level env evd (LocalAssum (make_annot Anonymous Sorts.Relevant, concl) :: ctx)) tys
   in sup_list min sorts
 
 let is_flexible_sort evd s = match s with
+| QSort _ -> assert false
 | Set | Prop | SProp -> false
 | Type u ->
   match Univ.Universe.level u with
@@ -237,6 +241,7 @@ let is_direct_sort_constraint s v = match s with
   match v with
   | Sorts.Prop | Sorts.Set | Sorts.SProp -> false
   | Sorts.Type v -> Univ.univ_level_mem u v
+  | Sorts.QSort _ -> assert false
 
 let solve_constraints_system levels level_bounds =
   let open Univ in
@@ -286,6 +291,7 @@ let inductive_levels env evd arities inds =
   | Prop | SProp -> None
   | Set -> Some Univ.Universe.type0
   | Type u -> Some u
+  | QSort _ -> assert false
   in
   let levels = List.map map destarities in
   let cstrs_levels, sizes =
@@ -316,9 +322,10 @@ let inductive_levels env evd arities inds =
   in
   let evd, arities =
     CList.fold_left3 (fun (evd, arities) cu (arity,(ctx,du)) len ->
+      let cu = EConstr.ESorts.make cu in
       if is_impredicative_sort env du then
         (* Any product is allowed here. *)
-        evd, (false, arity) :: arities
+        evd, arity :: arities
       else (* If in a predicative sort, or asked to infer the type,
               we take the max of:
               - indices (if in indices-matter mode)
@@ -328,22 +335,22 @@ let inductive_levels env evd arities inds =
         (* Constructors contribute. *)
         let evd =
           if Sorts.is_set du then
-            if not (Evd.check_leq evd cu Sorts.set) then
+            if not (Evd.check_leq evd cu (EConstr.ESorts.make Sorts.set)) then
               raise (InductiveError LargeNonPropInductiveNotInType)
             else evd
           else evd
         in
         let evd =
-          if len >= 2 && Sorts.is_prop cu then
+          if len >= 2 && EConstr.ESorts.is_prop evd cu then
            (* "Polymorphic" type constraint and more than one constructor,
                should not land in Prop. Add constraint only if it would
                land in Prop directly (no informative arguments as well). *)
-            Evd.set_leq_sort env evd Sorts.set du
+            Evd.set_leq_sort env evd EConstr.ESorts.set (EConstr.ESorts.make du)
           else evd
         in
-        let template_prop, evd, arity =
-          if not (Sorts.is_small du) && Sorts.equal cu du then
-            if is_flexible_sort evd du && not (Evd.check_leq evd Sorts.set du)
+        let evd, arity =
+          if not (Sorts.is_small du) && EConstr.ESorts.equal evd cu (EConstr.ESorts.make du) then
+            if is_flexible_sort evd du && not (Evd.check_leq evd EConstr.ESorts.set (EConstr.ESorts.make du))
             then if Term.isArity arity
             (* If not a syntactic arity, the universe may be used in a
                polymorphic instance and so cannot be lowered to Prop.
@@ -353,13 +360,13 @@ let inductive_levels env evd arities inds =
                    arities. In this situation we have no constraints from the
                    constructor so we cook up a new type and unify the unbound
                    universe to a dummy value. *)
-                let evd = Evd.set_eq_sort env evd Sorts.set du in
-                true, evd, Term.mkArity (ctx, Sorts.prop)
-              else false, Evd.set_eq_sort env evd Sorts.set du, arity
-            else false, evd, arity
-          else false, Evd.set_eq_sort env evd cu du, arity
+                let evd = Evd.set_eq_sort env evd EConstr.ESorts.set (EConstr.ESorts.make du) in
+                evd, Term.mkArity (ctx, Sorts.prop)
+              else Evd.set_eq_sort env evd EConstr.ESorts.set (EConstr.ESorts.make du), arity
+            else evd, arity
+          else Evd.set_eq_sort env evd cu (EConstr.ESorts.make du), arity
         in
-          (evd, (template_prop, arity) :: arities))
+          (evd, arity :: arities))
     (evd,[]) (Array.to_list levels') destarities sizes
   in evd, List.rev arities
 
@@ -374,7 +381,7 @@ let check_named {CAst.loc;v=na} = match na with
    (starting from the most recent and ignoring let-definitions) is not
    contributing to the inductive type's sort or is Some u_k if its level
    is u_k and is contributing. *)
-let template_polymorphic_univs ~ctor_levels uctx paramsctxt concl =
+let template_polymorphic_univs ~ctor_levels uctx paramsctxt u =
   let unbounded_from_below u cstrs =
     let open Univ in
     Univ.Constraints.for_all (fun (l, d, r) ->
@@ -383,37 +390,115 @@ let template_polymorphic_univs ~ctor_levels uctx paramsctxt concl =
         | Lt | Le -> not (Univ.Level.equal r u))
       cstrs
   in
+  let fold_params accu decl = match decl with
+  | LocalAssum (_, p) ->
+    let c = Term.strip_prod_decls p in
+    begin match Constr.kind c with
+    | Constr.Sort (Type u) ->
+      begin match Univ.Universe.level u with
+      | Some l -> Univ.Level.Set.add l accu
+      | None -> accu
+      end
+    | _ -> accu
+    end
+  | LocalDef _ -> accu
+  in
+  let paramslevels = List.fold_left fold_params Univ.Level.Set.empty paramsctxt in
   let check_level l =
     Univ.Level.Set.mem l (Univ.ContextSet.levels uctx) &&
+    Univ.Level.Set.mem l paramslevels &&
     (let () = assert (not @@ Univ.Level.is_set l) in true) &&
     unbounded_from_below l (Univ.ContextSet.constraints uctx) &&
     not (Univ.Level.Set.mem l ctor_levels)
   in
-  let univs = match concl with
-  | Prop | Set | SProp -> Univ.Level.Set.empty
-  | Type u -> Univ.Universe.levels u
-  in
+  let univs = Univ.Universe.levels u in
   let univs = Univ.Level.Set.filter (fun l -> check_level l) univs in
-  match concl with
-  | Prop -> true
-  | Set | SProp | Type _ -> not @@ Univ.Level.Set.is_empty univs
+  univs
 
-let template_polymorphism_candidate ~ctor_levels uctx params concl =
-  match uctx with
-  | UState.Monomorphic_entry uctx ->
-    let concltemplate = Option.cata (fun s -> not (Sorts.is_small s)) false concl in
-    if not concltemplate then false
-    else
-      let conclu = Option.default Sorts.prop concl in
-      template_polymorphic_univs ~ctor_levels uctx params conclu
-  | UState.Polymorphic_entry _ -> false
+let template_polymorphism_candidate uctx params entry concl = match concl with
+| None -> Univ.Level.Set.empty
+| Some (Set | SProp | Prop) -> Univ.Level.Set.empty
+| Some (Type u) ->
+  let ctor_levels =
+    let add_levels c levels = Univ.Level.Set.union levels (Vars.universes_of_constr c) in
+    let param_levels =
+      List.fold_left (fun levels d -> match d with
+          | LocalAssum _ -> levels
+          | LocalDef (_,b,t) -> add_levels b (add_levels t levels))
+        Univ.Level.Set.empty params
+    in
+    List.fold_left (fun levels c -> add_levels c levels)
+      param_levels entry.mind_entry_lc
+  in
+  let univs = template_polymorphic_univs ~ctor_levels uctx params u in
+  univs
+| Some (QSort _) -> assert false
+
+let split_universe_context subset (univs, csts) =
+  let subfilter (l, _, r) =
+    let () = assert (not @@ Univ.Level.Set.mem r subset) in
+    Univ.Level.Set.mem l subset
+  in
+  let subcst = Univ.Constraints.filter subfilter csts in
+  let rem = Univ.Level.Set.diff univs subset in
+  let remfilter (l, _, r) =
+    not (Univ.Level.Set.mem l subset) && not (Univ.Level.Set.mem r subset)
+  in
+  let remcst = Univ.Constraints.filter remfilter csts in
+  (subset, subcst), (rem, remcst)
+
+let warn_no_template_universe =
+  CWarnings.create ~name:"no-template-universe"
+    (fun () -> Pp.str "This inductive type has no template universes.")
+
+let compute_template_inductive ~user_template ~env_ar_params ~ctx_params ~univ_entry entry concl =
+match user_template, univ_entry with
+| Some false, UState.Monomorphic_entry uctx ->
+  Monomorphic_ind_entry, uctx
+| Some false, UState.Polymorphic_entry uctx ->
+  Polymorphic_ind_entry uctx, Univ.ContextSet.empty
+| Some true, UState.Monomorphic_entry uctx ->
+  let template_universes = template_polymorphism_candidate uctx ctx_params entry concl in
+  let template, global = split_universe_context template_universes uctx in
+  let () = if Univ.Level.Set.is_empty (fst template) then warn_no_template_universe () in
+  Template_ind_entry template, global
+| Some true, UState.Polymorphic_entry _ ->
+  user_err Pp.(strbrk "Template-polymorphism and universe polymorphism are not compatible.")
+| None, UState.Polymorphic_entry uctx ->
+  Polymorphic_ind_entry uctx, Univ.ContextSet.empty
+| None, UState.Monomorphic_entry uctx ->
+  (* Heuristic: the user has not written Prop explicitly in the return
+      arity, but inference has decided to lower it to Prop. *)
+  let templatearity =
+    if Term.isArity entry.mind_entry_arity then
+      let (_, s) = Reduction.dest_arity env_ar_params entry.mind_entry_arity in
+      if Sorts.is_prop s then match concl with
+      | None | Some (Type _ | Set)-> true
+      | Some Prop -> false
+      | Some SProp | Some (QSort _) -> assert false
+      else false
+    else false
+  in
+  if templatearity then
+    let template = should_auto_template entry.mind_entry_typename true in
+    (* Dummy template inductive. Matters for the shape of the induction principle *)
+    if template then Template_ind_entry Univ.ContextSet.empty, uctx
+    else Monomorphic_ind_entry, uctx
+  else
+    let template_candidate = template_polymorphism_candidate uctx ctx_params entry concl in
+    let has_template = not @@ Univ.Level.Set.is_empty template_candidate in
+    let template = should_auto_template entry.mind_entry_typename has_template in
+    if template then
+      let template, global = split_universe_context template_candidate uctx in
+      Template_ind_entry template, global
+    else Monomorphic_ind_entry, uctx
 
 let check_param = function
 | CLocalDef (na, _, _) -> check_named na
 | CLocalAssum (nas, Default _, _) -> List.iter check_named nas
 | CLocalAssum (nas, Generalized _, _) -> ()
 | CLocalPattern {CAst.loc} ->
-    Loc.raise ?loc (Stream.Error "pattern with quote not allowed here")
+    Loc.raise ?loc (Gramlib.Stream.Error "pattern with quote not allowed here")
 
 let restrict_inductive_universes sigma ctx_params arities constructors =
   let merge_universes_of_constr c =
@@ -443,6 +528,14 @@ let variance_of_entry ~cumulative ~variances uctx =
       assert (lvs <= lus);
       Some (Array.append variances (Array.make (lus - lvs) None))
 
+let warn_bad_set_minimization =
+  CWarnings.create ~name:"bad-set-minimization" ~category:Deprecation.Version.v8_18
+    Pp.(fun () -> strbrk "This inductive will be minimized to Set even though Universe Minimization ToSet is unset. This will change in a future version.")
+
+let warn_bad_set_minimization ?loc () =
+  if UnivMinim.get_set_minimization () then ()
+  else warn_bad_set_minimization ?loc ()
+
 let interp_mutual_inductive_constr ~sigma ~template ~udecl ~variances ~ctx_params ~indnames ~arities ~arityconcl ~constructors ~env_ar_params ~cumulative ~poly ~private_ind ~finite =
   (* Compute renewed arities *)
   let sigma = Evd.minimize_universes sigma in
@@ -450,66 +543,46 @@ let interp_mutual_inductive_constr ~sigma ~template ~udecl ~variances ~ctx_param
   let constructors = List.map (on_snd (List.map nf)) constructors in
   let arities = List.map EConstr.(to_constr sigma) arities in
   let sigma = List.fold_left make_anonymous_conclusion_flexible sigma arityconcl in
-  let sigma, arities = inductive_levels env_ar_params sigma arities constructors in
-  let sigma = Evd.minimize_universes sigma in
+  let sigma', arities = inductive_levels env_ar_params sigma arities constructors in
+  let sigma =
+    let sigma' = Evd.minimize_universes sigma' in
+    let () = List.iter (fun ty ->
+        let _, s = Reduction.dest_arity env_ar_params ty in
+        let s = EConstr.ESorts.make s in
+        if EConstr.ESorts.is_set sigma' s && not (EConstr.ESorts.is_set sigma s)
+        then warn_bad_set_minimization ())
+        arities
+    in
+    sigma'
+  in
   let nf = Evarutil.nf_evars_universes sigma in
-  let arities = List.map (on_snd nf) arities in
+  let arities = List.map nf arities in
   let constructors = List.map (on_snd (List.map nf)) constructors in
-  let ctx_params = List.map Termops.(map_rel_decl (EConstr.to_constr sigma)) ctx_params in
+  let ctx_params = List.map (fun d -> EConstr.to_rel_decl sigma d) ctx_params in
   let arityconcl = List.map (Option.map (fun (_anon, s) -> EConstr.ESorts.kind sigma s)) arityconcl in
-  let sigma = restrict_inductive_universes sigma ctx_params (List.map snd arities) constructors in
+  let sigma = restrict_inductive_universes sigma ctx_params arities constructors in
   let univ_entry, binders = Evd.check_univ_decl ~poly sigma udecl in
 
   (* Build the inductive entries *)
-  let entries = List.map4 (fun indname (templatearity, arity) concl (cnames,ctypes) ->
+  let entries = List.map3 (fun indname arity (cnames,ctypes) ->
       { mind_entry_typename = indname;
         mind_entry_arity = arity;
         mind_entry_consnames = cnames;
         mind_entry_lc = ctypes
       })
-      indnames arities arityconcl constructors
+      indnames arities constructors
   in
-  let template =
-    if not (Int.equal (List.length indnames) 1) then
-      let () = match template with
-        | Some true -> user_err Pp.(str "Template-polymorphism not allowed with mutual inductives.")
-        | _ -> ()
-      in
-      List.map (fun _ -> false) indnames
-    else
-      List.map4 (fun indname (templatearity, _) concl (_, ctypes) ->
-          let template_candidate () =
-            templatearity ||
-            let ctor_levels =
-              let add_levels c levels = Univ.Level.Set.union levels (Vars.universes_of_constr c) in
-              let param_levels =
-                List.fold_left (fun levels d -> match d with
-                    | LocalAssum _ -> levels
-                    | LocalDef (_,b,t) -> add_levels b (add_levels t levels))
-                  Univ.Level.Set.empty ctx_params
-              in
-              List.fold_left (fun levels c -> add_levels c levels)
-                param_levels ctypes
-            in
-            template_polymorphism_candidate ~ctor_levels univ_entry ctx_params concl
-          in
-          match template with
-          | Some template ->
-            if poly && template then user_err
-                Pp.(strbrk "Template-polymorphism and universe polymorphism are not compatible.");
-            template
-          | None ->
-            should_auto_template indname (template_candidate ())
-        )
-        indnames arities arityconcl constructors
-  in
-  let is_template = List.for_all (fun t -> t) template in
-  let univ_entry, ctx = match univ_entry with
-  | UState.Monomorphic_entry ctx ->
-    if is_template then Template_ind_entry ctx, Univ.ContextSet.empty
-    else Monomorphic_ind_entry, ctx
-  | UState.Polymorphic_entry uctx ->
-    Polymorphic_ind_entry uctx, Univ.ContextSet.empty
+  let univ_entry, ctx = match entries, arityconcl with
+  | [entry], [concl] ->
+    compute_template_inductive ~user_template:template ~env_ar_params ~ctx_params ~univ_entry entry concl
+  | _ ->
+    let () = match template with
+    | Some true -> user_err Pp.(str "Template-polymorphism not allowed with mutual inductives.")
+    | _ -> ()
+    in
+    match univ_entry with
+    | UState.Monomorphic_entry ctx -> Monomorphic_ind_entry, ctx
+    | UState.Polymorphic_entry uctx -> Polymorphic_ind_entry uctx, Univ.ContextSet.empty
   in
   let variance = variance_of_entry ~cumulative ~variances univ_entry in
   (* Build the mutual inductive entry *)
@@ -588,7 +661,7 @@ let interp_mutual_inductive_gen env0 ~template udecl (uparamsl,paramsl,indl) not
   let lift_ctx n ctx =
     let t = EConstr.it_mkProd_or_LetIn EConstr.mkProp ctx in
     let t = EConstr.Vars.lift n t in
-    let ctx, _ = EConstr.decompose_prod_assum sigma t in
+    let ctx, _ = EConstr.decompose_prod_decls sigma t in
     ctx
   in
   let ctx_params_lifted, fullarities =
@@ -666,7 +739,11 @@ let eq_params (up1,p1) (up2,p2) =
 
 let extract_coercions indl =
   let mkqid (_,({CAst.v=id},_)) = qualid_of_ident id in
-  let extract lc = List.filter (fun (iscoe,_) -> iscoe) lc in
+  let iscoe (_, coe, inst) = match inst with
+    (* remove BackInstanceWarning after deprecation phase *)
+    | Vernacexpr.(NoInstance | BackInstanceWarning) -> coe = Vernacexpr.AddCoercion
+    | _ -> user_err (Pp.str "'::' not allowed in inductives.") in
+  let extract lc = List.filter (fun (coe,_) -> iscoe coe) lc in
   List.map mkqid (List.flatten(List.map (fun (_,_,_,lc) -> extract lc) indl))
 
 let extract_params indl =
@@ -704,8 +781,9 @@ type t = {
   univ_binders : UnivNames.universe_binders;
   implicits : DeclareInd.one_inductive_impls list;
   uctx : Univ.ContextSet.t;
-  where_notations : Metasyntax.where_decl_notation list;
+  where_notations : Metasyntax.notation_interpretation_decl list;
   coercions : Libnames.qualid list;
+  indlocs : Loc.t option list;
 }
 
 end
@@ -715,9 +793,10 @@ let rec count_binder_expr = function
   | CLocalAssum(l,_,_) :: rest -> List.length l + count_binder_expr rest
   | CLocalDef _ :: rest -> 1 + count_binder_expr rest
   | CLocalPattern {CAst.loc} :: _ ->
-    Loc.raise ?loc (Stream.Error "pattern with quote not allowed here")
+    Loc.raise ?loc (Gramlib.Stream.Error "pattern with quote not allowed here")
 
 let interp_mutual_inductive ~env ~template udecl indl ~cumulative ~poly ?typing_flags ~private_ind ~uniform finite =
+  let indlocs = List.map (fun ((n,_,_,_),_) -> n.CAst.loc) indl in
   let (params,indl),coercions,ntns = extract_mutual_inductive_declaration_components indl in
   let where_notations = List.map Metasyntax.prepare_where_notation ntns in
   (* Interpret the types *)
@@ -730,12 +809,12 @@ let interp_mutual_inductive ~env ~template udecl indl ~cumulative ~poly ?typing_
   let env = Environ.update_typing_flags ?typing_flags env in
   let mie, univ_binders, implicits, uctx = interp_mutual_inductive_gen env ~template udecl indl where_notations ~cumulative ~poly ~private_ind finite in
   let open Mind_decl in
-  { mie; nuparams; univ_binders; implicits; uctx; where_notations; coercions }
+  { mie; nuparams; univ_binders; implicits; uctx; where_notations; coercions; indlocs }
 
 let do_mutual_inductive ~template udecl indl ~cumulative ~poly ?typing_flags ~private_ind ~uniform finite =
   let open Mind_decl in
   let env = Global.env () in
-  let { mie; univ_binders; implicits; uctx; where_notations; coercions } =
+  let { mie; univ_binders; implicits; uctx; where_notations; coercions; indlocs} =
     interp_mutual_inductive ~env ~template udecl indl ~cumulative ~poly ?typing_flags ~private_ind ~uniform finite in
   (* Slightly hackish global universe declaration due to template types. *)
   let binders = match mie.mind_entry_universes with
@@ -746,11 +825,11 @@ let do_mutual_inductive ~template udecl indl ~cumulative ~poly ?typing_flags ~pr
   (* Declare the global universes *)
   DeclareUctx.declare_universe_context ~poly:false uctx;
   (* Declare the mutual inductive block with its associated schemes *)
-  ignore (DeclareInd.declare_mutual_inductive_with_eliminations ?typing_flags mie binders implicits);
+  ignore (DeclareInd.declare_mutual_inductive_with_eliminations ?typing_flags ~indlocs mie binders implicits);
   (* Declare the possible notations of inductive types *)
   List.iter (Metasyntax.add_notation_interpretation ~local:false (Global.env ())) where_notations;
   (* Declare the coercions *)
-  List.iter (fun qid -> ComCoercion.try_add_new_coercion (Nametab.locate qid) ~local:false ~poly ~nonuniform:false ~reversible:true) coercions
+  List.iter (fun qid -> ComCoercion.try_add_new_coercion (Nametab.locate qid) ~local:false ~poly ~reversible:true) coercions
 
 (** Prepare a "match" template for a given inductive type.
     For each branch of the match, we list the constructor name
@@ -764,7 +843,7 @@ let do_mutual_inductive ~template udecl indl ~cumulative ~poly ?typing_flags ~pr
 
   The Show Match could also be made more robust, for instance in the
   presence of let in the branch of a constructor. A
-  decompose_prod_assum would probably suffice for that, but then, it
+  decompose_prod_decls would probably suffice for that, but then, it
   is a Context.Rel.Declaration.t which needs to be matched and not
   just a pair (name,type).
 
@@ -796,5 +875,5 @@ module Internal =
 struct
 
 let compute_constructor_level = compute_constructor_level
-
+let warn_bad_set_minimization = warn_bad_set_minimization
 end
