@@ -66,7 +66,9 @@ let rec contract3' env sigma a b c = function
   | OccurCheck (evk,d) ->
     let x,d = contract4 env sigma a b c d in x,OccurCheck(evk, d)
   | NotClean ((evk,args),env',d) ->
+      let args = Evd.expand_existential sigma (evk, args) in
       let env',d,args = contract1 env' sigma d args in
+      let args = SList.of_full_list args in
       contract3 env sigma a b c,NotClean((evk,args),env',d)
   | ConversionFailed (env',t1,t2) ->
       let (env',t1,t2) = contract2 env' sigma t1 t2 in
@@ -216,13 +218,26 @@ let rec pr_disjunction pr = function
   | a::l -> pr a ++ str "," ++ spc () ++ pr_disjunction pr l
   | [] -> assert false
 
-let explain_elim_arity env sigma ind c pj okinds =
+type arity_error =
+  | NonInformativeToInformative
+  | StrongEliminationOnNonSmallType
+  | WrongArity
+
+let error_elim_explain kp ki =
+  let open Sorts in
+  match kp,ki with
+  | (InType | InSet), InProp -> NonInformativeToInformative
+  | InType, InSet -> StrongEliminationOnNonSmallType (* if Set impredicative *)
+  | _ -> WrongArity
+
+let explain_elim_arity env sigma ind c okinds =
   let open EConstr in
   let env = make_all_name_different env sigma in
   let pi = pr_inductive env (fst ind) in
   let pc = pr_leconstr_env env sigma c in
   let msg = match okinds with
-  | Some(sorts,kp,ki,explanation) ->
+  | Some (pj, sorts, kp, ki) ->
+      let explanation = error_elim_explain kp ki in
       let sorts = Inductiveops.sorts_below sorts in
       let pki = Sorts.pr_sort_family ki in
       let pkp = Sorts.pr_sort_family kp in
@@ -234,7 +249,7 @@ let explain_elim_arity env sigma ind c pj okinds =
         | WrongArity ->
           "wrong arity" in
       let ppar = pr_disjunction (fun s -> quote (Sorts.pr_sort_family s)) sorts in
-      let ppt = pr_leconstr_env env sigma (snd (decompose_prod_assum sigma pj.uj_type)) in
+      let ppt = pr_leconstr_env env sigma (snd (decompose_prod_decls sigma pj.uj_type)) in
       hov 0
         (str "the return type has sort" ++ spc () ++ ppt ++ spc () ++
          str "while it" ++ spc () ++ str "should be " ++ ppar ++ str ".") ++
@@ -264,6 +279,9 @@ let explain_case_not_inductive env sigma cj =
           str "has type" ++ brk(1,1) ++ pct ++ spc () ++
           str "which is not a (co-)inductive type."
 
+let explain_case_on_private_ind env sigma ind =
+  str "Case analysis on private inductive "++pr_inductive env ind
+
 let explain_number_branches env sigma cj expn =
   let env = make_all_name_different env sigma in
   let pc = pr_leconstr_env env sigma cj.uj_val in
@@ -301,6 +319,7 @@ let explain_unification_error env sigma p1 p2 = function
         strbrk " with term " ++ pr_leconstr_env env sigma rhs ++
         strbrk " that would depend on itself"]
      | NotClean ((evk,args),env,c) ->
+        let args = Evd.expand_existential sigma (evk, args) in
         let env = make_all_name_different env sigma in
         [str "cannot instantiate " ++ quote (pr_existential_key env sigma evk)
         ++ strbrk " because " ++ pr_leconstr_env env sigma c ++
@@ -346,7 +365,7 @@ let explain_unification_error env sigma p1 p2 = function
         strbrk " is expected to have a functional type but it has type " ++ u]
      | UnifUnivInconsistency p ->
        [str "universe inconsistency: " ++
-        Univ.explain_universe_inconsistency (Termops.pr_evd_level sigma) p]
+        UGraph.explain_universe_inconsistency (Termops.pr_evd_level sigma) p]
      | CannotSolveConstraint ((pb,env,t,u),e) ->
         let env = make_all_name_different env sigma in
         (strbrk "cannot satisfy constraint " ++ pr_leconstr_env env sigma t ++
@@ -390,10 +409,11 @@ let explain_incorrect_primitive env sigma j exp =
   str "has type" ++ brk(1,1) ++ pct ++ spc () ++
   str "while it is expected to have type" ++ brk(1,1) ++ pt ++ str ".")
 
-let explain_cant_apply_bad_type env sigma (n,exptyp,actualtyp) rator randl =
+let explain_cant_apply_bad_type env sigma ?error (n,exptyp,actualtyp) rator randl =
   let randl = jv_nf_betaiotaevar env sigma randl in
   let actualtyp = Reductionops.nf_betaiota env sigma actualtyp in
   let env = make_all_name_different env sigma in
+  let error = explain_unification_error env sigma actualtyp exptyp error in
   let actualtyp, exptyp = pr_explicit env sigma actualtyp exptyp in
   let nargs = Array.length randl in
 (*  let pe = pr_ne_context_of (str "in environment") env sigma in*)
@@ -413,8 +433,8 @@ let explain_cant_apply_bad_type env sigma (n,exptyp,actualtyp) rator randl =
   str "cannot be applied to the " ++ term_string1 ++ fnl () ++
   str " " ++ v 0 appl ++ fnl () ++ term_string2 ++ str " has type" ++
   brk(1,1) ++ actualtyp ++ spc () ++
-  str "which should be coercible to" ++ brk(1,1) ++
-  exptyp ++ str "."
+  str "which should be a subtype of" ++ brk(1,1) ++
+  exptyp ++ str "." ++ error
 
 let explain_cant_apply_not_functional env sigma rator randl =
   let env = make_all_name_different env sigma in
@@ -435,10 +455,11 @@ let explain_cant_apply_not_functional env sigma rator randl =
   str "cannot be applied to the " ++ str (String.plural nargs "term") ++
   fnl () ++ str " " ++ v 0 appl
 
-let explain_unexpected_type env sigma actual_type expected_type =
+let explain_unexpected_type env sigma actual_type expected_type e =
   let pract, prexp = pr_explicit env sigma actual_type expected_type in
   str "Found type" ++ spc () ++ pract ++ spc () ++
-  str "where" ++ spc () ++ prexp ++ str " was expected."
+  str "where" ++ spc () ++ prexp ++ str " was expected" ++
+  explain_unification_error env sigma actual_type expected_type (Some e) ++ str"."
 
 let explain_not_product env sigma c =
   let pr = pr_econstr_env env sigma c in
@@ -446,24 +467,14 @@ let explain_not_product env sigma c =
   str "while it is expected to be" ++
   (if EConstr.isType sigma c then str " a sort" else (brk(1,1) ++ pr)) ++ str "."
 
-(* TODO: use the names *)
-(* (co)fixpoints *)
-let explain_ill_formed_rec_body env sigma err names i fixenv vdefj =
-  let pr_lconstr_env env sigma c = pr_leconstr_env env sigma c in
-  let prt_name i =
-    match names.(i).binder_name with
-        Name id -> str "Recursive definition of " ++ Id.print id
-      | Anonymous -> str "The " ++ pr_nth i ++ str " definition" in
-
-  let st = match err with
-
+let explain_ill_formed_fix_body env sigma names i = function
   (* Fixpoint guard errors *)
   | NotEnoughAbstractionInFixBody ->
       str "Not enough abstractions in the definition"
   | RecursionNotOnInductiveType c ->
-      str "Recursive definition on" ++ spc () ++ pr_lconstr_env env sigma c ++
+      str "Recursive definition on" ++ spc () ++ pr_leconstr_env env sigma c ++
       spc () ++ str "which should be a recursive inductive type"
-  | RecursionOnIllegalTerm(j,(arg_env, arg),le,lt) ->
+  | RecursionOnIllegalTerm(j,(arg_env, arg),le_lt) ->
       let arg_env = make_all_name_different arg_env sigma in
       let called =
         match names.(j).binder_name with
@@ -471,69 +482,80 @@ let explain_ill_formed_rec_body env sigma err names i fixenv vdefj =
           | Anonymous -> str "the " ++ pr_nth i ++ str " definition" in
       let pr_db x = quote (pr_db env x) in
       let vars =
-        match (lt,le) with
+        match Lazy.force le_lt with
             ([],[]) -> assert false
-          | ([],[x]) -> str "a subterm of " ++ pr_db x
-          | ([],_) -> str "a subterm of the following variables: " ++
+          | ([x],[]) -> str "a subterm of " ++ pr_db x
+          | (le,[]) -> str "a subterm of the following variables: " ++
               pr_sequence pr_db le
-          | ([x],_) -> pr_db x
-          | _ ->
+          | (_,[x]) -> pr_db x
+          | (_,lt) ->
               str "one of the following variables: " ++
               pr_sequence pr_db lt in
       str "Recursive call to " ++ called ++ spc () ++
       strbrk "has principal argument equal to" ++ spc () ++
-      pr_lconstr_env arg_env sigma arg ++ strbrk " instead of " ++ vars
-
+      pr_leconstr_env arg_env sigma arg ++ strbrk " instead of " ++ vars
   | NotEnoughArgumentsForFixCall j ->
       let called =
         match names.(j).binder_name with
             Name id -> Id.print id
           | Anonymous -> str "the " ++ pr_nth i ++ str " definition" in
      str "Recursive call to " ++ called ++ str " has not enough arguments"
+  | FixpointOnIrrelevantInductive ->
+    strbrk "Fixpoints on proof irrelevant inductive types should produce proof irrelevant values"
 
+let explain_ill_formed_cofix_body env sigma = function
   (* CoFixpoint guard errors *)
   | CodomainNotInductiveType c ->
-      str "The codomain is" ++ spc () ++ pr_lconstr_env env sigma c ++ spc () ++
+      str "The codomain is" ++ spc () ++ pr_leconstr_env env sigma c ++ spc () ++
       str "which should be a coinductive type"
   | NestedRecursiveOccurrences ->
       str "Nested recursive occurrences"
   | UnguardedRecursiveCall c ->
-      str "Unguarded recursive call in" ++ spc () ++ pr_lconstr_env env sigma c
+      str "Unguarded recursive call in" ++ spc () ++ pr_leconstr_env env sigma c
   | RecCallInTypeOfAbstraction c ->
       str "Recursive call forbidden in the domain of an abstraction:" ++
-      spc () ++ pr_lconstr_env env sigma c
+      spc () ++ pr_leconstr_env env sigma c
   | RecCallInNonRecArgOfConstructor c ->
       str "Recursive call on a non-recursive argument of constructor" ++
-      spc () ++ pr_lconstr_env env sigma c
+      spc () ++ pr_leconstr_env env sigma c
   | RecCallInTypeOfDef c ->
       str "Recursive call forbidden in the type of a recursive definition" ++
-      spc () ++ pr_lconstr_env env sigma c
+      spc () ++ pr_leconstr_env env sigma c
   | RecCallInCaseFun c ->
       str "Invalid recursive call in a branch of" ++
-      spc () ++ pr_lconstr_env env sigma c
+      spc () ++ pr_leconstr_env env sigma c
   | RecCallInCaseArg c ->
       str "Invalid recursive call in the argument of \"match\" in" ++ spc () ++
-      pr_lconstr_env env sigma c
+      pr_leconstr_env env sigma c
   | RecCallInCasePred c ->
       str "Invalid recursive call in the \"return\" clause of \"match\" in" ++
-      spc () ++ pr_lconstr_env env sigma c
+      spc () ++ pr_leconstr_env env sigma c
   | NotGuardedForm c ->
-      str "Sub-expression " ++ pr_lconstr_env env sigma c ++
+      str "Sub-expression " ++ pr_leconstr_env env sigma c ++
       strbrk " not in guarded form (should be a constructor," ++
       strbrk " an abstraction, a match, a cofix or a recursive call)"
   | ReturnPredicateNotCoInductive c ->
      str "The return clause of the following pattern matching should be" ++
      strbrk " a coinductive type:" ++
-     spc () ++ pr_lconstr_env env sigma c
-  | FixpointOnIrrelevantInductive ->
-    strbrk "Fixpoints on proof irrelevant inductive types should produce proof irrelevant values"
+     spc () ++ pr_leconstr_env env sigma c
+
+(* TODO: use the names *)
+(* (co)fixpoints *)
+let explain_ill_formed_rec_body env sigma err names i fixenv vdefj =
+  let prt_name i =
+    match names.(i).binder_name with
+        Name id -> str "Recursive definition of " ++ Id.print id
+      | Anonymous -> str "The " ++ pr_nth i ++ str " definition" in
+  let st = match err with
+    | FixGuardError err -> explain_ill_formed_fix_body env sigma names i err
+    | CoFixGuardError  err -> explain_ill_formed_cofix_body env sigma err
   in
   prt_name i ++ str " is ill-formed." ++ fnl () ++
   pr_ne_context_of (str "In environment") env sigma ++
   st ++ str "." ++ fnl () ++
   (try (* May fail with unresolved globals. *)
       let fixenv = make_all_name_different fixenv sigma in
-      let pvd = pr_lconstr_env fixenv sigma vdefj.(i).uj_val in
+      let pvd = pr_leconstr_env fixenv sigma vdefj.(i).uj_val in
         str"Recursive definition is:" ++ spc () ++ pvd ++ str "."
     with e when CErrors.noncritical e -> mt ())
 
@@ -611,15 +633,18 @@ let rec explain_evar_kind env sigma evk ty =
       str " for the variable " ++ Id.print id
   | Evar_kinds.SubEvar (where,evk') ->
       let rec find_source evk =
-        let evi = Evd.find sigma evk in
-        match snd evi.evar_source with
+        let EvarInfo evi = Evd.find sigma evk in
+        match snd (Evd.evar_source evi) with
         | Evar_kinds.SubEvar (_,evk) -> find_source evk
-        | src -> evi,src in
-      let evi,src = find_source evk' in
-      let pc = match evi.evar_body with
+        | src -> EvarInfo evi, src in
+      let EvarInfo evi, src = find_source evk' in
+      let pc = match Evd.evar_body evi with
       | Evar_defined c -> pr_leconstr_env env sigma c
       | Evar_empty -> assert false in
-      let ty' = evi.evar_concl in
+      let ty' = match Evd.evar_body evi with
+      | Evar_empty -> Evd.evar_concl evi
+      | Evar_defined b -> Retyping.get_type_of env sigma b
+      in
       pr_existential_key env sigma evk ++
       strbrk " in the partial instance " ++ pc ++
       strbrk " found for " ++
@@ -627,11 +652,11 @@ let rec explain_evar_kind env sigma evk ty =
       (pr_leconstr_env env sigma ty') src
 
 let explain_typeclass_resolution env sigma evi k =
-  match Typeclasses.class_of_constr env sigma evi.evar_concl with
+  match Typeclasses.class_of_constr env sigma (Evd.evar_concl evi) with
   | Some _ ->
     let env = Evd.evar_filtered_env env evi in
       fnl () ++ str "Could not find an instance for " ++
-      pr_leconstr_env env sigma evi.evar_concl ++
+      pr_leconstr_env env sigma (Evd.evar_concl evi) ++
       pr_trailing_ne_context_of env sigma
   | _ -> mt()
 
@@ -647,11 +672,11 @@ let explain_placeholder_kind env sigma c e =
 let explain_unsolvable_implicit env sigma evk explain =
   let evi = Evarutil.nf_evar_info sigma (Evd.find_undefined sigma evk) in
   let env = Evd.evar_filtered_env env evi in
-  let type_of_hole = pr_leconstr_env env sigma evi.evar_concl in
+  let type_of_hole = pr_leconstr_env env sigma (Evd.evar_concl evi) in
   let pe = pr_trailing_ne_context_of env sigma in
   strbrk "Cannot infer " ++
-  explain_evar_kind env sigma evk type_of_hole (snd evi.evar_source) ++
-  explain_placeholder_kind env sigma evi.evar_concl explain ++ pe
+  explain_evar_kind env sigma evk type_of_hole (snd (Evd.evar_source evi)) ++
+  explain_placeholder_kind env sigma (Evd.evar_concl evi) explain ++ pe
 
 let explain_var_not_found env id =
   str "The variable" ++ spc () ++ Id.print id ++
@@ -674,7 +699,7 @@ let explain_evar_not_found env sigma id =
 
 let explain_wrong_case_info env (ind,u) ci =
   let pi = pr_inductive env ind in
-  if Ind.CanOrd.equal ci.ci_ind ind then
+  if QInd.equal env ci.ci_ind ind then
     str "Pattern-matching expression on an object of inductive type" ++
     spc () ++ pi ++ spc () ++ str "has invalid information."
   else
@@ -756,8 +781,50 @@ let explain_disallowed_sprop () =
       ++ str "\"Allow StrictProp\""
       ++ strbrk " flag is off.")
 
-let explain_bad_relevance env =
-  strbrk "Bad relevance (maybe a bugged tactic)."
+let pr_relevance sigma r =
+  let r = Evarutil.nf_relevance sigma r in
+  match r with
+  | Sorts.Relevant -> str "relevant"
+  | Sorts.Irrelevant -> str "irrelevant"
+  | Sorts.RelevanceVar q -> str "a variable " ++ Sorts.QVar.pr q
+
+let pr_binder env sigma = function
+| LocalAssum (na, t) ->
+  Pp.hov 2 (str "(" ++ (Name.print @@ na.binder_name) ++ str " : " ++
+    pr_leconstr_env env sigma t ++ str ")")
+| LocalDef (na, b, t) ->
+  Pp.hov 2 (str "(" ++ (Name.print @@ na.binder_name) ++ str " : " ++
+    pr_leconstr_env env sigma t ++ str " := " ++ pr_leconstr_env env sigma b ++ str ")")
+
+let explain_bad_binder_relevance env sigma rlv decl =
+  strbrk "Binder" ++ spc () ++ pr_binder env sigma decl ++
+    strbrk " has relevance mark set to " ++ pr_relevance sigma (RelDecl.get_relevance decl) ++
+    strbrk " but was expected to be " ++ pr_relevance sigma rlv ++
+    spc () ++ str "(maybe a bugged tactic)."
+
+let explain_bad_case_relevance env sigma rlv case =
+  let (ci, _, _, _, _, _, _) = EConstr.destCase sigma case in
+  strbrk "Pattern-matching" ++ spc () ++ pr_leconstr_env env sigma case ++
+    strbrk " has relevance mark set to " ++ pr_relevance sigma ci.ci_relevance ++
+    strbrk " but was expected to be " ++ pr_relevance sigma rlv ++
+    spc () ++ str "(maybe a bugged tactic)."
+
+let explain_bad_relevance env sigma = function
+  | Typeops.BadRelevanceCase (r,c) -> explain_bad_case_relevance env sigma r c
+  | BadRelevanceBinder (r,d) -> explain_bad_binder_relevance env sigma r d
+
+let ecast_bad_relevance = let open Typeops in function
+  | BadRelevanceCase (r,c) -> BadRelevanceCase (r, EConstr.of_constr c)
+  | BadRelevanceBinder (r,d) -> BadRelevanceBinder (r, RelDecl.map_constr_het EConstr.of_constr d)
+
+let () =
+  CWarnings.register_printer Typeops.bad_relevance_msg
+    (fun (env, b) ->
+       let sigma = Evd.from_env env in
+       explain_bad_relevance env sigma (ecast_bad_relevance b))
+
+let () = CWarnings.register_printer Typing.bad_relevance_msg
+    (fun (env, sigma, b) -> explain_bad_relevance env sigma b)
 
 let explain_bad_invert env =
   strbrk "Bad case inversion (maybe a bugged tactic)."
@@ -780,10 +847,11 @@ let explain_type_error env sigma err =
       explain_bad_assumption env sigma c
   | ReferenceVariables (id,c) ->
       explain_reference_variables sigma id c
-  | ElimArity (ind, c, pj, okinds) ->
-      explain_elim_arity env sigma ind c pj okinds
+  | ElimArity (ind, c, okinds) ->
+      explain_elim_arity env sigma ind c okinds
   | CaseNotInductive cj ->
       explain_case_not_inductive env sigma cj
+  | CaseOnPrivateInd ind -> explain_case_on_private_ind env sigma ind
   | NumberBranches (cj, n) ->
       explain_number_branches env sigma cj n
   | IllFormedBranch (c, i, actty, expty) ->
@@ -809,7 +877,8 @@ let explain_type_error env sigma err =
   | UndeclaredUniverse l ->
      explain_undeclared_universe env sigma l
   | DisallowedSProp -> explain_disallowed_sprop ()
-  | BadRelevance -> explain_bad_relevance env
+  | BadBinderRelevance (rlv, decl) -> explain_bad_binder_relevance env sigma rlv decl
+  | BadCaseRelevance (rlv, case) -> explain_bad_case_relevance env sigma rlv case
   | BadInvert -> explain_bad_invert env
   | BadVariance {lev;expected;actual} -> explain_bad_variance env sigma ~lev ~expected ~actual
 
@@ -841,10 +910,10 @@ let explain_cannot_unify_occurrences env sigma nested ((cl2,pos2),t2) ((cl1,pos1
 let pr_constraints printenv env sigma evars cstrs =
   let (ev, evi) = Evar.Map.choose evars in
     if Evar.Map.for_all (fun ev' evi' ->
-      eq_named_context_val evi.evar_hyps evi'.evar_hyps) evars
+      eq_named_context_val (Evd.evar_hyps evi) (Evd.evar_hyps evi')) evars
     then
       let l = Evar.Map.bindings evars in
-      let env' = reset_with_named_context evi.evar_hyps env in
+      let env' = Evd.evar_env env evi in
       let pe =
         if printenv then
           pr_ne_context_of (str "In environment:") env' sigma
@@ -853,7 +922,7 @@ let pr_constraints printenv env sigma evars cstrs =
       let evs =
         prlist
         (fun (ev, evi) -> fnl () ++ pr_existential_key (Global.env ()) sigma ev ++
-            str " : " ++ pr_leconstr_env env' sigma evi.evar_concl ++ fnl ()) l
+            str " : " ++ pr_leconstr_env env' sigma (Evd.evar_concl evi) ++ fnl ()) l
       in
       h (pe ++ evs ++ pr_evar_constraints sigma cstrs)
     else
@@ -866,10 +935,7 @@ let explain_unsatisfiable_constraints env sigma constr comp =
   let undef = Evd.undefined_map sigma in
   (* Only keep evars that are subject to resolution and members of the given
      component. *)
-  let is_kept evk _ = match comp with
-  | None -> Evar.Set.mem evk tcs
-  | Some comp -> Evar.Set.mem evk tcs && Evar.Set.mem evk comp
-  in
+  let is_kept evk _ = Evar.Set.mem evk tcs && Evar.Set.mem evk comp in
   let undef =
     let m = Evar.Map.filter is_kept undef in
       if Evar.Map.is_empty m then undef
@@ -904,9 +970,9 @@ let rec explain_pretype_error env sigma err =
   | UnsolvableImplicit (evk,exp) -> explain_unsolvable_implicit env sigma evk exp
   | VarNotFound id -> explain_var_not_found env id
   | EvarNotFound id -> explain_evar_not_found env sigma id
-  | UnexpectedType (actual,expect) ->
+  | UnexpectedType (actual,expect,e) ->
     let env, actual, expect = contract2 env sigma actual expect in
-    explain_unexpected_type env sigma actual expect
+    explain_unexpected_type env sigma actual expect e
   | NotProduct c -> explain_not_product env sigma c
   | CannotUnify (m,n,e) ->
     let env, m, n = contract2 env sigma m n in
@@ -923,12 +989,32 @@ let rec explain_pretype_error env sigma err =
   | AbstractionOverMeta (m,n) -> explain_abstraction_over_meta env m n
   | NonLinearUnification (m,c) -> explain_non_linear_unification env sigma m c
   | TypingError t -> explain_type_error env sigma t
+  | CantApplyBadTypeExplained ((t, rator, randl),error) ->
+    explain_cant_apply_bad_type env sigma ~error t rator randl
   | CannotUnifyOccurrences (b,c1,c2,e) -> explain_cannot_unify_occurrences env sigma b c1 c2 e
   | UnsatisfiableConstraints (c,comp) -> explain_unsatisfiable_constraints env sigma c comp
   | DisallowedSProp -> explain_disallowed_sprop ()
 
 (* Module errors *)
 
+let pr_modpath mp =
+  Libnames.pr_qualid (Nametab.shortest_qualid_of_module mp)
+
+let pr_modtype_subpath upper mp =
+  let rec aux mp =
+    try
+      let (dir,id) = Libnames.repr_qualid (Nametab.shortest_qualid_of_modtype mp) in
+      Libnames.add_dirpath_suffix dir id, []
+    with Not_found ->
+      match mp with
+      | MPdot (mp',id) -> let mp, suff = aux mp' in mp, Label.to_id id::suff
+      | _ -> assert false
+  in
+  let mp, suff = aux mp in
+  (if suff = [] then mt ()
+   else str (if upper then "Module " else "module ") ++ DirPath.print (DirPath.make suff) ++ str " of ") ++
+  DirPath.print mp
+
 open Modops
 
 let explain_not_match_error = function
@@ -981,7 +1067,7 @@ let explain_not_match_error = function
         status (not b) ++ str" declaration was found"
   | IncompatibleUniverses incon ->
     str"the universe constraints are inconsistent: " ++
-      Univ.explain_universe_inconsistency UnivNames.(pr_with_global_universes empty_binders) incon
+      UGraph.explain_universe_inconsistency UnivNames.pr_with_global_universes incon
   | IncompatiblePolymorphism (env, t1, t2) ->
     str "conversion of polymorphic values generates additional constraints: " ++
       quote (Printer.safe_pr_lconstr_env env (Evd.from_env env) t1) ++ spc () ++
@@ -991,7 +1077,7 @@ let explain_not_match_error = function
       let open Univ in
     let pr_auctx auctx =
       let sigma = Evd.from_ctx
-          (UState.of_binders
+          (UState.of_names
              (Printer.universe_binders_with_opt_names auctx None))
       in
       let uctx = AbstractContext.repr auctx in
@@ -1006,23 +1092,42 @@ let explain_not_match_error = function
   | IncompatibleVariance ->
     str "incompatible variance information"
 
-let explain_signature_mismatch l spec why =
-  str "Signature components for label " ++ Label.print l ++
+let rec get_submodules acc = function
+  | [] -> acc, []
+  | Submodule l :: trace -> get_submodules (l::acc) trace
+  | (FunctorArgument _ :: _) as trace -> acc, trace
+
+let get_submodules trace =
+  let submodules, trace = get_submodules [] trace in
+  (String.concat "." (List.map Label.to_string submodules)), trace
+
+let rec print_trace = function
+  | [] -> assert false
+  | (Submodule _ :: _) as trace ->
+    let submodules, trace = get_submodules trace in
+    str submodules ++
+    (if List.is_empty trace then mt() else spc() ++ str "in " ++ print_trace trace)
+  | FunctorArgument n :: trace ->
+    str "the " ++ str (CString.ordinal n) ++ str " functor argument" ++
+    (if List.is_empty trace then mt() else spc() ++ str "of " ++ print_trace trace)
+
+let explain_signature_mismatch trace l why =
+  let submodules, trace = get_submodules trace in
+  let l = if String.is_empty submodules then Label.print l
+    else str submodules ++ str"." ++ Label.print l
+  in
+  str "Signature components for field " ++ l ++
+  (if List.is_empty trace then mt() else str " in " ++ print_trace trace) ++
   str " do not match:" ++ spc () ++ explain_not_match_error why ++ str "."
 
 let explain_label_already_declared l =
   str "The label " ++ Label.print l ++ str " is already declared."
 
-let explain_application_to_not_path _ =
-  strbrk "A module cannot be applied to another module application or " ++
-  strbrk "with-expression; you must give a name to the intermediate result " ++
-  strbrk "module first."
-
 let explain_not_a_functor () =
   str "Application of a non-functor."
 
-let explain_is_a_functor () =
-  str "Illegal use of a functor."
+let explain_is_a_functor mp =
+  pr_modtype_subpath true mp ++ str " not expected to be a functor."
 
 let explain_incompatible_module_types mexpr1 mexpr2 =
   let open Declarations in
@@ -1038,20 +1143,13 @@ let explain_incompatible_module_types mexpr1 mexpr2 =
   else str "Incompatible module types."
 
 let explain_not_equal_module_paths mp1 mp2 =
-  str "Non equal modules."
+  str "Module " ++ pr_modpath mp1 ++ strbrk " is not equal to " ++ pr_modpath mp2 ++ str "."
 
-let explain_no_such_label l =
-  str "No such label " ++ Label.print l ++ str "."
+let explain_no_such_label l mp =
+  str "No field named " ++ Label.print l ++ str " in " ++ pr_modtype_subpath false mp ++ str "."
 
-let explain_incompatible_labels l l' =
-  str "Opening and closing labels are not the same: " ++
-  Label.print l ++ str " <> " ++ Label.print l' ++ str "!"
-
-let explain_not_a_module s =
-  quote (str s) ++ str " is not a module."
-
-let explain_not_a_module_type s =
-  quote (str s) ++ str " is not a module type."
+let explain_not_a_module_label l =
+  Label.print l ++ str " is not the name of a module field."
 
 let explain_not_a_constant l =
   quote (Label.print l) ++ str " is not a constant."
@@ -1070,23 +1168,19 @@ let explain_label_missing l s =
   ++ str s ++ str "."
 
 let explain_include_restricted_functor mp =
-  let q = Nametab.shortest_qualid_of_module mp in
-  str "Cannot include the functor " ++ Libnames.pr_qualid q ++
+  str "Cannot include the functor " ++ pr_modpath mp ++
   strbrk " since it has a restricted signature. " ++
   strbrk "You may name first an instance of this functor, and include it."
 
 let explain_module_error = function
-  | SignatureMismatch (l,spec,err) -> explain_signature_mismatch l spec err
+  | SignatureMismatch (trace,l,err) -> explain_signature_mismatch trace l err
   | LabelAlreadyDeclared l -> explain_label_already_declared l
-  | ApplicationToNotPath mexpr -> explain_application_to_not_path mexpr
   | NotAFunctor -> explain_not_a_functor ()
-  | IsAFunctor -> explain_is_a_functor ()
+  | IsAFunctor mp -> explain_is_a_functor mp
   | IncompatibleModuleTypes (m1,m2) -> explain_incompatible_module_types m1 m2
   | NotEqualModulePaths (mp1,mp2) -> explain_not_equal_module_paths mp1 mp2
-  | NoSuchLabel l -> explain_no_such_label l
-  | IncompatibleLabels (l1,l2) -> explain_incompatible_labels l1 l2
-  | NotAModule s -> explain_not_a_module s
-  | NotAModuleType s -> explain_not_a_module_type s
+  | NoSuchLabel (l,mp) -> explain_no_such_label l mp
+  | NotAModuleLabel l -> explain_not_a_module_label l
   | NotAConstant l -> explain_not_a_constant l
   | IncorrectWithConstraint l -> explain_incorrect_label_constraint l
   | GenerativeModuleExpected l -> explain_generative_module_expected l
@@ -1101,8 +1195,14 @@ let explain_declaration_not_path _ =
 
 *)
 
-let explain_not_module_nor_modtype s =
-  quote (str s) ++ str " is not a module or module type."
+let explain_not_module_nor_modtype qid =
+  Libnames.pr_qualid qid ++ str " is not a module or module type."
+
+let explain_not_a_module qid =
+  Libnames.pr_qualid qid ++ str " is not a module."
+
+let explain_not_a_module_type qid =
+  Libnames.pr_qualid qid ++ str " is not a module type."
 
 let explain_incorrect_with_in_module () =
   str "The syntax \"with\" is not allowed for modules."
@@ -1111,7 +1211,9 @@ let explain_incorrect_module_application () =
   str "Illegal application to a module type."
 
 let explain_module_internalization_error = let open Modintern in function
-  | NotAModuleNorModtype s -> explain_not_module_nor_modtype s
+  | NotAModuleNorModtype qid -> explain_not_module_nor_modtype qid
+  | NotAModule qid -> explain_not_a_module qid
+  | NotAModuleType qid -> explain_not_a_module_type qid
   | IncorrectWithInModule -> explain_incorrect_with_in_module ()
   | IncorrectModuleApplication -> explain_incorrect_module_application ()
 
@@ -1130,13 +1232,6 @@ let explain_typeclass_error env sigma = function
 
 (* Refiner errors *)
 
-let explain_refiner_bad_type env sigma arg ty conclty =
-  let pm, pn = with_diffs (pr_lconstr_env env sigma ty) (pr_leconstr_env env sigma conclty) in
-  str "Refiner was given an argument" ++ brk(1,1) ++
-  pr_lconstr_env env sigma arg ++ spc () ++
-  str "of type" ++ brk(1,1) ++ pm ++ spc () ++
-  str "instead of" ++ brk(1,1) ++ pn ++ str "."
-
 let explain_refiner_unresolved_bindings l =
   str "Unable to find an instance for the " ++
   str (String.plural (List.length l) "variable") ++ spc () ++
@@ -1144,30 +1239,24 @@ let explain_refiner_unresolved_bindings l =
 
 let explain_refiner_cannot_apply env sigma t harg =
   str "In refiner, a term of type" ++ brk(1,1) ++
-  pr_lconstr_env env sigma t ++ spc () ++ str "could not be applied to" ++ brk(1,1) ++
-  pr_lconstr_env env sigma harg ++ str "."
+  pr_leconstr_env env sigma t ++ spc () ++ str "could not be applied to" ++ brk(1,1) ++
+  pr_leconstr_env env sigma harg ++ str "."
 
 let explain_intro_needs_product () =
   str "Introduction tactics needs products."
 
 let explain_non_linear_proof env sigma c =
-  str "Cannot refine with term" ++ brk(1,1) ++ pr_lconstr_env env sigma c ++
+  str "Cannot refine with term" ++ brk(1,1) ++ pr_leconstr_env env sigma c ++
   spc () ++ str "because a metavariable has several occurrences."
 
-let explain_meta_in_type env sigma c =
-  str "In refiner, a meta appears in the type " ++ brk(1,1) ++ pr_leconstr_env env sigma c ++
-  str " of another meta"
-
 let explain_no_such_hyp id =
   str "No such hypothesis: " ++ Id.print id
 
 let explain_refiner_error env sigma = function
-  | BadType (arg,ty,conclty) -> explain_refiner_bad_type env sigma arg ty conclty
   | UnresolvedBindings t -> explain_refiner_unresolved_bindings t
   | CannotApply (t,harg) -> explain_refiner_cannot_apply env sigma t harg
   | IntroNeedsProduct -> explain_intro_needs_product ()
   | NonLinearProof c -> explain_non_linear_proof env sigma c
-  | MetaInType c -> explain_meta_in_type env sigma c
   | NoSuchHyp id -> explain_no_such_hyp id
 
 (* Inductive errors *)
@@ -1238,11 +1327,10 @@ let error_large_non_prop_inductive_not_in_type () =
   str "Large non-propositional inductive types must be in Type."
 
 let error_inductive_missing_constraints (us,ind_univ) =
-  let pr_u = Univ.Universe.pr_with UnivNames.(pr_with_global_universes empty_binders) in
   str "Missing universe constraint declared for inductive type:" ++ spc()
   ++ v 0 (prlist_with_sep spc (fun u ->
-      hov 0 (pr_u u ++ str " <= " ++ pr_u ind_univ))
-      (Univ.Universe.Set.elements us))
+      hov 0 (Printer.pr_sort Evd.empty u ++ str " <= " ++ Printer.pr_sort Evd.empty ind_univ))
+      us)
 
 (* Recursion schemes errors *)
 
@@ -1258,7 +1346,7 @@ let error_not_allowed_dependent_analysis env isrec i =
   pr_inductive env i ++ str "."
 
 let error_not_mutual_in_scheme env ind ind' =
-  if Ind.CanOrd.equal ind ind' then
+  if QInd.equal env ind ind' then
     str "The inductive type " ++ pr_inductive env ind ++
     str " occurs twice."
   else
@@ -1425,7 +1513,7 @@ let wrap_unhandled f e =
 
 let explain_exn_default = function
   (* Basic interaction exceptions *)
-  | Stream.Error txt -> hov 0 (str "Syntax error: " ++ str txt ++ str ".")
+  | Gramlib.Stream.Error txt -> hov 0 (str "Syntax error: " ++ str txt ++ str ".")
   | CLexer.Error.E err -> hov 0 (str (CLexer.Error.to_string err))
   | Sys_error msg -> hov 0 (str "System error: " ++ quote (str msg))
   | Out_of_memory -> hov 0 (str "Out of memory.")
@@ -1438,12 +1526,12 @@ let explain_exn_default = function
 let _ = CErrors.register_handler (wrap_unhandled explain_exn_default)
 
 let rec vernac_interp_error_handler = function
-  | Univ.UniverseInconsistency i ->
+  | UGraph.UniverseInconsistency i ->
     str "Universe inconsistency." ++ spc() ++
-    Univ.explain_universe_inconsistency UnivNames.(pr_with_global_universes empty_binders) i ++ str "."
-  | TypeError(ctx,te) ->
+    UGraph.explain_universe_inconsistency UnivNames.pr_with_global_universes i ++ str "."
+  | TypeError(env,te) ->
     let te = map_ptype_error EConstr.of_constr te in
-    explain_type_error ctx Evd.empty te
+    explain_type_error env (Evd.from_env env) te
   | PretypeError(ctx,sigma,te) ->
     explain_pretype_error ctx sigma te
   | Notation.PrimTokenNotationError(kind,ctx,sigma,te) ->
