@@ -31,7 +31,7 @@ let make_non_locality = function Some false -> false | _ -> true
 let make_locality = function Some true -> true | _ -> false
 
 let warn_local_declaration =
-  CWarnings.create ~name:"local-declaration" ~category:"scope"
+  CWarnings.create ~name:"local-declaration" ~category:CWarnings.CoreCategories.vernacular
     Pp.(fun () ->
         Pp.strbrk "Interpreting this declaration as if " ++
         strbrk "a global declaration prefixed by \"Local\", " ++
@@ -43,7 +43,7 @@ let enforce_locality_exp locality_flag discharge =
   match locality_flag, discharge with
   | Some b, NoDischarge -> Global (importability_of_bool b)
   | None, NoDischarge -> Global ImportDefaultBehavior
-  | None, DoDischarge when not (Global.sections_are_opened ()) ->
+  | None, DoDischarge when not (Lib.sections_are_opened ()) ->
      (* If a Let/Variable is defined outside a section, then we consider it as a local definition *)
      warn_local_declaration ();
      Global ImportNeedQualified
@@ -59,7 +59,7 @@ let enforce_locality locality_flag =
    Local in sections is the default, Local not in section forces non-export *)
 
 let make_section_locality =
-  function Some b -> b | None -> Global.sections_are_opened ()
+  function Some b -> b | None -> Lib.sections_are_opened ()
 
 let enforce_section_locality locality_flag =
   make_section_locality locality_flag
@@ -72,7 +72,7 @@ let enforce_section_locality locality_flag =
 
 let make_module_locality = function
   | Some false ->
-      if Global.sections_are_opened () then
+      if Lib.sections_are_opened () then
         CErrors.user_err Pp.(str
           "This command does not support the Global option in sections.");
       false
