@@ -142,9 +142,7 @@ let pr_import_module (m,f) =
   | ImportNames ns -> surround (prlist_with_sep pr_comma pr_one_import_filter_name ns)
 
 let sep_end = function
-  | VernacBullet _
-  | VernacSubproof _
-  | VernacEndSubproof -> str""
+  | VernacSynPure (VernacBullet _ | VernacSubproof _ | VernacEndSubproof) -> str""
   | _ -> str"."
 
 let sep = fun _ -> spc()
@@ -186,6 +184,9 @@ let pr_notation_entry = function
   | InConstrEntry -> keyword "constr"
   | InCustomEntry s -> keyword "custom" ++ spc () ++ str s
 
+let pr_abbreviation pr (ids, c) =
+  pr c ++ spc () ++ prlist_with_sep spc pr_id ids
+
 let pr_at_level = function
   | NumLevel n -> spc () ++ keyword "at" ++ spc () ++ keyword "level" ++ spc () ++ int n
   | NextLevel -> spc () ++ keyword "at" ++ spc () ++ keyword "next" ++ spc () ++ keyword "level"
@@ -196,7 +197,7 @@ let level_of_pattern_level = function None -> DefaultLevel | Some n -> NumLevel
 let pr_constr_as_binder_kind = let open Notation_term in function
     | AsIdent -> spc () ++ keyword "as ident"
     | AsName -> spc () ++ keyword "as name"
-    | AsNameOrPattern -> spc () ++ keyword "as pattern"
+    | AsAnyPattern -> spc () ++ keyword "as pattern"
     | AsStrictPattern -> spc () ++ keyword "as strict pattern"
 
 let pr_strict b = if b then str "strict " else mt ()
@@ -385,7 +386,7 @@ let pr_module_binders l pr_c =
   prlist_strict (pr_module_vardecls pr_c) l
 
 let pr_type_option pr_c = function
-  | { v = CHole (k, Namegen.IntroAnonymous, _) } -> mt()
+  | { v = CHole (k, Namegen.IntroAnonymous) } -> mt()
   | _ as c -> brk(0,2) ++ str" :" ++ pr_c c
 
 let pr_binders_arg =
@@ -479,25 +480,27 @@ let pr_syntax_modifier = let open Gramlib.Gramext in CAst.with_val (function
     | SetEntryType (x,typ) -> str x ++ spc() ++ pr_set_simple_entry_type typ
     | SetOnlyPrinting -> keyword "only printing"
     | SetOnlyParsing -> keyword "only parsing"
-    | SetFormat (TextFormat s) -> keyword "format " ++ pr_ast qs s
-    | SetFormat (ExtraFormat (k,s)) -> keyword "format " ++ qs k ++ spc() ++ pr_ast qs s)
+    | SetFormat (TextFormat s) -> keyword "format " ++ pr_ast qs s)
 
 let pr_syntax_modifiers = function
   | [] -> mt()
   | l -> spc() ++
          hov 1 (str"(" ++ prlist_with_sep sep_v2 pr_syntax_modifier l ++ str")")
 
-let pr_decl_notation prc decl_ntn =
+let pr_notation_declaration ntn_decl =
   let open Vernacexpr in
   let
-    { decl_ntn_string = {CAst.loc;v=ntn};
-      decl_ntn_interp = c;
-      decl_ntn_modifiers = modifiers;
-      decl_ntn_scope = scopt } = decl_ntn in
-  fnl () ++ keyword "where " ++ qs ntn ++ str " := "
-  ++ Flags.without_option Flags.beautify prc c
+    { ntn_decl_string = {CAst.loc;v=ntn};
+      ntn_decl_interp = c;
+      ntn_decl_modifiers = modifiers;
+      ntn_decl_scope = scopt } = ntn_decl in
+  qs ntn ++ spc () ++ str ":=" ++ spc ()
+  ++ Flags.without_option Flags.beautify pr_constr c
   ++ pr_syntax_modifiers modifiers
-  ++ pr_opt (fun sc -> str ": " ++ str sc) scopt
+  ++ pr_opt (fun sc -> spc () ++ str ":" ++ spc () ++ str sc) scopt
+
+let pr_where_notation decl_ntn =
+  fnl () ++ keyword "where " ++ pr_notation_declaration decl_ntn
 
 let pr_rec_definition { fname; univs; rec_order; binders; rtype; body_def; notations } =
   let pr_pure_lconstr c = Flags.without_option Flags.beautify pr_lconstr c in
@@ -505,7 +508,7 @@ let pr_rec_definition { fname; univs; rec_order; binders; rtype; body_def; notat
   pr_ident_decl (fname,univs) ++ pr_binders_arg binders ++ annot
   ++ pr_type_option (fun c -> spc() ++ pr_lconstr_expr c) rtype
   ++ pr_opt (fun def -> str":=" ++ brk(1,2) ++ pr_pure_lconstr def) body_def
-  ++ prlist (pr_decl_notation @@ pr_constr) notations
+  ++ prlist pr_where_notation notations
 
 let pr_statement head (idpl,(bl,c)) =
   hov 2
@@ -523,28 +526,39 @@ let pr_lconstrarg c =
   spc () ++ pr_lconstr c
 let pr_intarg n = spc () ++ int n
 
-let pr_oc = function
-  | NoInstance -> str" :"
-  | BackInstance -> str" :>"
+let pr_vernac_attributes =
+  function
+  | [] -> mt ()
+  | flags ->  str "#[" ++ prlist_with_sep pr_comma Attributes.pr_vernac_flag flags ++ str "]" ++ cut ()
 
-let pr_record_field (x, { rf_subclass = oc ; rf_priority = pri ; rf_notation = ntn }) =
+let pr_oc coe ins = match coe, ins with
+  | NoCoercion, NoInstance -> str" :"
+  | AddCoercion, NoInstance -> str" :>"
+  | NoCoercion, BackInstance -> str" ::"
+  | AddCoercion, BackInstance -> str" ::>"
+  | _, BackInstanceWarning -> str" :>"  (* remove this line at end of deprecation phase *)
+
+let pr_record_field (x, { rfu_attrs = attr ; rfu_coercion = coe ; rfu_instance = ins ; rfu_priority = pri ; rfu_notation = ntn }) =
   let prx = match x with
     | AssumExpr (id,binders,t) ->
-      hov 1 (pr_lname id ++
+      hov 1 (pr_vernac_attributes attr ++
+             pr_lname id ++
              pr_binders_arg binders ++ spc() ++
-             pr_oc oc ++ spc() ++
+             pr_oc coe ins ++ spc() ++
              pr_lconstr_expr t)
     | DefExpr(id,binders,b,opt) -> (match opt with
         | Some t ->
-          hov 1 (pr_lname id ++
+          hov 1 (pr_vernac_attributes attr ++
+                 pr_lname id ++
                  pr_binders_arg binders ++ spc() ++
-                 pr_oc oc ++ spc() ++
+                 pr_oc coe ins ++ spc() ++
                  pr_lconstr_expr t ++ str" :=" ++ pr_lconstr b)
         | None ->
-          hov 1 (pr_lname id ++ str" :=" ++ spc() ++
+          hov 1 (pr_vernac_attributes attr ++
+                 pr_lname id ++ str" :=" ++ spc() ++
                  pr_lconstr b)) in
   let prpri = match pri with None -> mt() | Some i -> str "| " ++ int i in
-  prx ++ prpri ++ prlist (pr_decl_notation @@ pr_constr) ntn
+  prx ++ prpri ++ prlist pr_where_notation ntn
 
 let pr_record_decl c fs obinder =
   pr_opt pr_lident c ++ pr_record "{" "}" pr_record_field fs ++
@@ -556,9 +570,12 @@ let pr_printable = function
   | PrintSectionContext s ->
     keyword "Print Section" ++ spc() ++ Libnames.pr_qualid s
   | PrintGrammar ent ->
-    keyword "Print Grammar" ++ spc() ++ str ent
+    keyword "Print Grammar" ++ spc() ++
+    prlist_with_sep spc str ent
   | PrintCustomGrammar ent ->
     keyword "Print Custom Grammar" ++ spc() ++ str ent
+  | PrintKeywords ->
+    keyword "Print Keywords"
   | PrintLoadPath dir ->
     keyword "Print LoadPath" ++ pr_opt DirPath.print dir
   | PrintLibraries ->
@@ -573,8 +590,8 @@ let pr_printable = function
     keyword "Print Graph"
   | PrintClasses ->
     keyword "Print Classes"
-  | PrintTypeClasses ->
-    keyword "Print TypeClasses"
+  | PrintTypeclasses ->
+    keyword "Print Typeclasses"
   | PrintInstances qid ->
     keyword "Print Instances" ++ spc () ++ pr_smart_global qid
   | PrintCoercions ->
@@ -673,17 +690,9 @@ let pr_extend s cl =
   with Not_found ->
     hov 1 (str "TODO(" ++ str (fst s) ++ spc () ++ prlist_with_sep sep pr_arg cl ++ str ")")
 
-let pr_vernac_expr v =
+let pr_synpure_vernac_expr v =
   let return = tag_vernac v in
   match v with
-  | VernacLoad (f,s) ->
-    return (
-      keyword "Load"
-      ++ if f then
-        (spc() ++ keyword "Verbose" ++ spc())
-      else
-        spc() ++ qs s
-    )
 
   (* Proof management *)
   | VernacAbortAll ->
@@ -723,6 +732,8 @@ let pr_vernac_expr v =
   | VernacCheckGuard ->
     return (keyword "Guarded")
 
+  | VernacValidateProof -> return (keyword "Validate Proof")
+
   (* Resetting *)
   | VernacResetName id ->
     return (keyword "Reset" ++ spc() ++ pr_lident id)
@@ -757,36 +768,38 @@ let pr_vernac_expr v =
       keyword "Bind Scope" ++ spc () ++ str sc ++
       spc() ++ keyword "with" ++ spc () ++ prlist_with_sep spc pr_class_rawexpr cll
     )
-  | VernacNotation (infix,c,({v=s},l),opt) ->
+  | VernacEnableNotation (on,rule,interp,flags,scope) ->
+    let pr_flag = function
+      | EnableNotationEntry CAst.{v=InConstrEntry} -> str "in constr"
+      | EnableNotationEntry CAst.{v=InCustomEntry s} -> str "in custom " ++ str s
+      | EnableNotationOnly OnlyParsing -> str "only parsing"
+      | EnableNotationOnly OnlyPrinting -> str "only printing"
+      | EnableNotationOnly ParsingAndPrinting -> assert false
+      | EnableNotationAll -> str "all" in
+    let pr_flags = function
+      | [] -> mt ()
+      | l -> str "(" ++ prlist_with_sep pr_comma pr_flag l ++ str ")" in
+    let pr_rule = match rule with
+      | None -> mt ()
+      | Some (Inl ntn) -> quote (str ntn)
+      | Some (Inr abbrev) -> pr_abbreviation pr_qualid abbrev in
+    let pr_opt_scope = function
+      | None -> mt ()
+      | Some (NotationInScope s) -> spc () ++ str ": " ++ str s
+      | Some LastLonelyNotation -> str ":" ++ spc () ++ str "none" in
+    let pp = pr_rule ++ pr_flags flags ++ pr_opt_scope scope in
     return (
-      hov 2 (hov 0 (keyword (if infix then "Infix" else "Notation") ++ spc() ++ qs s ++
-             str " :=" ++ Flags.without_option Flags.beautify pr_constrarg c) ++
-             pr_syntax_modifiers l ++
-             (match opt with
-              | None -> mt()
-              | Some sc -> spc() ++ str":" ++ spc() ++ str sc))
-    )
-  | VernacReservedNotation (_, (s, l)) ->
-    return (
-      keyword "Reserved Notation" ++ spc() ++ pr_ast qs s ++
-      pr_syntax_modifiers l
-    )
-  | VernacNotationAddFormat(s,k,v) ->
-    return (
-      keyword "Format Notation " ++ qs s ++ spc () ++ qs k ++ spc() ++ qs v
-    )
-  | VernacDeclareCustomEntry s ->
-    return (
-      keyword "Declare Custom Entry " ++ str s
+      keyword (if on then "Enable Notation " else "Disable Notation ") ++ pp
     )
 
   (* Gallina *)
   | VernacDefinition ((discharge,kind),id,b) -> (* A verifier... *)
-    let pr_def_token dk =
+    let isgoal = Name.is_anonymous (fst id).v in
+    let pr_def_token =
       keyword (
-        if Name.is_anonymous (fst id).v
+        if isgoal
         then "Goal"
-        else string_of_definition_object_kind dk)
+        else string_of_definition_object_kind kind)
     in
     let pr_reduce = function
       | None -> mt()
@@ -795,24 +808,22 @@ let pr_vernac_expr v =
         pr_red_expr r ++
         keyword " in" ++ spc()
     in
-    let pr_def_body = function
+    let pr_def_body = match b with
       | DefineBody (bl,red,body,d) ->
         let ty = match d with
           | None -> mt()
           | Some ty -> spc() ++ str":" ++ pr_spc_lconstr ty
         in
-        (pr_binders_arg bl,ty,Some (pr_reduce red ++ pr_lconstr body))
+        pr_binders_arg bl  ++ ty ++ str " :=" ++ spc() ++ pr_reduce red ++ pr_lconstr body
       | ProveBody (bl,t) ->
-        let typ u = if (fst id).v = Anonymous then (assert (bl = []); u) else (str" :" ++ u) in
-        (pr_binders_arg bl, typ (pr_spc_lconstr t), None) in
-    let (binds,typ,c) = pr_def_body b in
+        let typ u = if isgoal then (assert (bl = []); u) else (str" :" ++ u) in
+        pr_binders_arg bl ++ typ (pr_spc_lconstr t)
+    in
     return (
       hov 2 (
-        pr_def_token kind ++ spc()
-        ++ pr_lname_decl id ++ binds ++ typ
-        ++ (match c with
-            | None -> mt()
-            | Some cc -> str" :=" ++ spc() ++ cc))
+        pr_def_token
+        ++ (if isgoal then mt() else spc() ++ pr_lname_decl id)
+        ++ pr_def_body)
     )
 
   | VernacStartTheoremProof (ki,l) ->
@@ -837,14 +848,13 @@ let pr_vernac_expr v =
     let n = List.length (List.flatten (List.map fst (List.map snd l))) in
     let pr_params (c, (xl, t)) =
       hov 2 (prlist_with_sep sep pr_ident_decl xl ++ spc() ++
-             (if c then str":>" else str":" ++ spc() ++ pr_lconstr_expr t)) in
+             str(match c with AddCoercion -> ":>" | NoCoercion -> ":") ++ spc() ++ pr_lconstr_expr t) in
     let assumptions = prlist_with_sep spc (fun p -> hov 1 (str "(" ++ pr_params p ++ str ")")) l in
     return (hov 2 (pr_assumption_token (n > 1) discharge kind ++
                    pr_non_empty_arg pr_assumption_inline t ++ spc() ++ assumptions))
   | VernacInductive (f,l) ->
-    let pr_constructor (coe,(id,c)) =
-      hov 2 (pr_lident id ++ str" " ++
-             (if coe then str":>" else str":") ++
+    let pr_constructor ((attr,coe,ins),(id,c)) =
+      hov 2 (pr_vernac_attributes attr ++ pr_lident id ++ pr_oc coe ins ++
              Flags.without_option Flags.beautify pr_spc_lconstr c)
     in
     let pr_constructor_list l = match l with
@@ -860,12 +870,12 @@ let pr_vernac_expr v =
     let pr_oneind key (((coe,iddecl),(indupar,indpar),s,lc),ntn) =
       hov 0 (
         str key ++ spc() ++
-        (if coe then str"> " else str"") ++ pr_cumul_ident_decl iddecl ++
+        str(match coe with AddCoercion -> "> " | NoCoercion -> "") ++ pr_cumul_ident_decl iddecl ++
         pr_and_type_binders_arg indupar ++
         pr_opt (fun p -> str "|" ++ spc() ++ pr_and_type_binders_arg p) indpar ++
         pr_opt (fun s -> str":" ++ spc() ++ pr_lconstr_expr s) s ++
         str" :=") ++ pr_constructor_list lc ++
-      prlist (pr_decl_notation @@ pr_constr) ntn
+      prlist pr_where_notation ntn
     in
     let kind =
       match f with
@@ -898,7 +908,7 @@ let pr_vernac_expr v =
       pr_ident_decl (fname,univs) ++ spc() ++ pr_binders binders ++ spc() ++ str":" ++
       spc() ++ pr_lconstr_expr rtype ++
       pr_opt (fun def -> str":=" ++ brk(1,2) ++ pr_lconstr def) body_def ++
-      prlist (pr_decl_notation @@ pr_constr) notations
+      prlist pr_where_notation notations
     in
     return (
       hov 0 (local ++ keyword "CoFixpoint" ++ spc() ++
@@ -931,35 +941,9 @@ let pr_vernac_expr v =
     )
 
   (* Gallina extensions *)
-  | VernacBeginSection id ->
-    return (hov 2 (keyword "Section" ++ spc () ++ pr_lident id))
-  | VernacEndSegment id ->
-    return (hov 2 (keyword "End" ++ spc() ++ pr_lident id))
   | VernacNameSectionHypSet (id,set) ->
     return (hov 2 (keyword "Collection" ++ spc() ++ pr_lident id ++ spc()++
                    str ":="++spc()++pr_using set))
-  | VernacExtraDependency(from,file,id) ->
-    return (
-      hov 2
-        (keyword "From" ++ spc () ++ pr_module from ++ spc () ++
-         keyword "Extra" ++ spc() ++ keyword "Dependency" ++ spc() ++ qs file ++
-         pr_opt (fun x -> spc() ++ keyword "as" ++ spc () ++ pr_id x) id)
-    )
-  | VernacRequire (from, exp, l) ->
-    let from = match from with
-      | None -> mt ()
-      | Some r -> keyword "From" ++ spc () ++ pr_module r ++ spc ()
-    in
-    return (
-      hov 2
-        (from ++ keyword "Require" ++ spc() ++ pr_require_token exp ++
-         prlist_with_sep sep pr_import_module l)
-    )
-  | VernacImport (f,l) ->
-    return (
-      pr_export_with_cats f ++ spc() ++
-      prlist_with_sep sep pr_import_module l
-    )
   | VernacCanonical q ->
     return (
       keyword "Canonical Structure" ++ spc() ++ pr_smart_global q
@@ -1032,63 +1016,6 @@ let pr_vernac_expr v =
       hov 1 (keyword "Existing" ++ spc () ++ keyword "Class" ++ spc () ++ pr_qualid id)
     )
 
-  (* Modules and Module Types *)
-  | VernacDefineModule (export,m,bl,tys,bd) ->
-    let b = pr_module_binders bl pr_lconstr in
-    return (
-      hov 2 (keyword "Module" ++ spc() ++ pr_require_token export ++
-             pr_lident m ++ b ++
-             pr_of_module_type pr_lconstr tys ++
-             (if List.is_empty bd then mt () else str ":= ") ++
-             prlist_with_sep (fun () -> str " <+")
-               (pr_module_ast_inl true pr_lconstr) bd)
-    )
-  | VernacDeclareModule (export,id,bl,m1) ->
-    let b = pr_module_binders bl pr_lconstr in
-    return (
-      hov 2 (keyword "Declare Module" ++ spc() ++ pr_require_token export ++
-             pr_lident id ++ b ++ str " :" ++
-             pr_module_ast_inl true pr_lconstr m1)
-    )
-  | VernacDeclareModuleType (id,bl,tyl,m) ->
-    let b = pr_module_binders bl pr_lconstr in
-    let pr_mt = pr_module_ast_inl true pr_lconstr in
-    return (
-      hov 2 (keyword "Module Type " ++ pr_lident id ++ b ++
-             prlist_strict (fun m -> str " <:" ++ pr_mt m) tyl ++
-             (if List.is_empty m then mt () else str ":= ") ++
-             prlist_with_sep (fun () -> str " <+ ") pr_mt m)
-    )
-  | VernacInclude (mexprs) ->
-    let pr_m = pr_module_ast_inl false pr_lconstr in
-    return (
-      hov 2 (keyword "Include" ++ spc() ++
-             prlist_with_sep (fun () -> str " <+ ") pr_m mexprs)
-    )
-
-  (* Auxiliary file and library management *)
-  | VernacAddLoadPath { implicit; physical_path; logical_path } ->
-    return (
-      hov 2
-        (keyword "Add" ++
-         (if implicit then spc () ++ keyword "Rec" ++ spc () else spc()) ++
-         keyword "LoadPath" ++ spc() ++ qs physical_path ++
-         spc() ++ keyword "as" ++ spc() ++ DirPath.print logical_path))
-  | VernacRemoveLoadPath s ->
-    return (keyword "Remove LoadPath" ++ qs s)
-  | VernacAddMLPath (s) ->
-    return (
-      keyword "Add"
-      ++ keyword "ML Path"
-      ++ qs s
-    )
-  | VernacDeclareMLModule (l) ->
-    return (
-      hov 2 (keyword "Declare ML Module" ++ spc() ++ prlist_with_sep sep qs l)
-    )
-  | VernacChdir s ->
-    return (keyword "Cd" ++ pr_opt qs s)
-
   (* Commands *)
   | VernacCreateHintDb (dbname,b) ->
     return (
@@ -1106,8 +1033,7 @@ let pr_vernac_expr v =
   | VernacSyntacticDefinition (id,(ids,c),l) ->
     return (
       hov 2
-        (keyword "Notation" ++ spc () ++ pr_lident id ++ spc () ++
-         prlist_with_sep spc pr_id ids ++ str":=" ++ pr_constrarg c ++
+        (keyword "Notation" ++ spc () ++ pr_abbreviation pr_lident (ids,id) ++ str":=" ++ pr_constrarg c ++
          pr_syntax_modifiers l)
     )
   | VernacArguments (q, args, more_implicits, mods) ->
@@ -1115,7 +1041,7 @@ let pr_vernac_expr v =
       hov 2 (
         keyword "Arguments" ++ spc() ++
         pr_smart_global q ++
-        let pr_s = function None -> str"" | Some {v=s} -> str "%" ++ str s in
+        let pr_s = prlist (fun {v=s} -> str "%" ++ str s) in
         let pr_if b x = if b then x else str "" in
         let pr_one_arg (x,k) = pr_if k (str"!") ++ Name.print x in
         let pr_br imp force x =
@@ -1128,11 +1054,11 @@ let pr_vernac_expr v =
           left ++ x ++ right
         in
         let get_arguments_like s imp tl =
-          if s = None && imp = Glob_term.Explicit then [], tl
+          if s = [] && imp = Glob_term.Explicit then [], tl
           else
             let rec fold extra = function
               | RealArg arg :: tl when
-                  Option.equal (fun a b -> String.equal a.CAst.v b.CAst.v) arg.notation_scope s
+                  List.equal (fun a b -> String.equal a.CAst.v b.CAst.v) arg.notation_scope s
                   && arg.implicit_status = imp ->
                 fold ((arg.name,arg.recarg_like) :: extra) tl
               | args -> List.rev extra, args
@@ -1218,12 +1144,6 @@ let pr_vernac_expr v =
       hov 1 (keyword "Strategy" ++ spc() ++
              hv 0 (prlist_with_sep sep pr_line l))
     )
-  | VernacSetOption (export, na,v) ->
-    let export = if export then keyword "Export" ++ spc () else mt () in
-    let set = if v == OptionUnset then "Unset" else "Set" in
-    return (
-      hov 2 (export ++ keyword set ++ spc() ++ pr_set_option na v)
-    )
   | VernacAddOption (na,l) ->
     return (
       hov 2 (keyword "Add" ++ spc() ++ pr_printoption na (Some l))
@@ -1296,9 +1216,6 @@ let pr_vernac_expr v =
          ++ prlist_with_sep sep (pr_comment pr_constr) l)
     )
 
-  (* For extension *)
-  | VernacExtend (s,c) ->
-    return (pr_extend s c)
   | VernacProof (None, None) ->
     return (keyword "Proof")
   | VernacProof (None, Some e) ->
@@ -1312,8 +1229,6 @@ let pr_vernac_expr v =
       keyword "using" ++ spc() ++ pr_using e ++ spc() ++
       keyword "with" ++ spc() ++ pr_gen te
     )
-  | VernacProofMode s ->
-    return (keyword "Proof Mode" ++ str s)
   | VernacBullet b ->
     (* XXX: Redundant with Proof_bullet.print *)
     return (let open Proof_bullet in begin match b with
@@ -1328,9 +1243,113 @@ let pr_vernac_expr v =
   | VernacEndSubproof ->
     return (str "}")
 
+let pr_synterp_vernac_expr v =
+  let return = tag_vernac v in
+  match v with
+  | VernacLoad (f,s) ->
+    return (
+      keyword "Load"
+      ++ if f then
+        (spc() ++ keyword "Verbose" ++ spc())
+      else
+        spc() ++ qs s
+    )
+
+  | VernacBeginSection id ->
+    return (hov 2 (keyword "Section" ++ spc () ++ pr_lident id))
+  | VernacEndSegment id ->
+    return (hov 2 (keyword "End" ++ spc() ++ pr_lident id))
+  | VernacNotation (infix,ntn_decl) ->
+    return (
+      hov 2 (hov 0 (keyword (if infix then "Infix" else "Notation") ++ spc() ++
+      pr_notation_declaration ntn_decl))
+    )
+  | VernacReservedNotation (_, (s, l)) ->
+    return (
+      keyword "Reserved Notation" ++ spc() ++ pr_ast qs s ++
+      pr_syntax_modifiers l
+    )
+  | VernacDeclareCustomEntry s ->
+    return (
+      keyword "Declare Custom Entry " ++ str s
+    )
+  | VernacRequire (from, exp, l) ->
+    let from = match from with
+      | None -> mt ()
+      | Some r -> keyword "From" ++ spc () ++ pr_module r ++ spc ()
+    in
+    return (
+      hov 2
+        (from ++ keyword "Require" ++ spc() ++ pr_require_token exp ++
+         prlist_with_sep sep pr_import_module l)
+    )
+  | VernacImport (f,l) ->
+    return (
+      pr_export_with_cats f ++ spc() ++
+      prlist_with_sep sep pr_import_module l
+    )
+  (* Modules and Module Types *)
+  | VernacDefineModule (export,m,bl,tys,bd) ->
+    let b = pr_module_binders bl pr_lconstr in
+    return (
+      hov 2 (keyword "Module" ++ spc() ++ pr_require_token export ++
+             pr_lident m ++ b ++
+             pr_of_module_type pr_lconstr tys ++
+             (if List.is_empty bd then mt () else str ":= ") ++
+             prlist_with_sep (fun () -> str " <+")
+               (pr_module_ast_inl true pr_lconstr) bd)
+    )
+  | VernacDeclareModule (export,id,bl,m1) ->
+    let b = pr_module_binders bl pr_lconstr in
+    return (
+      hov 2 (keyword "Declare Module" ++ spc() ++ pr_require_token export ++
+             pr_lident id ++ b ++ str " :" ++
+             pr_module_ast_inl true pr_lconstr m1)
+    )
+  | VernacDeclareModuleType (id,bl,tyl,m) ->
+    let b = pr_module_binders bl pr_lconstr in
+    let pr_mt = pr_module_ast_inl true pr_lconstr in
+    return (
+      hov 2 (keyword "Module Type " ++ pr_lident id ++ b ++
+             prlist_strict (fun m -> str " <:" ++ pr_mt m) tyl ++
+             (if List.is_empty m then mt () else str ":= ") ++
+             prlist_with_sep (fun () -> str " <+ ") pr_mt m)
+    )
+  | VernacInclude (mexprs) ->
+    let pr_m = pr_module_ast_inl false pr_lconstr in
+    return (
+      hov 2 (keyword "Include" ++ spc() ++
+             prlist_with_sep (fun () -> str " <+ ") pr_m mexprs)
+    )
+
+  (* Auxiliary file and library management *)
+  | VernacDeclareMLModule (l) ->
+    return (
+      hov 2 (keyword "Declare ML Module" ++ spc() ++ prlist_with_sep sep qs l)
+    )
+  | VernacChdir s ->
+    return (keyword "Cd" ++ pr_opt qs s)
+  | VernacSetOption (export, na,v) ->
+    let export = if export then keyword "Export" ++ spc () else mt () in
+    let set = if v == OptionUnset then "Unset" else "Set" in
+    return (
+      hov 2 (export ++ keyword set ++ spc() ++ pr_set_option na v)
+    )
+  | VernacExtraDependency(from,file,id) ->
+    return (
+      hov 2
+        (keyword "From" ++ spc () ++ pr_module from ++ spc () ++
+         keyword "Extra" ++ spc() ++ keyword "Dependency" ++ spc() ++ qs file ++
+         pr_opt (fun x -> spc() ++ keyword "as" ++ spc () ++ pr_id x) id)
+    )
+  | VernacExtend (s,c) ->
+    return (pr_extend s c)
+  | VernacProofMode s ->
+    return (keyword "Proof Mode" ++ str s)
+
 let pr_control_flag (p : control_flag) =
   let w = match p with
-    | ControlTime _ -> keyword "Time"
+    | ControlTime -> keyword "Time"
     | ControlRedirect s -> keyword "Redirect" ++ spc() ++ qs s
     | ControlTimeout n -> keyword "Timeout " ++ int n
     | ControlFail -> keyword "Fail"
@@ -1340,10 +1359,10 @@ let pr_control_flag (p : control_flag) =
 
 let pr_vernac_control flags = Pp.prlist pr_control_flag flags
 
-let pr_vernac_attributes =
-  function
-  | [] -> mt ()
-  | flags ->  str "#[" ++ prlist_with_sep pr_comma Attributes.pr_vernac_flag flags ++ str "]" ++ cut ()
+let pr_vernac_expr v =
+  match v with
+  | VernacSynPure e -> pr_synpure_vernac_expr e
+  | VernacSynterp e -> pr_synterp_vernac_expr e
 
 let pr_vernac ({v = {control; attrs; expr}} as v) =
   tag_vernac v
