@@ -36,7 +36,7 @@ type vernac_classification =
      don't modify the interpretation state. *)
   | VtQuery
   (* Commands that change the current proof mode *)
-  | VtProofMode of string
+  | VtProofMode of Pvernac.proof_mode
   (* To be removed *)
   | VtMeta
 and vernac_start = opacity_guarantee * Names.Id.t list
@@ -98,18 +98,9 @@ module OutProof = struct
   type _ t =
     | No : unit t
     | Close : unit t
-    | Yes : Declare.Proof.t t
+    | Update : Declare.Proof.t t
+    | New : Declare.Proof.t t
 
-  type result =
-    | Ignored
-    | Closed
-    | Open of Declare.Proof.t
-
-  let cast (type a) (x:a) (ty:a t) : result =
-    match ty with
-    | No -> Ignored
-    | Close -> Closed
-    | Yes -> Open x
 end
 
 type ('inprog,'outprog,'inproof,'outproof) vernac_type = {
@@ -128,8 +119,6 @@ type typed_vernac =
       run : pm:'inprog -> proof:'inproof -> 'outprog * 'outproof;
     } -> typed_vernac
 
-[@@@ocaml.warning "-40"]
-
 let vtdefault f =
   TypedVernac { inprog = Ignore; outprog = No; inproof = Ignore; outproof = No;
                 run = (fun ~pm:() ~proof:() ->
@@ -151,14 +140,14 @@ let vtcloseproof f =
               }
 
 let vtopenproof f =
-  TypedVernac { inprog = Ignore; outprog = No; inproof = Ignore; outproof = Yes;
+  TypedVernac { inprog = Ignore; outprog = No; inproof = Ignore; outproof = New;
                 run = (fun ~pm:() ~proof:() ->
                     let proof = f () in
                     (), proof)
               }
 
 let vtmodifyproof f =
-  TypedVernac { inprog = Ignore; outprog = No; inproof = Use; outproof = Yes;
+  TypedVernac { inprog = Ignore; outprog = No; inproof = Use; outproof = Update;
                 run = (fun ~pm:() ~proof ->
                     let proof = f ~pstate:proof in
                     (), proof)
@@ -193,21 +182,19 @@ let vtmodifyprogram f =
               }
 
 let vtdeclareprogram f =
-  TypedVernac { inprog = Use; outprog = No; inproof = Ignore; outproof = Yes;
+  TypedVernac { inprog = Use; outprog = No; inproof = Ignore; outproof = New;
                 run = (fun ~pm ~proof:() ->
                     let proof = f ~pm in
                     (), proof)
               }
 
 let vtopenproofprogram f =
-  TypedVernac { inprog = Use; outprog = Yes; inproof = Ignore; outproof = Yes;
+  TypedVernac { inprog = Use; outprog = Yes; inproof = Ignore; outproof = New;
                 run = (fun ~pm ~proof:() ->
                     let pm, proof = f ~pm in
                     pm, proof)
               }
 
-[@@@ocaml.warning "+40"]
-
 type vernac_command = ?loc:Loc.t -> atts:Attributes.vernac_flags -> unit -> typed_vernac
 
 type plugin_args = Genarg.raw_generic_argument list
@@ -217,28 +204,23 @@ let vernac_tab =
   (Hashtbl.create 211 :
     (Vernacexpr.extend_name, bool * (plugin_args -> vernac_command)) Hashtbl.t)
 
-let vinterp_add depr s f =
-  try
-    Hashtbl.add vernac_tab s (depr, f)
-  with Failure _ ->
-    user_err
-      (str"Cannot add the vernac command " ++ str (fst s) ++ str" twice.")
+let vinterp_add depr s f = Hashtbl.replace vernac_tab s (depr, f)
 
 let vinterp_map s =
   try
     Hashtbl.find vernac_tab s
-  with Failure _ | Not_found ->
+  with Not_found ->
     user_err
       (str"Cannot find vernac command " ++ str (fst s) ++ str".")
 
 let warn_deprecated_command =
   let open CWarnings in
-  create ~name:"deprecated-command" ~category:"deprecated"
+  create ~name:"deprecated-command" ~category:CWarnings.CoreCategories.deprecated
          (fun pr -> str "Deprecated vernacular command: " ++ pr)
 
 (* Interpretation of a vernac command *)
 
-let type_vernac opn converted_args ?loc ~atts () =
+let type_vernac opn converted_args ?loc ~atts =
   let depr, callback = vinterp_map opn in
   let () = if depr then
       let rules = Egramml.get_extend_vernac_rule opn in
@@ -249,8 +231,7 @@ let type_vernac opn converted_args ?loc ~atts () =
       let pr = pr_sequence pr_gram rules in
       warn_deprecated_command pr;
   in
-  let hunk = callback converted_args in
-  hunk ?loc ~atts ()
+  callback converted_args ?loc ~atts
 
 (** VERNAC EXTEND registering *)
 
@@ -316,9 +297,9 @@ let rec untype_command : type r s. (r, s) ty_sig -> r -> plugin_args -> vernac_c
 let rec untype_user_symbol : type s a b c. (a, b, c) Extend.ty_user_symbol -> (s, Gramlib.Grammar.norec, a) Pcoq.Symbol.t =
   let open Extend in function
   | TUlist1 l -> Pcoq.Symbol.list1 (untype_user_symbol l)
-  | TUlist1sep (l, s) -> Pcoq.Symbol.list1sep (untype_user_symbol l) (Pcoq.Symbol.tokens [Pcoq.TPattern (CLexer.terminal s)]) false
+  | TUlist1sep (l, s) -> Pcoq.Symbol.list1sep (untype_user_symbol l) (Pcoq.Symbol.tokens [Pcoq.TPattern (Pcoq.terminal s)]) false
   | TUlist0 l -> Pcoq.Symbol.list0 (untype_user_symbol l)
-  | TUlist0sep (l, s) -> Pcoq.Symbol.list0sep (untype_user_symbol l) (Pcoq.Symbol.tokens [Pcoq.TPattern (CLexer.terminal s)]) false
+  | TUlist0sep (l, s) -> Pcoq.Symbol.list0sep (untype_user_symbol l) (Pcoq.Symbol.tokens [Pcoq.TPattern (Pcoq.terminal s)]) false
   | TUopt o -> Pcoq.Symbol.opt (untype_user_symbol o)
   | TUentry a -> Pcoq.Symbol.nterm (Pcoq.genarg_grammar (Genarg.ExtraArg a))
   | TUentryl (a, i) -> Pcoq.Symbol.nterml (Pcoq.genarg_grammar (Genarg.ExtraArg a)) (string_of_int i)
@@ -331,7 +312,21 @@ let rec untype_grammar : type r s. (r, s) ty_sig -> 'a Egramml.grammar_prod_item
   let symb = untype_user_symbol tu in
   Egramml.GramNonTerminal (Loc.tag (t, symb)) :: untype_grammar ty
 
-let vernac_extend ~command ?classifier ?entry ext =
+let declare_dynamic_vernac_extend ~command ?entry ~depr cl ty f =
+  let cl = untype_classifier ty cl in
+  let f = untype_command ty f in
+  let r = untype_grammar ty in
+  let ext = command, 0 in
+  vinterp_add depr ext f;
+  Egramml.declare_vernac_command_grammar ~allow_override:true ext entry r;
+  declare_vernac_classifier command [|cl|];
+  ext
+
+let is_static_linking_done = ref false
+
+let static_linking_done () = is_static_linking_done := true
+
+let static_vernac_extend ~plugin ~command ?classifier ?entry ext =
   let get_classifier (TyML (_, ty, _, cl)) = match cl with
   | Some cl -> untype_classifier ty cl
   | None ->
@@ -367,10 +362,24 @@ let vernac_extend ~command ?classifier ?entry ext =
     let f = untype_command ty f in
     let r = untype_grammar ty in
     let () = vinterp_add depr (command, i) f in
-    Egramml.extend_vernac_command_grammar (command, i) entry r
+    let () = Egramml.declare_vernac_command_grammar ~allow_override:false (command, i) entry r in
+    let () = match plugin with
+      | None ->
+        let () =
+          if !is_static_linking_done
+          then CErrors.anomaly
+              Pp.(str "static_vernac_extend in dynlinked code must pass non-None plugin.")
+        in
+        Egramml.extend_vernac_command_grammar ~undoable:false (command, i)
+      | Some plugin ->
+        Mltop.add_init_function plugin (fun () ->
+            Egramml.extend_vernac_command_grammar ~undoable:true (command, i))
+    in
+    ()
   in
   let () = declare_vernac_classifier command cl in
-  List.iteri iter ext
+  let () = List.iteri iter ext in
+  ()
 
 (** VERNAC ARGUMENT EXTEND registering *)
 
@@ -383,7 +392,7 @@ type 'a vernac_argument = {
   arg_parsing : 'a argument_rule;
 }
 
-let vernac_argument_extend ~name arg =
+let vernac_argument_extend ~plugin ~name arg =
   let wit = Genarg.create_arg name in
   let entry = match arg.arg_parsing with
   | Arg_alias e ->
@@ -391,7 +400,10 @@ let vernac_argument_extend ~name arg =
     e
   | Arg_rules rules ->
     let e = Pcoq.create_generic_entry2 name (Genarg.rawwit wit) in
-    let () = Pcoq.grammar_extend e (Pcoq.Fresh (Gramlib.Gramext.First, [None, None, rules])) in
+    let plugin_uid = (plugin, "vernacargextend:"^name) in
+    let () = Egramml.grammar_extend ~plugin_uid e
+        (Pcoq.Fresh (Gramlib.Gramext.First, [None, None, rules]))
+    in
     e
   in
   let pr = arg.arg_printer in
