@@ -9,7 +9,17 @@
 (************************************************************************)
 
 (** The main interpretation function of vernacular expressions *)
-val interp : ?verbosely:bool -> st:Vernacstate.t -> Vernacexpr.vernac_control -> Vernacstate.t
+val interp
+  : ?verbosely:bool
+  -> st:Vernacstate.t
+  -> Vernacexpr.vernac_control
+  -> Vernacstate.t
+
+val interp_entry
+  : ?verbosely:bool
+  -> st:Vernacstate.t
+  -> Synterp.vernac_control_entry
+  -> Vernacstate.Interp.t
 
 (** Execute a Qed but with a proof_object which may contain a delayed
    proof and won't be forced *)
@@ -18,12 +28,4 @@ val interp_qed_delayed_proof
   -> st:Vernacstate.t
   -> control:Vernacexpr.control_flag list
   -> Vernacexpr.proof_end CAst.t
-  -> Vernacstate.t
-
-(** Flag set when the test-suite is called. Its only effect to display
-    verbose information for [Fail] *)
-val test_mode : bool ref
-
-(** Default proof mode set by `start_proof` *)
-val get_default_proof_mode : unit -> Pvernac.proof_mode
-val proof_mode_opt_name : string list
+  -> Vernacstate.Interp.t
