@@ -12,55 +12,107 @@ module Parser = struct
 
   type t = Pcoq.frozen_t
 
-  let init () = Pcoq.freeze ~marshallable:false
+  let init () = Pcoq.freeze ()
 
-  let cur_state () = Pcoq.freeze ~marshallable:false
+  let cur_state () = Pcoq.freeze ()
 
   let parse ps entry pa =
     Pcoq.unfreeze ps;
-    Flags.with_option Flags.we_are_parsing
-      (fun () -> Pcoq.Entry.parse entry pa)
-      ()
+    Pcoq.Entry.parse entry pa
 
 end
 
-module System : sig
-  type t
-  val protect : ('a -> 'b) -> 'a -> 'b
-  val freeze : marshallable:bool -> t
-  val unfreeze : t -> unit
+module System = struct
+
+  module Synterp = struct
+
+    type t = Lib.Synterp.frozen * Summary.Synterp.frozen
+
+    let freeze () =
+      (Lib.Synterp.freeze (), Summary.Synterp.freeze_summaries ())
+
+    let unfreeze (fl,fs) =
+      Lib.Synterp.unfreeze fl;
+      Summary.Synterp.unfreeze_summaries fs
+
+    module Stm = struct
+      let make_shallow (lib, summary) = Lib.Synterp.drop_objects lib, Summary.Synterp.make_marshallable summary
+      let lib = fst
+      let summary = snd
+    end
 
-  module Stm : sig
-    val make_shallow : t -> t
-    val lib : t -> Lib.frozen
-    val summary : t -> Summary.frozen
-    val replace_summary : t -> Summary.frozen -> t
   end
-end = struct
-  type t = Lib.frozen * Summary.frozen
 
-  let freeze ~marshallable =
-    (Lib.freeze (), Summary.freeze_summaries ~marshallable)
+  module Interp : sig
+    type t
+    val freeze : unit -> t
+    val unfreeze : t -> unit
+    module Stm : sig
+      val make_shallow : t -> t
+      val lib : t -> Lib.Interp.frozen
+      val summary : t -> Summary.Interp.frozen
+      val replace_summary : t -> Summary.Interp.frozen -> t
+    end
 
-  let unfreeze (fl,fs) =
-    Lib.unfreeze fl;
-    Summary.unfreeze_summaries fs
+  end = struct
+
+    type t = Lib.Interp.frozen * Summary.Interp.frozen
+
+    let freeze () =
+      (Lib.Interp.freeze (), Summary.Interp.freeze_summaries ())
+
+    let unfreeze (fl,fs) =
+      Lib.Interp.unfreeze fl;
+      Summary.Interp.unfreeze_summaries fs
+
+    (* STM-specific state manipulations *)
+    module Stm = struct
+      let make_shallow (lib, summary) = Lib.Interp.drop_objects lib, Summary.Interp.make_marshallable summary
+      let lib = fst
+      let summary = snd
+      let replace_summary (lib,_) summary = (lib,summary)
+    end
+  end
 
   let protect f x =
-    let st = freeze ~marshallable:false in
+    let synterp_st = Synterp.freeze () in
+    let interp_st = Interp.freeze () in
     try
-      let a = f x in unfreeze st; a
+      let a = f x in
+      Synterp.unfreeze synterp_st;
+      Interp.unfreeze interp_st;
+      a
     with reraise ->
       let reraise = Exninfo.capture reraise in
-      (unfreeze st; Exninfo.iraise reraise)
+      begin
+        Synterp.unfreeze synterp_st;
+        Interp.unfreeze interp_st;
+        Exninfo.iraise reraise
+      end
+
+end
+
+module Synterp = struct
+
+  type t =
+    { parsing : Parser.t
+    ; system : System.Synterp.t
+    }
+
+  let freeze () =
+    { parsing = Parser.cur_state ();
+      system = System.Synterp.freeze ();
+    }
+
+  let make_shallow s =
+    { s with system = System.Synterp.Stm.make_shallow s.system }
+
+  let init () = freeze ()
+
+  let unfreeze st =
+    System.Synterp.unfreeze st.system;
+    Pcoq.unfreeze st.parsing
 
-  (* STM-specific state manipulations *)
-  module Stm = struct
-    let make_shallow (lib, summary) = Lib.drop_objects lib, summary
-    let lib = fst
-    let summary = snd
-    let replace_summary (lib,_) summary = (lib,summary)
-  end
 end
 
 module LemmaStack = struct
@@ -89,19 +141,19 @@ module LemmaStack = struct
 
 end
 
-type t = {
-  parsing : Parser.t;
-  system  : System.t;              (* summary + libstack *)
-  lemmas  : LemmaStack.t option;   (* proofs of lemmas currently opened *)
-  program : Declare.OblState.t NeList.t;    (* obligations table *)
-  opaques : Opaques.Summary.t;     (* opaque proof terms *)
-  shallow : bool                   (* is the state trimmed down (libstack) *)
-}
-
 let s_cache = ref None
 let s_lemmas = ref None
 let s_program = ref (NeList.singleton Declare.OblState.empty)
 
+module Interp = struct
+
+type t = {
+  system  : System.Interp.t;              (* summary + libstack *)
+  lemmas  : LemmaStack.t option;   (* proofs of lemmas currently opened *)
+  program : Declare.OblState.t NeList.t;    (* obligations table *)
+  opaques : Opaques.Summary.t;     (* opaque proof terms *)
+}
+
 let invalidate_cache () =
   s_cache := None
 
@@ -117,26 +169,41 @@ let do_if_not_cached rf f v =
   | Some _ ->
     ()
 
-let freeze_interp_state ~marshallable =
-  { system = update_cache s_cache (System.freeze ~marshallable);
+let freeze_interp_state () =
+  { system = update_cache s_cache (System.Interp.freeze ());
     lemmas = !s_lemmas;
     program = !s_program;
-    opaques = Opaques.Summary.freeze ~marshallable;
-    shallow = false;
-    parsing = Parser.cur_state ();
+    opaques = Opaques.Summary.freeze ();
   }
 
-let unfreeze_interp_state { system; lemmas; program; parsing; opaques } =
-  do_if_not_cached s_cache System.unfreeze system;
+let make_shallow s =
+  { s with system = System.Interp.Stm.make_shallow s.system }
+
+let unfreeze_interp_state { system; lemmas; program; opaques } =
+  do_if_not_cached s_cache System.Interp.unfreeze system;
   s_lemmas := lemmas;
   s_program := program;
-  Opaques.Summary.unfreeze opaques;
-  Pcoq.unfreeze parsing
+  Opaques.Summary.unfreeze opaques
+
+end
+
+type t =
+  { synterp: Synterp.t
+  ; interp: Interp.t
+  }
+
+let freeze_full_state () =
+  { synterp = Synterp.freeze ();
+    interp = Interp.freeze_interp_state ();
+  }
+
+let unfreeze_full_state st =
+  Synterp.unfreeze st.synterp;
+  Interp.unfreeze_interp_state st.interp
 
 (* Compatibility module *)
 module Declare_ = struct
 
-  let get () = !s_lemmas
   let get_program () = !s_program
 
   let set (pstate,pm) =
@@ -146,7 +213,6 @@ module Declare_ = struct
   let get_pstate () =
     Option.map (LemmaStack.with_top ~f:(fun x -> x)) !s_lemmas
 
-  let freeze ~marshallable:_ = get ()
   let unfreeze x = s_lemmas := Some x
 
   exception NoCurrentProof
@@ -226,44 +292,46 @@ module Stm = struct
     int                                     (* Evd.evar_counter_summary_tag *)
 
   (* Parts of the system state that are morally part of the proof state *)
-  let pstate { lemmas; system } =
-    let st = System.Stm.summary system in
+  let pstate { interp = { lemmas; system }} =
+    let st = System.Interp.Stm.summary system in
     lemmas,
-    Summary.project_from_summary st Evarutil.meta_counter_summary_tag,
-    Summary.project_from_summary st Evd.evar_counter_summary_tag
+    Summary.Interp.project_from_summary st Evarutil.meta_counter_summary_tag,
+    Summary.Interp.project_from_summary st Evd.evar_counter_summary_tag
 
-  let set_pstate ({ lemmas; system } as s) (pstate,c1,c2) =
-    { s with
+  let set_pstate ({ interp = { lemmas; system } } as s) (pstate,c1,c2) =
+    { s with interp = { s.interp with
       lemmas =
-        Declare_.copy_terminators ~src:s.lemmas ~tgt:pstate
+        Declare_.copy_terminators ~src:s.interp.lemmas ~tgt:pstate
     ; system =
-        System.Stm.replace_summary s.system
+        System.Interp.Stm.replace_summary s.interp.system
           begin
-            let st = System.Stm.summary s.system in
-            let st = Summary.modify_summary st Evarutil.meta_counter_summary_tag c1 in
-            let st = Summary.modify_summary st Evd.evar_counter_summary_tag c2 in
+            let st = System.Interp.Stm.summary s.interp.system in
+            let st = Summary.Interp.modify_summary st Evarutil.meta_counter_summary_tag c1 in
+            let st = Summary.Interp.modify_summary st Evd.evar_counter_summary_tag c2 in
             st
           end
       }
+    }
 
-  type non_pstate = Summary.frozen * Lib.frozen
-  let non_pstate { system } =
-    let st = System.Stm.summary system in
-    let st = Summary.remove_from_summary st Evarutil.meta_counter_summary_tag in
-    let st = Summary.remove_from_summary st Evd.evar_counter_summary_tag in
-    st, System.Stm.lib system
+  type non_pstate = Summary.Synterp.frozen * Lib.Synterp.frozen * Summary.Interp.frozen * Lib.Interp.frozen
+  let non_pstate { synterp; interp } =
+    let system = interp.system in
+    let st = System.Interp.Stm.summary system in
+    let st = Summary.Interp.remove_from_summary st Evarutil.meta_counter_summary_tag in
+    let st = Summary.Interp.remove_from_summary st Evd.evar_counter_summary_tag in
+    System.Synterp.Stm.summary synterp.system, System.Synterp.Stm.lib synterp.system,
+      st, System.Interp.Stm.lib system
 
-  let same_env { system = s1 } { system = s2 } =
-    let s1 = System.Stm.summary s1 in
-    let e1 = Summary.project_from_summary s1 Global.global_env_summary_tag in
-    let s2 = System.Stm.summary s2 in
-    let e2 = Summary.project_from_summary s2 Global.global_env_summary_tag in
+  let same_env { interp = { system = s1 } } { interp = { system = s2 } } =
+    let s1 = System.Interp.Stm.summary s1 in
+    let e1 = Summary.Interp.project_from_summary s1 Global.global_env_summary_tag in
+    let s2 = System.Interp.Stm.summary s2 in
+    let e2 = Summary.Interp.project_from_summary s2 Global.global_env_summary_tag in
     e1 == e2
 
   let make_shallow st =
-    { st with
-      system = System.Stm.make_shallow st.system
-    ; shallow = true
+    { interp = Interp.make_shallow st.interp
+    ; synterp = Synterp.make_shallow st.synterp
     }
 
 end
